/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <errno.h>
#include <sys/queue.h>
#include "config.h"
#include "queue.h"
#include "resman.h"
#include "compiler.h"
#include "spdk_io_mgr.h"
#include "nvme_emu_log.h"
#include "memp_bit_array.h"
#include "nvme_emu_io_driver.h"
#include "memp.h"

#ifdef HAVE_SPDK
#include <spdk/thread.h>
#endif

enum memp_result {
    MEMP_ERROR,
    MEMP_COMPLETE,
    MEMP_IN_PROGRESS,
    MEMP_QUOTA_EXCEEDED
};

// Whole memory is divided onto chunks
// Later, these chunks will be distributed among threads
#define CHUNK_SIZE (1048576U*4ULL)

// Minimal allocation size, if client wishes to allocate less than this
// value, actual allocated size is round up to MIN_ALLOC_SIZE
#define MIN_ALLOC_SIZE 1024ULL

// Controls how many requests we allocate at once
#define REQ_GROW_STEP 0x100ULL

// Should be multiple of MIN_ALLOC_SIZE
#define GROUP_SIZE_ALIGNMENT	0x1000ULL
_Static_assert((GROUP_SIZE_ALIGNMENT % MIN_ALLOC_SIZE) == 0,
	       "GROUP_SIZE_ALIGNMENT Should be multiple of MIN_ALLOC_SIZE");

// Helper macros
#define ARRAY_SIZE(n) (sizeof(n) / sizeof(n[0]))
#define MAX_ITEMS_PER_CHUNK (CHUNK_SIZE / MIN_ALLOC_SIZE)

enum CHUNK_GROUP {
    CHUNK_GROUP_1K   = 0x00U,
    CHUNK_GROUP_2K   = 0x01U,
    CHUNK_GROUP_3K   = 0x02U,
    CHUNK_GROUP_4K   = 0x03U,
    CHUNK_GROUP_256K = 0x43U,
    CHUNK_GROUP_HUGE,
    CHUNK_GROUP_COUNT
};

typedef struct chunk {
    // Resource manager index
    resman_entry_t entry;

    // For indexing chunks and fast lookup depending on group
    size_t group_id;
    LIST_ENTRY(chunk) group;

    // Points to the beginning of the data buffer (a portion of
    // struct memp->buffer)
    uint8_t *data;

    // Represents the size of element in this chunk
    // The more the element size, the less elements the chunk holds
    size_t item_size;

    // Number of elements in the chunk. Also means number
    // of meaningful bits in memp_bit_array
    size_t total_items;

    // Number of available elements, should match to number of '1's
    // in memp_bit_array
    size_t num_free_items;

    // Owner thread
    size_t thread_id;

    // Bit array, each bit signals whether corresponding item is
    // available or not. We allocate elements for the worst case
    // In reality, the more item_size is, the less bits are required
    memp_bit_array items[MEM_BITS_ARRAY_SIZE(MAX_ITEMS_PER_CHUNK)];
} chunk_t;

// If we have no available memory to satisfy allocation request
// we use this structure to enqueue the request into a pending queue
typedef struct memp_async_req {
    // Provided by client at memp_alloc
    void *user;
    // Provided by client at memp_alloc
    size_t size;
    // Provided by client at memp_alloc
    memp_callback_t callback;
    // Used by memp
    STAILQ_ENTRY(memp_async_req) entry;
} memp_async_req_t;

// Per thread context
typedef struct thread_context
{
    memp_t *memp;
    // Keeps demand in bytes per request group
    size_t group_demand[CHUNK_GROUP_COUNT];
    // Keeps chunks this thread got from resman
    LIST_HEAD(chunks, chunk) chunks[CHUNK_GROUP_COUNT];
    LIST_HEAD(full_chunks, chunk) full_chunks[CHUNK_GROUP_COUNT];
    // Keeps idle requests
    STAILQ_HEAD(idle_queue, memp_async_req) idle_queue;
    // Keeps requests which are waiting for a memory to become available
    STAILQ_HEAD(pending_queue, memp_async_req) pending_queue;
    // Service queue, required to deallocate requests on termination
    STAILQ_HEAD(alloc_queue, memp_async_req) alloc_queue;
    size_t counters[MEMP_COUNTERS_COUNT];
} thread_context_t;

// Global pool context
typedef struct memp
{
    // Total pool size
    size_t size;
    // Beginning of a back memory buffer
    uint8_t *buffer;
    // The buffer is divided onto chunks (CHUNK_SIZE bytes each)
    chunk_t *chunks;
    // Size of chunks array
    size_t chunk_count;
    // Per thread context
    thread_context_t *threads;
    // Size of threads array
    size_t thread_count;
    // A resource management object which manages chunks
    resman_t resman;
} memp_t;

/**
 * memp_init() - allocate and init memory pool object
 *
 * @buf :           Underlying memory buffer
 * @buf_size:       size of @buf
 * @thread_count:   number of threads in the application
 *
 * Return:          memp_t object or NULL on error
 */
memp_t *memp_init(void *buf, size_t buf_size, size_t thread_count)
{
    memp_t *memp;
    chunk_t *chunk;
    thread_context_t *ctx;
    size_t thread_no, group_no, chunk_id;

    memp = calloc(1, sizeof(memp_t));
    if (!memp)
        goto err;

    memp->buffer = buf;
    memp->size = buf_size;
    memp->thread_count = thread_count;
    memp->chunk_count = buf_size / CHUNK_SIZE;

    memp->threads = calloc(memp->thread_count, sizeof(thread_context_t));
    if (!memp->threads)
        goto free_memp;

    for (thread_no = 0; thread_no < memp->thread_count; thread_no++) {
        ctx = &memp->threads[thread_no];
        ctx->memp = memp;
        for (group_no = 0; group_no < CHUNK_GROUP_COUNT; group_no++) {
            LIST_INIT(&ctx->chunks[group_no]);
            LIST_INIT(&ctx->full_chunks[group_no]);
        }
        STAILQ_INIT(&ctx->idle_queue);
        STAILQ_INIT(&ctx->alloc_queue);
        STAILQ_INIT(&ctx->pending_queue);
    }

    memp->chunks = calloc(memp->chunk_count, sizeof(chunk_t));
    if (!memp->chunks)
        goto free_threads;

    if (resman_init(&memp->resman, memp->thread_count))
        goto free_chunks;

    for (chunk_id = 0; chunk_id < memp->chunk_count; chunk_id++) {
        chunk = &memp->chunks[chunk_id];
        chunk->data = memp->buffer + CHUNK_SIZE * chunk_id;
        chunk->group_id = -1;
        resman_add(&memp->resman, &chunk->entry);
    }

    nvmx_info("Memory pool initialized for %zu threads, size %zu",
            memp->thread_count, memp->size);

    return memp;

free_chunks:
    free(memp->chunks);
free_threads:
    free(memp->threads);
free_memp:
    free(memp);
err:
    nvmx_info("Memory pool initialized failed");
    return NULL;
}

/**
 * memp_destroy() - destroy memory pool object
 *
 * @memp:       Memory pool object
 *
 * Return:      void
 */
void memp_destroy(memp_t *memp)
{
    size_t thread_no;
    thread_context_t *ctx;
    memp_async_req_t *req, *req_temp;

    for (thread_no = 0; thread_no < memp->thread_count; thread_no++) {
        ctx = &memp->threads[thread_no];
        STAILQ_FOREACH_SAFE(req, &ctx->alloc_queue, entry, req_temp) {
            nvmx_debug("Thread %zu free requests %p", thread_no, req);
            free(req);
        }
    }

    resman_clear(&memp->resman);

    free(memp->chunks);
    free(memp->threads);
    free(memp);

    nvmx_info("Memory pool destroyed");
}

void *memp_get_buffer(memp_t *memp)
{
    return memp->buffer;
}

size_t memp_get_size(memp_t *memp)
{
    return memp->size;
}

/**
 * memp_increment_group_demand() - signals to resman that a group needs less
 *              memory
 * @memp:       Memory pool object
 * @thread_id:  Normalized thread id
 * @group_id:   group id
 * @demand_in_bytes
 *              Desired decrement
 *
 * Return:      void
 */
static inline void memp_increment_group_demand(memp_t *memp, size_t thread_id,
                                      size_t group_id, size_t demand_in_bytes)
{
    thread_context_t *ctx = &memp->threads[thread_id];

    const size_t current_demand = ctx->group_demand[group_id];
    const size_t new_demand = current_demand + demand_in_bytes;

    const size_t rounded_current_demand =
                            (current_demand + CHUNK_SIZE - 1) & -CHUNK_SIZE;

    const size_t rounded_new_demand =
                            (new_demand + CHUNK_SIZE - 1) & -CHUNK_SIZE;

    const size_t current_demand_in_chunks =
                            rounded_current_demand / CHUNK_SIZE;

    const size_t new_demand_in_chunks = rounded_new_demand / CHUNK_SIZE;

    const size_t delta = new_demand_in_chunks - current_demand_in_chunks;

    ctx->group_demand[group_id] = new_demand;

    nvmx_debug("Thread %zu demand %zu %zu %zu %zu %zu %zu", thread_id,
            current_demand, new_demand, rounded_current_demand,
            rounded_new_demand, current_demand_in_chunks, new_demand_in_chunks,
            delta);

    if (delta)
        resman_inc_demand(&memp->resman, thread_id);
}

/**
 * memp_decrement_group_demand() - signals to resman that a group needs more
 *              memory
 * @memp:       Memory pool object
 * @thread_id:  Normalized thread id
 * @group_id:   group id
 * @demand_in_bytes
 *              Desired increment
 *
 * Return:      void
 */
static inline void memp_decrement_group_demand(memp_t *memp, size_t thread_id,
                                      size_t group_id, size_t demand_in_bytes)
{
    thread_context_t *ctx = &memp->threads[thread_id];

    const size_t current_demand = ctx->group_demand[group_id];
    const size_t new_demand = current_demand - demand_in_bytes;

    const size_t rounded_current_demand =
                            (current_demand + CHUNK_SIZE - 1) & -CHUNK_SIZE;

    const size_t rounded_new_demand =
                            (new_demand + CHUNK_SIZE - 1) & -CHUNK_SIZE;

    const size_t current_demand_in_chunks =
                            rounded_current_demand / CHUNK_SIZE;

    const size_t new_demand_in_chunks = rounded_new_demand / CHUNK_SIZE;

    const size_t delta = current_demand_in_chunks - new_demand_in_chunks;

    ctx->group_demand[group_id] = new_demand;

    nvmx_debug("Thread %zu demand %zu %zu %zu %zu %zu %zu", thread_id,
            current_demand, new_demand, rounded_current_demand,
            rounded_new_demand, current_demand_in_chunks, new_demand_in_chunks,
            delta);

    if (delta)
        resman_dec_demand(&memp->resman, thread_id);
}

/**
 * memp_get_req() - Find appropriate group for request of certain size
 * @size:       Allocation size
 *
 * All requests large than @CHUNK_GROUP_HUGE, are being processed on behalf
 * of the @CHUNK_GROUP_HUGE group
 *
 * Return:      group_id
 */
static inline size_t memp_get_group_id(size_t size)
{
    size_t group_id;

    if (likely(size >= GROUP_SIZE_ALIGNMENT)) {
        group_id = CHUNK_GROUP_4K + (size / GROUP_SIZE_ALIGNMENT) - ((size & (GROUP_SIZE_ALIGNMENT - 1)) ? 0:1);
        if (group_id > CHUNK_GROUP_HUGE)
            group_id = CHUNK_GROUP_HUGE;
    }
    else
        group_id = CHUNK_GROUP_1K + (size / 1024ULL) - 1;

    return group_id;
}

/**
 * memp_get_req() - Get a request object for asynchronous completion
 * @memp:       Memory pool object
 * @thread_id:  Normalized thread id
 *
 * Return:      NULL, error occurred
                pointer to memp_async_req_t
 */
static inline memp_async_req_t *memp_get_req(memp_t *memp, size_t thread_id)
{
    size_t i;
    memp_async_req_t *req;
    thread_context_t *ctx = &memp->threads[thread_id];

    // Try to get one from idle_queue
    req = STAILQ_FIRST(&ctx->idle_queue);
    if (likely(req)) {
        STAILQ_REMOVE_HEAD(&ctx->idle_queue, entry);
        return req;
    }

    // Allocate next portion of requests
    nvmx_debug("Thread %zu creating new portion of reqs", thread_id);
    req = malloc(REQ_GROW_STEP * sizeof(memp_async_req_t));
    if (unlikely(!req)) {
        nvmx_error("Thread %zu couldn't to allocate memory for requests",
                                                        thread_id);
        return NULL;
    }

    nvmx_debug("Thread %zu %zu requests created at %p", thread_id,
                                                REQ_GROW_STEP, req);

    // We reserve first request and enqueue it into alloc_queue
    // So that we'll be able to free the whole set at once on termination
    STAILQ_INSERT_TAIL(&ctx->alloc_queue, req, entry);
    req++;

    // Last request will be returned from this function. No need to
    // enqueue it to idle queue, hence REQ_GROW_STEP - 1
    for (i = 1; i < REQ_GROW_STEP - 1; i++) {
        STAILQ_INSERT_TAIL(&ctx->idle_queue, req, entry);
        req++;
    }

    return req;
}

/**
 * memp_try_alloc() - Attempt to allocate a buffer from memory pool
 * @memp:       Memory pool object
 * @thread_id:  Normalized thread id
 * @size:       Allocation size (in bytes)
 * @p:          Output, receives a pointer to the buffer

 *
 * Complexity:  O(1)
 *
 * Return:      MEMP_ERROR, error occurred
                MEMP_COMPLETE, allocation finished, this is the only
                    case when output @p is valid
                MEMP_IN_PROGRESS, no available memory, repeat later
                MEMP_QUOTA_EXCEEDED, quota exceeded, repeat later
 */
static enum memp_result memp_try_alloc(memp_t *memp, size_t thread_id,
                                                size_t size, void **p)
{
    chunk_t *chunk = NULL;;
    size_t group_id;
    resman_entry_t *entry;
    thread_context_t *ctx = &memp->threads[thread_id];

    // Stop processing of quota exceeded
    if (resman_quota_exceed(&memp->resman, thread_id))
        return MEMP_QUOTA_EXCEEDED;

    group_id = memp_get_group_id(size);

    if (likely(group_id < CHUNK_GROUP_HUGE))
        chunk = LIST_FIRST(&ctx->chunks[group_id]);
    else {
        LIST_FOREACH(chunk, &ctx->chunks[group_id], group) {
           if (chunk->item_size == size)
               break;
        }
    }

    if (chunk) {
        // Find available slot
        size_t item_no = find_bit(chunk->items,
                                MEM_BITS_ARRAY_SIZE(chunk->total_items));

        // Pointers math to get slot's base pointer
        *p = chunk->data + item_no * chunk->item_size;

        // Consume slot by clearing corresponding bit
        // and decrementing counter
        clr_bit(chunk->items, item_no);
        chunk->num_free_items--;

        if (!chunk->num_free_items) {
            LIST_REMOVE(chunk, group);
            LIST_INSERT_HEAD(&ctx->full_chunks[group_id], chunk, group);
        }

        nvmx_debug("Thread %zu found suitable chunk %p in group %zu, "
                "item_no: %zu", thread_id, chunk, group_id, item_no);

        return MEMP_COMPLETE;
    }

    // Local chunk wasn't found, so try to get one from external pool
    entry = resman_get(&memp->resman, thread_id);

    // No luck
    if (!entry)
        return MEMP_IN_PROGRESS;

    chunk = container_of(entry, chunk_t, entry);
    nvmx_debug("Thread %zu got chunk %p with base %p from pool",
                                    thread_id, chunk, chunk->data);

    // If the chunk is already initialized and meets our needs, simply take it
    if (chunk->group_id == group_id && group_id < CHUNK_GROUP_HUGE) {
        nvmx_debug("Thread %zu reuse chunk %p with %zu items, size %zu",
                thread_id, chunk, chunk->num_free_items, chunk->item_size);
    }
    else {
        // Initialize chunk
        chunk->group_id = group_id;
        chunk->item_size = size;
        chunk->total_items = CHUNK_SIZE / chunk->item_size;

        // We set 'chunk->total_items' amount of bits, indicating that
        // all slots are available
        set_bits(chunk->items, MEM_BITS_ARRAY_SIZE(chunk->total_items),
                                                    chunk->total_items);

        nvmx_debug("Thread %zu init chunk %p with %zu items, size %zu",
                thread_id, chunk, chunk->total_items, chunk->item_size);
    }

    // Consume first slot
    chunk->num_free_items = chunk->total_items - 1;
    clr_bit(chunk->items, 0);

    // No we own the chunk
    chunk->thread_id = thread_id;
    LIST_INSERT_HEAD(&ctx->chunks[group_id], chunk, group);
    ctx->counters[MEMP_COUNTER_CHUNKS]++;
    ctx->counters[MEMP_COUNTER_CHUNKS_ALLOC]++;

    // First slot points to the beginning of the chunk
    // No need for pointers math
    *p = chunk->data;

    nvmx_debug("Thread %zu complete: %p", thread_id, *p);

    return MEMP_COMPLETE;
}

/**
 * memp_thread_poll_impl() - Poller for the memory pool
 * @ptr:        Pointer to thread's context
 *
 * Complexity:  O(N) N - number of pending requests
 *
 * Return:      void
 */
static void memp_thread_poll_impl(void *ptr)
{
    void *p, *user;
    memp_async_req_t *req;
    memp_callback_t callback;
    thread_context_t *ctx = ptr;
    const size_t thread_id = ctx - ctx->memp->threads;

    // Try to complete pending requests
    while ((req = STAILQ_FIRST(&ctx->pending_queue))) {

        // Attempt to allocate requested size
        if (memp_try_alloc(ctx->memp, thread_id, req->size, &p) == MEMP_COMPLETE) {

            // Invalidate request, so that if someone calls memp_cancel,
            // nothing bad happens
            req->size = 0;

            // Release request before callback since callback may issue other
            // allocations. And we want to keep requests idle
            // as much as possible
            user = req->user;
            callback = req->callback;
            STAILQ_REMOVE_HEAD(&ctx->pending_queue, entry);
            STAILQ_INSERT_HEAD(&ctx->idle_queue, req, entry);

            // Complete request
            ctx->counters[MEMP_COUNTER_ASYNC_COMPLETE]++;
            nvmx_debug("Thread %zu completed request %p from poll with %p",
                                                    thread_id, req, p);
            callback(p, user);
        }
        else // If allocation failed, there is no need to attempt further
            break;
    }
}

/**
 * memp_alloc_impl() - Allocate a buffer from memory pool
 * @memp:       Memory pool object
 * @size:       Allocation size (in bytes)
 * @user:       Opaque pointer returned back to user
 * @cb:         Callback function being called if allocation is to be done
 *              asynchronously
 * @tag:        Optional parameter, assigned to a opaque pointer which
 *              may be used later with memp_cancel in order to cancel
 *              asynchronous allocation
 * @thread_id   Normalized thread id
 *
 * Complexity:  O(1) most of the time, worst case O(N) N - number of threads
 *
 * Return:      pointer to the buffer or NULL if request is to be completed
 *              asynchronously.
 *              If error occurred, NULL is returned and errno contains error
 *              code
 */
static inline void *memp_alloc_impl(memp_t *memp, size_t size, void *user,
                            memp_callback_t cb, void **tag, size_t thread_id)
{
    void *p;
    size_t group_id;
    enum memp_result rc;
    memp_async_req_t *req;
    thread_context_t *ctx = &memp->threads[thread_id];

    ctx->counters[MEMP_COUNTER_ALLOC]++;

    // Round up to MIN_ALLOC_SIZE
    const size_t rounded_size = (size + MIN_ALLOC_SIZE - 1) & -MIN_ALLOC_SIZE;

    /*
     * (#3660603)
     * Round up to the next 4k group, only if size is greater than 3k,
     * as there are groups of 1k, 2k and 3k size and after 3k,
     * groups are present at 4k intervals e.g., 4k, 8k, 12k, 16k ... 256k)
     */
    if (rounded_size > (GROUP_SIZE_ALIGNMENT))
        size = (rounded_size + GROUP_SIZE_ALIGNMENT - 1) & -GROUP_SIZE_ALIGNMENT;
    else
        size = rounded_size;

    nvmx_debug("Thread %zu allocation of %zu (%zu) bytes", thread_id, size,
                                                                rounded_size);

    if (size > CHUNK_SIZE) {
        nvmx_error("Thread %zu size exceed chunk limit %zu",
                                     thread_id, CHUNK_SIZE);
        errno = -EINVAL;
        ctx->counters[MEMP_COUNTER_ERROR]++;
        return NULL;
    }

    if (size == 0) {
        nvmx_error("Thread %zu zero allocation is not supported", thread_id);
        errno = -EINVAL;
        ctx->counters[MEMP_COUNTER_ERROR]++;
        return NULL;
    }

    group_id = memp_get_group_id(size);

    // Increment group demand by allocation size
    memp_increment_group_demand(memp, thread_id, group_id, size);

    // Check if we have something pending in order to reach more fair execution
    memp_thread_poll_impl(ctx);

    // Attempt to allocate requested size
    rc = memp_try_alloc(memp, thread_id, size, &p);

    switch (rc) {
    case MEMP_ERROR:
        // Something went wrong
        ctx->counters[MEMP_COUNTER_ERROR]++;
        break;
    case MEMP_COMPLETE:
        ctx->counters[MEMP_COUNTER_IMMEDIATE_COMPLETE]++;
        nvmx_debug("Thread %zu allocation completed with data %p",
                                                   thread_id, p);
        return p;
        break;
    case MEMP_IN_PROGRESS:
        ctx->counters[MEMP_COUNTER_PENDING_NOMEM]++;
        nvmx_debug("Thread %zu no available chunks", thread_id);
        /* no break */
    case MEMP_QUOTA_EXCEEDED:
        // The request will be accomplished asynchronously

        if (rc == MEMP_QUOTA_EXCEEDED) {
            ctx->counters[MEMP_COUNTER_PENDING_QUOTA]++;
            nvmx_debug("Thread %zu quota exceeded", thread_id);
        }

        // Obtain request record to track this allocation
        req = memp_get_req(memp, thread_id);
        if (!req) {
            ctx->counters[MEMP_COUNTER_ERROR]++;
            nvmx_error("Thread %zu failed to get request handle", thread_id);
            rc = -ENOMEM;
            break;
        }

        // Copy request parameters
        req->size = size;
        req->user = user;
        req->callback = cb;

        nvmx_debug("Thread %zu enqueue request %p", thread_id, req);
        STAILQ_INSERT_TAIL(&ctx->pending_queue, req, entry);

        // We need to poll ourself just in case if some memory was released
        // between call to memp_try_alloc and pushing req to pending_queue
        #ifdef HAVE_SPDK
        spdk_thread_send_msg(spdk_io_mgr_get_thread((int)thread_id - 1), memp_thread_poll_impl, ctx);
        #endif

        // The tag may be used to cancel this request
        if (tag)
            *tag = req;

        break;
    }

    errno = 0;

    return NULL;
}

/**
 * memp_alloc() - Allocate a buffer from memory pool
 * @memp:       Memory pool object
 * @size:       Allocation size (in bytes)
 * @user:       Opaque pointer returned back to user
 * @cb:         Callback function being called if allocation is to be done
 *              asynchronously
 * @tag:        Optional parameter, assigned to a opaque pointer which
 *              may be used later with memp_cancel in order to cancel
 *              asynchronous allocation
 * @tid         current I/O thread id, -1 for primary application thread
 *
 * Complexity:  O(1) most of the time, worst case O(N) N - number of threads
 *
 * Return:      pointer to the buffer or NULL if request is to be completed
 *              asynchronously.
 *              If error occurred, NULL is returned and errno contains error
 *              code
 */
void *memp_alloc(memp_t *memp, size_t size, void *user,
                            memp_callback_t cb, void **tag, int tid)
{
    const size_t thread_id = tid + 1;
    return memp_alloc_impl(memp, size, user, cb, tag, thread_id);
}

/**
 * memp_alloc_ex() - Allocate a buffer from memory pool
 * @req:        I/O request
 * @cb:         Callback function being called if allocation is to be done
 *              asynchronously, req is provided as user argument
 *
 * Complexity:  O(1) most of the time, worst case O(N) N - number of threads
 *
 * Operates with a fewer arguments count to speed up things with heavy load
 *
 * Return:      pointer to the buffer or NULL if request is to be completed
 *              asynchronously.
 *              If error occurred, NULL is returned and errno contains error
 *              code
 */
void *memp_alloc_ex(nvme_io_driver_req_t *req, memp_callback_t cb)
{
    const size_t thread_id = req->async_req.thread_id + 1;
    return memp_alloc_impl(req->async_req.memp, req->async_req.len, req, cb,
                                        NULL, thread_id);
}

void *memp_alloc_req(nvme_async_req_t *req, memp_callback_t cb)
{
    const size_t thread_id = req->thread_id + 1;
    return memp_alloc_impl(req->memp, req->len, req, cb,
                                        NULL, thread_id);
}

/**
 * memp_cancel() - Cancel asynchronous allocation
 * @memp:       Memory pool object
 * @tag:        Request tag
 * @tid:        current I/O thread id, -1 for primary application thread
 *
 * Complexity:  O(1)
 *
 * Return:      zero on success, error code in case of an error
 */
int memp_cancel(memp_t *memp, void *tag, int tid)
{
    memp_async_req_t *req = tag;
    const size_t thread_id = tid + 1;
    thread_context_t *ctx = &memp->threads[thread_id];

    // We use req->size as a valid marker
    if (req && req->size) {
        req->size = 0;
        STAILQ_REMOVE(&ctx->pending_queue, req, memp_async_req, entry);
        STAILQ_INSERT_HEAD(&ctx->idle_queue, req, entry);
        ctx->counters[MEMP_COUNTER_CANCEL]++;
    }

    return 0;
}

/**
 * memp_free() - Return a buffer to memory pool
 * @memp:       Memory pool object
 * @p:          Pointer to the buffer
 *
 * Complexity:  O(1) most of the time, worst case O(N) N - number of threads
 *
 * Return:      void
 */
void memp_free(memp_t *memp, void *p)
{
    int i;
    size_t item_id;
    chunk_t *chunk;
    size_t thread_id;
    thread_context_t *ctx;

    if (!p)
        return;

    // Simple pointers math to get chunk
    chunk = &memp->chunks[((uint8_t *)p - memp->buffer) / CHUNK_SIZE];
    thread_id = chunk->thread_id;
    ctx = &memp->threads[thread_id];

    ctx->counters[MEMP_COUNTER_FREE]++;

    // Decrement group demand as we need less memory now
    memp_decrement_group_demand(memp, thread_id, chunk->group_id,
                                                    chunk->item_size);

    // Same pointers math to get item slot inside chunk
    item_id = ((uint8_t *)p - chunk->data) / chunk->item_size;

    nvmx_debug("Thread %zu put item %zu to chunk %p", thread_id, item_id,
                                                                chunk);
    // Return item to chunk by setting appropriate bit
    set_bit(chunk->items, item_id);

    if (!chunk->num_free_items) {
        // Remove from full_chunks
        LIST_REMOVE(chunk, group);
        LIST_INSERT_HEAD(&ctx->chunks[chunk->group_id], chunk, group);
        chunk->num_free_items = 1;
    }
    else {
        chunk->num_free_items++;

        // If chunk is totally empty - return it to resman
        if (chunk->num_free_items == chunk->total_items) {

            nvmx_debug("Thread %zu put chunk %p to pool", thread_id, chunk);

            // Release our ownership
            LIST_REMOVE(chunk, group);
            ctx->counters[MEMP_COUNTER_CHUNKS]--;
            ctx->counters[MEMP_COUNTER_CHUNKS_FREE]++;

            // If entry was returned to local pool - check if we have pending items
            // Otherwise poll other threads in case they have something pending
            if (resman_put(&memp->resman, thread_id, &chunk->entry))
                memp_thread_poll_impl(&memp->threads[thread_id]);
            else {
                for (i = 0; resman_has_entry(&memp->resman) &&
                                            i < memp->thread_count; i++) {

                    // Ignore ourself
                    if (i == thread_id)
                        continue;

                    // Ignore threads which have no pending items
                    if (STAILQ_EMPTY(&memp->threads[i].pending_queue))
                        continue;

                    // Poll thread
                    #ifdef HAVE_SPDK
                    spdk_thread_send_msg(spdk_io_mgr_get_thread(i - 1),
                                memp_thread_poll_impl, &memp->threads[i]);
                    #endif
                }
            }
        }
    }
}

/**
 * memp_thread_poll() - Poller for the memory pool
 * @memp:       Memory pool object
 * @tid:        current I/O thread id, -1 for primary application thread
 *
 * Should be called in a loop for non-SPDK environment
 *
 * Complexity:  O(N) N - number of pending requests
 *
 * Return:      void
 */
void memp_thread_poll(memp_t *memp, int tid)
{
    // In case of SPDK we poll external polling is not required
#ifndef HAVE_SPDK
    const size_t thread_id = tid + 1;
    thread_context_t *ctx = &memp->threads[thread_id];

    memp_thread_poll_impl(ctx);
#endif
}

/**
 * memp_get_counters() - Get thread statistics
 * @memp:       Memory pool object
 * @tid:        current I/O thread id, -1 for primary application thread
 */
size_t *memp_get_counters(memp_t *memp, int tid)
{
    const size_t thread_id = tid + 1;
    thread_context_t *ctx = &memp->threads[thread_id];
    return ctx->counters;
}
