/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include <stdarg.h>
#include <Python.h>
#include "nvme_emu_log.h"
#include "json/json.h"
#include "nic_utils.h"

#define MLNX_SNAP_BFB_ADMIN_PATH "MLNX_SNAP_BFB_ADMIN_PATH"

struct nic_utils
{
    PyObject *module;
    PyObject *dict;
    PyObject *std_reader;
    size_t bytes_read;
    FILE *fp;
};

static const char *std_reader = "\
import io, sys\n\
class StdRedirect(io.TextIOBase):\n\
    def __init__(self):\n\
        self.value = ''\n\
    def write(self, line):\n\
        self.value += line\n\
std_reader = StdRedirect()\n\
sys.stderr = std_reader\n\
sys.stdout = std_reader\
";

static void nic_utils_handle_output(nic_utils_t *ctx)
{
    PyObject *output;
    size_t out_len;
    const char *out_str;

    if (PyErr_Occurred()) {
        PyErr_PrintEx(0);
        PyErr_Clear();
    }

    output = PyObject_GetAttrString(ctx->std_reader, "value");
    if (output) {
        out_str = (char *)PyUnicode_1BYTE_DATA(output);
        if (out_str) {
            out_len = strlen(out_str);
            if (out_len > ctx->bytes_read) {
                nvmx_info("Library reports: %s", out_str + ctx->bytes_read);
                ctx->bytes_read = out_len;
            }
        }
        Py_DECREF(output);
    }
}

nic_utils_t* nic_utils_open()
{
    const char *filename;
    int status;
    nic_utils_t *ctx;

    filename = getenv(MLNX_SNAP_BFB_ADMIN_PATH);
    if (!filename) {
        nvmx_error("Unable to locate bfb_admin script\n");
        goto out;
    }

    ctx = calloc(1, sizeof(nic_utils_t));
    if (!ctx)
        goto out;

    ctx->fp = fopen(filename, "r");
        if (!ctx->fp)
             goto free_mem;

    Py_Initialize();

    ctx->module = PyImport_AddModule("__main__");
    if (!ctx->module) {
        nvmx_error("Failed to create a module");
        goto close_file;
    }

    ctx->dict = PyModule_GetDict(ctx->module);
    if (!ctx->dict) {
        nvmx_error("Failed to create a dictionary");
        goto close_file;
    }

    if (PyRun_SimpleString(std_reader)) {
        nvmx_error("Failed to execute a reader script");
        goto close_file;
    }

    ctx->std_reader = PyObject_GetAttrString(ctx->module, "std_reader");
    if (!ctx->std_reader) {
        nvmx_error("Failed to create std reader");
        goto close_file;
    }

    status = PyRun_SimpleFile(ctx->fp, filename);

    nic_utils_handle_output(ctx);

    if (status)
        goto free_reader;

    return ctx;

free_reader:
    Py_DECREF(ctx->std_reader);
close_file:
    fclose(ctx->fp);
    Py_Finalize();
free_mem:
    free(ctx);
out:
    return NULL;
}

void nic_utils_close(nic_utils_t *ctx)
{
    if (ctx->fp)
        fclose(ctx->fp);
    Py_DECREF(ctx->std_reader);
    Py_Finalize();
    free(ctx);
}

char *nic_utils_exec(nic_utils_t *ctx, const char *func_name, const char *proto, ...)
{
    PyObject *function, *args = NULL, *value;
    char *ret = NULL;
    size_t num_args, i;
    char *str;
    va_list ap;

    function = PyDict_GetItemString(ctx->dict, func_name);
    if (!function) {
        nvmx_error("Function '%s' not found", func_name);
        goto err;
    }

    num_args = proto ? strlen(proto) : 0;

    if (num_args) {

        args = PyTuple_New(num_args);
        if (!args) {
            nvmx_error("Failed to create arguments tuple for function '%s'", func_name);
            goto err;
        }

        va_start(ap, proto);

        for (i = 0; i < num_args; i++) {
            switch (proto[i]) {
            case 's':
                value = PyUnicode_FromString(va_arg(ap, char *));
                break;
            case 'l':
                value = PyLong_FromLong(va_arg(ap, long));
                break;
            case 'b':
                value = PyBool_FromLong(va_arg(ap, int));
                break;
            default:
                va_end(ap);
                nvmx_error("Unknown argument type %c", proto[i]);
                goto free_args;
            }
            if (PyTuple_SetItem(args, i, value)) {
                va_end(ap);
                nvmx_error("Failed to apply argument of type %c at position %zu", proto[i], i);
                goto free_args;
            }
        }

        va_end(ap);
    }

    value = PyObject_CallObject(function, args);

    if (value) {
        str = (char *)PyUnicode_1BYTE_DATA(value);
        if (str) {
            nvmx_debug("Python output: %s", str);
            ret = strdup(str);
        }
        Py_DECREF(value);
    }

free_args:
    Py_XDECREF(args);
err:
    nic_utils_handle_output(ctx);

    return ret;
}

