/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define _GNU_SOURCE
#include "config.h"
#include "nvme_emu.h"
#include "nvme_emu_io_driver.h"
#include "nvme_controller.h"
#include "nvme_prp.h"
#include "compiler.h"
#include "json/json.h"
#include "json/nvme_json.h"
#include "utils.h"
#ifdef HAVE_SPDK
#include "nvme_spdk_subsystem.h"
#endif
#if HAVE_SNAP
#include "mlnx_snap_pci_manager.h"
#endif
#include "firmware.h"
#include "nic_utils.h"
#include <string.h>
#include <pthread.h>
#include <sys/sysinfo.h>
#include "snap_env.h"
#include "effdma/effdma_domain.h"

/* Unlimited burst. Set lpw, mpw and hpw to 64, note: zero based value */
#define NVME_DEFAULT_ARBITRATION (0x7 | (63 << 8) | (63 << 16) | (63 << 24))

#define NVME_AER_LIMIT 4

/* Default power state. WH = 0, PS = 0 */
#define NVME_DEFAULT_PM           0

#define ASYNC_REQ_RESULT(type, info, logpage) \
        (type) | (info << 8) | (logpage << 16)

#define ASYNC_EVENT_IS_MASKED(ctrl, logpage) \
        (bool)(ctrl->masked_logpage_ids & (1 << logpage))

#define ASYNC_EVENT_MASK(ctrl, logpage) \
        ctrl->masked_logpage_ids |= (1 << logpage)

#define ASYNC_EVENT_UNMASK(ctrl, logpage) \
        ctrl->masked_logpage_ids &= ~(1 << logpage)

/* how much time to wait before the next IO RETRY */
#define NVME_CTRL_WAIT_USECS  10
/* how many times to try to perform cmd on non-active endpoint */
#define NVME_CTRL_NRETRIES    5

#define NVME_SGL_SUPP "NVME_SGL_SUPP"
SNAP_ENV_REG_ENV_VARIABLE(NVME_SGL_SUPP, 1);

#define NVME_FW_SUPP "NVME_FW_SUPP"
SNAP_ENV_REG_ENV_VARIABLE(NVME_FW_SUPP, 0);

struct nvme_ctrl_pg {
    TAILQ_HEAD(sq_list_head, nvme_squeue) sq_list;
    pthread_spinlock_t lock;
};

static size_t *nvme_pg_usage = NULL;
static size_t nvme_pg_ref_count = 0;

static void nvme_discard_image(nvme_ctrl_t *ctrl);

static void nvme_free_bfb_info(nvme_ctrl_t *ctrl);

static void nvme_ctrl_pgs_free(nvme_ctrl_t *ctrl);
static int nvme_ctrl_pgs_alloc(nvme_ctrl_t *ctrl, int nthreads);

static void nvme_ctrl_sq_sched(nvme_ctrl_t *ctrl, nvme_squeue_t *sq);
static void nvme_ctrl_sq_desched(nvme_squeue_t *sq);

static inline int nvme_ctrl_progress_pending_cqes(nvme_squeue_t *sq);

static void nvme_progress_delete_sq(nvme_squeue_t *sq, nvme_ctrl_t *ctrl);
static void nvme_cq_free(nvme_cqueue_t *cq, nvme_ctrl_t *ctrl);
static void nvme_ctrl_clear_namespaces(nvme_ctrl_t *ctrl);
static nvme_namespace_t *nvme_find_namespace(nvme_ctrl_t *ctrl, uint32_t nsid);
static nvme_namespace_t *nvme_find_namespace_nolock(nvme_ctrl_t *ctrl, uint32_t nsid);
static void nvme_ctrl_queue_async_event(nvme_ctrl_t *ctrl,
                                        enum NvmeAsyncEventConfig event);
static void nvme_ctrl_push_async_event_request(nvme_ctrl_t *ctrl,
                                               nvme_request_t *req);
static void nvme_ctrl_abort_async_event_requests(nvme_ctrl_t *ctrl);
static inline size_t nvme_ctrl_num_async_event_requests(nvme_ctrl_t *ctrl);
static void nvme_ctrl_raise_async_events(nvme_ctrl_t *ctrl);
static int nvme_acmd_vs_json_rpc_req_do(nvme_request_t *req, enum nvme_emu_status rc);
static uint16_t nvme_acmd_fw_download_do(nvme_request_t *req, enum nvme_emu_status rc);
static void strpadcpy(char *buf, int buf_size, const char *str, char pad)
{
    int len = strnlen(str, buf_size);

    memcpy(buf, str, len);
    memset(buf + len, pad, buf_size - len);
}

static void nvme_command_dispatch(nvme_squeue_t *sq, const nvme_cmd_t *cmd,
                                  bool fabrics);

static void nvme_enqueue_req_completion(nvme_cqueue_t *cq, nvme_request_t *req);

static inline bool nvme_check_sqid_valid(nvme_ctrl_t *ctrl, uint16_t sqid)
{
    return sqid < ctrl->num_queues && ctrl->sq[sqid] != NULL;
}

static inline bool nvme_check_cqid_valid(nvme_ctrl_t *ctrl, uint16_t cqid)
{
    return cqid < ctrl->num_queues && ctrl->cq[cqid] != NULL;
}

static inline bool nvme_check_new_sqid_valid(nvme_ctrl_t *ctrl, uint16_t sqid)
{
    return sqid < ctrl->num_queues && ctrl->sq[sqid] == NULL;
}

static inline bool nvme_check_new_cqid_valid(nvme_ctrl_t *ctrl, uint16_t cqid)
{
    return cqid < ctrl->num_queues && ctrl->cq[cqid] == NULL;
}

static inline bool nvme_check_cqid_in_range(nvme_ctrl_t *ctrl, uint16_t cqid)
{
    return cqid && cqid < ctrl->num_queues;
}

static int nvme_ctrl_handle_cmd(nvme_queue_ops_t *ops, uint32_t db_addr,
                                const nvme_cmd_t *cmd, uint32_t len, bool fabrics)
{
    nvme_squeue_t *sq;
    int qid;

    sq = container_of(ops, nvme_squeue_t, ops);

    qid = (db_addr - 0x1000) >> 3;
    nvmx_debug("sq: %d SQE write", qid); 

    if (unlikely(!nvme_check_sqid_valid(sq->ctrl, qid))) {
        nvmx_error("submission queue doorbell write for nonexistent queue,"
                " sqid=%d ignoring", qid);
        return NVMX_ERR;
    }

    /* handle command */
    nvme_command_dispatch(sq, cmd, fabrics);
    return NVMX_OK;
}

/* SQE mode SQ/CQ */
static int nvme_ctrl_ops_sqe(nvme_queue_ops_t *ops, uint32_t db_addr,
        const void *cmd, uint32_t len)
{
    return nvme_ctrl_handle_cmd(ops, db_addr, cmd, len, false);
}

static int nvme_ctrl_ops_cq_bad_op(nvme_queue_ops_t *ops, uint32_t db_addr,
                                   void *cmd, uint32_t len)
{
    //nvme_cqueue_t *cq;

    nvmx_fatal("ops 0x%p callback called", ops);
    //cq = container_of(ops, nvme_cqueue_t, ops);
    /* should never get here */
    //nvmx_fatal("CQ%d callback called", cq->cqid);
    return NVMX_OK;
}

static int nvme_ctrl_ops_sq_bad_op(nvme_queue_ops_t *ops, uint32_t db_addr,
                                   const void *cmd, uint32_t len)
{
    //nvme_squeue_t *sq;

    /* should never get here */
    nvmx_fatal("ops 0x%p callback called", ops);
    //sq = container_of(ops, nvme_squeue_t, ops);

    return NVMX_OK;
}

static nvme_queue_ops_t ctrl_sqe_ops = {
    .process_sq       = nvme_ctrl_ops_sqe,
    .process_cq       = nvme_ctrl_ops_cq_bad_op
};

/* CC mode SQ/CQ */
static int nvme_ctrl_ops_cc(nvme_queue_ops_t *ops, uint32_t db_addr,
        const void *cmd, uint32_t len)
{
    nvmx_debug("CC callback called, ops 0x%p len %d", ops, len);

    return nvme_ctrl_handle_cmd(ops, db_addr, cmd, len, true);
}

static nvme_queue_ops_t ctrl_cc_ops = {
    .process_sq       = nvme_ctrl_ops_cc,
    .process_cq       = nvme_ctrl_ops_cq_bad_op
};

static nvme_queue_ops_t ctrl_cc_remote_ops = {
    .process_sq       = nvme_ctrl_ops_sq_bad_op,
    .process_cq       = nvme_ctrl_ops_cq_bad_op
};

static int nvme_post_cqes(nvme_cqueue_t *cq, nvme_squeue_t *sq);

void nvme_bar_write_cb(void *bar, nvme_register_t *reg, uint64_t val,
		       uint64_t prev_val);

void nvme_ctrl_bar_write_event(void *ctrl)
{
    nvme_emulator_t *dev = ((nvme_ctrl_t *)ctrl)->dev;
    int err;

    err = nvme_bar_update(&dev->bar, (nvme_bar_read_func_t)nvme_emu_mmio_read,
                          nvme_bar_write_cb);
    if (err != NVMX_OK) {
        nvmx_warn("failed to update bar");
    }
}

static inline void nvme_addr_read(nvme_squeue_t *sq, void *buf, uint64_t addr, size_t size)
{
    nvme_emu_dma_read(&sq->ops, buf, addr, size, 0);
}

static int nvme_cq_init(nvme_cqueue_t *cq, nvme_ctrl_t *ctrl, uint64_t dma_addr,
        uint16_t cqid, uint16_t vector, uint16_t size, uint16_t irq_enabled)
{
    int ret;
    int cq_period, cq_max_count;

    cq->ctrl = ctrl;
    cq->cqid = cqid;
    cq->size = size;
    cq->dma_addr = dma_addr;
    cq->phase = 1;
    cq->vector = vector;
    cq->head = cq->tail = 0;
    TAILQ_INIT(&cq->req_list);
    TAILQ_INIT(&cq->sq_list);

    /* SPEC: 5.14.1.8 states that moderation only applies to the IO queues */
    if (cqid) {
        cq_period    = ctrl->cq_period;
        cq_max_count = ctrl->cq_max_count;
    } else {
        cq_period    = 0;
        cq_max_count = 0;
    }

    ret = nvme_driver_create_cq(ctrl->driver, cqid, dma_addr, size, vector,
                                cq_period, cq_max_count, irq_enabled,
                                &cq->ops);
    if (ret != NVMX_OK)
        return ret;

    ctrl->cq[cqid] = cq;
    return NVMX_OK;
}

static void nvme_io_req_completion(void *user, enum nvme_emu_status rc)
{
    nvme_request_t *req = user;
    nvme_ctrl_t *ctrl;
    nvme_cqueue_t *cq;

    uint64_t len = req->io_req.async_req.len;
    ctrl = req->sq->ctrl;
    cq = ctrl->cq[req->sq->cqid];

    if (unlikely(req->rpc_buf)) {
        req->status = nvme_acmd_vs_json_rpc_req_do(req, rc);
        if (unlikely(req->status != NVME_SC_SUCCESS))
            nvme_enqueue_req_completion(cq, req);
    }
    else if (unlikely(req->fw_buf)) {
        req->status = nvme_acmd_fw_download_do(req, rc);
        nvme_enqueue_req_completion(cq, req);
    }
    else {
        if (unlikely(rc != NVMX_OK)) {
            nvmx_error("sq%d: io completion status %d", req->sq->sqid, rc);
            req->status = req->io_req.async_req.err_code;
            if (likely(req->ns_stat)) {
                req->ns_stat->err++;
                req->ns_stat->err_bytes_count += len;
            }
            if (likely(req->ctrl_stat)) {
                req->ctrl_stat->err++;
                req->ctrl_stat->err_bytes_count += len;
            }
        }
        else {
            req->status = NVME_SC_SUCCESS;
            if (likely(req->ns_stat)) {
                req->ns_stat->completed++;
                req->ns_stat->completed_bytes_count += len;
                req->ns_stat->completed_units_count += ROUND_UP(len, 512) / 512;
            }
            if (likely(req->ctrl_stat)) {
                req->ctrl_stat->completed++;
                req->ctrl_stat->completed_bytes_count += len;
                req->ctrl_stat->completed_units_count += ROUND_UP(len, 512) / 512;
            }
        }

        nvme_enqueue_req_completion(cq, req);
    }
}

static inline void nvme_io_req_completion_lock(nvme_request_t *req,
                                               nvme_ctrl_t *ctrl,
                                               enum nvme_emu_status rc)
{
    /*
     * Since we complete a request from a different context, and
     * asynchronously, we need to make sure we're not racing with other
     * pthread contexts. Currently the only race is between async event
     * and rpc server completions
     */
    pthread_mutex_lock(&ctrl->async_event_lock);
    req->io_req.comp_cb(req, rc);
    pthread_mutex_unlock(&ctrl->async_event_lock);
}

static int nvme_sq_init(nvme_squeue_t *sq, nvme_ctrl_t *ctrl, uint64_t dma_addr,
    uint16_t sqid, uint16_t cqid, uint16_t size)
{
    int i;
    nvme_cqueue_t *cq;
    int ret;
    int req_size;
    nvme_request_t *req;
    size_t req_data_len, req_aux_len;

    sq->dma_addr = dma_addr;
    sq->sqid = sqid;
    sq->size = size;
    sq->cqid = cqid;
    sq->head = sq->tail = 0;
    sq->ctrl = ctrl;
    cq = ctrl->cq[cqid];

    ret = nvme_driver_create_sq(ctrl->driver, sqid, dma_addr, size,
                                &cq->ops, &sq->ops);
    if (ret != NVMX_OK)
        return ret;

    /* same initialization is used for both admin and io queues even though
     * admin queue will use only async_req part of the request. It make code
     * much simpler.
     */
    req_size = nvme_driver_request_size(ctrl->driver) + sizeof(nvme_request_t) -
               sizeof(nvme_io_driver_req_t);
    sq->reqs_base = calloc(sq->size, req_size);
    if (!sq->reqs_base)
        goto free_sq;

    sq->ops.page_bits = ctrl->page_bits;
    TAILQ_INIT(&sq->req_list);
    TAILQ_INIT(&sq->out_req_list);

    /* Calculate values that are the same for all requests on req 0 */
    req = (nvme_request_t *)(sq->reqs_base);
    sq->mr = nvme_driver_sq_init(ctrl->driver, ctrl->memp, &sq->ops, size, &req->io_req);
    if (!sq->mr && !ctrl->memp)
        goto free_sq;

    req_data_len = req->io_req.async_req.data_len;
    req_aux_len = req->io_req.async_req.aux_len;
    sq->io_buf = req->io_req.async_req.io_buf;

    for (i = 0; i < sq->size; i++) {
        req = (nvme_request_t *)(sq->reqs_base + i * req_size);
        req->sq = sq;
        TAILQ_INSERT_TAIL(&(sq->req_list), req, entry);
        ret = nvme_driver_request_init(ctrl->driver, &sq->ops, &req->io_req,
                                 nvme_io_req_completion, req, i, ctrl->memp,
                                 req_data_len, req_aux_len, sq->io_buf, sq->mr);
        if (ret != NVMX_OK)
            goto free_reqs;

        /* for the admin queue */
        req->io_req.async_req.comp_cb = nvme_io_req_completion;
        req->io_req.async_req.comp_cb_user = req;
    }

    TAILQ_INSERT_TAIL(&(cq->sq_list), sq, entry);
    ctrl->sq[sqid] = sq;

    ctrl->sq_stat[sqid].valid = true;

    sq->state = NVME_SQ_RUNNING;
    nvme_ctrl_sq_sched(ctrl, sq);
    return NVMX_OK;

free_reqs:
    for (--i; i >= 0; i--) {
        req = (nvme_request_t *)(sq->reqs_base + i * req_size);
        if (i == 0) {
            nvme_driver_sq_reset(&req->io_req);
            sq->io_buf = NULL;
            sq->mr  = NULL;
        }
        nvme_driver_request_reset(&req->io_req);
    }
    free(sq->reqs_base);
free_sq:
    nvme_driver_delete_sq(ctrl->driver, &sq->ops, sq->sqid);
    return NVMX_ERR;
}

static void nvme_ctrl_id_uninit(nvme_ctrl_t *ctrl)
{
    memset(&ctrl->id_ctrl, 0, sizeof(ctrl->id_ctrl));
}

static void nvme_ctrl_id_init(nvme_ctrl_t *ctrl)
{
    nvme_ctrl_id_t *id = &ctrl->id_ctrl;
    int vf;

    memset(id, 0, sizeof(*id));
    id->vid = cpu_to_le16(ctrl->dev->vid);
    id->ssvid = cpu_to_le16(ctrl->dev->ssvid);
    id->cntlid = cpu_to_le16(snap_config_get_cntlid(ctrl->sconfig));

    strpadcpy((char *)id->mn, sizeof(id->mn), snap_config_get_mn(ctrl->sconfig),
              ' ');

    strpadcpy((char *)id->fr, sizeof(id->fr), "1.0", ' ');

    strpadcpy((char *)id->sn, sizeof(id->sn), snap_config_get_sn(ctrl->sconfig),
              ' ');


    id->rab = 6;
    id->ieee[0] = 0x00;
    id->ieee[1] = 0x02;
    id->ieee[2] = 0xb3;
    id->cntrltype = 0x1;

    if ((ctrl->quirks & NVME_QUIRK_FORCE_OACS_NS_MGMT)) {
        nvmx_info("Force OACS value to support Namespace Management");
        id->oacs.ns_manage = 1;
    }

    if (ctrl->fw_update_enable) {
        // Bit 2 if set to ‘1’, then the controller supports the
        // Firmware Commit and Firmware  Image Download commands
        id->oacs.firmware = 1;

        // Bits 3:1 indicate the number of firmware slots that the
        // controller supports
        id->frmw.num_slots = NVME_FW_SLOTS_COUNT;

        // Bit 0 if set to ‘1’ indicates that the first firmware slot
        // (i.e., slot 1) is read only
        id->frmw.slot1_ro = 0;

        // There is no granularity limitation
        id->fwug = 0xFF;
    }

    /* TODO: support namespace specific HEALTH log pages */
    id->lpa.celp = 1;

    id->oaes.raw = cpu_to_le32(NVME_SUPPORTED_OPTIONAL_AENS);
    /*
     * Async Event Request Limit (AERL) is zero-based value,
     * we currently support up to 1 parallel Async events.
     */
    id->aerl = NVME_AER_LIMIT - 1; // -1 for zero-based value

    id->sqes.min = id->sqes.max = snap_config_get_sqes(ctrl->sconfig);
    id->cqes.min = id->cqes.max = snap_config_get_cqes(ctrl->sconfig);

    id->psd[0].mp = cpu_to_le16(0x9c4);
    id->psd[0].enlat = cpu_to_le32(0x10);
    id->psd[0].exlat = cpu_to_le32(0x4);

    id->mdts = snap_config_get_mdts(ctrl->sconfig);
    id->oncs.raw = cpu_to_le16(snap_config_get_oncs(ctrl->sconfig));

    // Flush command for all namespaces is not supported
    id->vwc.flush_broadcast = 0x2;

    vf = snap_config_get_pci_vfunc(ctrl->sconfig);
    if (snap_config_get_subnqn(ctrl->sconfig)) {
        /* subnqn was set by the SPDK nvme subsystem */
        snprintf((char *)id->subnqn, sizeof(id->subnqn), "%s",
                 snap_config_get_subnqn(ctrl->sconfig));
        id->cmic.multi_ctrl = 1;
    } else {
        /* running in the stand alone mode. Create host unique subnqn */
        if (vf >= 0)
            snprintf((char *)id->subnqn, sizeof(id->subnqn), "%s%04x.%04x.%04x",
                     NVME_DEFAULT_SUBNQN_PREFIX, id->vid,
                     snap_config_get_pci_func(ctrl->sconfig), vf);
        else
            snprintf((char *)id->subnqn, sizeof(id->subnqn), "%s%04x.%04x",
                     NVME_DEFAULT_SUBNQN_PREFIX, id->vid,
                     snap_config_get_pci_func(ctrl->sconfig));
    }

    if (vf >= 0)
        id->cmic.sr_iov = 1;

    if ((snap_env_getenv(NVME_SGL_SUPP) != 0) &&
        snap_config_get_emu_mode(ctrl->sconfig) == SNAP_EMU_SQE_ONLY)
        id->sgls = cpu_to_le32(NVME_SGLS_SUPPORTED_NO_REQS | NVME_SGLS_LARGE_SGL);

    id->ver.raw = cpu_to_le32(ctrl->dev->bar.curr.vs);
    nvmx_info("NVM Express specification version: %u.%u.%u",
              id->ver.bits.mjr, id->ver.bits.mnr, id->ver.bits.ter);

    if (id->ver.raw >= NVME_VERSION(1, 4, 0)) {
        id->nn = snap_config_get_nn(ctrl->sconfig);
        if (snap_config_get_emu_mode(ctrl->sconfig) != SNAP_EMU_SQE_ONLY)
            id->nn = nvmx_min(id->nn, ctrl->dev->dev_caps.max_nsid);
        id->mnan = snap_config_get_mnan(ctrl->sconfig);
        id->mnan = nvmx_min(id->mnan,
                            snap_config_get_max_namespaces(ctrl->sconfig));
        if (snap_config_get_emu_mode(ctrl->sconfig) != SNAP_EMU_SQE_ONLY)
            id->mnan = nvmx_min(id->mnan, ctrl->dev->dev_caps.max_namespaces);
    } else {
        id->nn = snap_config_get_nn(ctrl->sconfig);
        id->nn = nvmx_min(id->nn,
                          snap_config_get_max_namespaces(ctrl->sconfig));
        if (snap_config_get_emu_mode(ctrl->sconfig) != SNAP_EMU_SQE_ONLY) {
            id->nn = nvmx_min(id->nn, ctrl->dev->dev_caps.max_nsid);
            id->nn = nvmx_min(id->nn, ctrl->dev->dev_caps.max_namespaces);
        }
        id->mnan = 0;
    }
    id->mnan = nvmx_min(id->mnan, id->nn);
    ctrl->max_namespaces = nvmx_min(id->mnan ? : id->nn,
                                 snap_config_get_max_namespaces(ctrl->sconfig));
    nvmx_debug("Maximal allowed NSID: %lu, Maximum allowed namespaces: %lu",
               id->nn, id->mnan);

    id->avscc.spec_format = 1;
    id->vs.json_rpc_2_0_mjr = cpu_to_le16(NVME_CTRL_JSON_RPC_2_0_MJR);
    id->vs.json_rpc_2_0_mnr = cpu_to_le16(NVME_CTRL_JSON_RPC_2_0_MNR);
    id->vs.json_rpc_2_0_ter = cpu_to_le16(NVME_CTRL_JSON_RPC_2_0_TER);

    if (ctrl->iova_mgmt_enable) {
        id->vs.iova_mmts = cpu_to_le16(NVME_CTRL_IOVA_MAP_UNMAP_RANGE | NVME_CTRL_IOVA_UNMAP_ALL);
        id->vs.iova_mms = 16;
    }

    nvmx_info("SNAP CLI tool version: %u.%u.%u",
              le32_to_cpu(id->vs.json_rpc_2_0_mjr),
              le32_to_cpu(id->vs.json_rpc_2_0_mnr),
              le32_to_cpu(id->vs.json_rpc_2_0_ter));
}

static void nvme_ctrl_update_changed_namespace(nvme_ctrl_t *ctrl, uint32_t nsid)
{
    int i;

    pthread_spin_lock(&ctrl->namespaces_lock);
    if (ctrl->num_changed_nsids < 1024) {
        for (i = 0; i < ctrl->num_changed_nsids; i++) {
            if (ctrl->changed_nsids[i] == nsid)
                break;
        }
        if (i == ctrl->num_changed_nsids)
            ctrl->changed_nsids[ctrl->num_changed_nsids++] = nsid;
    }
    pthread_spin_unlock(&ctrl->namespaces_lock);

    nvme_ctrl_queue_async_event(ctrl, NVME_NAMESPACE_ATTR_NOTICE);
}

static void nvme_ctrl_unmap_nsid(nvme_ctrl_t *ctrl, uint32_t nsid)
{
#if HAVE_SNAP
    const char *rdma_dev = snap_config_get_emu_manager(ctrl->sconfig);

    if (snap_config_get_emu_mode(ctrl->sconfig) != SNAP_EMU_SQE_ONLY)
        mlnx_snap_release_namespace_id(rdma_dev, nsid);
#endif
}

/* Returns src_nsid from dst_nsid, or 0 (invalid nsid value) on failure */
static uint32_t nvme_ctrl_map_nsid(nvme_ctrl_t *ctrl, uint32_t dst_nsid)
{
    uint32_t src_nsid;
#if HAVE_SNAP
    const char *rdma_dev = snap_config_get_emu_manager(ctrl->sconfig);
#endif

    /* Find src_nsid candidate */
    if (snap_config_get_emu_mode(ctrl->sconfig) == SNAP_EMU_SQE_ONLY) {
        /*
         * There is no limitation on mapping namespaces on sqe_only mode,
         * so just mirror back the same value
         */
        src_nsid = dst_nsid;
    } else {
#if HAVE_SNAP
        src_nsid = mlnx_snap_acquire_namespace_id(rdma_dev);
        if (!src_nsid) {
            nvmx_error("Failed to acquire valid PCI namespace ID");
            goto err;
        }
#else
        src_nsid = dst_nsid;
#endif
    }

    /* Verify candidate follows limitations */
    if (src_nsid > ctrl->id_ctrl.nn) {
        nvmx_error("Failed to add namespace. desired NSID (%lu) exceeds max allowed (%lu)",
                   src_nsid, ctrl->id_ctrl.nn);
        goto release_nsid;
    }

    if (src_nsid != dst_nsid)
        nvmx_info("nsid %lu is mapped to %lu", dst_nsid, src_nsid);
    return src_nsid;

release_nsid:
#if HAVE_SNAP
    mlnx_snap_release_namespace_id(rdma_dev, src_nsid);
#endif
err:
    return 0; // invalid NSID
}

static void nvme_ctrl_delete_namespace(nvme_ctrl_t *ctrl, nvme_namespace_t *ns)
{
    if (snap_config_get_emu_mode(ctrl->sconfig) != SNAP_EMU_SQE_ONLY) {
        if (ns->nvme_emu_ns)
            nvme_emu_dev_remove_namespace(ns->nvme_emu_ns);
    }
    nvme_ctrl_pgs_suspend(ctrl);
    pthread_spin_lock(&ctrl->namespaces_lock);
    ctrl->num_namespaces--;
    TAILQ_REMOVE(&ctrl->namespaces, ns, entry);
    pthread_spin_unlock(&ctrl->namespaces_lock);
    nvme_ctrl_pgs_resume(ctrl);
    nvmx_debug("Namespace nsid %d was removed from NVMe namespace list", ns->id);
    nvme_ctrl_unmap_nsid(ctrl, ns->id);
    free(ns->sq_stat);
    free(ns);
}

static int nvme_ctrl_add_namespace(nvme_ctrl_t *ctrl, uint32_t dst_nsid,
                                   uint64_t size_b, uint8_t block_order,
                                   uint16_t backend_metadata,
                                   nvme_id_ns_descriptor_t id_descs[], void *priv_ns)
{
    nvme_namespace_t *ns;
    nvme_id_ns_t *id_ns;
    int ret;
    int i;
    uint32_t src_nsid;

    if (ctrl->num_namespaces >= ctrl->max_namespaces) {
        nvmx_error("Failed to add namespace. "
                   "Already reached namespaces capacity (%lu)",
                   ctrl->max_namespaces);
        return -ENOMEM;
    }

    ns = calloc(1, sizeof(nvme_namespace_t));
    if (!ns)
        return -ENOMEM;

    ns->sq_stat = calloc(ctrl->num_queues, sizeof(*ns->sq_stat));
    if (!ns->sq_stat) {
        ret = -ENOMEM;
        goto free_ns;
    }

    src_nsid = nvme_ctrl_map_nsid(ctrl, dst_nsid);
    if (src_nsid <= 0 || src_nsid >= UINT_MAX) {
        nvmx_info("Failed to map NSID %lu", dst_nsid);
        ret = -EINVAL;
        goto free_sq_stat;
    }

    ns->id = src_nsid;
    ns->remote_id = dst_nsid;
    ns->size_b = size_b;
    ns->block_order = block_order;
    ns->backend_metadata = backend_metadata;
    ns->backend_ns = priv_ns;

    id_ns = &ns->id_ns;

    for (i = 0; i < NVME_NS_ID_TYPES_NUM; ++i) {
        nvme_id_ns_descriptor_t *id_dst = &ns->id_descs[i];
        nvme_id_ns_descriptor_t *id_src = &id_descs[i];

        memcpy(id_dst, id_src, sizeof(*id_dst));

        switch(id_src->nidt) {
        case NVME_NS_IDENTIFIER_EUI64:
            memcpy(id_ns->eui64, id_src->nid.eui64.raw, sizeof(id_ns->eui64));
            break;
        case NVME_NS_IDENTIFIER_NGUID:
            memcpy(id_ns->nguid, id_src->nid.nguid.raw, sizeof(id_ns->nguid));
            break;
        default:
            break;
        }
    }

    id_ns->nsfeat = 0;
    id_ns->nlbaf = 0;
    id_ns->flbas = 0;
    id_ns->mc = 0;
    id_ns->dpc = 0;
    id_ns->dps = 0;
    id_ns->lbaf[0].ds = block_order;
    id_ns->ncap  = id_ns->nuse = id_ns->nsze =
        cpu_to_le64(size_b >>
            id_ns->lbaf[NVME_ID_NS_FLBAS_INDEX(ns->id_ns.flbas)].ds);

    if (nvme_emu_dev_is_started(ctrl->dev)) {
        if (snap_config_get_emu_mode(ctrl->sconfig) != SNAP_EMU_SQE_ONLY) {
            ns->nvme_emu_ns = nvme_emu_dev_add_namespace(ctrl->dev, ns->id,
                                       ns->remote_id, backend_metadata,
                                       block_order);
            if (!ns->nvme_emu_ns) {
                nvmx_error("Failed to add src_nsid %d", ns->id);
                ret = -EINVAL;
                goto unmap_nsid;
            }
        }
    }

    nvme_ctrl_pgs_suspend(ctrl);
    pthread_spin_lock(&ctrl->namespaces_lock);
    TAILQ_INSERT_TAIL(&ctrl->namespaces, ns, entry);
    ctrl->num_namespaces++;
    pthread_spin_unlock(&ctrl->namespaces_lock);
    nvme_ctrl_pgs_resume(ctrl);

    nvmx_debug("Namespace nsid %d was added to NVMe namespace list", ns->id);
    return 0;

unmap_nsid:
    nvme_ctrl_unmap_nsid(ctrl, ns->id);
free_sq_stat:
    free(ns->sq_stat);
free_ns:
    free(ns);
    return ret;
}

static int
nvme_ctrl_driver_detach_namespace(void *drv_ctrl, uint32_t nsid)
{
    nvme_namespace_t *ns;
    nvme_ctrl_t *ctrl = (nvme_ctrl_t *)drv_ctrl;

    ns = nvme_find_namespace(ctrl, nsid);
    if (ns) {
        nvme_ctrl_delete_namespace(ctrl, ns);
        nvme_ctrl_update_changed_namespace(ctrl, nsid);
    }

    return 0;
}

static bool nvme_ctrl_is_ns_descs_equal(nvme_namespace_t *ns,
                                        nvme_id_ns_descriptor_t descs[])
{
    int i, j;
    nvme_id_ns_descriptor_t *desc, *ns_desc;

    for (i = 0; i < NVME_NS_ID_TYPES_NUM; ++i) {
        ns_desc = &ns->id_descs[i];

        for (j = 0; j < NVME_NS_ID_TYPES_NUM; j++) {
            desc = &descs[j];
            if (desc->nidt != ns_desc->nidt)
                continue;

            if (!is_global_id_equal(ns_desc, desc->nidt, &desc->nid))
                return false;
        }
    }

    return true;
}

static int nvme_ctrl_driver_verify_namespace(nvme_namespace_t *ns,
                                             uint32_t nsid, uint64_t size_b,
                                             uint8_t block_order,
                                             uint16_t backend_metadata,
                                             nvme_id_ns_descriptor_t id_descs[])
{
    if (nsid == ns->id &&
        size_b >= ns->size_b &&
        block_order == ns->block_order &&
        backend_metadata == ns->backend_metadata &&
        nvme_ctrl_is_ns_descs_equal(ns, id_descs))
        return 0;

    nvmx_error("Existing nsid %d instance doesn't match new attributes", nsid);
    return -EINVAL;
}

static int
nvme_ctrl_driver_attach_namespace(void *drv_ctrl, uint32_t nsid, uint64_t size_b,
                                  uint8_t block_order, uint16_t backend_metadata,
                                  nvme_id_ns_descriptor_t id_descs[], void *priv_ns)
{
    nvme_namespace_t *ns;
    nvme_ctrl_t *ctrl = (nvme_ctrl_t *)drv_ctrl;
    int rc = 0;

    ns = nvme_find_namespace(ctrl, nsid);
    if (ns) {
        rc = nvme_ctrl_driver_verify_namespace(ns, nsid, size_b,
                                               block_order,
                                               backend_metadata, id_descs);
        if (!rc)
            goto out;

        nvme_ctrl_delete_namespace(ctrl, ns);
        nvme_ctrl_update_changed_namespace(ctrl, nsid);
    }

    rc = nvme_ctrl_add_namespace(ctrl, nsid, size_b, block_order,
                                 backend_metadata, id_descs, priv_ns);
    if (!rc)
        nvme_ctrl_update_changed_namespace(ctrl, nsid);

out:
    return rc;
}

/**
 * Select queue type/operations based on external configuration 
 * TODO: config file */
static nvme_queue_ops_t *nvme_ctrl_qparams(nvme_ctrl_t *ctrl,
                                           bool is_adm, int *q_type)
{
    const char *emu_mode;
    nvme_queue_ops_t *ops;

    if (is_adm) {
        *q_type = NVME_EMU_QUEUE_SQE;
        ops = &ctrl_sqe_ops;
    } else {
        *q_type = NVME_EMU_QUEUE_NVME2NVMF_REMOTE;
        ops = &ctrl_cc_remote_ops;
    }

    emu_mode = snap_config_get_emu_mode_str(ctrl->sconfig);

    if (strncmp(emu_mode, "sqe_only", 8) == 0) {
        *q_type = NVME_EMU_QUEUE_SQE;
        ops = &ctrl_sqe_ops;
        nvmx_debug("using sqe_only mode");
        /* TODO: pass proper qtype to the sdk */
    } else if (strncmp(emu_mode, "sqe_cc", 6) == 0) {
        if (is_adm) {
            *q_type = NVME_EMU_QUEUE_SQE;
            ops = &ctrl_sqe_ops;
        } else {
           *q_type = NVME_EMU_QUEUE_NVME2NVMF_LOCAL;
           ops = &ctrl_cc_ops;
        }
        nvmx_debug("using sqe_cc mode");
    } else if (strncmp(emu_mode, "cc_remote", 9) == 0) {
        if (is_adm) {
            *q_type = NVME_EMU_QUEUE_SQE;
            ops = &ctrl_sqe_ops;
        } else {
           *q_type = NVME_EMU_QUEUE_NVME2NVMF_REMOTE;
           ops = &ctrl_cc_remote_ops;
        }
        nvmx_debug("using sqe_cc_remote mode");
    } else {
        nvmx_warn("unknown emulation type %s reverting to 'sqe_only' mode", emu_mode);
    }

    return ops;
}

/* choose queue type */
static int nvme_ctrl_setup_queues(nvme_ctrl_t *ctrl)
{
    int q_type;
    nvme_queue_ops_t *ops;
    int err;

    ops = nvme_ctrl_qparams(ctrl, true, &q_type);
    err = nvme_emu_queue_type_add(ctrl->dev, q_type, true, ops);
    if (err != NVMX_OK)
        goto err;

    ops = nvme_ctrl_qparams(ctrl, false, &q_type);
    err = nvme_emu_queue_type_add(ctrl->dev, q_type, false, ops);
    if (err != NVMX_OK)
        goto err_del_aq;

    ctrl->sq = calloc(ctrl->num_queues, sizeof(nvme_squeue_t *));
    if (!ctrl->sq) {
        nvmx_error("Failed to allocate ctrl sqs");
        err = -ENOMEM;
        goto err_del_ioq;
    }

    ctrl->cq = calloc(ctrl->num_queues, sizeof(nvme_cqueue_t *));
    if (!ctrl->cq) {
        nvmx_error("Failed to allocate ctrl cqs");
        err = -ENOMEM;
        goto err_free_sqs;
    }

    ctrl->sq_stat = calloc(ctrl->num_queues, sizeof(*ctrl->sq_stat));
    if (!ctrl->sq_stat) {
        nvmx_error("Failed to allocate ctrl sq_stat");
        err = -ENOMEM;
        goto err_free_cq;
    }

    return NVMX_OK;

err_free_cq:
    free(ctrl->cq);
err_free_sqs:
    free(ctrl->sq);
err_del_ioq:
    nvme_ctrl_qparams(ctrl, false, &q_type);
    nvme_emu_queue_type_del(ctrl->dev, q_type, false);
err_del_aq:
    nvme_ctrl_qparams(ctrl, true, &q_type);
    nvme_emu_queue_type_del(ctrl->dev, q_type, true);
err:
   return err;
}

static void nvme_ctrl_remove_queues(nvme_ctrl_t *ctrl)
{
    int q_type;

    free(ctrl->cq);
    free(ctrl->sq);
    free(ctrl->sq_stat);

    nvme_ctrl_qparams(ctrl, false, &q_type);
    nvme_emu_queue_type_del(ctrl->dev, q_type, false);

    nvme_ctrl_qparams(ctrl, true, &q_type);
    nvme_emu_queue_type_del(ctrl->dev, q_type, true);
}

int nvme_ctrl_progress_all_io(nvme_ctrl_t *ctrl)
{
    int i;
    int n;

    for (n = 0, i = 1; i < ctrl->num_queues; i++) {
        if (!ctrl->sq[i])
            continue;
        ctrl->sq[i]->thread_id = -1;
        n += nvme_emu_progress_sq(&ctrl->sq[i]->ops);
        nvme_ctrl_progress_pending_cqes(ctrl->sq[i]);
    }

    return n;
}

static void nvme_ctrl_progress_suspend(nvme_ctrl_t *ctrl)
{
    int i;

    for (i = 0; i < ctrl->num_queues; i++)
        if (ctrl->sq[i])
            return;

    nvmx_info("controller %d is moving to the SUSPENDED state",
              snap_config_get_cntlid(ctrl->sconfig));
    ctrl->state = NVME_CTRL_SUSPENDED;

    if (ctrl->should_stop)
        nvme_ctrl_stop(ctrl);
}

int nvme_ctrl_progress_admin(nvme_ctrl_t *ctrl)
{
    int events = 0;

    if (ctrl->sq[0]) {
        ctrl->sq[0]->thread_id = -1;
        events = nvme_emu_progress_sq(&ctrl->sq[0]->ops);
    }

    return events > 0;
}

void nvme_ctrl_progress_mmio(nvme_ctrl_t *ctrl)
{
    int new_num_vfs, old_num_vfs, pf_id, i;
    time_t cur_time;

    for (i = 0; i < ctrl->num_queues; i++)
        if (ctrl->sq[i] && ctrl->sq[i]->state == NVME_SQ_FLUSHING)
            nvme_progress_delete_sq(ctrl->sq[i], ctrl);

    if (ctrl->state == NVME_CTRL_SUSPENDING) {
        nvme_ctrl_progress_suspend(ctrl);
        return;
    }

    old_num_vfs = ctrl->dev->dev_caps.max_emulated_vfs;
    nvme_emu_progress_mmio(ctrl->dev);
    new_num_vfs = ctrl->dev->dev_caps.max_emulated_vfs;

    if (old_num_vfs != new_num_vfs) {
        nvmx_info("VF change detected: old %d new %d", old_num_vfs, new_num_vfs);
        if (ctrl->vf_change_cb) {
            pf_id = snap_config_get_pci_func(ctrl->sconfig);
            ctrl->vf_change_cb(ctrl->vf_change_ctx, pf_id, new_num_vfs);
        }
    }

    // Query FW for hotplug state to know when it is safe to unplug
    if (ctrl->dev->pci_hotplug_state_cur == MLX5_EMULATION_HOTPLUG_STATE_POWER_OFF &&
        ctrl->dev->pci_hotplug_state_prev != MLX5_EMULATION_HOTPLUG_STATE_POWER_OFF) {
        ctrl->hotunplug_device_cb(ctrl->hotunplug_device_ctx);
    } else if (ctrl->dev->pci_hotplug_state_cur == MLX5_EMULATION_HOTPLUG_STATE_HOTUNPLUG_PREPARE) {
        if (time(&cur_time) == (time_t)(-1)) {
            nvmx_error("Failed to get system time");
            return;
        }

        if (ctrl->hotunplug_timer_expire > 0 && ctrl->hotunplug_timer_expire <= cur_time) {
            ctrl->hotunplug_timer_expire = 0;
            if (ctrl->force_hotunplug_device)
                ctrl->hotunplug_device_cb(ctrl->hotunplug_device_ctx);
            else if (ctrl->state == NVME_CTRL_STARTED)
                ctrl->hotunplug_timeout_cb(ctrl->hotunplug_device_ctx);
            else
                ctrl->hotunplug_device_cb(ctrl->hotunplug_device_ctx);
        }
    }
}

#ifdef HAVE_SNAP

static void *nvme_ctrl_rpc_server_completion_thread(void *arg);

static int nvme_ctrl_rpc_server_set(nvme_ctrl_t *ctrl)
{
    int err = 0;
    const char *rpc_server;

    rpc_server = snap_config_get_rpc_server(ctrl->sconfig);
    if (!rpc_server || !strncmp(rpc_server, SNAP_CONFIG_RPC_SERVER_NONE,
                                sizeof(SNAP_CONFIG_RPC_SERVER_NONE))) {
        nvmx_info("No rpc_server was set. Host JSON-RPC commands "
                  "are not supported");
        err = -EINVAL;
        goto err;
    }
    ctrl->rpc_server = rpc_server;

    ctrl->rpc_client = snap_json_rpc_client_open(rpc_server);
    if (!ctrl->rpc_client) {
        nvmx_info("failed to create rpc client to %s", rpc_server);
        err = -EINVAL;
        goto clear_rpc_server;
    }

    err = sem_init(&ctrl->rpc_thread_kick, 0, 0);
    if (err)
        goto client_close;

    err = sem_init(&ctrl->rpc_thread_idle, 0, 0);
    if (err)
        goto sem1_destroy;

    err = pthread_create(&ctrl->rpc_thread, NULL,
                         nvme_ctrl_rpc_server_completion_thread, ctrl);
    if (err)
        goto sem2_destroy;

    ctrl->rpc_thread_req = NULL;
    return 0;

sem2_destroy:
    sem_destroy(&ctrl->rpc_thread_idle);
sem1_destroy:
    sem_destroy(&ctrl->rpc_thread_kick);
client_close:
    snap_json_rpc_client_close(ctrl->rpc_client);
clear_rpc_server:
    ctrl->rpc_server = NULL;
err:
    return err;
}

static void nvme_ctrl_rpc_server_unset(nvme_ctrl_t *ctrl)
{
    if (!ctrl->rpc_server)
        return;

    if (ctrl->rpc_rsp)
        snap_json_rpc_put_response(ctrl->rpc_rsp);
    pthread_cancel(ctrl->rpc_thread);
    pthread_join(ctrl->rpc_thread, NULL);
    sem_destroy(&ctrl->rpc_thread_idle);
    sem_destroy(&ctrl->rpc_thread_kick);
    if (ctrl->rpc_client)
        snap_json_rpc_client_close(ctrl->rpc_client);
    ctrl->rpc_server = NULL;
}

#define NVME_DEFAULT_CPU_MASK           0xF0

void nvme_set_thread_attr(void)
{
    int i;
    int ret;
    int cpu_mask;
    int cpu_count;
    cpu_set_t cpuset;
    pthread_t thread;

    thread = pthread_self();
    (void) pthread_setname_np(thread, "snap-rpc-client");

    /*
     * "CPU_MASK" is a ENV variable to set which CPU cores used
     *  by SPDK reactor, and in here we are trying to set
     *  snap-rpc-client thread cpu affinity to avoid it running
     *  on those CPU cores which pointed by "CPU_MASK".
     * Why need do this? It is because snap-rpc-client thread will
     *  handle 'storage_admin' RPC request, and inside 'storage_admin'
     *  RPC method will use fork()/exec()/wait(), and if snap-rpc-client
     *  thread running on one of "CPU_MASK" CPU cores then it will
     *  have risk to frozen that SPDK reactor.
     */
    if (getenv("CPU_MASK"))
        cpu_mask = strtol(getenv("CPU_MASK"), NULL, 0);
    else
        cpu_mask = NVME_DEFAULT_CPU_MASK;

    cpu_count = get_nprocs();

    CPU_ZERO(&cpuset);
    for (i = 0; i < cpu_count; i++) {
        if (!(1 << i & cpu_mask))
            CPU_SET(i, &cpuset);
    }

    if (!CPU_COUNT(&cpuset)) {
        /* CPU_MASK use all the CPU cores, then set
           snap-rpc-client to running on the last CPU core */
        CPU_SET(cpu_count - 1, &cpuset);
    }

    ret = pthread_setaffinity_np(thread, sizeof(cpuset), &cpuset);
    if (ret)
        nvmx_warn("Set RPC Client Completion Thread CPU Affinity Failed");
}

static inline void
nvme_ctrl_handle_rpc_pending_list(nvme_ctrl_t *ctrl)
{
    nvme_request_t *req;

    req = TAILQ_FIRST(&ctrl->rpc_pending_list);
    if (!req) {
        ctrl->rpc_busy = false;
        return;
    }

    TAILQ_REMOVE(&ctrl->rpc_pending_list, req, rpc_entry);

    ctrl->rpc_thread_req = req;
    sem_post(&ctrl->rpc_thread_kick);
}

static inline void
nvme_ctrl_json_rpc_req_completion(nvme_request_t *req, nvme_ctrl_t *ctrl,
                                  enum nvme_emu_status status, uint32_t result)
{
    /*
     * Hint driver about required response size by setting rsp->length
     * in cqe.result field (in dwords).
     */
    nvmx_debug("%lu dwords are ready to be read in JSON-RPC response",
               result);
    req->cqe.result = result;
    nvme_io_req_completion_lock(req, ctrl, status);
    nvme_ctrl_handle_rpc_pending_list(ctrl);
}

static void *nvme_ctrl_rpc_server_completion_thread(void *arg)
{
    int ret;
    nvme_ctrl_t *ctrl = arg;
    nvme_request_t *req;
    struct nvme_snap_method *rpc_method;

    nvme_set_thread_attr();

    while (true) {
        sem_post(&ctrl->rpc_thread_idle);
        sem_wait(&ctrl->rpc_thread_kick);

        req = ctrl->rpc_thread_req;

        /* send command to rpc server */
        ret = snap_json_rpc_client_send_req(ctrl->rpc_client,
                                        req->rpc_buf,
                                        req->io_req.async_req.len);

        free(req->rpc_buf);
        req->rpc_buf = NULL;

        if (ret) {
            nvme_ctrl_json_rpc_req_completion(req, ctrl, NVMX_ERR, 0);
            continue;
        }

        if (likely(ctrl->rpc_method_is_supported_cb)) {
            rpc_method = ctrl->rpc_method_is_supported_cb(ctrl);
            if (rpc_method) {
                if (likely(ctrl->rpc_method_handle_cb)) {
                    ret = ctrl->rpc_method_handle_cb(rpc_method, ctrl);
                } else {
                    nvmx_warn("rpc_method_is_supported_cb set,\
                               but rpc_method_handle_cb not set!");
                    ret = NVMX_ERR; /* treat it as Error */
                }
                /* Once returned 'rpc_method' is not NULL, that meaning
                 *  it is a storage_admin or network_admin RPC request,
                 *  either handle it or drop it, but cannot send to RPC Server.
                 */
                nvme_ctrl_json_rpc_req_completion(req, ctrl, ret,
                                             (ctrl->rpc_rsp->length + 3) >> 2);
                continue;
            }
        }

        /* Send RPC request to RPC Server handle, if:
         *   1. ctrl->rpc_method_is_supported_cb is NULL
         *   2. ctrl->rpc_method_is_supported_cb() return NULL
         */
        ret = snap_json_rpc_wait_for_response(ctrl->rpc_client);
        if (ret) {
            nvmx_error("json rpc failed to wait for response %d", ret);
            /* in case the pipe got broken, reconnect to the rpc server */
            if (ret == -EPIPE) {
                snap_json_rpc_client_close(ctrl->rpc_client);
                /* best effort to re-establish connection to rpc server */
                ctrl->rpc_client = snap_json_rpc_client_open(ctrl->rpc_server);
                if (unlikely(!ctrl->rpc_client)) {
                    nvme_ctrl_rpc_server_unset(ctrl);
                    nvmx_error("Failed to reopen RPC client. Thread Exit!!!");
                    break;
                }
            }
            nvme_ctrl_json_rpc_req_completion(req, ctrl, NVMX_ERR, 0);
            continue;
        }

        /* We support only 1 in-flight req/rsp pair. Override the previous rsp. */
        if (ctrl->rpc_rsp)
            snap_json_rpc_put_response(ctrl->rpc_rsp);
        ctrl->rpc_rsp = snap_json_rpc_get_response(ctrl->rpc_client);
        if (!ctrl->rpc_rsp) {
            nvmx_error("json rpc failed to get response");
            nvme_ctrl_json_rpc_req_completion(req, ctrl, NVMX_ERR, 0);
            continue;
        }

        nvme_ctrl_json_rpc_req_completion(req, ctrl, ret,
                                          (ctrl->rpc_rsp->length + 3) >> 2);
    }

    return NULL;
}
#endif

nvme_ctrl_t *nvme_ctrl_init(nvme_ctrl_init_param_t *param)
{
    int pf_id;
    nvme_ctrl_t *ctrl;
    int err = 0;

    if (!param)
        goto out;

    ctrl = calloc(1, sizeof(*ctrl));
    if (!ctrl)
        goto out;

    ctrl->sconfig = param->sconfig;
    ctrl->vf_change_cb = param->vf_change_cb;
    ctrl->vf_change_ctx = param->vf_change_ctx;
    ctrl->rpc_method_is_supported_cb = param->rpc_method_is_supported_cb;
    ctrl->rpc_method_handle_cb = param->rpc_method_handle_cb;
    ctrl->hotunplug_device_cb = param->hotunplug_device_cb;
    ctrl->hotunplug_device_ctx = param->hotunplug_device_ctx;
    ctrl->hotunplug_timeout_cb = param->hotunplug_timeout_cb;

    if (!strcmp(ctrl->sconfig->mem, "pool")) {
        ctrl->memp = param->memp;
        if (!ctrl->memp)
            nvmx_warn("Mempool mode requested but memory pool is not provided");
    }

    err = pthread_mutex_init(&ctrl->async_event_lock, NULL);
    if (err)
        goto free_ctrl;

#if HAVE_SNAP
    /*
     * Best effort to open RPC server for supporting JSON RPC
     * commands from host. Upon failure, we can still continue
     * without that feature (ctrl->rpc_server is NULL)
     */
    nvme_ctrl_rpc_server_set(ctrl);
    TAILQ_INIT(&ctrl->rpc_pending_list);
#endif

    ctrl->dev = nvme_emu_dev_open(snap_config_get_emu_manager(ctrl->sconfig),
                                  snap_config_get_rdma_dev(ctrl->sconfig),
                                  ctrl->sconfig, ctrl->memp, nvme_ctrl_bar_write_event,
                                  ctrl, &ctrl->num_queues);
    if (!ctrl->dev)
        goto destroy_async_mutex;

    if (snap_config_get_backends_cnt(ctrl->sconfig) < 1) {
        nvmx_error("Cannot find backends in the config file");
        goto dev_close;
    }

    ctrl->quirks = snap_config_get_quirks(ctrl->sconfig);

    TAILQ_INIT(&ctrl->namespaces);

    err = pthread_spin_init(&ctrl->namespaces_lock, 0);
    if (err)
        goto dev_close;

    pf_id = snap_config_get_pci_func(ctrl->sconfig);

    if (nvme_emu_is_pf(ctrl->dev)) {
        ctrl->fw_update_enable = snap_env_getenv(NVME_FW_SUPP) != 0;
        if (ctrl->fw_update_enable) {
            if (!getenv(NVME_FW_UPDATE_PERSISTENT_LOCATION)) {
                ctrl->fw_update_enable = false;
                nvmx_warn("Firmware update feature is enabled but the NVME_FW_UPDATE_PERSISTENT_LOCATION variable is not provided");
            }
        }
        if (snap_config_get_iova_mgmt(ctrl->sconfig)) {
            if (effdma_domain_find_or_create(pf_id, -1)) {
                nvmx_debug("IOVA_MGMT enabled");
                ctrl->iova_mgmt_enable = true;
            }
            else {
                nvmx_warn("IOVA_MGMT requested but initialization failed");
                ctrl->iova_mgmt_enable = false;
            }
        }
    }
    else {
        if (effdma_domain_find(pf_id, -1)) {
            if (!effdma_domain_find_or_create(pf_id,
                    snap_config_get_pci_vfunc(ctrl->sconfig))) {
                nvmx_warn("Failed to create effdma domain, pf_id: %d, vf_id: %d",
                        pf_id, snap_config_get_pci_vfunc(ctrl->sconfig));
            }
        }
    }

    nvme_ctrl_id_init(ctrl);

    err = nvme_ctrl_setup_queues(ctrl);
    if (err != NVMX_OK)
        goto ctrl_id_uninit;

    ctrl->driver = nvme_driver_create(ctrl->dev, ctrl->sconfig,
                                      ctrl->num_queues,
                                      &nvme_ctrl_driver_attach_namespace,
                                      &nvme_ctrl_driver_detach_namespace,
                                      0, ctrl->id_ctrl.ver.raw, ctrl->id_ctrl.mdts);
    if (!ctrl->driver)
        goto io_q_type_del;

    if (param->nthreads) {
        if (nvme_ctrl_pgs_alloc(ctrl, param->nthreads) != NVMX_OK) {
            nvmx_error("allocate poll groups failed");
            goto io_q_type_del;
        }
    }

    ctrl->fw_file = -1;
    ctrl->over_temperature_threshold = 0xFFFF;
    ctrl->under_temperature_threshold = 0;
    ctrl->temperature = 300;
    ctrl->async_event_mask = 0;

    return ctrl;

io_q_type_del:
    nvme_ctrl_remove_queues(ctrl);
ctrl_id_uninit:
    nvme_ctrl_id_uninit(ctrl);
    pthread_spin_destroy(&ctrl->namespaces_lock);
dev_close:
    nvme_emu_dev_close(ctrl->dev);
destroy_async_mutex:
#if HAVE_SNAP
    nvme_ctrl_rpc_server_unset(ctrl);
#endif
    pthread_mutex_destroy(&ctrl->async_event_lock);
free_ctrl:
    free(ctrl);
out:
    return NULL;
}

void nvme_ctrl_destroy(nvme_ctrl_t *ctrl)
{
    nvme_free_bfb_info(ctrl);
    nvme_ctrl_pgs_free(ctrl);
    nvme_driver_stop(ctrl->driver);
    nvme_driver_destroy(ctrl->driver);
    nvme_ctrl_remove_queues(ctrl);
    pthread_spin_destroy(&ctrl->namespaces_lock);
    pthread_mutex_destroy(&ctrl->async_event_lock);
    nvme_ctrl_id_uninit(ctrl);
    nvme_emu_dev_close(ctrl->dev);
    nvme_discard_image(ctrl);
#if HAVE_SNAP
    nvme_ctrl_rpc_server_unset(ctrl);
#endif
    free(ctrl);
}

static void nvme_get_default_moderation(nvme_ctrl_t *ctrl, int *cq_period,
                                        int *cq_max_count)
{
    *cq_max_count = snap_config_get_cq_max_count(ctrl->sconfig);
    *cq_period = snap_config_get_cq_period(ctrl->sconfig);
}

static void nvme_ctrl_ftrs_init(nvme_ctrl_t *ctrl)
{
    ctrl->arbitration        = NVME_DEFAULT_ARBITRATION;
    ctrl->power_management   = NVME_DEFAULT_PM;

    pthread_mutex_lock(&ctrl->async_event_lock);
    /* Compliant driver is supposed to get a bitmask of supported
     * optional async events via id_ctrl.oaes. Then use SET_FTRS to
     * to select events it wants to receive.
     */
    if (ctrl->quirks & NVME_QUIRK_ASYNC_EV_CONFIG)
        /* some drivers ex Centos 7.5 know how to deal with namespace
         * change events but they don't use SET_FTRS to enable it
         */
        ctrl->async_event_config = NVME_SUPPORTED_ASYNC_EVENTS;
    else
        ctrl->async_event_config = 0;

    ctrl->num_aers = 0;
    TAILQ_INIT(&ctrl->async_event_reqs);
    pthread_mutex_unlock(&ctrl->async_event_lock);

    nvme_get_default_moderation(ctrl, &ctrl->cq_period, &ctrl->cq_max_count);
}

static void nvme_ctrl_reset(nvme_ctrl_t *ctrl)
{
    nvme_ctrl_ftrs_init(ctrl);
}

static void nvme_ctrl_start_namespaces(nvme_ctrl_t *ctrl)
{
    nvme_namespace_t *ns, *next;

    nvmx_assertv_always(nvme_emu_dev_is_started(ctrl->dev), "dev isn't ready");

    if (snap_config_get_emu_mode(ctrl->sconfig) != SNAP_EMU_SQE_ONLY) {
        nvme_ctrl_pgs_suspend(ctrl);
        pthread_spin_lock(&ctrl->namespaces_lock);
        TAILQ_FOREACH_SAFE(ns, &ctrl->namespaces, entry, next) {
            if (ns->nvme_emu_ns)
                continue;

            ns->nvme_emu_ns = nvme_emu_dev_add_namespace(ctrl->dev, ns->id,
                                       ns->remote_id, ns->backend_metadata,
                                       ns->block_order);
            if (!ns->nvme_emu_ns) {
                nvmx_error("Failed to create namespace %d object. "
                           "Removing it from namespace list", ns->id);
                ctrl->num_namespaces--;
                TAILQ_REMOVE(&ctrl->namespaces, ns, entry);
                nvme_ctrl_unmap_nsid(ctrl, ns->id);
                free(ns->sq_stat);
                free(ns);
            }
        }
        pthread_spin_unlock(&ctrl->namespaces_lock);
        nvme_ctrl_pgs_resume(ctrl);
    }
}

static void nvme_ctrl_clear_namespaces(nvme_ctrl_t *ctrl)
{
    while (!TAILQ_EMPTY(&ctrl->namespaces))
        nvme_ctrl_delete_namespace(ctrl, TAILQ_FIRST(&ctrl->namespaces));
}

static int nvme_ctrl_start(nvme_ctrl_t *ctrl)
{
    nvme_emulator_t *dev = ctrl->dev;
    nvme_bar_t *bar = &dev->bar.curr;
    uint32_t page_bits;
    uint32_t page_size;
    int ret = NVMX_ERR, err;

    err = nvme_driver_start(ctrl->driver);
    if (err)
        goto out_err;

    err = nvme_emu_dev_start(ctrl->dev);
    if (err)
        goto driver_stop;

    nvme_ctrl_start_namespaces(ctrl);

    page_bits = nvme_reg_cc_mps(bar->cc) + 12;
    page_size = 1 << page_bits;

    if (unlikely(ctrl->cq[0]))
        goto dev_stop;

    if (unlikely(ctrl->sq[0]))
        goto dev_stop;

    if (unlikely(!bar->asq))
        goto dev_stop;

    if (unlikely(!bar->acq))
        goto dev_stop;

    if (unlikely(bar->asq & (page_size - 1)))
        goto dev_stop;

    if (unlikely(bar->acq & (page_size - 1)))
        goto dev_stop;

    if (unlikely(nvme_reg_cc_mps(bar->cc) <
                 nvme_reg_cap_mpsmin(bar->cap)))
        goto dev_stop;

    if (unlikely(nvme_reg_cc_mps(bar->cc) >
                 nvme_reg_cap_mpsmax(bar->cap)))
        goto dev_stop;

    if (unlikely(nvme_reg_cc_iocqes(bar->cc) < ctrl->id_ctrl.cqes.min))
        goto dev_stop;

    if (unlikely(nvme_reg_cc_iocqes(bar->cc) > ctrl->id_ctrl.cqes.max))
        goto dev_stop;

    if (unlikely(nvme_reg_cc_iosqes(bar->cc) < ctrl->id_ctrl.sqes.min))
        goto dev_stop;

    if (unlikely(nvme_reg_cc_iosqes(bar->cc) > ctrl->id_ctrl.sqes.max))
        goto dev_stop;

    if (unlikely(!nvme_reg_aqa_asqs(bar->aqa)))
        goto dev_stop;

    if (unlikely(!nvme_reg_aqa_acqs(bar->aqa)))
        goto dev_stop;

    ctrl->page_bits = page_bits;
    ctrl->page_size = page_size;
    ctrl->page_size_min = 1 << (12 + nvme_reg_cap_mpsmin(bar->cap));
    ctrl->max_prp_ents = ctrl->page_size / sizeof(uint64_t);
    ctrl->cqe_size = 1 << nvme_reg_cc_iocqes(bar->cc);
    ctrl->sqe_size = 1 << nvme_reg_cc_iosqes(bar->cc);
    ret = nvme_cq_init(&ctrl->admin_cq, ctrl, bar->acq, 0, 0,
                       nvme_reg_aqa_acqs(bar->aqa) + 1, 1);
    if (ret != NVMX_OK)
        goto dev_stop;

    nvme_ctrl_reset(ctrl);

    ret = nvme_sq_init(&ctrl->admin_sq, ctrl, bar->asq, 0, 0,
                       nvme_reg_aqa_asqs(bar->aqa) + 1);
    if (ret != NVMX_OK)
        goto out_free_cq;

    ctrl->allow_set_num_qs = true;
    ctrl->should_stop = false;
    ctrl->state = NVME_CTRL_STARTED;

    nvmx_info("*** controller enabled page_size=%d cqes=%d sqes=%d ***", 
              ctrl->page_size, ctrl->cqe_size, ctrl->sqe_size);
    return NVMX_OK;

out_free_cq:
    nvme_cq_free(&ctrl->admin_cq, ctrl);
dev_stop:
    nvme_ctrl_clear_namespaces(ctrl);
    nvme_emu_dev_stop(ctrl->dev);
driver_stop:
    nvme_driver_stop(ctrl->driver);
out_err:
    return ret;
}

static void nvme_sq_suspend(nvme_squeue_t *sq)
{
    struct nvme_ctrl_pg *pg = sq->pg;

    nvmx_info("SQ %d is moving to the FLUSHING state", sq->sqid);
    if (!pg || sq->sqid == 0) {
        sq->state = NVME_SQ_FLUSHING;
    } else {
        pthread_spin_lock(&pg->lock);
        sq->state = NVME_SQ_FLUSHING;
        pthread_spin_unlock(&pg->lock);
    }
}

static void nvme_progress_delete_sq(nvme_squeue_t *sq, nvme_ctrl_t *ctrl)
{
    int i, req_size;
    nvme_request_t *req, *next;
    nvme_cqueue_t *cq = NULL;

    nvme_ctrl_pgs_suspend(ctrl);
    if (!TAILQ_EMPTY(&sq->out_req_list)) {
        nvme_ctrl_pgs_resume(ctrl);
        return;
    }

    if (nvme_check_cqid_valid(ctrl, sq->cqid)) {
        cq = ctrl->cq[sq->cqid];

        TAILQ_FOREACH_SAFE(req, &cq->req_list, entry, next) {
            if (req->sq == sq) {
                nvme_ctrl_pgs_resume(ctrl);
                return;
            }
        }
    }
    nvme_ctrl_pgs_resume(ctrl);

    nvme_ctrl_sq_desched(sq);

    if (sq == &ctrl->admin_sq)
        nvme_ctrl_abort_async_event_requests(ctrl);

    if (cq)
        TAILQ_REMOVE(&cq->sq_list, sq, entry);
    sq->state = NVME_SQ_SUSPENDED;

    req_size = nvme_driver_request_size(ctrl->driver) + sizeof(nvme_request_t) -
               sizeof(nvme_io_driver_req_t);

    req = (nvme_request_t *)(sq->reqs_base);
    nvme_driver_sq_reset(&req->io_req);
    sq->io_buf = NULL;
    sq->mr  = NULL;

    for (i = 0; i < sq->size; i++) {
        req = (nvme_request_t *)(sq->reqs_base + i * req_size);
        nvme_driver_request_reset(&req->io_req);
    }

    nvme_driver_delete_sq(ctrl->driver, &sq->ops, sq->sqid);

    ctrl->sq[sq->sqid] = NULL;
    free(sq->reqs_base);

    if (sq == &ctrl->admin_sq)
        return;

    if (sq->delete_req) {
        nvmx_debug("complete admin delete sq %d command", sq->sqid);
        sq->delete_req->status = NVMX_OK;
        nvme_enqueue_req_completion(&ctrl->admin_cq, sq->delete_req);
    }

    nvmx_assertv_always(sq->sqid != 0, "Invalid sqid");
    free(sq);
}

static void nvme_cq_free(nvme_cqueue_t *cq, nvme_ctrl_t *ctrl)
{
    //nvme_emu_db_off(ctrl->dev, cq->cqid, 1);

    nvme_driver_delete_cq(ctrl->driver, &cq->ops, cq->cqid);

    ctrl->cq[cq->cqid] = NULL;

    if (cq == &ctrl->admin_cq)
        return;

    nvmx_assertv_always(cq->cqid != 0, "Invalid cqid");
    free(cq);
}

int nvme_ctrl_stop(nvme_ctrl_t *ctrl)
{
    int i, ret;

    if (ctrl->state == NVME_CTRL_STOPPED) {
        nvmx_debug("controller %d is already at STOPPED state",
                  snap_config_get_cntlid(ctrl->sconfig));
        return NVMX_OK;
    }

    if (ctrl->state != NVME_CTRL_SUSPENDED) {
        nvmx_error("controller %d must be at SUSPENDED state",
                  snap_config_get_cntlid(ctrl->sconfig));
        return NVMX_ERR;
    }

    for (i = 0; i < ctrl->num_queues; i++) {
        if (ctrl->cq[i])
            nvme_cq_free(ctrl->cq[i], ctrl);
    }

    nvme_ctrl_clear_namespaces(ctrl);

    memset(ctrl->sq_stat, 0, sizeof(*ctrl->sq_stat) * ctrl->num_queues);

    ret = nvme_emu_dev_stop(ctrl->dev);
    if (ret == NVMX_ERR)
        nvmx_warn("Failed to stop controller gracefully");

    nvme_driver_stop(ctrl->driver);
    nvme_discard_image(ctrl);

    nvmx_info("controller %d is moving to the STOPPED state",
              snap_config_get_cntlid(ctrl->sconfig));
    ctrl->state = NVME_CTRL_STOPPED;

    if (ctrl->should_stop) {
        nvme_emulator_t *dev = ctrl->dev;
        nvme_bar_t *bar = &dev->bar.curr;

        bar->csts = 0;
        nvme_emu_mmio_write(dev, &bar->csts, SNAP_NVME_REG_CSTS, 4);
        ctrl->should_stop = false;
    }

    return ret;
}

void nvme_ctrl_suspend(nvme_ctrl_t *ctrl)
{
    int i;

    if (ctrl->state == NVME_CTRL_SUSPENDING)
        return;

    if (ctrl->state != NVME_CTRL_STARTED) {
        nvmx_debug("controller %d is not at STARTED state",
                  snap_config_get_cntlid(ctrl->sconfig));
        return;
    }

    nvmx_info("controller %d is moving to the SUSPENDING state",
              snap_config_get_cntlid(ctrl->sconfig));
    ctrl->state = NVME_CTRL_SUSPENDING;

    pthread_mutex_lock(&ctrl->async_event_lock);
    ctrl->async_event_config = 0;
    ctrl->masked_logpage_ids = 0;
    pthread_mutex_unlock(&ctrl->async_event_lock);

    for (i = 0; i < ctrl->num_queues; i++) {
        if (ctrl->sq[i])
            nvme_sq_suspend(ctrl->sq[i]);
    }
}

bool nvme_ctrl_is_suspended(nvme_ctrl_t *ctrl)
{
    return ctrl->state == NVME_CTRL_SUSPENDED ||
           ctrl->state == NVME_CTRL_STOPPED;
}

/* change controller configuration */
static void nvme_ctrl_modify_config(nvme_ctrl_t *ctrl, uint32_t cur_cfg, uint32_t prev_cfg)
{
    nvme_emulator_t *dev = ctrl->dev;
    nvme_bar_t *bar = &dev->bar.curr;
    int ret;

    /* TODO: need to check this one more carefully */
    /* Windows first sends data, then sends enable bit */
    if (!nvme_reg_cc_en(cur_cfg) && 
        !nvme_reg_cc_en(prev_cfg) &&
        !nvme_reg_cc_shn(cur_cfg) &&
        !nvme_reg_cc_shn(prev_cfg)) {
        return;
    }

    /* enable command:
     * NOTE that SPEC 3.1.5 allows modifying both CC.SHN and CC.EN
     * however it does not clarify what is expected behaviour
     *
     * SPEC 3.1.5 CC.EN:
     * Setting this field from a '0' to a '1' when CSTS.RDY is a '1' or
     * setting this field from a '1' to a '0' when CSTS.RDY is a '0'
     * has undefined results.
     *
     * We don't see 0 -> 0 or 1 -> 1 transitions because we only pickup
     * bar changes.
     *
     * We must always set or clear RDY bit even if we failed to start or
     * stop the controller. If we don't following can happen:
     * - snap is not running
     * - modprobe nvme, modprobe timeouts, bus master disabled. EN=1, RDY=0
     * - start snap. start fails because bus master disabled. EN=1, RDY=0
     * - modprobe nvme fails because
     *      - driver does EN=1 -> EN:0 and immediately sees RDY=0, writes EN=1
     *        we miss EN=1 -> EN:0 because we are in the polling mode
     *      - driver is spec compliant and just does EN=1, which we ignore
     */
    if (nvme_reg_cc_en(cur_cfg) && !nvme_reg_cc_en(prev_cfg)) {
        /* controller start:
         * If we fail assume it is because driver tried enable, got timeout
         * and disabled bus master. Act as if we are ready but also raise CFS.
         *
         * Next time driver will try to reset either because EN == 1 && RDY == 1
         * or because CFS == 1 or because there is an admin command timeout.
         */
        ret = nvme_ctrl_start(ctrl);
        bar->csts = NVME_CSTS_READY;
        if (ret != NVMX_OK)
            bar->csts |= NVME_CSTS_FAILED;

        nvme_emu_mmio_write(dev, &bar->csts, SNAP_NVME_REG_CSTS, 4);
    } else if (!nvme_reg_cc_en(cur_cfg) && nvme_reg_cc_en(prev_cfg)) {

        nvme_ctrl_reset(ctrl);
        /* controller stop:
         * SPEC 7.3.2 "Controller Level Reset" says that
         * All other controller registers defined in section 3 and internal
         * controller state are reset.
         * Reset CSTS
         *
         * SPEC does not say that we can raise CFS if reset fails. So we don't
         * raise it.
         */
        if (ctrl->state == NVME_CTRL_STOPPED) {
            bar->csts = 0;
            nvme_emu_mmio_write(dev, &bar->csts, SNAP_NVME_REG_CSTS, 4);
        } else {
            ctrl->should_stop = true;
            nvme_ctrl_suspend(ctrl);
        }
    }

    /* shutdown command */
    if (nvme_reg_cc_shn(cur_cfg) && !nvme_reg_cc_shn(prev_cfg)) {
        bar->csts |= NVME_CSTS_SHST_COMPLETE;
        nvme_emu_mmio_write(dev, &bar->csts, SNAP_NVME_REG_CSTS, 4);
    } else if (!nvme_reg_cc_shn(cur_cfg) && nvme_reg_cc_shn(prev_cfg)) {
        bar->csts &= ~NVME_CSTS_SHST_COMPLETE;
        nvme_emu_mmio_write(dev, &bar->csts, SNAP_NVME_REG_CSTS, 4);
    }
}

static inline const char *nvme_ctrl_reg_str(unsigned reg_base)
{
    switch (reg_base) {
        case SNAP_NVME_REG_CAP:
            return "Controller Capabilities (CAP)";
        case SNAP_NVME_REG_VS:
            return "Controller Version (VS)";
        case SNAP_NVME_REG_INTMS:
            return "Interrupt Mask Set (INTMS)";
        case SNAP_NVME_REG_INTMC:
            return "Interrupt Mask Set (INTMC)";
        case SNAP_NVME_REG_CC:
            return "Controller Configuration (CC)";
        case SNAP_NVME_REG_CSTS:
            return "Controller Status (CSTS)";
        case SNAP_NVME_REG_NSSR:
            return "NVM Subsystem Reset (NSSR)";
        case SNAP_NVME_REG_AQA:
            return "Admin Queue Attributes (AQA)";
        case SNAP_NVME_REG_ASQ:
            return "Admin Submission Queue Base Address (ASQ)";
        case SNAP_NVME_REG_ACQ:
            return "Admin Completion Queue Base Address (ACQ)";
        case SNAP_NVME_REG_CMBLOC:
            return "Controller Memory Buffer Location (CMBLOC)";
        case SNAP_NVME_REG_CMBSZ:
            return "Controller Memory Buffer Size (CMBSZ)";
        case SNAP_NVME_REG_BPINFO:
            return "Boot Partition Information (BPINFO)";
        case SNAP_NVME_REG_BPRSEL:
            return "Boot Partition Select (BPRSEL)";
        case SNAP_NVME_REG_BPMBL:
            return "Boot Partition Memory Buffer Location (BPMBL)";
        default:
            return "unrecognized register";
    }
}

void nvme_bar_write_cb(void *bar, nvme_register_t *reg, uint64_t val,
		       uint64_t prev_val)
{
    nvme_bar_instance_t *dev_bar = bar;
    nvme_emulator_t *dev = container_of(dev_bar, nvme_emulator_t, bar);

    nvmx_debug("%s [0x%llx] <- 0x%llx bar write detected",
               nvme_ctrl_reg_str(reg->reg_base),
               (unsigned long long)prev_val,
               (unsigned long long)val);

    switch (reg->reg_base) {
        case SNAP_NVME_REG_CC:
            nvme_ctrl_modify_config((nvme_ctrl_t *)(dev->ctrl), (uint32_t)val, (uint32_t)prev_val);
            break;
        case SNAP_NVME_REG_AQA:
        case SNAP_NVME_REG_ASQ:
        case SNAP_NVME_REG_ACQ:
            break;
        case SNAP_NVME_REG_CSTS:
            /* ignore it until fully decoupled from qemu */
            nvmx_warn("CSTS change ignored");
            break;
        default:
            nvmx_fatal("0x%X <- 0x%llx unsupported bar write detected", reg->reg_base, (unsigned long long)val);
    }
}

static inline void nvme_cq_inc_tail(nvme_cqueue_t *cq)
{
    cq->tail++;
    if (cq->tail >= cq->size) {
        cq->tail = 0;
        cq->phase = !cq->phase;
    }
}

static inline uint16_t nvme_sq_inc_head(nvme_squeue_t *sq)
{
    return (sq->head + 1) % sq->size;
}

static inline int to_nvme_err(int rc)
{
    if (likely(rc == NVMX_INPROGRESS))
        return NVME_SC_NO_COMPLETE;
    if (rc == NVMX_OK)
        return NVME_SC_SUCCESS;
    return NVME_SC_INTERNAL_DEV_ERROR;
}

static uint16_t nvme_acmd_identify_ctrl(nvme_ctrl_t *ctrl, const nvme_cmd_identify_t *c,
                                        nvme_request_t *req)
{
    uint64_t prp1 = le64_to_cpu(c->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(c->common.dptr.prp2);
    int rc;

    rc = nvme_prp_write_bcopy(&req->io_req.async_req, prp1, prp2, &ctrl->id_ctrl,
                              sizeof(ctrl->id_ctrl));
    return to_nvme_err(rc);
}

static nvme_namespace_t *nvme_find_namespace(nvme_ctrl_t *ctrl, uint32_t nsid)
{
    nvme_namespace_t *ns;

    pthread_spin_lock(&ctrl->namespaces_lock);
    ns = nvme_find_namespace_nolock(ctrl, nsid);
    pthread_spin_unlock(&ctrl->namespaces_lock);

    return ns;
}

static inline nvme_namespace_t *nvme_find_namespace_nolock(nvme_ctrl_t *ctrl, uint32_t nsid)
{
    nvme_namespace_t *ns;

    TAILQ_FOREACH(ns, &ctrl->namespaces, entry) {
        if (ns->id == nsid)
            return ns;
    }

    return NULL;
}

static uint16_t nvme_acmd_identify_ns(nvme_ctrl_t *ctrl, const nvme_cmd_identify_t *c,
                                      nvme_request_t *req)
{
    uint32_t nsid = le32_to_cpu(c->common.nsid);
    uint64_t prp1 = le64_to_cpu(c->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(c->common.dptr.prp2);
    nvme_id_ns_t id_ns = {};
    nvme_namespace_t *ns;
    int rc;

    if (nsid == NVME_NAMESPACE_GLOBAL)
        return NVME_SC_INVALID_NSID | NVME_SC_DNR;

    ns = nvme_find_namespace(ctrl, nsid);
    if (!ns)
        rc = nvme_prp_write_bcopy(&req->io_req.async_req, prp1, prp2, &id_ns,
                                  sizeof(id_ns));
    else
        rc = nvme_prp_write_bcopy(&req->io_req.async_req, prp1, prp2, &ns->id_ns,
                                  sizeof(ns->id_ns));

    return to_nvme_err(rc);
}

static uint16_t nvme_acmd_identify_nslist(nvme_ctrl_t *ctrl, const nvme_cmd_identify_t *c,
                                          nvme_request_t *req)
{
    static const int data_len = 4096;
    uint32_t min_nsid = le32_to_cpu(c->common.nsid);
    uint64_t prp1 = le64_to_cpu(c->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(c->common.dptr.prp2);
    uint32_t *list;
    int j = 0;
    nvme_namespace_t *ns;
    int rc;

    list = calloc(data_len, 1);
    if (!list)
        return NVME_SC_INTERNAL_DEV_ERROR;

    pthread_spin_lock(&ctrl->namespaces_lock);
    TAILQ_FOREACH(ns, &ctrl->namespaces, entry) {
        /*
         * By NVMe spec., we should return only nsids that are greater than
         * the given min_nsid
         */
        if (ns->id <= min_nsid)
            continue;

        list[j++] = cpu_to_le32(ns->id);
        if (j == data_len / sizeof(uint32_t))
            break;
    }
    pthread_spin_unlock(&ctrl->namespaces_lock);

    /* According to NVMe spec., we have to return list in increasing order */
    qsort(list, j, sizeof(uint32_t), compare_uint32_t);
    rc = nvme_prp_write_bcopy(&req->io_req.async_req, prp1, prp2, list, data_len);
    free(list);
    return to_nvme_err(rc);
}

static uint16_t nvme_acmd_identify_ns_id_descriptor_list(nvme_ctrl_t *ctrl,
                                                         const nvme_cmd_identify_t *c,
                                                         nvme_request_t *req)
{
    static const int data_len = 4096;
    uint32_t nsid = le32_to_cpu(c->common.nsid);
    uint64_t prp1 = le64_to_cpu(c->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(c->common.dptr.prp2);
    nvme_namespace_t *ns;
    char *id_desc_list;
    size_t offset = 0;
    int i, rc;

    ns = nvme_find_namespace(ctrl, nsid);
    if (!ns)
        return NVME_SC_NS_NOT_READY;

    id_desc_list = (char *) calloc(data_len, 1);
    if (!id_desc_list)
        return NVME_SC_INTERNAL_DEV_ERROR;

    pthread_spin_lock(&ctrl->namespaces_lock);

    for (i = 0; i < NVME_NS_ID_TYPES_NUM; ++i) {
        nvme_id_ns_descriptor_t *id_desc = &ns->id_descs[i];

        memcpy(id_desc_list + offset, id_desc, id_desc->nidl + 4);
        offset += (id_desc->nidl + 4);
    }

    pthread_spin_unlock(&ctrl->namespaces_lock);

    rc = nvme_prp_write_bcopy(&req->io_req.async_req, prp1, prp2, id_desc_list,
                              data_len);
    free(id_desc_list);
    return to_nvme_err(rc);
}

static inline bool
nvme_controller_has_spdk_backend(snap_conf_t *sconfig)
{
    const snap_config_backend_t *backend;

    backend = snap_config_get_backend(sconfig, 0);
    if (!backend)
        return false;

    return (strncmp(backend->be_type, "spdk_bdev", 9) == 0);
}

#ifdef HAVE_SPDK
static uint16_t nvme_acmd_identify_allocated_nslist(nvme_ctrl_t *ctrl,
                                                    const nvme_cmd_identify_t *c,
                                                    nvme_request_t *req)
{
    static const int ns_list_sz = 1024;
    uint32_t min_nsid = le32_to_cpu(c->common.nsid);
    uint64_t prp1 = le64_to_cpu(c->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(c->common.dptr.prp2);
    uint32_t *ns_list;
    int rc;

    /*
     * In case backend is not spdk-based, there is no spdk subsystem,
     * and no spdk namespaces, so whole command is undefined
     */
    if (!nvme_controller_has_spdk_backend(ctrl->sconfig))
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;

    ns_list = calloc(ns_list_sz, sizeof(uint32_t));
    if (!ns_list)
        return NVME_SC_INTERNAL_DEV_ERROR;

    rc = nvme_subsys_ns_list_allocated_nsids(ctrl->sconfig->spdk.subsys,
                                             ns_list, ns_list_sz, min_nsid);
    if (rc) {
        free(ns_list);
        return NVME_SC_INTERNAL_DEV_ERROR;
    }

    rc = nvme_prp_write_bcopy(&req->io_req.async_req, prp1, prp2,
                              ns_list, ns_list_sz * sizeof(uint32_t));
    free(ns_list);
    return to_nvme_err(rc);
}
#endif

static uint16_t nvme_acmd_identify(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd,
                                   nvme_request_t *req)
{
    const nvme_cmd_identify_t *c = (nvme_cmd_identify_t *)cmd;

    switch (le32_to_cpu(c->cns)) {
    case NVME_IDENTIFY_CNS_NAMESPACE:
        return nvme_acmd_identify_ns(ctrl, c, req);
    case NVME_IDENTIFY_CNS_CTRL:
        return nvme_acmd_identify_ctrl(ctrl, c, req);
    case NVME_IDENTIFY_CNS_ACTIVE_NS_LIST:
        return nvme_acmd_identify_nslist(ctrl, c, req);
    case NVME_IDENTIFY_CNS_NS_ID_DESCRIPTOR_LIST:
        return nvme_acmd_identify_ns_id_descriptor_list(ctrl, c, req);
#ifdef HAVE_SPDK
    case NVME_IDENTIFY_CNS_ALLOCATED_NS_LIST:
        return nvme_acmd_identify_allocated_nslist(ctrl, c, req);
#endif
    default:
        nvmx_warn("invalid identify command cns 0x%x", le32_to_cpu(c->cns));
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }
}

static uint16_t nvme_acmd_create_cq(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd)
{
    nvme_emulator_t *dev = ctrl->dev;
    nvme_cqueue_t *cq;
    const nvme_cmd_create_cq_t *c = (nvme_cmd_create_cq_t *)cmd;

    uint16_t cqid = le16_to_cpu(c->qid);
    uint16_t vector = le16_to_cpu(c->iv);
    uint16_t qsize = le16_to_cpu(c->qsize);
    uint64_t prp1 = le64_to_cpu(c->common.dptr.prp1);
    nvme_bar_t *bar = &dev->bar.curr;
    int ret;

    if (unlikely(!cqid || !nvme_check_new_cqid_valid(ctrl, cqid))) {
        nvmx_error("cqid %d failed cqid check", cqid);
        return NVME_SC_INVALID_QID | NVME_SC_DNR;
    }
    if (unlikely(!qsize || qsize > nvme_reg_cap_mqes(bar->cap))) {
        nvmx_error("qsize %d is incorrect", qsize);
        return NVME_SC_MAX_QSIZE_EXCEEDED | NVME_SC_DNR;
    }
    if (unlikely(vector > ctrl->num_queues)) {
        return NVME_SC_INVALID_IRQ_VECTOR | NVME_SC_DNR;
    }
    if (unlikely(!c->pc)) {
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    cq = calloc(1, sizeof(*cq));
    if (unlikely(!cq))
        return NVME_SC_INTERNAL_DEV_ERROR | NVME_SC_DNR;

    ret = nvme_cq_init(cq, ctrl, prp1, cqid, vector, qsize + 1,c->ien);
    if (ret != NVMX_OK) {
        free(cq);
        return NVME_SC_INTERNAL_DEV_ERROR | NVME_SC_DNR;
    }

    ctrl->allow_set_num_qs = false;
    nvmx_debug("cq: %d created", cqid);
    return NVME_SC_SUCCESS;
}

static uint16_t nvme_acmd_delete_cq(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd)
{
    const nvme_cmd_delete_q_t *c = (nvme_cmd_delete_q_t *)cmd;
    nvme_cqueue_t *cq;
    uint16_t qid = le16_to_cpu(c->qid);

    if (unlikely(!qid || !nvme_check_cqid_valid(ctrl, qid))) {
        return NVME_SC_INVALID_CQID | NVME_SC_DNR;
    }

    cq = ctrl->cq[qid];
    if (unlikely(!TAILQ_EMPTY(&cq->sq_list))) {
        return NVME_SC_INVALID_QUEUE_DEL;
    }

    nvme_cq_free(cq, ctrl);
    return NVME_SC_SUCCESS;
}

static uint16_t nvme_acmd_create_sq(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd)
{
    nvme_emulator_t *dev = ctrl->dev;
    nvme_squeue_t *sq;
    const nvme_cmd_create_sq_t *c = (nvme_cmd_create_sq_t *)cmd;

    uint16_t cqid = le16_to_cpu(c->cqid);
    uint16_t sqid = le16_to_cpu(c->qid);
    uint16_t qsize = le16_to_cpu(c->qsize);
    uint64_t prp1 = le64_to_cpu(c->common.dptr.prp1);
    nvme_bar_t *bar = &dev->bar.curr;
    int ret;

    if (unlikely(!sqid || !nvme_check_cqid_in_range(ctrl, cqid) ||
                          !nvme_check_new_sqid_valid(ctrl, sqid))) {
        return NVME_SC_INVALID_QID | NVME_SC_DNR;
    }
    if (unlikely(!nvme_check_cqid_valid(ctrl, cqid))) {
        return NVME_SC_INVALID_CQID | NVME_SC_DNR;
    }
    if (unlikely(!qsize || qsize > nvme_reg_cap_mqes(bar->cap))) {
        return NVME_SC_MAX_QSIZE_EXCEEDED | NVME_SC_DNR;
    }
    if (unlikely(!prp1 || prp1 & (ctrl->page_size - 1))) {
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }
    if (unlikely(!c->pc)) {
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    sq = calloc(1, sizeof(*sq));
    if (unlikely(!sq))
        return NVME_SC_INTERNAL_DEV_ERROR | NVME_SC_DNR;

    ret = nvme_sq_init(sq, ctrl, prp1, sqid, cqid, qsize + 1);
    if (ret != NVMX_OK) {
        free(sq);
        return NVME_SC_INTERNAL_DEV_ERROR | NVME_SC_DNR;
    }

    ctrl->allow_set_num_qs = false;
    nvmx_debug("sqid:%d with cqid:%d at pg %zu created", sqid, cqid, sq->pg - ctrl->pgs);
    return NVME_SC_SUCCESS;
}

static uint16_t nvme_acmd_delete_sq(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd,
                                    nvme_request_t *req)
{
    const nvme_cmd_delete_q_t *c = (nvme_cmd_delete_q_t *)cmd;
    uint16_t qid = le16_to_cpu(c->qid);
    nvme_squeue_t *sq;

    if (unlikely(!qid || !nvme_check_sqid_valid(ctrl, qid))) {
        return NVME_SC_INVALID_QID | NVME_SC_DNR;
    }

    sq = ctrl->sq[qid];
    if (sq->delete_req) {
        nvmx_error("sqid:%d old delete SQ is in progress", qid);
        return NVME_SC_INVALID_QID | NVME_SC_DNR;
    }

    sq->delete_req = req;
    nvme_sq_suspend(sq);
    return NVME_SC_NO_COMPLETE;
}

static uint16_t nvme_set_feature_arbitration(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd)
{
    uint32_t nsid = le32_to_cpu(cmd->common.nsid);
    uint32_t dw11 = le32_to_cpu(cmd->cdw11);

    if (nsid != 0 && nsid != 0xFFFFFFFF)
        return NVME_SC_FID_NOT_NSID_SPEC;

    ctrl->arbitration = dw11;

    return NVME_SC_SUCCESS;
}

static uint16_t nvme_set_feature_interrupt_coalescing(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd)
{
    int aggr_time, aggr_thresh;
    uint32_t nsid = le32_to_cpu(cmd->common.nsid);
    uint32_t dw11 = le32_to_cpu(cmd->cdw11);

    if (nsid != 0 && nsid != 0xFFFFFFFF)
        return NVME_SC_FID_NOT_NSID_SPEC;

    /* Spec 5.14.1.8, zero based value, 0 - no moderation.
     * Ours is 1 based value where 0 and 1 mean moderation is disabled */
    ctrl->cq_max_count = NVME_FTR_AGGR_THRESH(dw11);
    if (ctrl->cq_max_count > 0)
        ctrl->cq_max_count++;

    /* Spec 5.14.1.8, Aggregation time is in 100 microseconds increments */
    ctrl->cq_period = 100 * NVME_FTR_AGGR_TIME(dw11);

    /* Treat our defaults as a hard minimum value */
    nvme_get_default_moderation(ctrl, &aggr_time, &aggr_thresh);
    if (aggr_time > ctrl->cq_period)
        ctrl->cq_period = aggr_time;

    if (aggr_thresh > ctrl->cq_max_count)
        ctrl->cq_max_count = aggr_thresh;

    nvmx_info("SET cq_max_count: %d cq_max_period: %d",
              ctrl->cq_max_count, ctrl->cq_period);

    return NVME_SC_SUCCESS;
}

static uint16_t nvme_set_feature_asynchronous_event_conf(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd)
{
    uint32_t nsid = le32_to_cpu(cmd->common.nsid);
    uint32_t dw11 = le32_to_cpu(cmd->cdw11);

    if (nsid != 0 && nsid != 0xFFFFFFFF)
        return NVME_SC_FID_NOT_NSID_SPEC;

    if (dw11 & ~NVME_SUPPORTED_ASYNC_EVENTS)
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;

    pthread_mutex_lock(&ctrl->async_event_lock);
    ctrl->async_event_config = dw11;
    if (ctrl->quirks & NVME_QUIRK_ASYNC_EV_CONFIG) {
        /*
         * some drivers ex Windows driver know how to handle this
         * events, though not negotiate the feature bit
         */
        ctrl->async_event_config |= NVME_SUPPORTED_ASYNC_EVENTS;
    }
    pthread_mutex_unlock(&ctrl->async_event_lock);

    nvme_ctrl_raise_async_events(ctrl);
    return NVME_SC_SUCCESS;
}

static void nvme_get_feature_arbitration(nvme_ctrl_t *ctrl,
                                         nvme_cqe_t *cqe,
                                         uint8_t ftr_val)
{

    switch (ftr_val) {
    case NVME_FEATURE_CURRENT:
        cqe->result = cpu_to_le32(ctrl->arbitration);
        break;
    case NVME_FEATURE_DEFAULT:
    case NVME_FEATURE_SAVED:
        cqe->result = cpu_to_le32(NVME_DEFAULT_ARBITRATION);
        break;
    case NVME_FEATURE_CAPS:
        cqe->result = cpu_to_le32(NVME_FEATURE_CHANGEABLE);
        break;
    }
}

/* power management */
static void nvme_get_feature_pm(nvme_ctrl_t *ctrl,
                                nvme_cqe_t *cqe,
                                uint8_t ftr_val)
{

    switch (ftr_val) {
    case NVME_FEATURE_CURRENT:
        cqe->result = cpu_to_le32(ctrl->power_management);
        break;
    case NVME_FEATURE_DEFAULT:
    case NVME_FEATURE_SAVED:
        cqe->result = cpu_to_le32(NVME_DEFAULT_PM);
        break;
    case NVME_FEATURE_CAPS:
        cqe->result = cpu_to_le32(NVME_FEATURE_CHANGEABLE);
        break;
    }
}

static void nvme_get_feature_null(nvme_cqe_t *cqe, uint8_t ftr_val)
{
    switch (ftr_val) {
    case NVME_FEATURE_CURRENT:
    case NVME_FEATURE_DEFAULT:
    case NVME_FEATURE_SAVED:
    case NVME_FEATURE_CAPS:
        cqe->result = 0; /* unchangeable */
        break;
    }
}


/* error recovery */
static void nvme_get_feature_error_recovery(nvme_ctrl_t *ctrl,
                                            nvme_cqe_t *cqe,
                                            uint8_t ftr_val)
{
    nvme_get_feature_null(cqe, ftr_val);
}

static uint16_t
nvme_set_feature_num_queues(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd, nvme_cqe_t *cqe)
{
    uint32_t nsid = le32_to_cpu(cmd->common.nsid);
    const nvme_cmd_set_feature_t *sf = (nvme_cmd_set_feature_t *)cmd;

    if (nsid != 0 && nsid != 0xFFFFFFFF)
        return NVME_SC_FID_NOT_NSID_SPEC | NVME_SC_DNR;

    if ((sf->dw11 & 0xFFFF) == 0xFFFF) {
        nvmx_error("Invalid requested number of SQs %u\n",
                   sf->dw11 & 0xFFFF);
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    if (((sf->dw11 >> 16) & 0xFFFF) == 0xFFFF) {
        nvmx_error("Invalid requested number of CQs %u\n",
                   (sf->dw11 >> 16) & 0xFFFF);
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    /* According to the spec 5.14.1.7 it is ok to always
     * return our max number of queues and ignore user
     * supplied number.
     *
     * Zero based, and num queues includes the admin.
     * If a Set Features command is issued after creation
     * of any I/O Queues, then the command shall fail */
    if (ctrl->allow_set_num_qs == false)
        return NVME_SC_CMD_SEQ_ERROR | NVME_SC_DNR;

    cqe->result = cpu_to_le32((ctrl->num_queues - 2) |
                              ((ctrl->num_queues - 2) << 16));
    return NVME_SC_SUCCESS;
}

static void
nvme_get_feature_num_queues(nvme_ctrl_t *ctrl, nvme_cqe_t *cqe, uint8_t ftr_val)
{
    switch (ftr_val) {
    case NVME_FEATURE_CURRENT:
    case NVME_FEATURE_DEFAULT:
    case NVME_FEATURE_SAVED:
        cqe->result = cpu_to_le32((ctrl->num_queues - 2) |
                                  ((ctrl->num_queues - 2) << 16));
        break;
    case NVME_FEATURE_CAPS:
        cqe->result = cpu_to_le32(NVME_FEATURE_CHANGEABLE);
        break;
    }
}

static uint32_t cq_period2agg_time(int cq_period)
{
    int result;

    /* TODO:
     * SPEC 5.14.1.8
     * aggregation time is defined in 100 of micros.
     * Our hardware works with micros. So we have to convert our
     * value to 100 of micros. Possible conversions:
     * Round up:
     * result = (cq_period + 100 - 1)/100;
     * Round down
     */
    result = cq_period/100;

    /* if cannot fit into one byte report max (25500 usec!!!!) */
    if (result > 0xff)
        result = 0xff;

    return result;
}

static uint32_t cq_max_count2agg_threshold(int cq_max_count)
{
    /* we keep it as one base val, 0 and 1 mean 'disabled' */
    return cq_max_count > 0 ? (uint32_t)(cq_max_count - 1) : 0;
}

static void nvme_get_feature_interrupt_coalescing(nvme_ctrl_t *ctrl,
                                                  nvme_cqe_t *cqe,
                                                  uint8_t ftr_val)
{
    int aggr_time;
    int aggr_thresh;

    switch (ftr_val) {
    case NVME_FEATURE_CURRENT:
        aggr_time   = cq_period2agg_time(ctrl->cq_period);
        aggr_thresh = cq_max_count2agg_threshold(ctrl->cq_max_count);
        cqe->result = cpu_to_le32((aggr_time << NVME_FTR_AGGR_TIME_SHIFT)|
                                   aggr_thresh);
        break;
    case NVME_FEATURE_DEFAULT:
    case NVME_FEATURE_SAVED:
        nvme_get_default_moderation(ctrl, &aggr_time, &aggr_thresh);
        aggr_time   = cq_period2agg_time(aggr_time);
        aggr_thresh = cq_max_count2agg_threshold(aggr_thresh);
        cqe->result = cpu_to_le32((aggr_time << NVME_FTR_AGGR_TIME_SHIFT)|
                                   aggr_thresh);
        break;
    case NVME_FEATURE_CAPS:
        cqe->result = cpu_to_le32(NVME_FEATURE_CHANGEABLE);
        break;
    }
}

/* interrupt vector config */
static void nvme_get_feature_iv_conf(nvme_ctrl_t *ctrl, nvme_cqe_t *cqe,
                                     uint8_t ftr_val)
{
    nvme_get_feature_null(cqe, ftr_val);
}

/* write atomicity */
static void nvme_get_feature_write_atomicity(nvme_ctrl_t *ctrl, nvme_cqe_t *cqe,
                                             uint8_t ftr_val)
{
    nvme_get_feature_null(cqe, ftr_val);
}


static void nvme_get_feature_asynchronous_event_conf(nvme_ctrl_t *ctrl,
                                                     nvme_cqe_t *cqe,
                                                     uint8_t ftr_val)
{
    switch (ftr_val) {
    case NVME_FEATURE_CURRENT:
        cqe->result = cpu_to_le32(ctrl->async_event_config);
        break;
    case NVME_FEATURE_DEFAULT:
    case NVME_FEATURE_SAVED:
        cqe->result = cpu_to_le32(0);
        break;
    case NVME_FEATURE_CAPS:
        cqe->result = cpu_to_le32(NVME_FEATURE_CHANGEABLE);
        break;
    }
}

static uint32_t get_temp_threshold(nvme_ctrl_t *ctrl, uint32_t dw11)
{
	if (dw11 & (3 << 20))
		return cpu_to_le32(ctrl->under_temperature_threshold);
	return cpu_to_le32(ctrl->over_temperature_threshold);
}

static void nvme_get_feature_temp_threshold(nvme_ctrl_t *ctrl,
                                            nvme_cqe_t *cqe,
                                            uint32_t dw11,
                                            uint8_t ftr_val)
{
    switch (ftr_val) {
    case NVME_FEATURE_CURRENT:
        cqe->result = get_temp_threshold(ctrl, dw11);
        break;
    case NVME_FEATURE_DEFAULT:
    case NVME_FEATURE_SAVED:
        cqe->result = get_temp_threshold(ctrl, dw11);
        break;
    case NVME_FEATURE_CAPS:
        cqe->result = cpu_to_le32(0x0); /* The feature is not changeable */
        break;
    }
}

static uint16_t
nvme_set_feature_temp_threshold(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd)
{
    uint32_t nsid = le32_to_cpu(cmd->common.nsid);
    uint32_t dw11 = le32_to_cpu(cmd->cdw11);

    if (nsid != 0 && nsid != 0xFFFFFFFF)
        return NVME_SC_FID_NOT_NSID_SPEC;

    if (dw11 & (3 << 20))
        ctrl->under_temperature_threshold = NVME_FTR_TEMP_THRESH(dw11);
    else
        ctrl->over_temperature_threshold = NVME_FTR_TEMP_THRESH(dw11);

    if (ctrl->temperature <= ctrl->under_temperature_threshold ||
            ctrl->temperature >= ctrl->over_temperature_threshold)
        nvme_ctrl_queue_async_event(ctrl, NVME_TEMPERATURE_ERR);

    return NVME_SC_SUCCESS;
}

/*
 * According to the NVMe spec (section 5.1)- it's implementation specific
 * when a controller chooses to complete the abort command
 * in case if the command to abort is not found,
 * so the controller always says that the command
 * with cid to abort wasn't found and completes the abort immediately
 */
static uint16_t nvme_acmd_abort(nvme_request_t *req)
{
    /* cid wasn't aborted */
    req->cqe.result = cpu_to_le32(NVME_ABORT_FAILED);

    return NVME_SC_SUCCESS;
}

static uint16_t nvme_set_feature_pm(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd)
{
    uint8_t npss;
    uint8_t pstate;
    uint32_t nsid = le32_to_cpu(cmd->common.nsid);
    uint32_t dw11 = le32_to_cpu(cmd->cdw11);

    if (nsid != 0 && nsid != 0xFFFFFFFF)
        return NVME_SC_FID_NOT_NSID_SPEC;

    npss   = ctrl->id_ctrl.npss;
    pstate = NVME_FTR_PM_PSTATE(dw11);

    /* check sanity */
    if (pstate > npss) {
        nvmx_error("power state %d is not supported", pstate);
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    ctrl->power_management = dw11;
    return NVME_SC_SUCCESS;
}

static uint16_t nvme_set_feature_vector_conf(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd)
{
    uint32_t nsid = le32_to_cpu(cmd->common.nsid);

    if (nsid != 0 && nsid != 0xFFFFFFFF)
        return NVME_SC_FID_NOT_NSID_SPEC;

    return NVME_SC_SUCCESS;
}

static uint16_t nvme_set_feature_unchangeable()
{
    return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
}

static uint16_t nvme_acmd_set_feature(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd,
                                      nvme_request_t *req)
{
    const nvme_cmd_set_feature_t *c = (nvme_cmd_set_feature_t *)cmd;

    if (c->sv & (1 << 7)) {
        return NVME_SC_FID_NOT_SAVEABLE | NVME_SC_DNR;
    }

    switch (c->fid) {
    case NVME_ARBITRATION:
        return nvme_set_feature_arbitration(ctrl, cmd);
    case NVME_POWER_MANAGEMENT:
        return nvme_set_feature_pm(ctrl, cmd);
    case NVME_ERROR_RECOVERY:
        return nvme_set_feature_unchangeable();
    case NVME_INTERRUPT_COALESCING:
        return nvme_set_feature_interrupt_coalescing(ctrl, cmd);
    case NVME_INTERRUPT_VECTOR_CONF:
        return nvme_set_feature_vector_conf(ctrl, cmd);
    case NVME_WRITE_ATOMICITY:
        return nvme_set_feature_unchangeable();
    case NVME_ASYNCHRONOUS_EVENT_CONF:
        return nvme_set_feature_asynchronous_event_conf(ctrl, cmd);
    case NVME_VOLATILE_WRITE_CACHE:
        break;
    case NVME_NUMBER_OF_QUEUES:
        return nvme_set_feature_num_queues(ctrl, cmd, &req->cqe);
    case NVME_TEMPERATURE_THRESHOLD:
        return nvme_set_feature_temp_threshold(ctrl, cmd);
    default:
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    return NVME_SC_SUCCESS;
}

static uint16_t nvme_acmd_get_feature(nvme_ctrl_t *ctrl,
                                      const nvme_cmd_t *cmd,
                                      nvme_request_t *req)
{
    uint32_t dw11 = le32_to_cpu(cmd->cdw11);
    const nvme_cmd_get_feature_t *c = (nvme_cmd_get_feature_t *)cmd;
    uint8_t ftr_val = c->sel;

    switch (c->fid) {
    case NVME_ARBITRATION:
        nvme_get_feature_arbitration(ctrl, &req->cqe, dw11);
        break;
    case NVME_POWER_MANAGEMENT:
        nvme_get_feature_pm(ctrl, &req->cqe, ftr_val);
        break;
    case NVME_LBA_RANGE_TYPE:
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    case NVME_TEMPERATURE_THRESHOLD:
        nvme_get_feature_temp_threshold(ctrl, &req->cqe, dw11, ftr_val);
        break;
    case NVME_ERROR_RECOVERY:
        nvme_get_feature_error_recovery(ctrl, &req->cqe, ftr_val);
        break;
    case NVME_VOLATILE_WRITE_CACHE:
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    case NVME_NUMBER_OF_QUEUES:
        nvme_get_feature_num_queues(ctrl, &req->cqe, ftr_val);
        break;
    case NVME_INTERRUPT_COALESCING:
        nvme_get_feature_interrupt_coalescing(ctrl, &req->cqe, ftr_val);
        break;
    case NVME_INTERRUPT_VECTOR_CONF:
        nvme_get_feature_iv_conf(ctrl, &req->cqe, ftr_val);
        break;
    case NVME_WRITE_ATOMICITY:
        nvme_get_feature_write_atomicity(ctrl, &req->cqe, ftr_val);
        break;
    case NVME_ASYNCHRONOUS_EVENT_CONF:
        nvme_get_feature_asynchronous_event_conf(ctrl, &req->cqe, ftr_val);
        break;
    case NVME_AUTONOMOUS_POWER_STATE_TRANSIT:
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    case NVME_SOFTWARE_PROGRESS_MARKER:
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    case NVME_HOST_IDENTIFIER:
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    default:
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    return NVME_SC_SUCCESS;
}

static size_t dwords2bytes(uint16_t num_dwords, size_t type_size)
{
    /* note that num_dwords is zero based value */
    size_t size = 4 * (num_dwords + 1);
    return size > type_size ? type_size : size;
}

static uint16_t nvme_log_error_info(nvme_ctrl_t *ctrl, uint64_t prp1, uint64_t prp2,
                                    uint16_t num_dwords, nvme_request_t *req)
{
    nvme_error_log_t elog;
    int rc;

    memset(&elog, 0, sizeof(elog));
    rc = nvme_prp_write_bcopy(&req->io_req.async_req, prp1, prp2, &elog,
                              dwords2bytes(num_dwords, sizeof(elog)));
    return to_nvme_err(rc);
}

static uint16_t nvme_log_smart_info(nvme_ctrl_t *ctrl,
                                    const nvme_cmd_get_page_log_t *c,
                                    nvme_request_t *req)
{
    uint64_t prp1 = le64_to_cpu(c->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(c->common.dptr.prp2);
    uint32_t nsid = le32_to_cpu(c->common.nsid);
    nvme_smart_log_t smart_log = {};
    int rc;
    uint16_t sqid;
    const nvme_stat_t *stat;
    uint32_t num_dwords = nvme_acmd_get_log_page_cmd_num_dwords(c);

    if (nsid != 0 && nsid != 0xFFFFFFFF)
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;

    smart_log.temperature = ctrl->temperature;

    if (ctrl->temperature <= ctrl->under_temperature_threshold ||
            ctrl->temperature >= ctrl->over_temperature_threshold)
        smart_log.critical_warning.temperature = 1;
    else
        smart_log.critical_warning.temperature = 0;

    for (sqid = 1; sqid < ctrl->num_queues; sqid++) {
        stat = &ctrl->sq_stat[sqid];

        smart_log.host_read_commands[0] += stat->read.completed;
        smart_log.host_write_commands[0] += stat->write.completed;
        smart_log.data_units_read[0] += stat->read.completed_units_count;
        smart_log.data_units_written[0] += stat->write.completed_units_count;
    }
    smart_log.data_units_read[0] = ROUND_UP(smart_log.data_units_read[0], 1000) / 1000;
    smart_log.data_units_written[0] = ROUND_UP(smart_log.data_units_written[0], 1000) / 1000;

    rc = nvme_prp_write_bcopy(&req->io_req.async_req, prp1, prp2, &smart_log,
                              dwords2bytes(num_dwords, sizeof(smart_log)));

    if (!c->rae)
        ASYNC_EVENT_UNMASK(ctrl, NVME_LOG_SMART_INFO);

    return to_nvme_err(rc);
}

static uint16_t nvme_log_fw_slot_info(nvme_ctrl_t *ctrl,
                                      const nvme_cmd_get_page_log_t *c,
                                      nvme_request_t *req)
{
    nic_utils_t *ctx;
    nvme_json_t *json;
    nvme_fw_slot_info_log_t fw_log;
    int i, rc;
    bool success, valid, bool_val;
    size_t slot_no;
    uint64_t prp1 = le64_to_cpu(c->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(c->common.dptr.prp2);
    uint32_t num_dwords = nvme_acmd_get_log_page_cmd_num_dwords(c);
    char version[16];

    memset(&fw_log, 0, sizeof fw_log);

    if (ctrl->fw_update_enable) {
        ctx = nic_utils_open();
        if (!ctx) {
            rc = NVMX_ERR;
            nvmx_error("open nic utils failed");
            goto out;
        }

        for (slot_no = 1; slot_no <= NVME_FW_SLOTS_COUNT; slot_no++) {
            json = fw_get_bfb_info(ctrl, ctx, slot_no);
            if (!json)
                continue;

            if (nvme_json_get_value(json, "success", &success, sizeof(success))) {
                nvmx_warn("Failed to retrieve success flag for slot:%d", slot_no);
                goto next;
            }

            if (!success)
                goto next;

            if (nvme_json_get_value(json, "valid", &valid, sizeof(valid))) {
                nvmx_warn("Failed to retrieve valid flag for slot:%d", slot_no);
                goto next;
            }

            if (!valid)
                goto next;

            memset(version, 0, sizeof(version));
            if (nvme_json_get_value(json, "version", version, sizeof(version)))
                nvmx_warn("Failed to retrieve bfb version for slot:%d", slot_no);

            for (i = 0; i < sizeof(fw_log.frs[slot_no - 1]); i++)
                fw_log.frs[slot_no - 1][i] = version[i];

            if (nvme_json_get_value(json, "next", &bool_val, sizeof(bool_val)))
                nvmx_warn("Failed to retrieve bfb next flag for slot:%d", slot_no);
            if (bool_val)
                fw_log.afi_next = slot_no;

            if (nvme_json_get_value(json, "active", &bool_val, sizeof(bool_val)))
                nvmx_warn("Failed to retrieve bfb active flag for slot:%d", slot_no);
            if (bool_val)
                fw_log.afi_current = slot_no;
next:
            nvme_json_free(json);
        }

        nic_utils_close(ctx);
    } else {
        fw_log.afi_current = (1 << 0); //Suppose FW loaded on the first slot
        fw_log.frs[0][0] = 0x1; //Let's say FW ver. N% 1
    }

    rc = nvme_prp_write_bcopy(&req->io_req.async_req, prp1, prp2, &fw_log,
                              dwords2bytes(num_dwords, sizeof(fw_log)));

out:
    return to_nvme_err(rc);
}

static uint16_t nvme_log_commands_effects_info(nvme_ctrl_t *ctrl,
                                               uint64_t prp1, uint64_t prp2,
                                               uint16_t num_dwords,
                                               nvme_request_t *req)
{
    nvme_commands_and_effects_log_t effects_log;
    nvme_cmds_supported_and_effects_entry_t *cmd_entry;
    int rc;

    memset(&effects_log, 0, sizeof(effects_log));

    /* Standard admin commands */
    cmd_entry = &effects_log.acs[NVME_ADMIN_DELETE_SQ];
    cmd_entry->csupp = 1;
    cmd_entry = &effects_log.acs[NVME_ADMIN_CREATE_SQ];
    cmd_entry->csupp = 1;
    cmd_entry = &effects_log.acs[NVME_ADMIN_DELETE_CQ];
    cmd_entry->csupp = 1;
    cmd_entry = &effects_log.acs[NVME_ADMIN_CREATE_CQ];
    cmd_entry->csupp = 1;
    cmd_entry = &effects_log.acs[NVME_ADMIN_IDENTIFY];
    cmd_entry->csupp = 1;
    cmd_entry = &effects_log.acs[NVME_ADMIN_SET_FEATURES];
    cmd_entry->csupp = 1;
    cmd_entry = &effects_log.acs[NVME_ADMIN_GET_FEATURES];
    cmd_entry->csupp = 1;
    cmd_entry = &effects_log.acs[NVME_ADMIN_GET_LOG_PAGE];
    cmd_entry->csupp = 1;
    cmd_entry = &effects_log.acs[NVME_ADMIN_ASYNC_EV_REQ];
    cmd_entry->csupp = 1;
    if (ctrl->fw_update_enable) {
        cmd_entry = &effects_log.acs[NVME_ADMIN_FW_COMMIT];
        cmd_entry->csupp = 1;
        cmd_entry = &effects_log.acs[NVME_ADMIN_FW_DOWNLOAD];
        cmd_entry->csupp = 1;
    }

    /* Standard IO commands */
    cmd_entry = &effects_log.acs[NVME_EFFECTS_LOG_IO_OFF + NVME_IO_FLUSH];
    cmd_entry->csupp = 1;
    cmd_entry = &effects_log.acs[NVME_EFFECTS_LOG_IO_OFF + NVME_IO_WRITE];
    cmd_entry->csupp = 1;
    cmd_entry = &effects_log.acs[NVME_EFFECTS_LOG_IO_OFF + NVME_IO_READ];
    cmd_entry->csupp = 1;

    /* Vendor-specific admin commands */
    if (ctrl->iova_mgmt_enable) {
        cmd_entry = &effects_log.acs[NVME_ADMIN_VS_IOVA_MGMT];
        cmd_entry->csupp = 1;
    }
#if HAVE_SNAP
    if (ctrl->rpc_server) {
        cmd_entry = &effects_log.acs[NVME_ADMIN_VS_JSON_RPC_2_0_REQ];
        cmd_entry->csupp = 1;
        cmd_entry->lbcc = 0;
        cmd_entry->ncc = 0;
        cmd_entry->nic = 1;
        cmd_entry->ccc = 0;
        cmd_entry->cse = 0;

        cmd_entry = &effects_log.acs[NVME_ADMIN_VS_JSON_RPC_2_0_RSP];
        cmd_entry->csupp = 1;
    }
#endif

    rc = nvme_prp_write_bcopy(&req->io_req.async_req, prp1, prp2, &effects_log,
                              dwords2bytes(num_dwords, sizeof(effects_log)));
    return to_nvme_err(rc);

}

static uint16_t nvme_log_changed_ns_info(nvme_ctrl_t *ctrl,
                                         const nvme_cmd_get_page_log_t *c,
                                         nvme_request_t *req)
{
    uint64_t prp1 = le64_to_cpu(c->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(c->common.dptr.prp2);
    uint32_t num_dwords = nvme_acmd_get_log_page_cmd_num_dwords(c);
    nvme_changed_nslist_log_t nslog;
    int rc;

    memset(&nslog, 0, sizeof(nslog));
    pthread_spin_lock(&ctrl->namespaces_lock);
    if (ctrl->num_changed_nsids >= 1024) {
        /*
         * According to spec., If there 1024 (or more) changes,
         * return single elemented list, with value 0xFFFFFFFF
         */
        nslog.ns_list[0] = NVME_NAMESPACE_GLOBAL;
    } else {
        memcpy(nslog.ns_list, ctrl->changed_nsids, sizeof(nslog.ns_list));
    }
    memset(ctrl->changed_nsids, 0, sizeof(ctrl->changed_nsids));
    ctrl->num_changed_nsids = 0;
    rc = nvme_prp_write_bcopy(&req->io_req.async_req, prp1, prp2, &nslog,
                              dwords2bytes(num_dwords, sizeof(nslog)));
    if (!c->rae)
        ASYNC_EVENT_UNMASK(ctrl, NVME_LOG_CHANGED_NS_LIST_INFO);
    pthread_spin_unlock(&ctrl->namespaces_lock);

    return to_nvme_err(rc);
}

static uint16_t nvme_log_bf_pkg_info(nvme_ctrl_t *ctrl, uint64_t prp1, uint64_t prp2,
                                      uint16_t num_dwords, nvme_request_t *req)
{
    nvme_bf_pkg_info_log_t bf_pkg_info;
    int rc;

    fw_bf_pkg_info_init(ctrl, &bf_pkg_info);

    rc = nvme_prp_write_bcopy(&req->io_req.async_req, prp1, prp2, &bf_pkg_info,
                              dwords2bytes(num_dwords, sizeof(bf_pkg_info)));
    return to_nvme_err(rc);
}

static uint16_t
nvme_acmd_get_log_page(nvme_ctrl_t *ctrl,
                      const nvme_cmd_t *cmd,
                      nvme_request_t *req)
{
    const nvme_cmd_get_page_log_t *c = (nvme_cmd_get_page_log_t *)cmd;
    uint32_t num_dwords = nvme_acmd_get_log_page_cmd_num_dwords(c);

    switch (c->lid) {
    case NVME_LOG_ERROR_INFO:
        return nvme_log_error_info(ctrl, c->common.dptr.prp1, c->common.dptr.prp2, num_dwords, req);
    case NVME_LOG_SMART_INFO:
        return nvme_log_smart_info(ctrl, c, req);
    case NVME_LOG_FW_SLOT_INFO:
        return nvme_log_fw_slot_info(ctrl, c, req);
    case NVME_LOG_CHANGED_NS_LIST_INFO:
        return nvme_log_changed_ns_info(ctrl, c, req);
    case NVME_LOG_COMMANDS_EFFECTS_LOG:
        return nvme_log_commands_effects_info(ctrl, c->common.dptr.prp1, c->common.dptr.prp2, num_dwords, req);
    case NVME_LOG_BF_PKG_INFO:
        return nvme_log_bf_pkg_info(ctrl, c->common.dptr.prp1, c->common.dptr.prp2, num_dwords, req);
    default:
        return NVME_SC_INVALID_LOG_ID | NVME_SC_DNR;
    }
}

static uint16_t nvme_acmd_async_ev(nvme_ctrl_t *ctrl, nvme_request_t *req)
{
    static int aen_quirk_warn;
    static int log_page_quirk_warn;

    /* pretend that we support the command.
       But in reality we never complete it */
    pthread_mutex_lock(&ctrl->async_event_lock);
    if (nvme_ctrl_num_async_event_requests(ctrl) >= NVME_AER_LIMIT) {
        pthread_mutex_unlock(&ctrl->async_event_lock);
        nvmx_error("Num of sent async event requests exceeds AERL");
        return NVME_SC_AER_LIMIT_EXCEEDED;
    }

    req->is_async_event = true;
    TAILQ_REMOVE(&req->sq->out_req_list, req, entry);
    nvme_ctrl_push_async_event_request(ctrl, req);

    if (!ctrl->async_event_config) {
        /* warn once about a possibility of a buggy driver */
        if (!aen_quirk_warn) {
            nvmx_warn("Optional async events are not configured. "
                      "Namespace changes will not be reported");
            nvmx_warn("Consider setting 'ctrl.quirks' if the host "
                      "nvme driver supports 'namespace change' event");
            aen_quirk_warn = 1;
        }
    }

    /* asking for a new async event before clearing a pending event is
     * an indication of a potential driver bug
     */
    if (ASYNC_EVENT_IS_MASKED(ctrl, NVME_LOG_CHANGED_NS_LIST_INFO)) {
        if (ctrl->quirks & NVME_QUIRK_NS_CHANGE_FORCE_UNMASK) {
            ASYNC_EVENT_UNMASK(ctrl, NVME_LOG_CHANGED_NS_LIST_INFO);
        } else if (!log_page_quirk_warn) {
            nvmx_warn("Async event requested while 'namespace change' is pending. "
                      "Consider setting 'ctrl.quirks'");
            log_page_quirk_warn = 1;
        }
    }

    pthread_mutex_unlock(&ctrl->async_event_lock);
    /* Maybe we have pending async events waiting to be informed */
    nvme_ctrl_raise_async_events(ctrl);

    return NVME_SC_NO_COMPLETE;
}

static uint16_t nvme_acmd_fw_download_do(nvme_request_t *req, enum nvme_emu_status rc)
{
    int status;
    off_t position;
    ssize_t bytes_written;
    nvme_ctrl_t *ctrl = req->sq->ctrl;

    if (unlikely(rc != NVMX_OK)) {
        nvmx_error("sq%d: io completion status %d", req->sq->sqid, rc);
        goto err;
    }

    if (ctrl->fw_file == -1) {

        status = fw_get_slot_filename(0, ctrl->filename, sizeof(ctrl->filename));
        if (status)
            goto err;

        ctrl->fw_file = creat(ctrl->filename, S_IRUSR | S_IWUSR);
        if (ctrl->fw_file == -1) {
            nvmx_error("Create file '%s' operation failed with code %d",
                    ctrl->filename, errno);
            goto err;
        }
    }

    position = lseek(ctrl->fw_file, le32_to_cpu(req->fw_offset) << 2U, SEEK_SET);

    if (position == -1) {
        nvmx_error("File seek operation failed with code %d", errno);
        goto err;
    }

    bytes_written = write(ctrl->fw_file, req->fw_buf, req->fw_size);

    if (bytes_written == -1) {
        nvmx_error("File write operation failed with code %d", errno);
        goto err;
    }

    if (bytes_written != req->fw_size) {
        nvmx_error(
                "File write operation failed, written %u bytes instead of %u bytes",
                bytes_written, req->fw_size);
        goto err;
    }

    free(req->fw_buf);
    req->fw_buf = NULL;

    return NVME_SC_SUCCESS;

err:
    free(req->fw_buf);
    req->fw_buf = NULL;

    nvme_discard_image(ctrl);

    return NVME_SC_INTERNAL_DEV_ERROR;
}

static uint16_t nvme_acmd_fw_download(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd, nvme_request_t *req)
{
    int rc;
    size_t count;
    const nvme_cmd_fw_download_t *c = (const nvme_cmd_fw_download_t*) cmd;

    /* Dword to bytes */
    count = (le32_to_cpu(c->numd) + 1U) << 2;

    req->fw_buf = calloc(count, 1);
    if (!req->fw_buf)
        goto err_mem;

    req->fw_size = count;
    req->fw_offset = c->ofst;

    /* read buffer from user */
    rc = nvme_prp_read_bcopy(&req->io_req.async_req, le64_to_cpu(cmd->common.dptr.prp1),
                        le64_to_cpu(cmd->common.dptr.prp2),
                        req->fw_buf, count);

    if (rc != NVMX_INPROGRESS) {
        free(req->fw_buf);
        req->fw_buf = NULL;
    }

    return to_nvme_err(rc);

err_mem:
    return NVME_SC_INTERNAL_DEV_ERROR;
}

static uint16_t nvme_copy_image_to_slot(nvme_ctrl_t *ctrl, uint32_t target_slot_no)
{
    int status;
    char target_slot_name[PATH_MAX];

    status = fw_get_slot_filename(target_slot_no, target_slot_name, sizeof(target_slot_name));
    if (status)
        return NVME_SC_INTERNAL_DEV_ERROR;

    if (ctrl->fw_file != -1) {

        close(ctrl->fw_file);
        ctrl->fw_file = -1;

        status = remove(target_slot_name);

        if (status != 0 && errno != ENOENT) {
            nvmx_error("Remove operation failed with errno: %d", errno);
            return NVME_SC_INTERNAL_DEV_ERROR;
        }

        status = rename(ctrl->filename, target_slot_name);

        if (status != 0) {
            nvmx_error("Rename operation failed with errno: %d", errno);
            return NVME_SC_INTERNAL_DEV_ERROR;
        }

        /* clear and free the cached bfb_info */
        if (ctrl->bfb_info[target_slot_no]) {
            free(ctrl->bfb_info[target_slot_no]);
            ctrl->bfb_info[target_slot_no] = NULL;
        }
    }
    else
        return NVME_SC_INVALID_FW_IMAGE;

    return NVME_SC_SUCCESS;
}

static void nvme_discard_image(nvme_ctrl_t *ctrl)
{
    if (ctrl->fw_file != -1) {

        close(ctrl->fw_file);
        ctrl->fw_file = -1;

        remove(ctrl->filename);
    }
}

static uint16_t nvme_validate_fw_image(nvme_ctrl_t *ctrl, const char *filename, uint32_t slot_no)
{
    nvme_json_t *json;
    nic_utils_t *ctx;
    char *bfb_info;
    bool success, bool_val;
    int fw_image, ret;

    fw_image = open(filename, O_RDONLY);

    if (fw_image == -1)
        return errno == ENOENT ? NVME_SC_INVALID_FW_IMAGE : NVME_SC_INTERNAL_DEV_ERROR;

    close(fw_image);

    ret = NVME_SC_INTERNAL_DEV_ERROR;

    ctx = nic_utils_open();
    if (ctx) {
        fw_image = open(filename, O_RDONLY);
        if (fw_image != -1) {
            close(fw_image);
            bfb_info = nic_utils_exec(ctx, "fw_get_bfb_info", "s", filename);
            if (bfb_info) {
                json = nvme_str_to_json(bfb_info);
                if (json) {
                    if (nvme_json_get_value(json, "success", &success, sizeof(success)) == 0) {
                        if (success) {
                            if (nvme_json_get_value(json, "valid", &bool_val, sizeof(bool_val)))
                                nvmx_error("Failed to retrieve valid flag");
                            ret = bool_val ? NVME_SC_SUCCESS : NVME_SC_INVALID_FW_IMAGE;
                            ctrl->bfb_info[slot_no] = bfb_info;
                        } else nvmx_error("Nic utils operation finished but error occurred");
                    } else nvmx_error("Failed to retrieve success flag");
                    nvme_json_free(json);
                } else nvmx_error("Failed to convert string to json");
            } else nvmx_error("Nic utils operation failed")
        }
        nic_utils_close(ctx);
    }

    return ret;
}

static uint16_t nvme_activate_firmware(nvme_ctrl_t *ctrl, uint32_t slot_no, bool now)
{
    nvme_json_t *json;
    nic_utils_t *ctx;
    char *active_bfb;
    bool success, bool_val;
    int ret;
    char filename[PATH_MAX];

    ret = fw_get_slot_filename(slot_no, filename, sizeof(filename));
    if (ret)
        return NVME_SC_INTERNAL_DEV_ERROR;

    ret = nvme_validate_fw_image(ctrl, filename, slot_no);

    if (ret != NVME_SC_SUCCESS)
        goto err;

    ret = NVME_SC_INTERNAL_DEV_ERROR;

    ctx = nic_utils_open();
    if (ctx) {
        active_bfb = nic_utils_exec(ctx, "fw_activate_bfb", "sb", filename, now);
        if (active_bfb) {
            json = nvme_str_to_json(active_bfb);
            if (json) {
                if (nvme_json_get_value(json, "success", &success, sizeof(success)) == 0) {
                    if (success) {
                        if (nvme_json_get_value(json, "reset_required", &bool_val, sizeof(bool_val)))
                            nvmx_error("Failed to retrieve reset_required flag");
                        ret = bool_val ? NVME_SC_FW_REQ_CONV_RESET : NVME_SC_SUCCESS;
                    } else nvmx_error("Nic utils operation finished but error occurred");
                } else nvmx_error("Failed to retrieve success flag");
                nvme_json_free(json);
            } else nvmx_error("Failed to convert string to json");
        } else nvmx_error("Nic utils operation failed")
        nic_utils_close(ctx);
    }
err:
    return ret;
}

static uint16_t nvme_acmd_fw_commit(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd)
{
    const nvme_cmd_fw_commit_t *c = (const nvme_cmd_fw_commit_t*) cmd;
    uint16_t status;
    int firmware_slot = c->firmware_slot;

    if (firmware_slot == 0) {
        // TODO If the value specified is 0h, then the controller shall choose the firmware slot
        // (i.e., slot 1 to slot 7) to use for the operation
        nvme_discard_image(ctrl);
        return NVME_SC_INVALID_FW_SLOT;
    }

    if (firmware_slot > NVME_FW_SLOTS_COUNT) {
        nvme_discard_image(ctrl);
        return NVME_SC_INVALID_FW_SLOT;
    }

    switch (c->commit_action) {
    case COMMIT_ACTION_DOWNLOAD:
        return nvme_copy_image_to_slot(ctrl, firmware_slot);
    case COMMIT_ACTION_DOWNLOAD_AND_ACTIVATE:
        status = nvme_copy_image_to_slot(ctrl, firmware_slot);
        if (status != NVME_SC_SUCCESS)
            return status;
        return nvme_activate_firmware(ctrl, firmware_slot, false);
    case COMMIT_ACTION_ACTIVATE:
        nvme_discard_image(ctrl);
        return nvme_activate_firmware(ctrl, firmware_slot, false);
    case COMMIT_ACTION_DOWNLOAD_AND_ACTIVATE_NOW:
        (void)nvme_copy_image_to_slot(ctrl, firmware_slot);
        return nvme_activate_firmware(ctrl, firmware_slot, true);
    case COMMIT_ACTION_REPLACE_BOOTP:
    case COMMIT_ACTION_ACTIVATE_BOOTP:
        nvme_discard_image(ctrl);
        return NVME_SC_BOOTP_WRITE_PROHIB;
    default:
        return NVME_SC_INVALID_FIELD;
    }

    return NVME_SC_SUCCESS;
}

static uint16_t nvme_acmd_fw_recover(nvme_ctrl_t *ctrl,
                const nvme_cmd_t *cmd, nvme_request_t *req)
{
    const nvme_cmd_fw_recover_t *c = (const nvme_cmd_fw_recover_t*) cmd;
    int rc, active_slot_no;
    uint8_t fw_ver_in_bfb[64] = {0};
    uint8_t fw_ver_running[64] = {0};
    nvme_json_t *json;
    nic_utils_t *ctx;
    char *recover;
    bool success;

    if (c->pid != 1) {
        nvmx_error("Invalid PID(%d) from fw-recover cmd", c->pid);
        req->cqe.result = 0x7; /* Invalid Package Identifier return 0x7*/
        return NVME_SC_INVALID_FIELD;
    }

    active_slot_no = fw_get_version_in_bfb(ctrl, fw_ver_in_bfb, 64);
    if (active_slot_no < 0) {
        nvmx_error("Failed to get active BFB slot no");
        return NVME_SC_INTERNAL_DEV_ERROR;
    }
    nvmx_debug("NIC FW version in running BFB is:%s", fw_ver_in_bfb);

    rc = fw_get_version_in_running(fw_ver_running, 64);
    if (rc) {
        nvmx_error("Failed to get current running FW version");
        return NVME_SC_INTERNAL_DEV_ERROR;
    }
    nvmx_debug("Current running NIC FW version is:%s", fw_ver_running);

    if (!strcmp((const char *)fw_ver_running, (const char *)fw_ver_in_bfb)) {
        nvmx_error("Running NIC FW version is same as FW version in running BFB");
        /*
         * If the current running FW version is same with
         * the FW version in running BFB image, according
         * to SNAP Bluefield Image update doc, this is a
         * Invalid Request, should return status 0x06.
         */
        req->cqe.result = 0x6;
        return NVME_SC_INTERNAL_DEV_ERROR;
    }

    ctx = nic_utils_open();
    if (!ctx) {
        nvmx_error("Failed to open nic utils");
        return NVME_SC_INTERNAL_DEV_ERROR;
    }

    recover = nic_utils_exec(ctx, "fw_recover", "");
    if (!recover) {
        nvmx_error("Failed to exec 'fw_recover'");
        goto close_utils;
    }

    json = nvme_str_to_json(recover);
    if (!json) {
        nvmx_error("Failed to convert string to json");
        goto close_utils;
    }

    if (nvme_json_get_value(json, "success", &success, sizeof(success))) {
        nvmx_error("Failed to get 'success' flag");
        goto free_json;
    }

    nvmx_info("FW Recover execute %s", success ? "succeed" : "Failed");

    nvme_json_free(json);
    nic_utils_close(ctx);

    /* fw-recover execute success, return 0xB to tell user do Conventional Reset */
    req->cqe.result = 0x0B;
    return NVME_SC_SUCCESS;

free_json:
    nvme_json_free(json);

close_utils:
    nic_utils_close(ctx);

    return NVME_SC_INTERNAL_DEV_ERROR;
}

#if HAVE_SNAP
static uint16_t nvme_check_vs_json_rpc_fields(nvme_ctrl_t *ctrl,
                                              const nvme_cmd_t *cmd)
{
    const nvme_vendor_specific_cmd_t *c = (nvme_vendor_specific_cmd_t *)cmd;

    if (!ctrl->rpc_client) {
        nvmx_error("vendor specific json rpc commands not supported\n");
        return NVME_SC_INVALID_OPCODE | NVME_SC_DNR;
    }

    /* Currently we support only PRP based data transfers */
    if (cmd->common.psdt != NVME_PSDT_PRP) {
        nvmx_error("wrong psdt for vs command 0x%x\n", cmd->common.psdt);
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    /* Currently we don't support FUSED operations */
    if (cmd->common.fuse) {
        nvmx_error("wrong fuse value for vs command 0x%x\n", cmd->common.fuse);
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    /* This command shouldn't apply on namespaces */
    if (cmd->common.nsid) {
        nvmx_error("wrong nsid value for vs command 0x%x\n", cmd->common.nsid);
        return NVME_SC_INVALID_NSID | NVME_SC_DNR;
    }

    if (cmd->common.mptr) {
        nvmx_error("wrong metadata pointer value for vs command 0x%x\n",
                   cmd->common.mptr);
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    if (!c->ndt) {
        nvmx_error("wrong ndt value for vs command 0x%x\n", c->ndt);
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    if (c->ndm) {
        nvmx_error("wrong ndm value for json rpc vs command 0x%x\n",
                   le32_to_cpu(c->ndm));
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    return NVME_SC_SUCCESS;
}

static int nvme_acmd_vs_json_rpc_req_do(nvme_request_t *req, enum nvme_emu_status rc)
{
    nvme_ctrl_t *ctrl = req->sq->ctrl;

    if (unlikely(rc != NVMX_OK)) {
        free(req->rpc_buf);
        req->rpc_buf = NULL;
        nvmx_error("sq%d: io completion status %d", req->sq->sqid, rc);
        return NVME_SC_INTERNAL_DEV_ERROR;
    }

    if (ctrl->rpc_busy) {
        TAILQ_INSERT_TAIL(&ctrl->rpc_pending_list, req, rpc_entry);
        return NVME_SC_SUCCESS;
    }

    ctrl->rpc_busy = true;

    sem_wait(&ctrl->rpc_thread_idle);
    ctrl->rpc_thread_req = req;
    sem_post(&ctrl->rpc_thread_kick);

    return NVME_SC_SUCCESS;
}

static uint16_t
nvme_acmd_vs_json_rpc_req(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd,
                          nvme_request_t *req)
{
    const nvme_vendor_specific_cmd_t *c = (nvme_vendor_specific_cmd_t *)cmd;
    size_t count;
    uint16_t status;
    int rc;

    if (!ctrl->rpc_server)
        return NVME_SC_INVALID_OPCODE | NVME_SC_DNR;

    status = nvme_check_vs_json_rpc_fields(ctrl, cmd);
    if (status != NVME_SC_SUCCESS)
        return status;

    /* Dword to bytes */
    count = le32_to_cpu(c->ndt) << 2;

    req->rpc_buf = malloc(count);
    if (!req->rpc_buf) {
        nvmx_error("json rpc req failed to alloc memory");
        return NVME_SC_INTERNAL_DEV_ERROR;
    }

    /* read buffer from user */
    rc = nvme_prp_read_bcopy(&req->io_req.async_req, le64_to_cpu(cmd->common.dptr.prp1),
                        le64_to_cpu(cmd->common.dptr.prp2),
                        req->rpc_buf, count);

    if (rc != NVMX_INPROGRESS) {
        free(req->rpc_buf);
        req->rpc_buf = NULL;
    }

    return to_nvme_err(rc);
}

static uint16_t
nvme_acmd_vs_json_rpc_rsp(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd,
                          nvme_request_t *req)
{
    const nvme_vendor_specific_cmd_t *c = (nvme_vendor_specific_cmd_t *)cmd;
    size_t count;
    int rc;
    uint16_t status;
    char *rsp_buf;

    if (!ctrl->rpc_server)
        return NVME_SC_INVALID_OPCODE | NVME_SC_DNR;

    status = nvme_check_vs_json_rpc_fields(ctrl, cmd);
    if (status != NVME_SC_SUCCESS)
        return status;

    if (!ctrl->rpc_rsp) {
        nvmx_error("json rpc response is not found");
        return NVME_SC_INTERNAL_DEV_ERROR;
    }

    if ((ctrl->rpc_rsp->length + 3) >> 2 != le32_to_cpu(c->ndt)) {
        nvmx_error("host buffer for json rpc response is not matched");
        return NVME_SC_INVALID_FIELD;
    }

    /* Dword to bytes */
    count = le32_to_cpu(c->ndt) << 2;

    /*
     * rsp rsp->buf may be smaller than buffer size the user expects, so it
     * might not fit well into PRP conventions (using PRP2 as a data
     * rather than a prp list).To avoid this, we better copy our cached
     * output into a buffer which matches user's expectations
     */
    rsp_buf = calloc(count, sizeof(uint8_t));
    if (!rsp_buf) {
        nvmx_error("Failed to allocate resp buffer");
        return NVME_SC_INTERNAL_DEV_ERROR;
    }
    memcpy(rsp_buf, ctrl->rpc_rsp->buf, ctrl->rpc_rsp->length);

    nvmx_debug("JSON-RPC response:\n%s", rsp_buf);
    /* write buffer to the host */
    rc = nvme_prp_write_bcopy(&req->io_req.async_req, le64_to_cpu(cmd->common.dptr.prp1),
                                  le64_to_cpu(cmd->common.dptr.prp2), rsp_buf, count);

    free(rsp_buf);

    return to_nvme_err(rc);
}
#endif

static uint16_t
nvme_acmd_iova_mgmt(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd,
                          nvme_request_t *req)
{
    int rc;
    size_t size;
    struct effdma_domain *domain;
    address_space_t *address_space;
    const nvme_cmd_vs_iova_mgmt_t *iova_mgmt_cmd = (nvme_cmd_vs_iova_mgmt_t *)cmd;
    const int pf_id = snap_config_get_pci_func(ctrl->sconfig);
    const int vf_id = (int)iova_mgmt_cmd->fid - 1;

    if (vf_id >= (int)ctrl->dev->dev_caps.max_emulated_vfs)
        return NVME_IOVA_INVALID_FUNCTION_ID;

    domain = effdma_domain_find(pf_id, vf_id);
    if (!domain) {
        nvmx_error("Couldn't find memory domain");
        return NVME_SC_INTERNAL_DEV_ERROR;
    }

    address_space = effdma_domain_transaction_begin(domain);
    if (!address_space) {
        nvmx_error("Couldn't start update transaction");
        return NVME_SC_INTERNAL_DEV_ERROR;
    }

    switch (iova_mgmt_cmd->opm) {
    case NVME_IOVA_OPM_UNMAP_ALL:
        rc = address_space_unmap_all(address_space);
        break;
    case NVME_IOVA_OPM_UNMAP_RANGE:
        size = effdma_normalize_size(iova_mgmt_cmd->size, iova_mgmt_cmd->szu);
        rc = address_space_unmap(address_space, iova_mgmt_cmd->siova, size);
        break;
    case NVME_IOVA_OPM_MAP_RANGE:
        if (iova_mgmt_cmd->szu >= NVME_IOVA_SZU_RESERVED) {
            nvmx_error("Unsupported szu value %u", iova_mgmt_cmd->szu);
            return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
        }
        size = effdma_normalize_size(iova_mgmt_cmd->size, iova_mgmt_cmd->szu);
        rc = address_space_map(address_space, iova_mgmt_cmd->siova,
                                              iova_mgmt_cmd->tiova,
                                              size);
        break;
    default:
        nvmx_error("Unsupported opm %u", iova_mgmt_cmd->opm);
        return NVME_IOVA_UNSUPPORTED_MODE_TYPE;
    }

    if (rc) {
        effdma_domain_transaction_abort(domain);
        switch (rc) {
        case -EINVAL:
            return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
        default:
            return NVME_SC_INTERNAL_DEV_ERROR;
        }
    }

    effdma_domain_transaction_commit(domain);
    return NVME_SC_SUCCESS;
}

static uint16_t nvme_admin_cmd(nvme_ctrl_t *ctrl, const nvme_cmd_t *cmd,
        nvme_request_t *req)
{
    nvme_admin_cmd_dump(cmd);

    if (unlikely(ctrl->state != NVME_CTRL_STARTED)) {
        nvmx_warn("opcode: 0x%x got admin command on controller %d teardown",
                  cmd->common.opc, snap_config_get_cntlid(ctrl->sconfig));
        return NVME_SC_INTERNAL_DEV_ERROR | NVME_SC_DNR;
    }

    switch (cmd->common.opc) {
    case NVME_ADMIN_DELETE_SQ:
        return nvme_acmd_delete_sq(ctrl, cmd, req);
    case NVME_ADMIN_CREATE_SQ:
        return nvme_acmd_create_sq(ctrl, cmd);
    case NVME_ADMIN_DELETE_CQ:
        return nvme_acmd_delete_cq(ctrl, cmd);
    case NVME_ADMIN_CREATE_CQ:
        return nvme_acmd_create_cq(ctrl, cmd);
    case NVME_ADMIN_IDENTIFY:
        return nvme_acmd_identify(ctrl, cmd, req);
    case NVME_ADMIN_ABORT:
        return nvme_acmd_abort(req);
    case NVME_ADMIN_SET_FEATURES:
        return nvme_acmd_set_feature(ctrl, cmd, req);
    case NVME_ADMIN_GET_FEATURES:
        return nvme_acmd_get_feature(ctrl, cmd, req);
    case NVME_ADMIN_GET_LOG_PAGE:
        return nvme_acmd_get_log_page(ctrl, cmd, req);
    case NVME_ADMIN_ASYNC_EV_REQ:
        return nvme_acmd_async_ev(ctrl, req);
    case NVME_ADMIN_FW_DOWNLOAD:
        if (ctrl->fw_update_enable)
            return nvme_acmd_fw_download(ctrl, cmd, req);
        break;
    case NVME_ADMIN_FW_COMMIT:
        if (ctrl->fw_update_enable)
            return nvme_acmd_fw_commit(ctrl, cmd);
        break;
    case NVME_ADMIN_FW_RECOVER:
        return nvme_acmd_fw_recover(ctrl, cmd, req);
#if HAVE_SNAP
    case NVME_ADMIN_VS_JSON_RPC_2_0_REQ:
        return nvme_acmd_vs_json_rpc_req(ctrl, cmd, req);
    case NVME_ADMIN_VS_JSON_RPC_2_0_RSP:
        return nvme_acmd_vs_json_rpc_rsp(ctrl, cmd, req);
#endif
    case NVME_ADMIN_VS_IOVA_MGMT:
        if (ctrl->iova_mgmt_enable)
            return nvme_acmd_iova_mgmt(ctrl, cmd, req);
        break;
    }

    nvmx_warn("opcode: 0x%x unsupported admin command", cmd->common.opc);
    return NVME_SC_INVALID_OPCODE | NVME_SC_DNR;
}

static inline bool
nvme_rw_cmd_exceeds_mdts(nvme_ctrl_t *ctrl, nvme_namespace_t *ns,
                         const nvme_cmd_rw_t *rw)
{
    const uint64_t effective_size = (rw->nlb + 1) * (1 << ns->block_order);
    const uint64_t mdts_cap = (1 << ctrl->id_ctrl.mdts) * ctrl->page_size_min;

    if (ctrl->id_ctrl.mdts == 0)
        return false;

    return effective_size > mdts_cap;
}

static inline uint16_t nvme_rw(nvme_ctrl_t *ctrl, nvme_namespace_t *ns,
        int sqid, const nvme_cmd_t *cmd, nvme_request_t *req, uint8_t opcode)
{
    const nvme_cmd_rw_t *rw = (nvme_cmd_rw_t *)cmd;
    int rc;

    // nlb + 1 for zero-based value
    if (unlikely(rw->slba + (rw->nlb + 1) > ns->id_ns.nsze || rw->slba > UINT64_MAX - (rw->nlb + 1))) {
        nvmx_error("IO sent with invalid range");
        return NVME_SC_LBA_RANGE;
    }

    if (unlikely(nvme_rw_cmd_exceeds_mdts(ctrl, ns, rw))) {
        nvmx_error("IO sent exceeds MDTS capacity");
        return NVME_SC_INVALID_FIELD;
    }

    if (cmd->common.psdt == NVME_PSDT_PRP) {
        if (opcode == NVME_IO_READ)
            rc = nvme_driver_read_prp_nb(&req->io_req, rw);
        else
            rc = nvme_driver_write_prp_nb(&req->io_req, rw);
    } else if (cmd->common.psdt == NVME_PSDT_SGL_MPTR_CONTIG) {
        if (opcode == NVME_IO_READ)
            rc = nvme_driver_read_sgl_nb(&req->io_req, rw);
        else
            rc = nvme_driver_write_sgl_nb(&req->io_req, rw);
    } else {
        nvmx_error("wrong psdt for 0x%x\n", cmd->common.psdt);
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    /* io started but was not completed */
    if (rc == NVMX_INPROGRESS)
        return NVME_SC_NO_COMPLETE;

    if (rc) {
        int ret = NVME_SC_INTERNAL_DEV_ERROR;

        if (rc != NVMX_ERR_NOTREADY)
            ret |= NVME_SC_DNR;

        return ret;
    }

    return NVME_SC_SUCCESS;
}

static uint16_t nvme_flush(nvme_ctrl_t *ctrl, nvme_namespace_t *ns,
                           const nvme_cmd_t *cmd, nvme_request_t *req)
{
    uint32_t nsid = le32_to_cpu(cmd->common.nsid);

    nvmx_debug("nvme flush command");

    if (ctrl->id_ctrl.vwc.flush_broadcast == 0x2 && nsid == 0xFFFFFFFF)
        return NVME_SC_INVALID_NSID;

    return NVME_SC_SUCCESS;
}

static uint16_t nvme_io_cmd(nvme_ctrl_t *ctrl, unsigned sqid, const nvme_cmd_t *cmd,
        nvme_request_t *req, bool fabrics)
{
    nvme_namespace_t *ns;
    uint32_t nsid = le32_to_cpu(cmd->common.nsid);
    int retry_count = NVME_CTRL_NRETRIES;
    int rc;
    uint64_t len;

    nvme_io_cmd_dump(cmd);

    if (unlikely(nsid > ctrl->id_ctrl.nn)) {
        nvmx_error("IOQ %d got R/W command with invalid NSID", sqid, nsid);
        return NVME_SC_INVALID_NSID;
    }

    /*
     * Do not take namespace lock in the multi threaded mode. Instead
     * when namespace list is modified, all polling groups will be stopped.
     *
     * However we must take the lock in the single threaded mode because
     * namespace list modification can happen from other thread context.
     */
    if (likely(ctrl->npgs))
        ns = nvme_find_namespace_nolock(ctrl, nsid);
    else
        ns = nvme_find_namespace(ctrl, nsid);

    if (!unlikely(ns)) {
        nvmx_error("IOQ %d got R/W command with inactive NSID", sqid, nsid);
        return NVME_SC_INVALID_FIELD;
    }

    req->io_req.ns = ns->backend_ns;

    if (fabrics) {
        if (cmd->common.psdt != NVME_PSDT_SGL_MPTR_CONTIG) {
            nvmx_error("wrong psdt for fabrics 0x%x\n", cmd->common.psdt);
            return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
        }
        if (cmd->common.dptr.sgl.type != NVME_SGL_TYPE_KEYED_DATA_BLOCK) {
            nvmx_error("wrong type for fabrics 0x%x\n", cmd->common.dptr.sgl.type);
            return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
        }

        if (cmd->common.dptr.sgl.subtype != NVME_SGL_SUBTYPE_ADDRESS) {
            nvmx_error("wrong subtype for fabrics 0x%x\n",
                       cmd->common.dptr.sgl.subtype);
            return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
        }
    }

    /* Currently we support only keyed SGL for fabrics */
    if (!fabrics && cmd->common.psdt != NVME_PSDT_PRP && cmd->common.psdt != NVME_PSDT_SGL_MPTR_CONTIG) {
        nvmx_error("wrong psdt for non fabrics 0x%x\n", cmd->common.psdt);
        return NVME_SC_INVALID_FIELD | NVME_SC_DNR;
    }

    req->ns_stat = NULL;
    req->ctrl_stat = NULL;

retry:
    switch (cmd->common.opc) {
    case NVME_IO_FLUSH:
        req->ns_stat = &ns->sq_stat[sqid].flush;
        req->ctrl_stat = &ctrl->sq_stat[sqid].flush;
        rc = nvme_flush(ctrl, ns, cmd, req);
        break;
#if 0
    case NVME_IO_WRITE_ZEROS:
        rc = nvme_write_zeros(ctrl, ns, cmd, req);
        break;
#endif
    case NVME_IO_WRITE:
        req->ns_stat = &ns->sq_stat[sqid].write;
        req->ctrl_stat = &ctrl->sq_stat[sqid].write;
        rc = nvme_rw(ctrl, ns, sqid, cmd, req, NVME_IO_WRITE);
        break;
    case NVME_IO_READ:
        req->ns_stat = &ns->sq_stat[sqid].read;
        req->ctrl_stat = &ctrl->sq_stat[sqid].read;
        rc = nvme_rw(ctrl, ns, sqid, cmd, req, NVME_IO_READ);
        break;
    case NVME_IO_DSM:
        return NVME_SC_INTERNAL_DEV_ERROR | NVME_SC_DNR;
    default:
        return NVME_SC_INVALID_OPCODE | NVME_SC_DNR;
    }

    /*
     * It's possible that ep is not passed into the active state yet
     * by the driver thread, so let's try a couple of times
     * in case DNR is not set.
     */
    if (rc == NVME_SC_INTERNAL_DEV_ERROR && retry_count) {
        --retry_count;
        usleep(NVME_CTRL_WAIT_USECS);
        goto retry;
    }

    if (req->ns_stat && req->ctrl_stat) {
        req->ns_stat->total++;
        req->ctrl_stat->total++;
        if (rc == NVME_SC_SUCCESS) {
            len = req->io_req.async_req.len;
            req->ns_stat->completed++;
            req->ns_stat->completed_bytes_count += len;
            req->ns_stat->completed_units_count += ROUND_UP(len, 512) / 512;
            req->ctrl_stat->completed++;
            req->ctrl_stat->completed_bytes_count += len;
            req->ctrl_stat->completed_units_count += ROUND_UP(len, 512) / 512;
        }
        else if (rc != NVME_SC_NO_COMPLETE) {
            req->ns_stat->err++;
            req->ctrl_stat->err++;
        }
    }

    return rc;
}

static inline nvme_squeue_t *get_comp_sq(nvme_cqueue_t *cq, nvme_squeue_t *sq)
{
    nvme_squeue_t *comp_sq;

    /* in nvme2nvmf mode completions must be always posted
     * on the original sq */
    if (cq->ops.q_type != NVME_EMU_QUEUE_SQE)
        return sq;

    /* Always post cqes on the first SQ attached to the CQ.
     * This way we make sure that cqes are not reordered.
     *
     * NOTE: this is unfair to others SQs that are attached to
     * the same CQ.
     */
    comp_sq = TAILQ_FIRST(&cq->sq_list);
    if (unlikely(!comp_sq))
        nvmx_fatal("CQ %d has no SQs attached and cannot post completions", cq->cqid);
    return comp_sq;
}

static inline int nvme_post_cqe(nvme_cqueue_t *cq, nvme_request_t *req)
{
    nvme_squeue_t *sq;
    nvme_squeue_t *comp_sq;
    uint16_t newhead;
    int ret;

    sq = req->sq;
    comp_sq = get_comp_sq(cq, sq);
    if (comp_sq == NULL)
        return NVMX_ERR;

    /* write the phase bit only for RAW modes */
    if (cq->ops.q_type == NVME_EMU_QUEUE_SQE)
        req->cqe.status = cpu_to_le16((req->status << 1) | cq->phase);
    else
        req->cqe.status = cpu_to_le16((req->status << 1));

    /* Advance sq head only when request is completed and we are
     * posting cqe. This ensures that sq->size requests will be enough.
     * Otherwise number of needed requests will be bound by the io backend
     * queue size and service time. It may be much bigger than the
     * actual SQ size.
     */
    newhead = nvme_sq_inc_head(sq);
    req->cqe.sq_id = cpu_to_le16(sq->sqid);
    req->cqe.sq_head = cpu_to_le16(newhead);

    ret = nvme_emu_send(&comp_sq->ops, (void *)&req->cqe, sizeof(req->cqe));
    if (ret != NVMX_OK)
        return ret;

    nvme_cq_inc_tail(cq);
    sq->head = newhead;
    req->io_req.async_req.len = 0;

    nvmx_debug("CQ %d posted completion: tag %d status 0x%x phase %d sq_id %d sq_head 0x%x",
               cq->cqid, req->cqe.cid, req->status, le16_to_cpu(req->cqe.status) & 0x1 ,
               req->cqe.sq_id, req->cqe.sq_head);

    return NVMX_OK;
}

static inline int nvme_post_cqes(nvme_cqueue_t *cq, nvme_squeue_t *sq)
{
    int n = 0;
    nvme_request_t *req, *next;
    int ret;

    TAILQ_FOREACH_SAFE(req, &cq->req_list, entry, next) {
        if (unlikely(sq && req->sq != sq))
            continue;
        ret = nvme_post_cqe(cq, req);
        if (ret) {
            if (unlikely(ret != NVMX_ERR_NOTREADY))
                nvmx_error("CQ %d failed to post cqe", cq->cqid);
            return n;
        }
        TAILQ_REMOVE(&cq->req_list, req, entry);
        TAILQ_INSERT_HEAD(&req->sq->req_list, req, entry);
        req = TAILQ_FIRST(&cq->req_list);
        n++;
    }
    return n;
}

static inline void nvme_req_push_to_cq(nvme_cqueue_t *cq, nvme_request_t *req)
{
    int ret;

    nvmx_assertv_always(cq->cqid == req->sq->cqid, "bad request");

    /*
     * Try to post if there are no others pending cqes.
     * If posting fails add to the pending list
     */
    if (unlikely(!TAILQ_EMPTY(&cq->req_list))) {
        TAILQ_INSERT_TAIL(&cq->req_list, req, entry);
        return;
    }

    ret = nvme_post_cqe(cq, req);
    if (ret) {
        if (unlikely(ret != NVMX_ERR_NOTREADY))
            nvmx_error("CQ %d failed to post cqe", cq->cqid);
        TAILQ_INSERT_TAIL(&cq->req_list, req, entry);
        return;
    }

    TAILQ_INSERT_HEAD(&req->sq->req_list, req, entry);
}

static void nvme_enqueue_req_completion(nvme_cqueue_t *cq, nvme_request_t *req)
{
    TAILQ_REMOVE(&req->sq->out_req_list, req, entry);
    nvme_req_push_to_cq(cq, req);
}

/* Must be called under ctrl->async_event_lock */
static inline nvme_request_t *
nvme_ctrl_peek_async_event_request(nvme_ctrl_t *ctrl)
{
    return TAILQ_FIRST(&ctrl->async_event_reqs);
}

/* Must be called under ctrl->async_event_lock */
static nvme_request_t *nvme_ctrl_pop_async_event_request(nvme_ctrl_t *ctrl)
{
    nvme_request_t *req;

    ctrl->num_aers--;
    req = TAILQ_FIRST(&ctrl->async_event_reqs);
    nvmx_assertv_always(req, "No Async events credits left");
    TAILQ_REMOVE(&ctrl->async_event_reqs, req, entry);
    return req;
}

/* Must be called under ctrl->async_event_lock */
static void nvme_ctrl_push_async_event_request(nvme_ctrl_t *ctrl,
                                               nvme_request_t *req)
{
    TAILQ_INSERT_HEAD(&ctrl->async_event_reqs, req, entry);
    ctrl->num_aers++;
}

static void nvme_ctrl_abort_async_event_requests(nvme_ctrl_t *ctrl)
{
    nvme_request_t *req;
    nvme_cqueue_t *cq;

    pthread_mutex_lock(&ctrl->async_event_lock);
    while (nvme_ctrl_peek_async_event_request(ctrl) != NULL) {
        req = nvme_ctrl_pop_async_event_request(ctrl);

        cq = ctrl->cq[req->sq->cqid];
        req->status = NVME_SC_CMD_ABORT_REQ;
        req->is_async_event = false;
        nvme_req_push_to_cq(cq, req);
    }
    pthread_mutex_unlock(&ctrl->async_event_lock);
}

/* Must be called under ctrl->async_event_lock */
static inline size_t nvme_ctrl_num_async_event_requests(nvme_ctrl_t *ctrl)
{
    return ctrl->num_aers;
}

static void nvme_ctrl_complete_async_event(nvme_ctrl_t *ctrl, uint32_t result)
{
    nvme_request_t *req;
    nvme_cqueue_t *cq;

    req = nvme_ctrl_pop_async_event_request(ctrl);
    cq = ctrl->cq[req->sq->cqid];
    req->status = NVME_SC_SUCCESS;
    req->cqe.result = cpu_to_le32(result);
    /* we are under async_event lock here, safe to change state */
    req->is_async_event = false;
    nvme_req_push_to_cq(cq, req);
}

static void nvme_ctrl_raise_async_event(nvme_ctrl_t *ctrl,
                                        enum NvmeAsyncEventConfig event)
{
    uint32_t result = 0;
    uint8_t log_page_id;
    char *ev_str="UNKNOWN";
    nvme_request_t *req;

    switch (event) {
    case NVME_NAMESPACE_ATTR_NOTICE:
        ev_str = "Namespace Change";
        log_page_id = NVME_LOG_CHANGED_NS_LIST_INFO;
        result = ASYNC_REQ_RESULT(NVME_ASYNC_EVENT_TYPE_NOTICE,
                                  NVME_ASYNC_EVENT_NS_ATTR_CHANGED,
                                  NVME_LOG_CHANGED_NS_LIST_INFO);
        break;
    case NVME_TEMPERATURE_ERR:
        ev_str = "Temperature Warning";
        log_page_id = NVME_LOG_SMART_INFO;
        result = ASYNC_REQ_RESULT(NVME_ASYNC_EVENT_TYPE_SMART,
                                  NVME_ASYNC_EVENT_TEMPERATURE_THRESHOLD,
                                  NVME_LOG_SMART_INFO);
        break;
    default:
        nvmx_warn("Unknown Async Event 0x%x is raised", event);
        return;
    }

    if (ASYNC_EVENT_IS_MASKED(ctrl, log_page_id)) {
        nvmx_info("ASYNC EVENT %s (0x%x) is masked", ev_str, event);
        return;
    }

    req = nvme_ctrl_peek_async_event_request(ctrl);
    nvmx_info("ASYNC EVENT %s (0x%x) tag %d result 0x%x", ev_str, event,
              req->cqe.cid, result);

    ASYNC_EVENT_MASK(ctrl, log_page_id);
    nvme_ctrl_complete_async_event(ctrl, result);
    ctrl->pending_aens &= ~event;
}

static void nvme_ctrl_raise_async_events(nvme_ctrl_t *ctrl)
{
    enum NvmeAsyncEventConfig event;

    pthread_mutex_lock(&ctrl->async_event_lock);
    event = 1;
    while (event && !TAILQ_EMPTY(&ctrl->async_event_reqs)) {
        if ((event & ctrl->async_event_config & ctrl->pending_aens)
            && !(event & ctrl->async_event_mask))
            nvme_ctrl_raise_async_event(ctrl, event);
        event <<= 1;
    }
    pthread_mutex_unlock(&ctrl->async_event_lock);
}

static void nvme_ctrl_queue_async_event(nvme_ctrl_t *ctrl,
                                        enum NvmeAsyncEventConfig event)
{
    pthread_mutex_lock(&ctrl->async_event_lock);
    ctrl->pending_aens |= event;
    pthread_mutex_unlock(&ctrl->async_event_lock);
    nvme_ctrl_raise_async_events(ctrl);
}

static void nvme_command_dispatch(nvme_squeue_t *sq, const nvme_cmd_t *cmd,
                                  bool fabrics)
{
    nvme_ctrl_t *ctrl = sq->ctrl;
    nvme_cqueue_t *cq = ctrl->cq[sq->cqid];
    uint16_t status;
    nvme_request_t *req;

    req = TAILQ_FIRST(&sq->req_list);
    if (unlikely(!req)) {
        /* driver will see command timeout and eventually will
         * either destroy the SQ or reset the whole controller */
        nvmx_error("SQ %d is out of requests, command dropped", sq->sqid);
        return;
    }

    if (unlikely(sq->state == NVME_SQ_FLUSHING)) {
        nvmx_debug("drop req %p id %d", req, req->io_req.id);
        if (TAILQ_EMPTY(&sq->out_req_list))
            nvme_post_cqes(cq, sq);
        return;
    }

    TAILQ_REMOVE(&sq->req_list, req, entry);
    TAILQ_INSERT_TAIL(&sq->out_req_list, req, entry);
    memset(&req->cqe, 0, sizeof(req->cqe));
    req->cqe.cid = cmd->common.cid;
    req->io_req.async_req.thread_id = sq->thread_id;

    status = likely(sq->sqid) ? nvme_io_cmd(ctrl, sq->sqid, cmd, req, fabrics) :
        nvme_admin_cmd(ctrl, cmd, req);
    /* optimize for async io which does not complete immediately */
    if (unlikely(status != NVME_SC_NO_COMPLETE)) {
        req->status = status;
        nvme_enqueue_req_completion(cq, req);
    }
}

static void nvme_free_bfb_info(nvme_ctrl_t *ctrl)
{
    int i;

    for (i = 0; i <= NVME_FW_SLOTS_COUNT; i++) {
        if (ctrl->bfb_info[i])
            free(ctrl->bfb_info[i]);
    }
}

static void nvme_ctrl_pgs_free(nvme_ctrl_t *ctrl)
{
    int i;

    if (ctrl->npgs == 0)
        return;

    for (i = 0; i < ctrl->npgs; i++)
        pthread_spin_destroy(&ctrl->pgs[i].lock);

    free(ctrl->pgs);

    nvme_pg_ref_count--;
    if (!nvme_pg_ref_count) {
        free(nvme_pg_usage);
        nvme_pg_usage = NULL;
    }
}

static int nvme_ctrl_pgs_alloc(nvme_ctrl_t *ctrl, int nthreads)
{
    int i;

    ctrl->pgs = calloc(nthreads, sizeof(struct nvme_ctrl_pg));
    if (!ctrl->pgs)
        return NVMX_ERR;

    if (!nvme_pg_usage) {
        nvme_pg_usage = calloc(nthreads, sizeof(*nvme_pg_usage));
        if (!nvme_pg_usage) {
            free(ctrl->pgs);
            return NVMX_ERR;
        }
        nvme_pg_ref_count = 0;
    }
    nvme_pg_ref_count++;

    ctrl->npgs = nthreads;
    for (i = 0; i < nthreads; i++) {
        pthread_spin_init(&ctrl->pgs[i].lock, PTHREAD_PROCESS_PRIVATE);
        TAILQ_INIT(&ctrl->pgs[i].sq_list);
    }
    return NVMX_OK;
}

void nvme_ctrl_pgs_suspend(nvme_ctrl_t *ctrl)
{
    int i;

    for (i = 0; i < ctrl->npgs; i++)
        pthread_spin_lock(&ctrl->pgs[i].lock);
}

void nvme_ctrl_pgs_resume(nvme_ctrl_t *ctrl)
{
    int i;

    for (i = ctrl->npgs - 1; i >= 0; i--)
        pthread_spin_unlock(&ctrl->pgs[i].lock);
}

static void nvme_ctrl_sq_sched(nvme_ctrl_t *ctrl, nvme_squeue_t *sq)
{
    struct nvme_ctrl_pg *pg;
    nvme_cqueue_t *cq;
    size_t pg_index = 0, i;

    if (ctrl->npgs <= 0)
        return;

    if (sq->sqid == 0)
        return;

	/*
	 * It is possible that cq is already used by another
	 * sq. In that case we must schedule both sqs on the
	 * same polling group. Doing this will eliminate need
	 * of the cq lock.
	 */
    cq = ctrl->cq[sq->cqid];
    if (!cq->pg) {
        for (i = 1; i < ctrl->npgs; i++)
            if (nvme_pg_usage[i] < nvme_pg_usage[pg_index])
                pg_index = i;
        pg = &ctrl->pgs[pg_index];
        cq->pg = pg;
    } else {
        pg = cq->pg;
    }

    pthread_spin_lock(&pg->lock);
    TAILQ_INSERT_TAIL(&pg->sq_list, sq, pg_entry);
    sq->pg = pg;
    nvme_pg_usage[sq->pg - sq->ctrl->pgs]++;
    pthread_spin_unlock(&pg->lock);
}

static void nvme_ctrl_sq_desched(nvme_squeue_t *sq)
{
    struct nvme_ctrl_pg *pg = sq->pg;

    if (!pg)
        return;

    if (sq->sqid == 0)
        return;

    pthread_spin_lock(&pg->lock);
    TAILQ_REMOVE(&pg->sq_list, sq, pg_entry);
    nvme_pg_usage[sq->pg - sq->ctrl->pgs]--;
    sq->pg = NULL;
    pthread_spin_unlock(&pg->lock);
}

static inline int nvme_ctrl_progress_pending_cqes(nvme_squeue_t *sq)
{
    nvme_cqueue_t *cq;

    cq = sq->ctrl->cq[sq->cqid];
    /* if sq is also used to post cqes then try to push pending cqes.
     * Optimize for the case when there is a single sq per cq
     */
    if (likely(sq == get_comp_sq(cq, sq)))
        return nvme_post_cqes(cq, NULL);

    return 0;
}

int nvme_ctrl_progress_io(nvme_ctrl_t *ctrl, int thread_id)
{
    int n = 0;
    struct nvme_ctrl_pg *pg;
    nvme_squeue_t *sq;

    pg = &ctrl->pgs[thread_id];

    pthread_spin_lock(&pg->lock);
    TAILQ_FOREACH(sq, &pg->sq_list, pg_entry) {
        sq->thread_id = thread_id;
        n += nvme_emu_progress_sq(&sq->ops);
        n += nvme_ctrl_progress_pending_cqes(sq);
    }
    pthread_spin_unlock(&pg->lock);

    return n;
}

void nvme_ctrl_query_caps(nvme_ctrl_t *ctrl, struct nvme_ctrl_caps *caps)
{
    snprintf(caps->version, 16, "%u.%u.%u", ctrl->id_ctrl.ver.bits.mjr,
             ctrl->id_ctrl.ver.bits.mnr, ctrl->id_ctrl.ver.bits.ter);
    caps->max_nsid = ctrl->id_ctrl.nn;
    caps->max_namespaces = nvmx_min(ctrl->id_ctrl.mnan ? : ctrl->id_ctrl.nn,
                                snap_config_get_max_namespaces(ctrl->sconfig));
    caps->offload = snap_config_get_offload(ctrl->sconfig);
    caps->mempool = (ctrl->memp != NULL);
}

void nvme_ctrl_mask_unmask_async_event(nvme_ctrl_t *ctrl,
                enum NvmeAsyncEventConfig event, bool mask)
{
    pthread_mutex_lock(&ctrl->async_event_lock);
    if (mask)
        ctrl->async_event_mask |= event;
    else
        ctrl->async_event_mask &= ~event;
    pthread_mutex_unlock(&ctrl->async_event_lock);

    if (!mask)
        nvme_ctrl_raise_async_events(ctrl);
}
