#include "config.h"
#include "json/json.h"
#include "json/nvme_json.h"
#include "nvme_emu_ep.h"
#include "nvme_emu_io_driver.h"
#include "nvme_prp.h"
#include "nvme_sgl.h"
#include "snap_conf.h"

#define CROSS_RDMA_DEV_SUPPORT_ENV "CROSS_RDMA_DEV_SUPPORT"

static int nvme_emu_driver_ep_index(nvme_emu_ep_t *ep)
{
    return ep - ep->io_driver->endpoints;
}

/*
 * Must be called with ep_lock held
 */
static nvme_emu_ep_t *nvme_driver_get_active_ep(nvme_emu_driver_t *driver)
{
    if (likely(driver->active_endpoint >= 0 &&
               driver->active_endpoint < driver->nr_endpoints))
        return &driver->endpoints[driver->active_endpoint];
    else
        return NULL;
}

/*
 * Must be called with ep_lock held
 */
static int nvme_emu_driver_change_ep_state(nvme_emu_ep_t *ep,
                                           enum nvme_emu_ep_state new_state)
{
    nvme_emu_driver_t *driver = ep->io_driver;
    enum nvme_emu_ep_state old_state = ep->state;

    switch (old_state) {
    case NVME_EMU_EP_NEW:
        switch (new_state) {
        case NVME_EMU_EP_ADMIN_START:
            ep->state = new_state;
            break;
        default:
            goto out_err;
        }
        break;
    case NVME_EMU_EP_ADMIN_START:
        switch (new_state) {
        case NVME_EMU_EP_START:
        /* fallthrough */
        case NVME_EMU_EP_DELETING:
            ep->state = new_state;
            break;
        default:
            goto out_err;
        }
        break;
    case NVME_EMU_EP_START:
        switch (new_state) {
        case NVME_EMU_EP_ACTIVE:
            driver->active_endpoint = nvme_emu_driver_ep_index(ep);
        /* fallthrough */
        case NVME_EMU_EP_ADMIN_START:
        /* fallthrough */
        case NVME_EMU_EP_DELETING:
            ep->state = new_state;
            break;
        default:
            goto out_err;
        }
        break;
    case NVME_EMU_EP_ACTIVE:
        switch (new_state) {
        case NVME_EMU_EP_DELETING:
        /* fallthrough */
        case NVME_EMU_EP_ADMIN_START:
            driver->active_endpoint = -1;
            ep->state = new_state;
            break;
        default:
            goto out_err;
        }
        break;
    case NVME_EMU_EP_DELETING:
        switch (new_state) {
        case NVME_EMU_EP_NEW:
            ep->state = new_state;
            break;
        default:
            goto out_err;
        }
        break;
    default:
        goto out_err;
    }

    nvmx_debug("ep 0x%p: move from state %d to state %d", ep, old_state,
               new_state);

    return 0;

out_err:
    nvmx_error("ep 0x%p: can't move from state %d to state %d", ep,
               old_state, new_state);
    return NVMX_ERR;
}

void nvme_driver_destroy(nvme_emu_driver_t *driver)
{
    nvme_emu_ep_t *ep;
    int i;

    pthread_mutex_lock(&driver->ep_lock);
    for (i = 0; i < driver->nr_endpoints; i++) {
        ep = &driver->endpoints[i];

        nvme_emu_ep_destroy_backend(ep);
        free(ep->cqs);
        free(ep->sqs);
    }
    pthread_mutex_unlock(&driver->ep_lock);
    /* TODO: move the code */
    free(driver->endpoints);
    sem_destroy(&driver->sem);
    pthread_mutex_destroy(&driver->ep_lock);
    pthread_mutex_destroy(&driver->state_lock);
    free(driver->cqs);
    free(driver->sqs);
    free(driver);
}

/*
 * Must be called with ep_lock held
 *
 * This function will activate the non-active CQs and/or SQs in a best effort
 * approach.
 *
 */
static void nvme_emu_driver_activate_ep(nvme_emu_driver_t *driver,
                                        nvme_emu_ep_t *ep)
{
   bool activate = false;
   int ret;

   if (driver->num_created_io_cqueues != ep->num_created_io_cqueues ||
       driver->num_created_io_squeues != ep->num_created_io_squeues)
       activate = true;

   /* start and/or activate missing cqueues on active ep */
   if (driver->num_created_io_cqueues != ep->num_created_io_cqueues) {
       ret = nvme_emu_ep_start_cqs(ep);
       if (ret)
           nvmx_warn("couldn't start %d completion queues", ret);
   }
   /* start and/or activate missing squeues on active ep */
   if (driver->num_created_io_squeues != ep->num_created_io_squeues) {
       ret = nvme_emu_ep_start_sqs(ep);
       if (ret)
           nvmx_warn("couldn't start %d submission queues", ret);
   }

   if (activate)
       nvme_emu_ep_activate_backend(ep, NULL, NULL);
}

/*
 * Must be called with ep_lock held
 */
static void nvme_emu_driver_progress_ep(nvme_emu_driver_t *driver,
                                        nvme_emu_ep_t *ep)
{
    nvme_emu_ep_t *active_ep;
    enum nvme_emu_ep_state new_state;
    int ret;

    active_ep = nvme_driver_get_active_ep(driver);
    if (ep == active_ep) {
        /* try to activate non-active queue */
        nvme_emu_driver_activate_ep(driver, ep);
        return;
    }

retry:
    ret = -1; //change to zero on successful progress
    switch (ep->state) {
    case NVME_EMU_EP_NEW:
        ret = nvme_emu_ep_start_backend(ep);
        if (!ret)
            new_state = NVME_EMU_EP_ADMIN_START;
        break;
    case NVME_EMU_EP_ADMIN_START:
        if (driver->num_created_io_cqueues && driver->num_created_io_squeues) {
            nvme_emu_ep_start_cqs(ep);
            nvme_emu_ep_start_sqs(ep);
            if (ep->num_created_io_cqueues >= 1 &&
                ep->num_created_io_squeues >= 1) {
                new_state = NVME_EMU_EP_START;
                ret = 0;
            }
        }
        break;
    case NVME_EMU_EP_DELETING:
        nvme_emu_ep_stop_sqs(ep);
        nvme_emu_ep_stop_cqs(ep);
        nvme_emu_ep_stop_backend(ep, false);
        new_state = NVME_EMU_EP_NEW;
        ret = 0;
        break;
    case NVME_EMU_EP_START:
        if (driver->num_created_io_cqueues > ep->num_created_io_cqueues) {
            nvme_emu_ep_start_cqs(ep);
        }
        if (driver->num_created_io_squeues > ep->num_created_io_squeues) {
            nvme_emu_ep_start_sqs(ep);
        }
        if (!active_ep) {
            int activated_cqs = 0;
            int activated_sqs = 0;

            nvme_emu_ep_activate_backend(ep, &activated_cqs, &activated_sqs);
            if (activated_cqs && activated_sqs) {
                new_state = NVME_EMU_EP_ACTIVE;
                ret = 0;//at least 1 SQ and 1 CQ were activated, we can move state.
            } else if ((activated_cqs && !activated_sqs) ||
                       (!activated_cqs && activated_sqs)) {
                /*
                 * If couldn't activate an SQ+CQ pair, make sure
                 * there are no activated SQ/CQ left behind
                 */
                nvme_emu_ep_deactivate_backend(ep);
            }
        }
        break;
    default:
        nvmx_debug("ep 0x%p is not progressing from state %d", ep, ep->state);
        return;
    }

    if (!ret) {
        ret = nvme_emu_driver_change_ep_state(ep, new_state);
        if (!ret) {
            if (new_state == NVME_EMU_EP_START ||
                new_state == NVME_EMU_EP_ADMIN_START)
                goto retry;
        } //TODO: how we handle change state failure ??
    }
}

static void *driver_thread(void *arg)
{
    nvme_emu_driver_t *driver = arg;
    struct timespec ts;
    int i;

    while (1) {
        if (clock_gettime(CLOCK_REALTIME, &ts) == -1) {
            nvmx_error("failed to get system time, exiting..");
            return NULL;
        }
        ts.tv_sec += 5; //TODO: set configurable

        sem_timedwait(&driver->sem, &ts);

        pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
        pthread_mutex_lock(&driver->ep_lock);
        for (i = 0; i < driver->nr_endpoints; i++)
            nvme_emu_driver_progress_ep(driver, &driver->endpoints[i]);
        pthread_mutex_unlock(&driver->ep_lock);
        pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
    }

    return NULL;
}

nvme_emu_driver_t *nvme_driver_create(nvme_emulator_t *dev,
                                      struct snap_conf *sconfig,
                                      int num_queues,
                                      attach_namespace_fn_t *attach_namespace_fn,
                                      detach_namespace_fn_t *detach_namespace_fn,
                                      int index, uint32_t ver, uint8_t mdts)
{
    nvme_emu_driver_t *driver;
    nvme_emu_ep_t *ep;
    int nr_endpoints;
    enum nvme_emu_ep_fam ep_fam;
    enum nvme_emu_ep_type ep_type;
    char port[NVME_EMU_MAX_PORT_LEN + 1] = {0};
    int i, j, rc;
    char param[64];
    const snap_config_backend_t *backend;
    const snap_config_be_path_t *path;

    driver = calloc(1, sizeof(*driver));
    if (!driver)
        goto out;

    driver->index = index;
    driver->dev = dev;
    driver->sconfig = sconfig;
    driver->num_queues = num_queues;
    driver->mdts = mdts;
    driver->attach_namespace_cb = attach_namespace_fn;
    driver->detach_namespace_cb = detach_namespace_fn;
    driver->sqs = calloc(num_queues, sizeof(*driver->sqs));
    if (!driver->sqs)
        goto free_driver;
    driver->cqs = calloc(num_queues, sizeof(*driver->cqs));
    if (!driver->cqs)
        goto free_sqs;

    driver->ver = ver;
    backend = snap_config_get_backend(sconfig, index);

    if (backend)
        nr_endpoints = snap_config_get_paths_cnt(backend);
    else
        nr_endpoints = 0;

    if (nr_endpoints == 1) {
        driver->type = NVME_EMU_SINGLE_EP_DRIVER_TYPE;
    } else if (nr_endpoints == 2) {
        driver->type = NVME_EMU_MULTIPLE_EP_DRIVER_TYPE;
    } else if (nr_endpoints < 0) {
        nvmx_error("the path %s doesn't exist in the config file", param);
        goto free_cqs;
    } else {
        nvmx_error("num of endpoints %d is not supported", nr_endpoints);
        goto free_cqs;
    }

    driver->active_endpoint = -1;
    rc = pthread_mutex_init(&driver->state_lock, NULL);
    if (rc)
        goto free_cqs;
    rc = pthread_mutex_init(&driver->ep_lock, NULL);
    if (rc)
        goto free_state_lock;
    rc = sem_init(&driver->sem, 0, 0);
    if (rc)
        goto free_ep_lock;

    driver->nr_endpoints = nr_endpoints;
    driver->endpoints = calloc(driver->nr_endpoints,
                               sizeof(*(driver->endpoints)));
    if (!driver->endpoints)
        goto free_sem;

    if (strncmp(snap_config_get_emu_manager(sconfig),
                snap_config_get_rdma_dev(sconfig), MAX_RDMA_DEV_LEN))
        driver->cross_rdma_dev_support = true;

    if (strncmp(backend->be_type, "posix_io", 8) == 0) {
        ep_fam = NVME_EMU_ADRFAM_NONE;
        ep_type = NVME_EMU_EP_TYPE_POSIX;

    } else if (strncmp(backend->be_type, "memdisk", 7) == 0) {
        ep_fam = NVME_EMU_ADRFAM_NONE;
        ep_type = NVME_EMU_EP_TYPE_MEM;

    } else if (strncmp(backend->be_type, "spdk_bdev", 9) == 0) {
        ep_fam = NVME_EMU_ADRFAM_NONE;
        ep_type = NVME_EMU_EP_TYPE_SPDK;

    } else if (strncmp(backend->be_type, "nvmf_rdma", 9) == 0) {
        ep_fam = NVME_EMU_ADRFAM_IPV4;
        ep_type = NVME_EMU_EP_TYPE_NVMF_RDMA;

    } else {
        nvmx_error("Bad endpoint type specification: %s", backend->be_type);
        goto free_driver_endpoints;
    }

    if (ep_type != NVME_EMU_EP_TYPE_NVMF_RDMA && driver->nr_endpoints > 1) {
        nvmx_error("Multiple endpoints are not supported for ep_type %s",
                   backend->be_type);
        goto free_driver_endpoints;
    }

    for (i = 0; i < driver->nr_endpoints; i++) {
        ep = &driver->endpoints[i];
        ep->io_driver = driver;
        ep->ep_type = ep_type;
        ep->ep_fam = ep_fam;
        ep->state = NVME_EMU_EP_NEW;
        ep->sqs = calloc(num_queues, sizeof(*ep->sqs));
        if (!ep->sqs)
            goto free_eps;
        ep->cqs = calloc(num_queues, sizeof(*ep->cqs));
        if (!ep->cqs) {
            free(ep->sqs);
            goto free_eps;
        }

        path = snap_config_get_be_path(backend, i);
        snprintf(port, NVME_EMU_MAX_PORT_LEN, "%d", path->port);

        memcpy(driver->endpoints[i].name, backend->name, NVME_EMU_MAX_NAME_LEN);
        memcpy(driver->endpoints[i].addr, path->addr, NVME_EMU_ADDR_MAX_LEN);
        memcpy(driver->endpoints[i].port, port, NVME_EMU_MAX_PORT_LEN);
        memcpy(driver->endpoints[i].nqn, path->nqn, NVME_EMU_MAX_NQN_LEN);
        /* after this call we expect ep_ops and backend to be initialized */
        rc = nvme_emu_ep_create_backend(&driver->endpoints[i]);
        if (rc) {
            free(ep->cqs);
            free(ep->sqs);
            goto free_eps;
        }
    }

    nvmx_debug("%s %d endpoints", backend->be_type, nr_endpoints);

    return driver;

free_eps:
    for (j = 0; j < i; j++) {
        nvme_emu_ep_destroy_backend(&driver->endpoints[j]);
        free(ep->cqs);
        free(ep->sqs);
    }
free_driver_endpoints:
    free(driver->endpoints);
free_sem:
    sem_destroy(&driver->sem);
free_ep_lock:
    pthread_mutex_destroy(&driver->ep_lock);
free_state_lock:
    pthread_mutex_destroy(&driver->state_lock);
free_cqs:
    free(driver->cqs);
free_sqs:
    free(driver->sqs);
free_driver:
    free(driver);
out:
    return NULL;
}

int nvme_driver_start(nvme_emu_driver_t *driver)
{
    int rc = 0;

    pthread_mutex_lock(&driver->state_lock);
    if (driver->state == NVME_EMU_DRIVER_ACTIVE)
        goto out;

    rc = pthread_create(&driver->d_thread, NULL, driver_thread, driver);
    if (rc)
        goto out;

    // start looking for active endpoint asynchronously
    sem_post(&driver->sem);
    driver->state = NVME_EMU_DRIVER_ACTIVE;
out:
    pthread_mutex_unlock(&driver->state_lock);
    return rc;
}

void nvme_driver_stop(nvme_emu_driver_t *driver)
{
    nvme_emu_ep_t *ep;
    int i;

    pthread_mutex_lock(&driver->state_lock);
    if (driver->state == NVME_EMU_DRIVER_NEW)
        goto out;

    /* we don't want to keep progress when driver is stopped */
    pthread_cancel(driver->d_thread);
    pthread_join(driver->d_thread, NULL);

    pthread_mutex_lock(&driver->ep_lock);
    for (i = 0; i < driver->nr_endpoints; i++) {
        ep = &driver->endpoints[i];

        if (ep == nvme_driver_get_active_ep(driver))
            nvme_emu_ep_deactivate_backend(ep);
        /**
         * We stopped d_thread from progressing ep state machine,
         * so we need to explicitly advance the states as done in
         * DELETING->NEW flow, to properly prepare driver to
         * future possible driver_start() call
         */
        if (ep->state != NVME_EMU_EP_NEW) {
            nvme_emu_driver_change_ep_state(ep, NVME_EMU_EP_DELETING);
            nvme_emu_ep_stop_sqs(ep);
            nvme_emu_ep_stop_cqs(ep);
            nvme_emu_ep_stop_backend(ep, true);
            nvme_emu_driver_change_ep_state(ep, NVME_EMU_EP_NEW);
        }
    }
    pthread_mutex_unlock(&driver->ep_lock);

    driver->state = NVME_EMU_DRIVER_NEW;
out:
    pthread_mutex_unlock(&driver->state_lock);
}

int nvme_driver_create_sq(nvme_emu_driver_t *driver, uint16_t sqid,
                          uint64_t dma_addr, uint16_t size,
                          nvme_queue_ops_t *cq_ops, nvme_queue_ops_t *sq_ops)
{
    int ret;
    struct nvme_emu_create_sq_attr nvme_sq_attr = {};

    pthread_mutex_lock(&driver->ep_lock);

    nvme_emu_qops_clone(driver->dev, sq_ops, sqid);

    nvme_sq_attr.id = sqid;
    nvme_sq_attr.base_addr = dma_addr;
    nvme_sq_attr.size = size;
    nvme_sq_attr.cq_ops = cq_ops;
    nvme_sq_attr.fe_only = driver->cross_rdma_dev_support;
    ret = nvme_emu_create_sq(driver->dev, &nvme_sq_attr, sq_ops);
    if (ret)
        goto err;

    driver->sqs[sqid].driver = driver;
    driver->sqs[sqid].depth = size;
    driver->sqs[sqid].qid = sqid;
    driver->sqs[sqid].cqid = cq_ops->qid;
    driver->sqs[sqid].type = NVME_EMU_DRIVER_SQ;

    /* SQ for EPs will be created during progress thread */
    if (sqid)
        driver->num_created_io_squeues++;

    nvmx_assertv_always(driver->num_created_io_squeues < driver->num_queues,
                        "Creating more SQ's than supported");
    /* wake up the progress thread for fast creation of the EPs SQ */
    sem_post(&driver->sem);
    pthread_mutex_unlock(&driver->ep_lock);

    return 0;

err:
    pthread_mutex_unlock(&driver->ep_lock);
    return ret;
}

int nvme_driver_create_cq(nvme_emu_driver_t *driver, int cqid,
                          uint64_t dma_addr,uint16_t size, uint16_t vector,
                          int cq_period, int cq_max_count,
                          uint16_t irq_enabled, nvme_queue_ops_t *cq_ops)
{
    int ret;
    struct nvme_emu_create_cq_attr cq_attr = {};

    pthread_mutex_lock(&driver->ep_lock);

    cq_attr.qid = cqid;
    cq_attr.base_addr = dma_addr;
    cq_attr.size = size;
    cq_attr.vector = vector;
    cq_attr.cq_period = cq_period;
    cq_attr.cq_max_count = cq_max_count;
    cq_attr.irq_disable = !irq_enabled;
    nvme_emu_qops_clone(driver->dev, cq_ops, cqid);
    ret = nvme_emu_create_cq(driver->dev, cq_ops, &cq_attr);
    if (ret)
        goto err;

    driver->cqs[cqid].driver = driver;
    driver->cqs[cqid].depth = size;
    driver->cqs[cqid].qid = cqid;
    driver->cqs[cqid].type = NVME_EMU_DRIVER_CQ;

    /* CQ for EPs will be created during progress thread */
    if (cqid)
        driver->num_created_io_cqueues++;

    nvmx_assertv_always(driver->num_created_io_cqueues < driver->num_queues,
                        "Creating more CQ's than supported");
    /* wake up the progress thread for fast creation of the EPs CQ */
    sem_post(&driver->sem);
    pthread_mutex_unlock(&driver->ep_lock);

    return 0;

err:
    pthread_mutex_unlock(&driver->ep_lock);
    return ret;
}

/*
 * Must be called with ep_lock held
 */
static void nvme_emu_driver_delete_io_sq(nvme_emu_driver_t *driver,
                                        uint16_t sqid)
{
    nvme_emu_ep_t *ep;
    int i;

    driver->num_created_io_squeues--;

    nvmx_assertv_always(driver->num_created_io_squeues >= 0,
                        "deleting more SQ's than created");

    for (i = 0; i < driver->nr_endpoints; i++) {
        ep = &driver->endpoints[i];
        if (ep->sqs[sqid].state != NVME_EMU_EP_QUEUE_NEW) {
            if (ep->sqs[sqid].state == NVME_EMU_EP_QUEUE_ACTIVE)
                nvme_emu_ep_deactivate_sq(ep, sqid);
            nvme_emu_ep_delete_sq(ep, sqid);
        }
    }
}

void nvme_driver_delete_sq(nvme_emu_driver_t *driver, nvme_queue_ops_t *sq_ops,
                           int sqid)
{
    pthread_mutex_lock(&driver->ep_lock);

    /* first destroy the EPs sq */
    if (sqid)
        nvme_emu_driver_delete_io_sq(driver, sqid);

    driver->sqs[sqid].driver = NULL;
    driver->sqs[sqid].depth = 0;
    driver->sqs[sqid].qid = 0;

    nvme_emu_delete_sq(sq_ops, sqid);

    pthread_mutex_unlock(&driver->ep_lock);
}

/*
 * Must be called with ep_lock held
 */
static void nvme_emu_driver_delete_io_cq(nvme_emu_driver_t *driver,
                                         uint16_t cqid)
{
    nvme_emu_ep_t *ep;
    int i;

    driver->num_created_io_cqueues--;

    nvmx_assertv_always(driver->num_created_io_cqueues >= 0,
                        "deleting more CQ's than created");

    for (i = 0; i < driver->nr_endpoints; i++) {
        ep = &driver->endpoints[i];
        if (ep->cqs[cqid].state != NVME_EMU_EP_QUEUE_NEW) {
            if (ep->cqs[cqid].state == NVME_EMU_EP_QUEUE_ACTIVE)
                nvme_emu_ep_deactivate_cq(ep, cqid);
            nvme_emu_ep_delete_cq(ep, cqid);
        }
        if (!driver->num_created_io_squeues &&
            !driver->num_created_io_cqueues &&
            (ep->state == NVME_EMU_EP_START || ep->state == NVME_EMU_EP_ACTIVE))
            nvme_emu_driver_change_ep_state(ep, NVME_EMU_EP_ADMIN_START);
    }
}

void nvme_driver_delete_cq(nvme_emu_driver_t *driver, nvme_queue_ops_t *cq_ops,
                           int cqid)
{
    pthread_mutex_lock(&driver->ep_lock);

    /* first destroy the EPs cq */
    if (cqid)
        nvme_emu_driver_delete_io_cq(driver, cqid);

    driver->cqs[cqid].driver = NULL;
    driver->cqs[cqid].depth = 0;
    driver->cqs[cqid].qid = 0;

    nvme_emu_delete_cq(cq_ops, cqid);

    pthread_mutex_unlock(&driver->ep_lock);
}

void nvme_driver_fail_ep(nvme_emu_driver_t *driver, nvme_emu_ep_t *ep)
{
    int ret;

    pthread_mutex_lock(&driver->ep_lock);
    if (ep == nvme_driver_get_active_ep(driver))
        nvme_emu_ep_deactivate_backend(ep);
    /* change to DELETING state should always succeed */
    ret = nvme_emu_driver_change_ep_state(ep, NVME_EMU_EP_DELETING);
    nvmx_assertv_always(ret == 0, "failed to move to DELETING state");
    sem_post(&driver->sem);
    pthread_mutex_unlock(&driver->ep_lock);
}

int nvme_driver_attach_ep_volume(nvme_emu_driver_t *driver, uint32_t nsid,
                                 uint64_t size_b, uint8_t block_order,
                                 uint16_t backend_metadata,
                                 nvme_id_ns_descriptor_t id_descs[], void *priv_ns)
{
    /* dynamic namespaces for multiple andpoints is not supported */
    if (driver->nr_endpoints > 1)
        return -ENOTSUP;

    /* Best effort to add the namespace to NVMe controller's list */
    if (driver->attach_namespace_cb)
        return driver->attach_namespace_cb(driver->dev->ctrl, nsid, size_b, block_order,
                                           backend_metadata, id_descs, priv_ns);

    return 0;
}

void nvme_driver_detach_ep_volume(nvme_emu_driver_t *driver, uint32_t nsid)
{
    /* dynamic namespaces for multiple andpoints is not supported */
    if (driver->nr_endpoints > 1)
        return;

    if (driver->detach_namespace_cb)
        driver->detach_namespace_cb(driver->dev->ctrl, nsid);
}

int nvme_driver_request_size(nvme_emu_driver_t *driver)
{
    int size;
    nvme_emu_ep_t *ep;

    size = nvme_emu_request_size(driver->dev) + sizeof(nvme_io_driver_req_t);

    /* nvme_emu_request_size and sizeof both count req->async_req size
     * so adjust for it */
    size -= sizeof(nvme_async_req_t);

    /**
     * plus extra size required by the backends
     */
    ep = &driver->endpoints[0];
    return size + driver->nr_endpoints * ep->backend_req_size;
}

struct ibv_mr *nvme_driver_sq_init(nvme_emu_driver_t *driver, memp_t *memp,
                nvme_queue_ops_t *ops, uint16_t q_size,  nvme_io_driver_req_t *req)
{
    nvme_emu_ep_t *ep = &driver->endpoints[0];
    struct ibv_mr *mr = NULL;
    size_t io_buf_len;
    int res;

    req->async_req.data_len = PRP_PAGE_SIZE(ops->page_bits) * (1 << driver->mdts);
    req->async_req.ops = ops;
    req->async_req.aux_len = nvmx_max(nvme_prp_list_size(&req->async_req), SGL_SEGMENT_SIZE);
    io_buf_len = (req->async_req.data_len + req->async_req.aux_len) * q_size;

    /* let endpoint init sq memory buffer */
    if (ep->ep_ops.io_buf_create) {
        res = ep->ep_ops.io_buf_create(io_buf_len, memp, req);
        if (res != NVMX_OK)
            return NULL;
    }

     // No extra allocations when memory pool is active
    if (!memp) {
        /* if no memory was allocated by the ep, allocate
         * generic buffer
         */
        if (!req->async_req.io_buf) {
            req->async_req.io_buf = malloc(io_buf_len);
            if (!req->async_req.io_buf) {
                nvmx_error("Failed to allocate io_buf for sq");
                return NULL;
            }
        }

        mr = nvme_emu_sq_init(ops, req->async_req.io_buf, io_buf_len);
        if (!mr) {
            req->ep  = ep;
            nvme_driver_sq_reset(req);
        }
   }
    return mr;
}

int nvme_driver_request_init(nvme_emu_driver_t *driver, nvme_queue_ops_t *ops,
                             nvme_io_driver_req_t *req, nvme_req_cb_t comp_cb,
                             void *comp_cb_user, int reqid, memp_t *memp,
                             size_t data_len, size_t aux_len, uint8_t *io_buf,
                             struct ibv_mr *mr)
{
    nvme_emu_ep_t *ep;

    ep = &driver->endpoints[0];
    memset(req, 0, sizeof(*req));
    req->comp_cb = comp_cb;
    req->comp_cb_user = comp_cb_user;
    req->ep = ep;
    req->id = reqid;

    req->async_req.data_len = data_len;
    req->async_req.ops = ops;
    req->async_req.memp = memp;


    req->async_req.aux_len = aux_len;

    // No extra allocations when memory pool is active
    if (memp) {
        req->async_req.io_buf = NULL;
        req->async_req.aux_buf = NULL;
    } else {
        req->async_req.io_buf = io_buf + (reqid * (data_len + aux_len));
        req->async_req.aux_buf = req->async_req.io_buf + req->async_req.data_len;
    }

    return nvme_emu_request_init(ops, &req->async_req, mr);
}

void nvme_driver_sq_reset(nvme_io_driver_req_t *req)
{
    nvme_emu_ep_t *ep = req->ep;

    nvme_emu_sq_reset(&req->async_req);

    if (ep->ep_ops.sq_reset)
        ep->ep_ops.sq_reset(req);

    if (req->async_req.io_buf)
        free(req->async_req.io_buf);
}

void nvme_driver_request_reset(nvme_io_driver_req_t *req)
{
    nvme_emu_ep_t *ep;

    nvme_emu_request_reset(&req->async_req);

    ep = req->ep;
    if (ep->ep_ops.request_reset)
        ep->ep_ops.request_reset(req);

}

int nvme_driver_write_prp_nb(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw)
{
    nvme_emu_ep_t *active_ep;
    int rc;

    /* TODO: handle failover */
    active_ep = req->ep;

    nvmx_debug("write prp: slb %lu nlb %u prp1: 0x%llx prp2: 0x%llx pbits %u",
               le64_to_cpu(rw->slba), le16_to_cpu(rw->nlb) + 1,
               (unsigned long long)le64_to_cpu(rw->common.dptr.prp1),
               (unsigned long long)le64_to_cpu(rw->common.dptr.prp2),
               req->async_req.ops->page_bits);

    if (likely(active_ep->ep_ops.write_prp_nb))
        return active_ep->ep_ops.write_prp_nb(req, rw);

    rc = active_ep->ep_ops.write_prp(active_ep, req, rw);
    req->comp_cb(req->comp_cb_user, rc);

    return NVMX_INPROGRESS;
}

int nvme_driver_read_prp_nb(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw)
{
    nvme_emu_ep_t *active_ep;
    int rc;

    /* TODO: handle failover */
    active_ep = req->ep;

    nvmx_debug("read prp: slb %lu nlb %u prp1: 0x%llx prp2: 0x%llx pbits %u",
               le64_to_cpu(rw->slba), le16_to_cpu(rw->nlb) + 1,
               (unsigned long long)le64_to_cpu(rw->common.dptr.prp1),
               (unsigned long long)le64_to_cpu(rw->common.dptr.prp2),
               req->async_req.ops->page_bits);

    if (likely(active_ep->ep_ops.read_prp_nb))
        return active_ep->ep_ops.read_prp_nb(req, rw);

    rc = active_ep->ep_ops.read_prp(active_ep, req, rw);
    if(!req->async_req.is_tcp_rx_zcopy)
        req->comp_cb(req->comp_cb_user, rc);

    return NVMX_INPROGRESS;
}

int nvme_driver_write_sgl_nb(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw)
{
    nvme_emu_ep_t *active_ep;
    int rc;

    active_ep = req->ep;

    nvmx_debug("start async write_sgl: slb %lu nlb %u type 0x%x sub: 0x%x pbits %u",
               le64_to_cpu(rw->slba), le16_to_cpu(rw->nlb) + 1,
               rw->common.dptr.sgl.type, rw->common.dptr.sgl.subtype, req->async_req.ops->page_bits);

    if (likely(active_ep->ep_ops.write_sgl_nb))
        return active_ep->ep_ops.write_sgl_nb(req, rw);

    rc = active_ep->ep_ops.write_sgl(active_ep, req, rw);
    req->comp_cb(req->comp_cb_user, rc);

    return NVMX_INPROGRESS;
}

int nvme_driver_read_sgl_nb(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw)
{
    int rc;
    nvme_emu_ep_t *active_ep;

    active_ep = req->ep;

    nvmx_debug("start async read_sgl: slb %lu nlb %u type 0x%x sub: 0x%x pbits %u",
               le64_to_cpu(rw->slba), le16_to_cpu(rw->nlb) + 1,
               rw->common.dptr.sgl.type, rw->common.dptr.sgl.subtype, req->async_req.ops->page_bits);

    if (likely(active_ep->ep_ops.read_sgl_nb))
        return active_ep->ep_ops.read_sgl_nb(req, rw);

    rc = active_ep->ep_ops.read_sgl(active_ep, req, rw);
    if(!req->async_req.is_tcp_rx_zcopy)
        req->comp_cb(req->comp_cb_user, rc);

    return NVMX_INPROGRESS;
}
