#include "config.h"
#include "nvme_mem_ctrl.h"
#include "nvme_prp.h"
#include "nvme_sgl.h"
#include "nvme_emu_ep.h"
#include "snap_conf.h"

/* setting to 0 will change memdisk to null disk which can be
 * useful for benchmarking
 */
#define MEMDISK_DO_COPY 1

static void read_prp_completion(void *user, int status)
{
    nvme_io_driver_req_t *io_req = user;
    nvme_mem_ctrl_t *ctrl = (nvme_mem_ctrl_t *)io_req->ep->ctrl;

    if (MEMDISK_DO_COPY && likely(status == NVMX_OK))
        memcpy(ctrl->memdisk_base + io_req->disk_offset, io_req->async_req.io_buf,
                io_req->async_req.len);

    io_req->comp_cb(io_req->comp_cb_user, status);
}

static int nvme_emu_mem_write_prp_nb(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw)
{
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    uint64_t slba = le64_to_cpu(rw->slba);
    uint64_t prp1 = le64_to_cpu(rw->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(rw->common.dptr.prp2);
    size_t count, offset;
    nvme_mem_ctrl_t *ctrl;

    ctrl = (nvme_mem_ctrl_t *)req->ep->ctrl;
    count  = (size_t)nlb << ctrl->disk_block_order;
    offset = slba << ctrl->disk_block_order;

    req->disk_offset = offset;
    req->async_req.comp_cb = read_prp_completion;
    req->async_req.comp_cb_user = req;
    return nvme_prp_read(&req->async_req, prp1, prp2, count);
}

static int nvme_emu_mem_read_prp_nb(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw)
{
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    uint64_t slba = le64_to_cpu(rw->slba);
    uint64_t prp1 = le64_to_cpu(rw->common.dptr.prp1);
    uint64_t prp2 = le64_to_cpu(rw->common.dptr.prp2);
    size_t count, offset;
    nvme_mem_ctrl_t *ctrl;

    ctrl = (nvme_mem_ctrl_t *)req->ep->ctrl;
    count  = (size_t)nlb << ctrl->disk_block_order;
    offset = slba << ctrl->disk_block_order;

    if (MEMDISK_DO_COPY)
        memcpy(req->async_req.io_buf, ctrl->memdisk_base + offset, count);

    return nvme_prp_write(&req->async_req, prp1, prp2, count);
}

static void read_sgl_completion(void *user, int status)
{
    nvme_io_driver_req_t *io_req = user;
    nvme_mem_ctrl_t *ctrl = (nvme_mem_ctrl_t *)io_req->ep->ctrl;

    if (MEMDISK_DO_COPY && likely(status == NVMX_OK))
        memcpy(ctrl->memdisk_base + io_req->disk_offset, io_req->async_req.io_buf,
                io_req->async_req.len);

    io_req->comp_cb(io_req->comp_cb_user, status);
}

static int nvme_emu_mem_write_sgl_nb(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw)
{
    nvme_mem_ctrl_t *ctrl = (nvme_mem_ctrl_t *)req->ep->ctrl;
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    uint64_t slba = le64_to_cpu(rw->slba);
    const nvme_sgl_desc_t *sgl = &rw->common.dptr.sgl;
    size_t count, offset;

    count  = (size_t)nlb << ctrl->disk_block_order;
    offset = slba << ctrl->disk_block_order;

    req->disk_offset = offset;
    req->async_req.comp_cb = read_sgl_completion;
    req->async_req.comp_cb_user = req;

    return nvme_sgl_read_nb(&req->async_req, sgl, count);
}

static int nvme_emu_mem_read_sgl_nb(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw)
{
    nvme_mem_ctrl_t *ctrl = (nvme_mem_ctrl_t *)req->ep->ctrl;
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    uint64_t slba = le64_to_cpu(rw->slba);
    const nvme_sgl_desc_t *sgl = &rw->common.dptr.sgl;
    size_t count, offset;

    count  = (size_t)nlb << ctrl->disk_block_order;
    offset = slba << ctrl->disk_block_order;

    if (MEMDISK_DO_COPY)
        memcpy(req->async_req.io_buf, ctrl->memdisk_base + offset, count);

    return nvme_sgl_write_nb(&req->async_req, sgl, count);
}

static int nvme_emu_mem_read_sgl(nvme_emu_ep_t *ep,
                                 nvme_io_driver_req_t *req,
                                 const nvme_cmd_rw_t *rw)
{
    nvme_mem_ctrl_t *ctrl = (nvme_mem_ctrl_t *)ep->ctrl;
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    uint64_t slba = le64_to_cpu(rw->slba);
    const nvme_sgl_desc_t *sgl = &rw->common.dptr.sgl;
    size_t count, offset;
    nvme_queue_ops_t *ops = req->async_req.ops;

    count  = (size_t)nlb << ctrl->disk_block_order;
    offset = slba << ctrl->disk_block_order;
    req->async_req.len = count;

    nvme_sgl_write(ops, sgl, ctrl->memdisk_base + offset, count);
    return NVMX_OK;
}

static int nvme_emu_mem_write_sgl(nvme_emu_ep_t *ep,
                                  nvme_io_driver_req_t *req,
                                  const nvme_cmd_rw_t *rw)
{
    nvme_mem_ctrl_t *ctrl = (nvme_mem_ctrl_t *)ep->ctrl;
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    uint64_t slba = le64_to_cpu(rw->slba);
    const nvme_sgl_desc_t *sgl = &rw->common.dptr.sgl;
    size_t count, offset;
    nvme_queue_ops_t *ops = req->async_req.ops;

    count  = (size_t)nlb << ctrl->disk_block_order;
    offset = slba << ctrl->disk_block_order;
    req->async_req.len = count;

    nvme_sgl_read(ops, sgl, ctrl->memdisk_base + offset, count);
    return NVMX_OK;
}

static int nvme_emu_mem_create_cq(nvme_emu_ep_t *ep, uint16_t cqid,
                                   uint16_t size)
{
    return 0;
}

static void nvme_emu_mem_delete_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
}

static void nvme_emu_mem_deactivate_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
}

static int nvme_emu_mem_activate_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
    return 0;
}

static int nvme_emu_mem_create_sq(nvme_emu_ep_t *ep, uint16_t sqid,
                                   uint16_t size, uint16_t cqid)
{
    return 0;
}

static void nvme_emu_mem_delete_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
}

static void nvme_emu_mem_deactivate_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
}

static int nvme_emu_mem_activate_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
    return 0;
}

void nvme_emu_delete_mem_ctrl(nvme_mem_ctrl_t *ctrl)
{
    free(ctrl->memdisk_base);
    free(ctrl);
}

int nvme_emu_create_mem_ctrl(nvme_emu_ep_t *ep)
{
    nvme_mem_ctrl_t *ctrl;
    snap_conf_t *sconfig = ep->io_driver->sconfig;
    const snap_config_backend_t *backend = snap_config_get_backend(sconfig, 0);

    ctrl = calloc(1, sizeof(*ctrl));
    if (!ctrl)
        return NVMX_ERR;

    ctrl->disk_size_b = snap_config_get_backend_size_b(backend);
    ctrl->disk_block_order = snap_config_get_backend_block_order(backend);

    ctrl->memdisk_base = calloc(ctrl->disk_size_b,
                                sizeof(*ctrl->memdisk_base));
    if (!ctrl->memdisk_base) {
        nvmx_error("failed to allocate memdisk sized %zd",
                   ctrl->disk_size_b);
        goto out_err;
    }

    ep->ctrl = ctrl;
    ep->ep_ops.read_prp_nb = nvme_emu_mem_read_prp_nb;
    ep->ep_ops.write_prp_nb = nvme_emu_mem_write_prp_nb;
    ep->ep_ops.read_sgl_nb  = nvme_emu_mem_read_sgl_nb;
    ep->ep_ops.write_sgl_nb = nvme_emu_mem_write_sgl_nb;
    ep->ep_ops.read_sgl  = nvme_emu_mem_read_sgl;
    ep->ep_ops.write_sgl = nvme_emu_mem_write_sgl;
    ep->ep_ops.create_cq = nvme_emu_mem_create_cq;
    ep->ep_ops.delete_cq = nvme_emu_mem_delete_cq;
    ep->ep_ops.activate_cq = nvme_emu_mem_activate_cq;
    ep->ep_ops.deactivate_cq = nvme_emu_mem_deactivate_cq;
    ep->ep_ops.create_sq = nvme_emu_mem_create_sq;
    ep->ep_ops.delete_sq = nvme_emu_mem_delete_sq;
    ep->ep_ops.activate_sq = nvme_emu_mem_activate_sq;
    ep->ep_ops.deactivate_sq = nvme_emu_mem_deactivate_sq;

    ctrl->ep = ep;

    return NVMX_OK;

out_err:
    free(ctrl);
    return NVMX_ERR;
}

int nvme_start_mem_ctrl(nvme_emu_ep_t *ep)
{
    nvme_mem_ctrl_t *ctrl = ep->ctrl;

    /* Build namespace, if needed */
    if (ctrl->disk_size_b > 0) {
        nvme_id_ns_descriptor_t ns_desc[2];

        ns_desc[0].nidt = NVME_NS_IDENTIFIER_UUID;
        ns_desc[0].nidl = 0x10;
        memset(&ns_desc[0].nid.uuid, 0, ns_desc[0].nidl);
        memset(&ns_desc[1], 0, sizeof(ns_desc[1]));
        nvme_emu_ep_attach_volume(ep, 1, ctrl->disk_size_b,
                                  ctrl->disk_block_order, 0, ns_desc, ctrl);
    }

    return NVMX_OK;
}

void nvme_stop_mem_ctrl(nvme_emu_ep_t *ep)
{
    nvme_emu_ep_detach_volume(ep, 1);
}
