#include "config.h"
#include "nvme_spdk_ctrl.h"
#include "mlnx_snap_pci_manager.h"
#include "nvmx_util.h"

#include <ctype.h>
#ifdef HAVE_SPDK
#include <spdk/bdev_module.h>
#include <spdk/env.h>
#include <spdk/rpc.h>
#include <spdk/json.h>
#include <spdk/jsonrpc.h>
#include <uuid/uuid.h>

#include "utils.h"
#include "nvme.h"
#include "nvme_prp.h"
#include "nvme_sgl.h"
#include "nvme_emu_io_driver.h"
#include "snap_conf.h"
#include "nvme_spdk_subsystem.h"
#include "snap_env.h"
#include "memp.h"
#include "spdk_ext_io/spdk_ext_io.h"
#include "snap_nvme_emulator.h"

#define NVME_SPDK_BUF_ALIGN_MIN 4096

#define NVME_SNAP_ZCOPY "NVME_SNAP_ZCOPY"
SNAP_ENV_REG_ENV_VARIABLE(NVME_SNAP_ZCOPY, 1);

/* Rx ZCOPY is disable by default for Rel1-2022 */
#define NVME_SNAP_TCP_RX_ZCOPY "NVME_SNAP_TCP_RX_ZCOPY"
SNAP_ENV_REG_ENV_VARIABLE(NVME_SNAP_TCP_RX_ZCOPY, 0);

struct nvme_spdk_namespace_io_context {
    struct spdk_io_channel *io_channel;
    struct spdk_thread *io_thread;
    struct nvme_spdk_namespace *ns;
    struct spdk_thread *deleter_thread;
    spdk_ext_io_ctx_t extio_ctx;

    uint32_t inflight_io;
};

struct nvme_spdk_rpc_namespace_attach_attr {
    uint32_t nsid;
    char *bdev_type;
    char *bdev_name;
    char *emu_name;
    char *subnqn;
    int cntlid;
    char *qn;
    char *protocol;
    char *uuid;
    char *nguid;
    char *eui64;
};

/*
 * Generally, when multiple NVMe emulations are supported,
 * there might be multiple nvme_spdk_ctrl instances.
 * All controllers need to be directly reached via RPC,
 * so they are kept in a static ctrl list, and can be
 * identified with a user-readable name.
 */
TAILQ_HEAD(, nvme_spdk_ctrl) nvme_spdk_ctrl_list =
                              TAILQ_HEAD_INITIALIZER(nvme_spdk_ctrl_list);
pthread_mutex_t nvme_spdk_ctrl_list_lock = PTHREAD_MUTEX_INITIALIZER;

static void nvme_spdk_namespace_delete_cb(void *arg);
static inline int nvme_spdk_rw_do(nvme_io_driver_req_t *req);
static void nvme_spdk_write_to_host_done(void *user, int status);

#ifdef HAVE_SPDK_BDEV_OPEN_EXT
static void nvme_spdk_namespace_event_cb(enum spdk_bdev_event_type type,
                                         struct spdk_bdev *bdev,
                                         void *event_ctx);
#endif

/* param @attach_fail indicate if destroy ns because of try to attach ns error.
 * if attach_err is true, then dont need namespaces_lock protection and need
 * to delete from subsystem ns list */
static void
nvme_spdk_namespace_delete_destroy(nvme_spdk_namespace_t *ns, bool attach_fail)
{
    nvme_spdk_ctrl_t *ctrl = ns->ctrl;

    if (attach_fail)
        nvme_subsys_ns_del(ctrl->ep->io_driver->sconfig->spdk.subsys, ns);

    if (!attach_fail)
        pthread_spin_lock(&ctrl->namespaces_lock);
    TAILQ_REMOVE(&ctrl->namespaces, ns, entry);
    if (!attach_fail)
        pthread_spin_unlock(&ctrl->namespaces_lock);

    nvmx_info("Namespace nsid %d (%s) was removed from namespace list",
              ns->nsid, spdk_bdev_get_name(ns->bdev));

    if (ns->bdev_desc)
        spdk_bdev_close(ns->bdev_desc);
    if (ns->deleter_cb)
        ns->deleter_cb(ns->deleter_cb_arg);

    free(ns->io_contexts);
    free(ns);
}

static void nvme_spdk_namespace_delete_end(void *arg)
{
    nvme_spdk_namespace_t *ns = arg;

    ns->num_io_contexts--;

    /* If not the last one to be destroyed, do nothing */
    if (ns->num_io_contexts)
        return;

    nvme_spdk_namespace_delete_destroy(ns, false);
}

static void nvme_spdk_namespace_put_channel(void *arg)
{
    struct nvme_spdk_namespace_io_context *io_context = arg;

    if (io_context->inflight_io)
        return;

    /*
     * this function can be queued from multiple contexts
     * like delete sq and delete namespace. Only need to
     * put channel once, so check for NULL.
     */
    if (io_context->io_channel) {
        spdk_put_io_channel(io_context->io_channel);
        io_context->io_channel = NULL;
        io_context->io_thread = NULL;
    }

    spdk_ext_io_context_clear(&io_context->extio_ctx);

#ifdef HAVE_SPDK_EXT_IO
    if (io_context->ns->zcopy || io_context->ns->tcp_tx_zcopy) {
        nvme_emu_destroy_zcopy_key_table(io_context->extio_ctx.zcopy_ctx.key_table);
        io_context->extio_ctx.zcopy_ctx.key_table = NULL;
    }
#endif

    if (io_context->deleter_thread) {
        spdk_thread_send_msg(io_context->deleter_thread,
                             nvme_spdk_namespace_delete_end,
                             io_context->ns);
    }
}

/* Must be called with ctrl->namespaces_lock held */
static void nvme_spdk_namespace_delete(void *arg,
                                       nvme_spdk_delete_cb_t deleter_cb,
                                       void *deleter_cb_arg)
{
    nvme_spdk_namespace_t *ns = arg;
    nvme_spdk_ctrl_t *ctrl = ns->ctrl;
    int i;

    ns->in_deletion = true;
    ns->deleter_cb = deleter_cb;
    ns->deleter_cb_arg = deleter_cb_arg;

    nvme_subsys_ns_del(ctrl->ep->io_driver->sconfig->spdk.subsys, ns);

    for (i = 0; i < ns->num_io_contexts; i++) {
        if (ns->io_contexts[i].io_thread) {
            ns->io_contexts[i].deleter_thread = spdk_get_thread();
            spdk_thread_send_msg(ns->io_contexts[i].io_thread,
                                 nvme_spdk_namespace_put_channel,
                                 &ns->io_contexts[i]);
        } else {
            spdk_thread_send_msg(spdk_get_thread(),
                                 nvme_spdk_namespace_delete_end, ns);
        }
    }
}

static int common_hex_parse(const char *in, size_t inlen,
                            uint8_t  *out, size_t outlen)
{
    int out_idx, in_idx = 0;

    for (out_idx = 0; out_idx < outlen; out_idx++) {
        if (in_idx + 2 > inlen)
            return -EINVAL;

        if (!isxdigit(in[in_idx]) || !isxdigit(in[in_idx + 1]))
            return -EINVAL;

        out[out_idx] = (char_to_hex(in[in_idx]) << 4) |
                        char_to_hex(in[in_idx + 1]);

	in_idx += 2;
        if (in[in_idx] == '-' || in[in_idx] == ':')
            in_idx++;
    }

    return 0;
}

static int eui64_parse(const char *eui64_str, nvme_eui64_t *eui64)
{
    return common_hex_parse(eui64_str, strlen(eui64_str), eui64->raw, 8);
}

static int nguid_parse(const char *nguid_str, nvme_ngiud_t *nguid)
{
    return common_hex_parse(nguid_str, strlen(nguid_str), nguid->raw, 16);
}

static int
nvme_spdk_namespace_set_ids(nvme_spdk_namespace_t *ns,
                            struct spdk_bdev *bdev,
                            struct nvme_spdk_rpc_namespace_attach_attr *attr)
{
    int i = 0;

    if (attr->uuid) {
        nvmx_debug("using UUID Identifier: %s", attr->uuid);
        if (uuid_parse(attr->uuid, ns->ns_desc[i].nid.uuid.raw)) {
            nvmx_error("Invalid Host UUID Identifier \"%s\". "
                       "Identifier UUID is a hex string of the form "
                       "1b4e28ba-2fa1-11d2-883f-b9a761bde3fb",
                       attr->uuid);
            return -EINVAL;
        } else {
            ns->ns_desc[i].nidt = NVME_NS_IDENTIFIER_UUID;
            ns->ns_desc[i].nidl = 0x10;
            i++;
	}
    }

    if (attr->nguid) {
        nvmx_debug("using NGUID Identifier: %s", attr->nguid);
        if (nguid_parse(attr->nguid, &ns->ns_desc[i].nid.nguid)) {
            nvmx_error("Invalid Host NGUID Identifier \"%s\". "
                       "Identifier NGUID is a hex string of the form "
                       "1b4e28ba-2fa1-11d2-883f-b9a761bde3fb",
                       attr->nguid);
            return -EINVAL;
        } else {
            ns->ns_desc[i].nidt = NVME_NS_IDENTIFIER_NGUID;
            ns->ns_desc[i].nidl = 0x10;
            i++;
	}
    }

    if (attr->eui64) {
        nvmx_debug("using EUI64 Identifier: %s", attr->eui64);
        if (eui64_parse(attr->eui64, &ns->ns_desc[i].nid.eui64)) {
            nvmx_error("Invalid Host EUI64 Identifier \"%s\". "
                       "Identifier EUI64 is a hex string of the form "
                       "1b4e-28ba-2fa1-11d2",
                       attr->eui64);
            return -EINVAL;
        } else {
            ns->ns_desc[i].nidt = NVME_NS_IDENTIFIER_EUI64;
            ns->ns_desc[i].nidl = 0x8;
            i++;
	}
    }

    /* Set bdev's UUID, NGUID in case none of the IDs were provided */
    if (i == 0) {
        ns->ns_desc[i].nidt = NVME_NS_IDENTIFIER_UUID;
        ns->ns_desc[i].nidl = 0x10;
        memcpy(&ns->ns_desc[i].nid.uuid.raw, spdk_bdev_get_uuid(bdev),
               ns->ns_desc[i].nidl);
        i++;

	ns->ns_desc[i].nidt = NVME_NS_IDENTIFIER_NGUID;
        ns->ns_desc[i].nidl = 0x10;
        memcpy(&ns->ns_desc[i].nid.nguid, spdk_bdev_get_uuid(bdev),
               ns->ns_desc[i].nidl);
        i++;
    }

    /* Always finish with NULL entry */
    memset(&ns->ns_desc[i], 0, sizeof(ns->ns_desc[i]));
    return 0;
}

/* Must be called with ctrl->namespaces_lock held */
static nvme_spdk_namespace_t*
nvme_spdk_namespace_create(nvme_spdk_ctrl_t *ctrl,
                           struct spdk_bdev *bdev,
                           struct nvme_spdk_rpc_namespace_attach_attr *attr)
{
    nvme_spdk_namespace_t *ns;
    int err;
    uint32_t nsid = attr->nsid;
    const char *bdev_type = attr->bdev_type;
    const char *qn = attr->qn;
    const char *protocol = attr->protocol;

    ns = calloc(1, sizeof(*ns));
    if (!ns)
        goto err;

#ifdef HAVE_SPDK_BDEV_OPEN_EXT
    err = spdk_bdev_open_ext(spdk_bdev_get_name(bdev), true,
                             nvme_spdk_namespace_event_cb, ns, &ns->bdev_desc);
#else
    err = spdk_bdev_open(bdev, true, nvme_spdk_namespace_delete_cb, ns,
                         &ns->bdev_desc);
#endif
    if (err)
        goto free_ns;

    ns->num_io_contexts = ctrl->ep->io_driver->num_queues;
    ns->io_contexts = calloc(ns->num_io_contexts, sizeof(*ns->io_contexts));
    if (!ns->io_contexts)
        goto close_bdev;

    ns->ctrl = ctrl;
    ns->bdev = bdev;
    ns->bdev_name = spdk_bdev_get_name(bdev);
    ns->nsid = nsid;
    ns->block_size = spdk_bdev_get_block_size(bdev);
    ns->num_blocks = spdk_bdev_get_num_blocks(bdev);
    ns->md_size = spdk_bdev_get_md_size(bdev);
    ns->zcopy = ctrl->zcopy && spdk_ext_io_zcopy_status(bdev) == 0;
    ns->effdma = ctrl->effdma_domain && spdk_ext_io_effdma_status(bdev) == 0;
#ifdef HAVE_SPDK_DMA_DEVICE_TYPE_RDMA_TCP
    ns->tcp_tx_zcopy = ctrl->zcopy && spdk_ext_io_get_bdev_status(bdev, SPDK_DMA_DEVICE_TYPE_RDMA_TCP) == 0;
#else
    ns->tcp_tx_zcopy = 0;
#endif

    strncpy(ns->bdev_type, bdev_type, sizeof(ns->bdev_type) - 1);

    /* SPDK_BDEV_IO_TYPE_ZCOPY will be set only in case of NVMeTCP tcp_rx_zcopy is 
     * supported 
     */
    if (ctrl->tcp_rx_zcopy &&
        spdk_bdev_io_type_supported(bdev, SPDK_BDEV_IO_TYPE_ZCOPY)) {
        ns->tcp_rx_zcopy = true;
        nvmx_info("Namespace nsid %d (%s) supports RX ZCOPY",
                  ns->nsid, spdk_bdev_get_name(ns->bdev));
    } else {
        ns->tcp_rx_zcopy = false;
        nvmx_info("Namespace nsid %d (%s) does not support RX ZCOPY",
                  ns->nsid, spdk_bdev_get_name(ns->bdev));
    }

    if (nvme_spdk_namespace_set_ids(ns, bdev, attr))
        goto free_io_contexts;

    if (qn)
        memcpy(ns->qn, qn, SPDK_NVME_NQN_FIELD_SIZE);

    if (protocol)
        strncpy(ns->protocol, protocol, sizeof(ns->protocol) - 1);

    TAILQ_INSERT_TAIL(&ctrl->namespaces, ns, entry);

    if (nvme_subsys_ns_add(ctrl->ep->io_driver->sconfig->spdk.subsys, ns))
        goto remove_ns;

    nvmx_info("Namespace nsid %d (%s) was added to namespace list",
              ns->nsid, spdk_bdev_get_name(ns->bdev));

    return ns;

remove_ns:
    TAILQ_REMOVE(&ctrl->namespaces, ns, entry);
free_io_contexts:
    free(ns->io_contexts);
close_bdev:
    spdk_bdev_close(ns->bdev_desc);
free_ns:
    free(ns);
err:
    return NULL;
}

/* Must be called with ctrl->namespaces_lock held */
static int nvme_spdk_namespace_attach(nvme_spdk_namespace_t *ns,
                                      nvme_spdk_ctrl_t *ctrl)
{
    return nvme_emu_ep_attach_volume(ctrl->ep, ns->nsid, ns->num_blocks * ns->block_size,
                                     ulog2(ns->block_size), ns->md_size,
                                     ns->ns_desc, ns);
}

/* Must be called with ctrl->namespaces_lock held */
static void nvme_spdk_namespace_detach(nvme_spdk_namespace_t *ns,
                                       nvme_spdk_ctrl_t *ctrl)
{
    nvme_emu_ep_detach_volume(ctrl->ep, ns->nsid);
}

static void nvme_spdk_namespace_delete_cb(void *arg)
{
    nvme_spdk_namespace_t *ns = arg;
    nvme_spdk_ctrl_t *ctrl = ns->ctrl;

    if (ns->in_deletion) {
        nvmx_warn("Got bdev event when namespace already in delete process");
        return;
    }

    pthread_spin_lock(&ctrl->namespaces_lock);
    if (ctrl->expose_namespaces)
        nvme_spdk_namespace_detach(ns, ctrl);
    nvme_spdk_namespace_delete(ns, NULL, NULL);
    pthread_spin_unlock(&ctrl->namespaces_lock);

    nvmx_info("nsid %d was hot removed\n", ns->nsid);
}

#ifdef HAVE_SPDK_BDEV_OPEN_EXT
static void nvme_spdk_namespace_resize(void *arg)
{
    uint64_t new_num_blocks;
    nvme_spdk_namespace_t *ns = arg;
    nvme_spdk_ctrl_t *ctrl = ns->ctrl;
    struct spdk_bdev *bdev = ns->bdev;

    new_num_blocks = spdk_bdev_get_num_blocks(bdev);
    if (unlikely(new_num_blocks == ns->num_blocks)) {
        nvmx_warn("nsid %d: resize to a same size\n", ns->nsid);
        return;
    }

    nvmx_info("New size detected for ns %s: %lu bytes",
              spdk_bdev_get_name(bdev), new_num_blocks * ns->block_size);
    pthread_spin_lock(&ctrl->namespaces_lock);
    if (ctrl->expose_namespaces)
        nvme_spdk_namespace_detach(ns, ctrl);
    ns->num_blocks = new_num_blocks;
    if (ctrl->expose_namespaces)
        nvme_spdk_namespace_attach(ns, ctrl);
    pthread_spin_unlock(&ctrl->namespaces_lock);

    nvmx_info("nsid %d was resized\n", ns->nsid);
}

static void
nvme_spdk_namespace_event_cb(enum spdk_bdev_event_type type,
                             struct spdk_bdev *bdev, void *event_ctx)
{
    nvmx_info("received bdev event, type: %d\n", type);

    switch (type) {
    case SPDK_BDEV_EVENT_REMOVE:
        nvme_spdk_namespace_delete_cb(event_ctx);
        break;
    case SPDK_BDEV_EVENT_RESIZE:
        nvme_spdk_namespace_resize(event_ctx);
        break;
    default:
        nvmx_error("Unsupported bdev event: type %d\n", type);
        return;
    }
}
#endif

/* TODO: Replace with rwlock to acoid lock contention on fastpath */
/* Must be called with ctrl->namespaces_lock held */
static nvme_spdk_namespace_t *
nvme_spdk_namespace_find(nvme_spdk_ctrl_t *ctrl, uint32_t nsid)
{
    nvme_spdk_namespace_t *ns;

    TAILQ_FOREACH(ns, &ctrl->namespaces, entry) {
        if(ns->nsid == nsid)
            return ns;
    }

    return NULL;
}

static inline struct nvme_spdk_namespace_io_context *
nvme_spdk_namespace_get_io_context(nvme_spdk_namespace_t *ns, uint32_t qid)
{
    /*
     * Each namespace (bdev) has its own IO channels, and each IO
     * channel is bound to a thread. In case we have #threads = #queues,
     * each namespace needs to hold io_channel-per-queue.
     * Cache these io channels because obtaining it every time requires
     * lock and scan of all bdevs - expensive.
     * Also we must get io channel from the context of the polling thread,
     * which means that we can not set it up during sq creation
     * (sq creation is done on admin thread).
     */
    if (unlikely(!ns->io_contexts[qid].io_channel)) {
        void *key_table = NULL;
        memzero_ops_t *memzero_ops;
        nvme_emulator_t *dev = ns->ctrl->ep->io_driver->dev;
        snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);

        ns->io_contexts[qid].io_channel = spdk_bdev_get_io_channel(ns->bdev_desc);
        ns->io_contexts[qid].io_thread = spdk_get_thread();
        ns->io_contexts[qid].ns = ns;

        if (ns->zcopy || ns->tcp_tx_zcopy) {
            char name[32];

            if (ns->ctrl->vf_id < 0) {
                snprintf(name, sizeof(name), "p-k-pf%d-ns%d-qid%d",
                        ns->ctrl->pf_id, ns->nsid, qid);
            } else {
                snprintf(name, sizeof(name), "p-k-pf%dvf%d-ns%d-qid%d",
                        ns->ctrl->pf_id, ns->ctrl->vf_id, ns->nsid, qid);
            }

            key_table = nvme_emu_create_zcopy_key_table(name);
        }

        memzero_ops = nvme_emu_dma_create_memzero(dev, qid);

        spdk_ext_io_context_init(&ns->io_contexts[qid].extio_ctx, snap_emu->sdev, key_table,
                                 ns->ctrl->effdma_domain, memzero_ops);
    }
    return &ns->io_contexts[qid];
}

static void nvme_spdk_check_inflight_io(nvme_io_driver_req_t *req)
{
    nvme_spdk_namespace_t *ns;
    struct nvme_spdk_namespace_io_context *io_context;

    ns = (nvme_spdk_namespace_t *)req->ns;
    io_context = nvme_spdk_namespace_get_io_context(ns, req->async_req.ops->qid);
    io_context->inflight_io--;

    if (unlikely(ns->in_deletion && io_context->inflight_io == 0))
        nvme_spdk_namespace_put_channel(io_context);
}

static inline void nvme_spdk_memory_release(nvme_async_req_t *req)
{
    if (req->memp) {
        memp_free(req->memp, req->aux_buf);
        req->aux_buf = NULL;
        memp_free(req->memp, req->io_buf);
        req->io_buf = NULL;
    }
}

static int nvme_spdk_io_buf_create(size_t len, memp_t *memp, nvme_io_driver_req_t *req)
{
    if (!memp) {
        req->async_req.io_buf = spdk_dma_malloc(len, NVME_SPDK_BUF_ALIGN_MIN, NULL);
        if (!req->async_req.io_buf)
            return NVMX_ERR;
    }

    return NVMX_OK;
}

static void nvme_spdk_sq_reset(nvme_io_driver_req_t *req)
{
     if (!req->async_req.memp) {
        spdk_dma_free(req->async_req.io_buf);
        req->async_req.io_buf = NULL;
    }
}

static void nvme_spdk_request_reset(nvme_io_driver_req_t *req)
{
    /* When not using mempool io_buf is freed once per sq */
    if (req->async_req.memp)
        nvme_spdk_memory_release(&req->async_req);
    else
        req->async_req.io_buf = NULL;
}

static void nvme_spdk_zcopy_end_done(struct spdk_bdev_io *bdev_io,
				     bool success, void *cb_arg)
{
    nvme_io_driver_req_t *io_req = (nvme_io_driver_req_t*)cb_arg;
    nvme_async_req_t *async_req = &io_req->async_req;

    nvmx_debug("RX zcopy end done: bio %p, success %d", bdev_io, success);
    async_req->comp_cb = nvme_spdk_write_to_host_done;
    async_req->comp_cb(io_req, async_req->tcp_rx_zcopy.req_status);

    spdk_bdev_free_io(bdev_io);
    io_req->bdev_io = NULL;
} 

/* Send completion on request */
void nvme_spdk_zcopy_dma_done(void *user, int status)
{
    int rc;
    nvme_io_driver_req_t *req = (nvme_io_driver_req_t*)user;
    struct spdk_bdev_io *bdev_io = req->bdev_io;

    nvmx_debug("RX zcopy DMA done: bio %p, status %d", bdev_io, status);

    req->async_req.tcp_rx_zcopy.req_status = (enum nvme_emu_status)status;

    /* Releases XLIO buffers back to their pool. */
    rc = spdk_bdev_zcopy_end(bdev_io, false, nvme_spdk_zcopy_end_done, req);
    if (rc)
        nvme_spdk_zcopy_end_done(bdev_io, false, req);
}

#ifdef HAVE_SPDK_DMA_DEVICE_TYPE_RDMA_TCP
/* Translating XLIO buffers into SGL, and send it back to the host 
 * through a new snap_dma_q_writev_lv() API extension.
 */
static void nvme_spdk_zcopy_populate_from_remote_done(struct spdk_bdev_io *bdev_io,
                                                      bool success, void *cb_arg)
{
    nvme_io_driver_req_t *req = cb_arg;
    nvme_spdk_ctrl_t *ctrl = req->ep->ctrl;
    struct iovec *iovs = NULL;
    int iovcnt = 0;
    int rc = 0;
    int i;
    int total_length = 0;

    nvmx_debug("Populated RX zcopy: bio %p, success %d", bdev_io, success);
    if (unlikely(!success)) {
        nvmx_error("success != 0");
        goto err;
    }

    spdk_bdev_io_get_iovec(bdev_io, &iovs, &iovcnt);
    if (iovcnt > SGL_MAX_DESCRIPTORS) {
        nvmx_error("Number of zcopy iovs (%d) is larger than max sgl (%d)",
                   iovcnt, SGL_MAX_DESCRIPTORS);
        goto err;
    }

    for (i = 0; i < iovcnt; i++) {
        struct ibv_sge *sge = &req->async_req.tcp_rx_zcopy.sgl[i];
        struct ibv_mr *mr;
        uint64_t length = iovs[i].iov_len;
        total_length += length;
        mr = (struct ibv_mr *)spdk_mem_map_translate(ctrl->mem_map, (uint64_t)iovs[i].iov_base, &length);
        if (!mr || length != iovs[i].iov_len) {
            nvmx_error("Translation failed: mr %p, iov_len %d, len %lu",
                       mr, iovs[i].iov_len, length);
            goto err;
        }
        sge->addr = (uint64_t)iovs[i].iov_base;
        sge->length = iovs[i].iov_len;
        sge->lkey = mr->lkey;
        nvmx_debug("SGE[%d] translation: addr %p, len %d, lkey %u",
                   i, sge->addr, sge->length, sge->lkey);
    }

    req->async_req.tcp_rx_zcopy.iov = iovs;
    req->async_req.tcp_rx_zcopy.iov_cnt = iovcnt;
    req->async_req.len = total_length;
    req->bdev_io = bdev_io;

    if (req->async_req.is_prp == false) {
        nvme_sgl_rw_iov_data_nb(&req->async_req);
        return;
    }
    else {
        rc = nvme_prp_rw(&req->async_req);
        if (unlikely(rc != NVMX_INPROGRESS)) {
            nvmx_error("NVMX is not INPROGRESS");
            goto err;
        }

        return;
    }

err:
    nvmx_error("nvme_spdk_zcopy_populate_from_remote_done ended with error. bdev_io = %p", bdev_io);
    req->bdev_io = bdev_io;
    req->async_req.tcp_rx_zcopy.req_status = NVMX_ERR;
    if (!success) {
        nvme_spdk_zcopy_end_done(bdev_io, false, req);
    } else {
        rc = spdk_bdev_zcopy_end(bdev_io, false, nvme_spdk_zcopy_end_done, req);
        if (rc)
            nvme_spdk_zcopy_end_done(bdev_io, false, req);
    }
}
#endif

static void nvme_spdk_write_to_host_done(void *user, int status)
{
    nvme_io_driver_req_t *req = user;

    nvme_spdk_memory_release(&req->async_req);
    req->comp_cb(req->comp_cb_user, status);
    nvme_spdk_check_inflight_io(req);
}

static void nvme_spdk_read_from_remote_done(struct spdk_bdev_io *bdev_io,
                                            bool success, void *cb_arg)
{
    nvme_io_driver_req_t *req = cb_arg;
    int rc;

    if (unlikely(!success)) {
        rc = NVMX_ERR;
        goto err;
    }

    if (req->async_req.is_prp == false)
        rc = nvme_sgl_rw_nb(&req->async_req);
    else
        rc = nvme_prp_rw(&req->async_req);

    if (unlikely(rc != NVMX_INPROGRESS))
        goto err;

    spdk_bdev_free_io(bdev_io);
    return;

err:
    nvme_spdk_memory_release(&req->async_req);
    req->comp_cb(req->comp_cb_user, NVMX_ERR);
    spdk_bdev_free_io(bdev_io);
    nvme_spdk_check_inflight_io(req);
}

static void nvme_spdk_write_to_remote_done(struct spdk_bdev_io *bdev_io,
                                           bool success, void *cb_arg)
{
    nvme_io_driver_req_t *req = cb_arg;

    nvme_spdk_memory_release(&req->async_req);
    req->comp_cb(req->comp_cb_user, success ? NVMX_OK : NVMX_ERR);
    spdk_bdev_free_io(bdev_io);
    nvme_spdk_check_inflight_io(req);
}

static void nvme_spdk_read_from_host_done(void *user, int status)
{
    int rc;
    nvme_io_driver_req_t *io_req = user;
    nvme_async_req_t *req = &io_req->async_req;
    nvme_queue_ops_t *ops = req->ops;
    nvme_spdk_namespace_t *ns = io_req->ns;
    struct spdk_io_channel *spdk_io_ch;
    struct nvme_spdk_namespace_io_context *io_context = nvme_spdk_namespace_get_io_context(ns, ops->qid);

    spdk_io_ch = io_context->io_channel;
    if (unlikely(!spdk_io_ch)) {
        nvmx_error("Cannot get SPDK IO channel");
        goto err;
    }

    rc = spdk_bdev_write(ns->bdev_desc,
                         spdk_io_ch,
                         req->io_buf, io_req->disk_offset, req->len,
                         nvme_spdk_write_to_remote_done, io_req);
    if (rc)
        goto err;

    return;

err:
    nvme_spdk_memory_release(&io_req->async_req);
    io_req->comp_cb(io_req->comp_cb_user, NVMX_ERR);
    nvme_spdk_check_inflight_io(io_req);
}

#ifdef HAVE_SPDK_EXT_IO

static void nvme_spdk_remote_iov_done(struct spdk_bdev_io *bdev_io,
                                      bool success, void *cb_arg)
{
    nvme_io_driver_req_t *req = cb_arg;

    req->comp_cb(req->comp_cb_user, success ? NVMX_OK : NVMX_ERR);
    spdk_bdev_free_io(bdev_io);
    nvme_spdk_check_inflight_io(req);
}

static void nvme_spdk_write_remote_iov(void *user, int status)
{
    int rc;
    nvme_io_driver_req_t *io_req = user;
    nvme_async_req_t *req = &io_req->async_req;
    nvme_spdk_namespace_t *ns = io_req->ns;
    nvme_queue_ops_t *ops = req->ops;
    struct spdk_io_channel *spdk_io_ch;
    struct nvme_spdk_namespace_io_context *io_context = nvme_spdk_namespace_get_io_context(ns, ops->qid);

    if (!req->is_prp) {
        snap_conf_t *config = ns->ctrl->ep->io_driver->sconfig;
        const size_t alignment = snap_config_get_zcopy_align(config);

        if (unlikely(status != NVMX_OK) ||
            unlikely(!spdk_ext_io_validate_iov(ns->bdev, req->iov, req->iov_cnt, alignment))) {
            req->dma_cmd.op = NVME_REQ_DMA_FROM_HOST;
            req->comp_cb = nvme_spdk_read_from_host_done;
            io_req->ext_io = false;
            nvme_spdk_rw_do(io_req);
            return;
        }
    }

    if (unlikely(status != NVMX_OK))
        goto err;

    status = NVMX_ERR;

    spdk_io_ch = io_context->io_channel;
    if (unlikely(!spdk_io_ch)) {
        nvmx_error("Cannot get SPDK IO channel");
        goto err;
    }

    rc = spdk_bdev_writev_blocks_ext(ns->bdev_desc,
                                spdk_io_ch,
                                req->iov,
                                req->iov_cnt,
                                io_req->disk_offset / ns->block_size,
                                req->len / ns->block_size,
                                nvme_spdk_remote_iov_done,
                                io_req,
                                &io_context->extio_ctx.ext_io_opts);
    if (rc == 0)
        return;

err:
    io_req->comp_cb(io_req->comp_cb_user, status);
    nvme_spdk_check_inflight_io(io_req);
}

static void nvme_spdk_read_remote_iov(void *user, int status)
{
    int rc;
    nvme_io_driver_req_t *io_req = user;
    nvme_async_req_t *req = &io_req->async_req;
    nvme_spdk_namespace_t *ns = io_req->ns;
    nvme_queue_ops_t *ops = req->ops;
    struct spdk_io_channel *spdk_io_ch;
    struct nvme_spdk_namespace_io_context *io_context = nvme_spdk_namespace_get_io_context(ns, ops->qid);

    if (!req->is_prp) {
        snap_conf_t *config = ns->ctrl->ep->io_driver->sconfig;
        const size_t alignment = snap_config_get_zcopy_align(config);

        if (unlikely(status != NVMX_OK) ||
            unlikely(!spdk_ext_io_validate_iov(ns->bdev, req->iov, req->iov_cnt, alignment))) {
            req->dma_cmd.op = NVME_REQ_DMA_TO_HOST;
            req->comp_cb = nvme_spdk_write_to_host_done;
            io_req->ext_io = false;
            nvme_spdk_rw_do(io_req);
            return;
        }
    }

    if (unlikely(status != NVMX_OK))
        goto err;

    status = NVMX_ERR;

    spdk_io_ch = io_context->io_channel;
    if (unlikely(!spdk_io_ch)) {
        nvmx_error("Cannot get SPDK IO channel");
        goto err;
    }

    rc = spdk_bdev_readv_blocks_ext(ns->bdev_desc,
                               spdk_io_ch,
                               req->iov,
                               req->iov_cnt,
                               io_req->disk_offset / ns->block_size,
                               req->len / ns->block_size,
                               nvme_spdk_remote_iov_done,
                               io_req,
                               &io_context->extio_ctx.ext_io_opts);
    if (rc == 0)
        return;

err:
    io_req->comp_cb(io_req->comp_cb_user, status);
    nvme_spdk_check_inflight_io(io_req);
}

static inline bool ext_io_check(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw, bool prp)
{
    nvme_spdk_namespace_t *ns = req->ns;
    snap_conf_t *config = ns->ctrl->ep->io_driver->sconfig;
    struct iovec iov;

    if (!ns->zcopy && !ns->effdma && !ns->tcp_tx_zcopy)
        return false;

    if (prp) {
        iov.iov_base = (void *)le64_to_cpu(rw->common.dptr.prp1);

        /* cannot use zcopy if the first prp is not bdev aligned */
        if (!spdk_ext_io_validate_iov(ns->bdev, &iov, 1, snap_config_get_zcopy_align(config)))
            return false;
    }

    return true;
}
#else
static void nvme_spdk_read_remote_iov(void *user, int status)
{

}
static void nvme_spdk_write_remote_iov(void *user, int status)
{

}

static inline bool ext_io_check(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw, bool prp)
{
    return false;
}
#endif

static int nvme_spdk_read(nvme_io_driver_req_t *req)
{
    int rc;
    nvme_queue_ops_t *ops = req->async_req.ops;
    nvme_spdk_namespace_t *ns = req->ns;
    struct spdk_io_channel *spdk_io_ch;
    struct nvme_spdk_namespace_io_context *io_context = nvme_spdk_namespace_get_io_context(ns, ops->qid);

    spdk_io_ch = io_context->io_channel;
    if (unlikely(!spdk_io_ch)) {
        nvmx_error("Cannot get SPDK IO channel");
        return NVMX_ERR;
    }

    rc = spdk_bdev_read(ns->bdev_desc,
                        spdk_io_ch,
                        req->async_req.io_buf, req->disk_offset, req->async_req.len,
                        nvme_spdk_read_from_remote_done, req);
    if (rc == 0)
        io_context->inflight_io++;

    return rc ? NVMX_ERR : NVMX_INPROGRESS;
}

static int nvme_spdk_write(nvme_io_driver_req_t *req)
{
    int ret;
    struct nvme_spdk_namespace_io_context *io_context;

    if (req->async_req.is_prp == false)
        ret = nvme_sgl_rw_nb(&req->async_req);
    else
        ret = nvme_prp_rw(&req->async_req);

    if (ret == NVMX_INPROGRESS) {
        io_context = nvme_spdk_namespace_get_io_context(req->ns, req->async_req.ops->qid);
        io_context->inflight_io++;
    }

    return ret;
}

static void nvme_spdk_mem_ready(void *data, void *user)
{
    nvme_io_driver_req_t *req = user;
    int rc;

    req->async_req.io_buf = data;
    if (req->async_req.dma_cmd.op == NVME_REQ_DMA_TO_HOST)
        rc = nvme_spdk_read(req);
    else
        rc = nvme_spdk_write(req);

    if (rc != NVMX_INPROGRESS) {
        nvme_spdk_memory_release(&req->async_req);
        req->comp_cb(req->comp_cb_user, NVMX_ERR);
    }
}

static inline int nvme_spdk_rw_do(nvme_io_driver_req_t *req)
{
    void *mem;

    if (req->ext_io) {
        int ret;
        struct nvme_spdk_namespace_io_context *io_context;

        if (req->async_req.is_prp) {
             ret = nvme_prp_rw(&req->async_req);
        } else {
            nvme_sgl_rw_iov_nb(&req->async_req);
            ret = NVMX_INPROGRESS;
        }
        if (ret == NVMX_INPROGRESS) {
            io_context = nvme_spdk_namespace_get_io_context(req->ns, req->async_req.ops->qid);
            io_context->inflight_io++;
        }
        return ret;
    } else {
        if (!req->async_req.memp) {
            if (req->async_req.dma_cmd.op == NVME_REQ_DMA_TO_HOST)
                return nvme_spdk_read(req);
            else
                return nvme_spdk_write(req);
        }
        else {
            mem = memp_alloc_ex(req, nvme_spdk_mem_ready);
            if (mem) {
                req->async_req.io_buf = mem;
                if (req->async_req.dma_cmd.op == NVME_REQ_DMA_TO_HOST)
                    return nvme_spdk_read(req);
                else
                    return nvme_spdk_write(req);
            } else if (errno)
                return NVMX_ERR;
            return NVMX_INPROGRESS;
        }
    }
}

static void nvme_spdk_aux_mem_ready(void *data, void *user)
{
    int rc;
    nvme_io_driver_req_t *req = user;

    req->async_req.aux_buf = data;
    rc = nvme_spdk_rw_do(req);
    if (rc != NVMX_INPROGRESS) {
        nvme_spdk_memory_release(&req->async_req);
        req->comp_cb(req->comp_cb_user, NVMX_ERR);
    }
}

static int nvme_spdk_rw(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw, bool read, bool prp)
{
    uint16_t nlb  = le16_to_cpu(rw->nlb) + 1;
    uint64_t slba = le64_to_cpu(rw->slba);
    nvme_spdk_namespace_t *ns = req->ns;
    uint64_t offset, count;
    bool need_aux;
    void *mem;


    offset = slba * ns->block_size;
    count = nlb * ns->block_size;
    
    req->async_req.is_tcp_rx_zcopy = false;
    
    req->ext_io = ext_io_check(req, rw, prp);
    if (read) {
#ifdef HAVE_SPDK_DMA_DEVICE_TYPE_RDMA_TCP
        if (ns->tcp_rx_zcopy) {
            struct nvme_spdk_namespace_io_context *io_context;
            nvme_async_req_t *async_req = &req->async_req;
            nvme_queue_ops_t *ops = async_req->ops;
            struct spdk_io_channel *spdk_io_ch;
            int rc;

            req->async_req.dma_cmd.op = NVME_REQ_DMA_TO_HOST;
            req->async_req.dma_cmd.writev2v = true;
            req->async_req.comp_cb = nvme_spdk_zcopy_dma_done;

            req->async_req.is_tcp_rx_zcopy = true;
            io_context = nvme_spdk_namespace_get_io_context(ns, ops->qid);
            spdk_io_ch = io_context->io_channel;
            if (unlikely(!spdk_io_ch)) {
                nvmx_error("Cannot get SPDK IO channel");
                return NVMX_ERR;
            }
		
            /* Prepare write request to host, will happen after read completion */
            req->disk_offset = offset;
            req->async_req.is_prp = prp;
            req->async_req.len = count;
            req->async_req.comp_cb_user = req;

            if (prp) {
                req->async_req.prp.prp1 = le64_to_cpu(rw->common.dptr.prp1);
                req->async_req.prp.prp2 = le64_to_cpu(rw->common.dptr.prp2);
            }
            else {
                memcpy(&req->async_req.sgl.sgl, &rw->common.dptr.sgl, sizeof(nvme_sgl_desc_t));
            }

            rc = spdk_bdev_zcopy_start(ns->bdev_desc, spdk_io_ch, NULL, 0, 
                                        slba, nlb, 
                                        true,  
                                        nvme_spdk_zcopy_populate_from_remote_done, req);
            if (rc) {
                nvmx_error("spdk_bdev_zcopy_start failed: rc = %d", rc);
                return NVMX_ERR;
            }
            io_context->inflight_io++;

            return NVMX_INPROGRESS;
        }
#endif
        /* writev2v() is currently used only with tcp_rx_zcopy. */
        req->async_req.dma_cmd.writev2v = false;

        /* Read flow should use IOV only for RDMA, not for TCP. */
        if (req->ext_io && !ns->tcp_tx_zcopy) {
            req->async_req.dma_cmd.op = NVME_REQ_IOV;
            req->async_req.comp_cb = nvme_spdk_read_remote_iov;
        }
        else {
            req->async_req.dma_cmd.op = NVME_REQ_DMA_TO_HOST;
            req->async_req.comp_cb = nvme_spdk_write_to_host_done;
        }
    }
    else {
        if (req->ext_io) {
            req->async_req.dma_cmd.op = NVME_REQ_IOV;
            req->async_req.comp_cb = nvme_spdk_write_remote_iov;
        }
        else {
            req->async_req.dma_cmd.op = NVME_REQ_DMA_FROM_HOST;
            req->async_req.comp_cb = nvme_spdk_read_from_host_done;
        }
    }

    /* Prepare write request to host, will happen after read completion */
    req->disk_offset = offset;
    req->async_req.is_prp = prp;
    req->async_req.len = count;
    req->async_req.comp_cb_user = req;

    if (prp) {
        req->async_req.prp.prp1 = le64_to_cpu(rw->common.dptr.prp1);
        req->async_req.prp.prp2 = le64_to_cpu(rw->common.dptr.prp2);
        need_aux = nvme_prp_needs_aux(&req->async_req);
    }
    else {
        memcpy(&req->async_req.sgl.sgl, &rw->common.dptr.sgl, sizeof(nvme_sgl_desc_t));
        need_aux = nvme_sgl_needs_aux(&req->async_req);
    }

    if (req->async_req.memp && need_aux) {
        mem = memp_alloc(req->async_req.memp, req->async_req.aux_len,
                            req, nvme_spdk_aux_mem_ready, NULL,
                            req->async_req.thread_id);
        if (!mem)
            return errno ? NVMX_ERR : NVMX_INPROGRESS;
        req->async_req.aux_buf = mem;
    }

    return nvme_spdk_rw_do(req);
}

static int nvme_spdk_read_prp_nb(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw)
{
    return nvme_spdk_rw(req, rw, true, true);
}

static int nvme_spdk_write_prp_nb(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw)
{
    return nvme_spdk_rw(req, rw, false, true);
}

static int nvme_spdk_read_sgl_nb(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw)
{
    return nvme_spdk_rw(req, rw, true, false);
}

static int nvme_spdk_write_sgl_nb(nvme_io_driver_req_t *req, const nvme_cmd_rw_t *rw)
{
    return nvme_spdk_rw(req, rw, false, false);
}

static int nvme_spdk_read_prp(nvme_emu_ep_t *ep,
                              nvme_io_driver_req_t *req,
                              const nvme_cmd_rw_t *rw)
{
    return NVMX_ERR_NOTSUPPORTED;
}

static int nvme_spdk_read_sgl(nvme_emu_ep_t *ep,
                              nvme_io_driver_req_t *req,
                              const nvme_cmd_rw_t *rw)
{
    return NVMX_ERR_NOTSUPPORTED;
}

static int nvme_spdk_write_prp(nvme_emu_ep_t *ep,
                               nvme_io_driver_req_t *req,
                               const nvme_cmd_rw_t *rw)
{
    return NVMX_ERR_NOTSUPPORTED;
}

static int nvme_spdk_write_sgl(nvme_emu_ep_t *ep,
                               nvme_io_driver_req_t *req,
                               const nvme_cmd_rw_t *rw)
{
    return NVMX_ERR_NOTSUPPORTED;
}

static int nvme_spdk_create_cq(nvme_emu_ep_t *ep, uint16_t cqid,
                               uint16_t size)
{
    return NVMX_OK;
}

static void nvme_spdk_delete_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
}

static int nvme_spdk_activate_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
    return NVMX_OK;
}

static void nvme_spdk_deactivate_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
}

static int nvme_spdk_create_sq(nvme_emu_ep_t *ep, uint16_t sqid,
                               uint16_t size, uint16_t cqid)
{
    return NVMX_OK;
}

static void nvme_spdk_delete_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
    nvme_spdk_ctrl_t *ctrl = ep->ctrl;
    nvme_spdk_namespace_t *ns;

    /* Iterate over all namespaces and put sq io channel.
     * Next time sq may be created on another thread.
     * For example driver creates sq per cq but
     * next time it creates 2 sqs per cq
     */
    pthread_spin_lock(&ctrl->namespaces_lock);
    TAILQ_FOREACH(ns, &ctrl->namespaces, entry) {
        if (ns->io_contexts[sqid].io_thread)
            spdk_thread_send_msg(ns->io_contexts[sqid].io_thread,
                                 nvme_spdk_namespace_put_channel,
                                 &ns->io_contexts[sqid]);
    }
    pthread_spin_unlock(&ctrl->namespaces_lock);
}

static int nvme_spdk_activate_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
    return NVMX_OK;
}

static void nvme_spdk_deactivate_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
}

/* Must be called with &nvme_spdk_ctrl_list_lock held */
static nvme_spdk_ctrl_t *nvme_spdk_find_ctrl_by_name(const char *name)
{
    nvme_spdk_ctrl_t *ctrl;

    TAILQ_FOREACH(ctrl, &nvme_spdk_ctrl_list, entry) {
        if(!strncmp(ctrl->name, name, NVME_SPDK_NAMESPACE_NAME_MAXLEN))
            return ctrl;
    }

    return NULL;
}

/* Must be called with &nvme_spdk_ctrl_list_lock held */
static nvme_spdk_ctrl_t *
nvme_spdk_find_ctrl_by_nqn_cntlid(const char *subnqn, int cntlid)
{
    nvme_spdk_ctrl_t *ctrl;
    struct nvme_subsystem *subsys;
    snap_conf_t *config;

    /* A cntlid is valid only when the NVMe subsystem
       which include this cntlid is exist */
    subsys = subsystem_nvme_find(subnqn);
    if (!subsys) {
        nvmx_error("subnqn: %s is not exist", subnqn);
        return NULL;
    }

    TAILQ_FOREACH(ctrl, &nvme_spdk_ctrl_list, entry) {
        config = ctrl->ep->io_driver->sconfig;
        if (ctrl->cntlid == cntlid &&
            !strcmp(subnqn, snap_config_get_subnqn(config)))
            return ctrl;
    }

    return NULL;
}

/* For backward compatible consideration, we support two approach
 * to find controller: by controller name or by nqn/controller id
 * pair. this function can be removed when only support to find
 * controller by nqn and controller id pair.
 *
 * Must be called with &nvme_spdk_ctrl_list_lock held */
static nvme_spdk_ctrl_t *nvme_spdk_find_ctrl(const char *subnqn,
                                             int cntlid,
                                             const char *name)
{
    if (cntlid != -1)
        return nvme_spdk_find_ctrl_by_nqn_cntlid(subnqn, cntlid);
    else
        return nvme_spdk_find_ctrl_by_name(name);
}

static int nvme_spdk_ctrl_mem_notify(void *ctx,
                                     struct spdk_mem_map *map,
                                     enum spdk_mem_map_notify_action action,
                                     void *vaddr, size_t size)
{
    nvme_spdk_ctrl_t *ctrl = ctx;
    struct ibv_mr *mr;
    int rc = -1;

    nvmx_info("SPDK memory event: action %d, addr %p, size %lu", action, vaddr, size);
    switch (action) {
    case SPDK_MEM_MAP_NOTIFY_REGISTER:
	    /* Only LOCAL_READ access is required for RX zcopy */
	    mr = ibv_reg_mr(ctrl->pd, vaddr, size, 0);
	    if (!mr) {
		    nvmx_error("Failed to register MR");
		    return -1;
	    }

	    nvmx_debug("Registered MR for SPDK memory: lkey %u", mr->lkey);
	    rc = spdk_mem_map_set_translation(map, (uint64_t)vaddr, size, (uint64_t)mr);
	    if (rc) {
		    nvmx_error("Failed to set translation");
		    ibv_dereg_mr(mr);
	    }
        break;
    case SPDK_MEM_MAP_NOTIFY_UNREGISTER:
	    mr = (struct ibv_mr *)spdk_mem_map_translate(map, (uint64_t)vaddr, NULL);
	    if (!mr) {
		    nvmx_error("Failed to find MR to unregister");
		    return -1;
	    }

	    nvmx_debug("Unregister MR for SPDK memory: lkey %u", mr->lkey);
	    rc = ibv_dereg_mr(mr);
        break;
    }

    return rc;
}

/* are_contiguous() operation should check if two addresses belong to contiguous 
 * mapping. For RDMA buffers being contiguous means that they should belong to 
 * the same MR. In our implementation we translate addresses to MRs. 
 * “addr1” and “addr2” passed to “are_contiguous()” are already translated 
 * addresses, i.e. MRs. So, it is enough to compare for equality to find out 
 * if it is the same MR or not.
*/
static int nvme_spdk_ctrl_mem_check_contiguous(uint64_t addr1, uint64_t addr2)
{
    return addr1 == addr2;
}

int nvme_spdk_create_ctrl(nvme_emu_ep_t *ep)
{
    nvme_spdk_ctrl_t *ctrl;
    int err;
    const char *rdma_dev;
    struct snap_context *sctx;

    if (ep->io_driver->nr_endpoints > 1) {
        nvmx_error("SPDK backend only supports one endpoint");
        err = -ENOTSUP;
        goto err;
    }

    ctrl = calloc(1, sizeof(*ctrl));
    if (!ctrl) {
        err = -ENOMEM;
        goto err;
    }

    ctrl->ep = ep;
    ctrl->cntlid = snap_config_get_cntlid(ep->io_driver->sconfig);

    rdma_dev = snap_config_get_emu_manager(ep->io_driver->sconfig);
    ctrl->pf_id = snap_config_get_pci_func(ep->io_driver->sconfig);
    ctrl->vf_id = snap_config_get_pci_vfunc(ep->io_driver->sconfig);

    ctrl->effdma_domain = effdma_domain_find(ctrl->pf_id, ctrl->vf_id);

    sctx = mlnx_snap_get_snap_context(rdma_dev);
    if ((snap_env_getenv(NVME_SNAP_ZCOPY) != 0) && sctx->nvme_caps.crossing_vhca_mkey)
        ctrl->zcopy = true;

    if (snap_env_getenv(NVME_SNAP_TCP_RX_ZCOPY) != 0) {
        ctrl->tcp_rx_zcopy = true;
        ctrl->pd = nvme_emu_get_pd(ctrl->ep->io_driver->dev);
        if (!ctrl->pd) {
		    nvmx_error("Failed to get PD for controller %p", ctrl);
            goto free_ctrl;
        }

        const struct spdk_mem_map_ops mem_map_ops = {
            .notify_cb = nvme_spdk_ctrl_mem_notify,
            .are_contiguous = nvme_spdk_ctrl_mem_check_contiguous
        };

        ctrl->mem_map = spdk_mem_map_alloc(0, &mem_map_ops, ctrl);
        if (!ctrl->mem_map) {
            nvmx_error("Failed to allocate SPDK memory map");
            err = -ENOMEM;
            goto free_ctrl;
        }

        nvmx_info("RX zcopy is enabled");
    } else {
        nvmx_info("RX zcopy is disabled");
        ctrl->tcp_rx_zcopy = false;
    }
    
    if (ctrl->vf_id >= 0)
        snprintf(ctrl->name, NVME_EMU_NAME_MAXLEN,
                "%s%dpf%dvf%d", NVME_EMU_NAME_PREFIX,
                rdma_dev_name_to_id(rdma_dev), ctrl->pf_id, ctrl->vf_id);
    else
        snprintf(ctrl->name, NVME_EMU_NAME_MAXLEN,
                "%s%dpf%d", NVME_EMU_NAME_PREFIX,
                rdma_dev_name_to_id(rdma_dev), ctrl->pf_id);

    TAILQ_INIT(&ctrl->namespaces);
    err = pthread_spin_init(&ctrl->namespaces_lock, 0);
    if (err)
        goto free_ctrl;

    /* Initialize ep ops */
    ep->ctrl = ctrl;

    ep->ep_ops.read_prp_nb  = nvme_spdk_read_prp_nb;
    ep->ep_ops.write_prp_nb = nvme_spdk_write_prp_nb;
    ep->ep_ops.read_sgl_nb  = nvme_spdk_read_sgl_nb;
    ep->ep_ops.write_sgl_nb = nvme_spdk_write_sgl_nb;
    ep->ep_ops.io_buf_create = nvme_spdk_io_buf_create;
    ep->ep_ops.request_reset  = nvme_spdk_request_reset;
    ep->ep_ops.sq_reset = nvme_spdk_sq_reset;

    ep->ep_ops.read_prp  = nvme_spdk_read_prp;
    ep->ep_ops.write_prp = nvme_spdk_write_prp;
    ep->ep_ops.read_sgl  = nvme_spdk_read_sgl;
    ep->ep_ops.write_sgl = nvme_spdk_write_sgl;

    ep->ep_ops.create_cq     = nvme_spdk_create_cq;
    ep->ep_ops.delete_cq     = nvme_spdk_delete_cq;
    ep->ep_ops.activate_cq   = nvme_spdk_activate_cq;
    ep->ep_ops.deactivate_cq = nvme_spdk_deactivate_cq;

    ep->ep_ops.create_sq     = nvme_spdk_create_sq;
    ep->ep_ops.delete_sq     = nvme_spdk_delete_sq;
    ep->ep_ops.activate_sq   = nvme_spdk_activate_sq;
    ep->ep_ops.deactivate_sq = nvme_spdk_deactivate_sq;

    pthread_mutex_lock(&nvme_spdk_ctrl_list_lock);
    TAILQ_INSERT_TAIL(&nvme_spdk_ctrl_list, ctrl, entry);
    pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);

    return 0;

free_ctrl:
    if (ctrl->mem_map)
        spdk_mem_map_free(&ctrl->mem_map);
    free(ctrl);
err:
    return err;
}

void nvme_spdk_delete_ctrl_end(void *arg)
{
    nvme_spdk_ctrl_t *ctrl = arg;

    /* Only when all namespaces are released, it is safe to destroy ctrl */
    if (!TAILQ_EMPTY(&ctrl->namespaces))
        return;

    pthread_spin_destroy(&ctrl->namespaces_lock);

    if (ctrl->mem_map) {
	    nvmx_debug("Deallocate SPDK memory map %p", ctrl->mem_map);
	    spdk_mem_map_free(&ctrl->mem_map);
    }

    free(ctrl);
}

void nvme_spdk_delete_ctrl(nvme_spdk_ctrl_t *ctrl)
{
    nvme_spdk_namespace_t *ns;
    bool delete_sync;

    pthread_mutex_lock(&nvme_spdk_ctrl_list_lock);
    TAILQ_REMOVE(&nvme_spdk_ctrl_list, ctrl, entry);
    pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);

    pthread_spin_lock(&ctrl->namespaces_lock);
    delete_sync = !!TAILQ_EMPTY(&ctrl->namespaces);
    TAILQ_FOREACH(ns, &ctrl->namespaces, entry)
        nvme_spdk_namespace_delete(ns, nvme_spdk_delete_ctrl_end, ctrl);
    pthread_spin_unlock(&ctrl->namespaces_lock);

    if (delete_sync)
        nvme_spdk_delete_ctrl_end(ctrl);
}

int nvme_spdk_start_ctrl(nvme_emu_ep_t *ep)
{
    nvme_spdk_ctrl_t *ctrl = ep->ctrl;
    nvme_spdk_namespace_t *ns;

    pthread_spin_lock(&ctrl->namespaces_lock);
    TAILQ_FOREACH(ns, &ctrl->namespaces, entry) {
        if (nvme_spdk_namespace_attach(ns, ctrl)) {
            nvmx_warn("Failed to attach namespace %d. Deleting it.", ns->nsid);
            nvme_spdk_namespace_delete_destroy(ns, true);
        }
    }
    ctrl->expose_namespaces = true;
    pthread_spin_unlock(&ctrl->namespaces_lock);

    return 0;
}

void nvme_spdk_stop_ctrl(nvme_emu_ep_t *ep)
{
    nvme_spdk_ctrl_t *ctrl = ep->ctrl;
    nvme_spdk_namespace_t *ns;

    pthread_spin_lock(&ctrl->namespaces_lock);
    TAILQ_FOREACH(ns, &ctrl->namespaces, entry)
        nvme_spdk_namespace_detach(ns, ctrl);

    ctrl->expose_namespaces = false;
    pthread_spin_unlock(&ctrl->namespaces_lock);
}

static const struct spdk_json_object_decoder
nvme_spdk_rpc_namespace_attach_decoder[] = {
    {
        "nsid",
        offsetof(struct nvme_spdk_rpc_namespace_attach_attr, nsid),
        spdk_json_decode_uint32
    },
    {
        "bdev_type",
        offsetof(struct nvme_spdk_rpc_namespace_attach_attr, bdev_type),
        spdk_json_decode_string
    },
    {
        "bdev",
        offsetof(struct nvme_spdk_rpc_namespace_attach_attr, bdev_name),
        spdk_json_decode_string
    },
    {
        "ctrl",
        offsetof(struct nvme_spdk_rpc_namespace_attach_attr, emu_name),
        spdk_json_decode_string,
        true
    },
    {
        "subnqn",
        offsetof(struct nvme_spdk_rpc_namespace_attach_attr, subnqn),
        spdk_json_decode_string,
        true
    },
    {
        "cntlid",
        offsetof(struct nvme_spdk_rpc_namespace_attach_attr, cntlid),
        spdk_json_decode_int32,
        true
    },
    {
        "qn",
        offsetof(struct nvme_spdk_rpc_namespace_attach_attr, qn),
        spdk_json_decode_string,
        true
    },
    {
        "protocol",
        offsetof(struct nvme_spdk_rpc_namespace_attach_attr, protocol),
        spdk_json_decode_string,
        true
    },
    {
        "uuid",
        offsetof(struct nvme_spdk_rpc_namespace_attach_attr, uuid),
        spdk_json_decode_string,
        true
    },
    {
        "nguid",
        offsetof(struct nvme_spdk_rpc_namespace_attach_attr, nguid),
        spdk_json_decode_string,
        true
    },
    {
        "eui64",
        offsetof(struct nvme_spdk_rpc_namespace_attach_attr, eui64),
        spdk_json_decode_string,
        true
    },
};

static void nvme_spdk_rpc_namespace_attach(struct spdk_jsonrpc_request *request,
                                           const struct spdk_json_val *params)
{
    struct nvme_spdk_rpc_namespace_attach_attr *attr = NULL;
    struct spdk_json_write_ctx *w;
    nvme_spdk_ctrl_t *ctrl;
    nvme_spdk_namespace_t *ns;
    struct spdk_bdev *bdev;
    int err;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    // Set invalid index, to identify when value was not decoded
    attr->cntlid = -1;

    if (spdk_json_decode_object(params,
                             nvme_spdk_rpc_namespace_attach_decoder,
                             SPDK_COUNTOF(nvme_spdk_rpc_namespace_attach_decoder),
                             attr)) {
        nvmx_error("Failed to decode nvme_spdk_rpc_namespace_attach parameters");
        goto free_attr;
    }

    if (strncmp(attr->bdev_type, "spdk", 4)) {
        nvmx_error("BDEV type %s is not supported", attr->bdev_type);
        goto free_attr_fields;
    }

    if (attr->nsid < 1) {
        nvmx_error("Invalid NSID %d is provided", attr->nsid);
        goto free_attr_fields;
    }

    pthread_mutex_lock(&nvme_spdk_ctrl_list_lock);
    ctrl = nvme_spdk_find_ctrl(attr->subnqn, attr->cntlid, attr->emu_name);
    if (!ctrl) {
        nvmx_error("No controller instance detected");
        pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);
        goto free_attr_fields;
    }

    pthread_spin_lock(&ctrl->namespaces_lock);
    ns = nvme_spdk_namespace_find(ctrl, attr->nsid);
    if (ns) {
        nvmx_error("Namespace %d is already in use by Bdev %s",
                   ns->nsid, ns->bdev_name);
        pthread_spin_unlock(&ctrl->namespaces_lock);
        pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);
        goto free_attr_fields;
    }

    /* SPEC: nsids must be unique per subsystem */
    if (nvme_subsys_ns_exists(ctrl->ep->io_driver->sconfig->spdk.subsys, attr->nsid)) {
        nvmx_error("Namespace %d already exists", attr->nsid);
        pthread_spin_unlock(&ctrl->namespaces_lock);
        pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);
        goto free_attr_fields;
    }

    bdev = spdk_bdev_get_by_name(attr->bdev_name);
    if (!bdev) {
        nvmx_error("Bdev %s cannot be found", attr->bdev_name);
        pthread_spin_unlock(&ctrl->namespaces_lock);
        pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);
        goto free_attr_fields;
    }

    ns = nvme_spdk_namespace_create(ctrl, bdev, attr);
    if (!ns) {
        nvmx_error("Namespace %d cannot be created", attr->nsid);
        pthread_spin_unlock(&ctrl->namespaces_lock);
        pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);
        goto free_attr_fields;
    }

    if (ctrl->expose_namespaces) {
        err = nvme_spdk_namespace_attach(ns, ctrl);
        if (err) {
            nvmx_error("Failed to attach namespace %d", attr->nsid);
            nvme_spdk_namespace_delete_destroy(ns, true);
            pthread_spin_unlock(&ctrl->namespaces_lock);
            pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);
            goto free_attr_fields;
        }
    }

    pthread_spin_unlock(&ctrl->namespaces_lock);
    pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_bool(w, true);
    spdk_jsonrpc_end_result(request, w);

    free(attr->bdev_name);
    free(attr->bdev_type);
    free(attr->emu_name);
    free(attr->subnqn);
    free(attr->qn);
    free(attr->protocol);
    free(attr->uuid);
    free(attr->nguid);
    free(attr->eui64);
    free(attr);
    return;

free_attr_fields:
    free(attr->bdev_name);
    free(attr->bdev_type);
    free(attr->emu_name);
    free(attr->subnqn);
    free(attr->qn);
    free(attr->protocol);
    free(attr->uuid);
    free(attr->nguid);
    free(attr->eui64);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}
SPDK_RPC_REGISTER("controller_nvme_namespace_attach",
                  nvme_spdk_rpc_namespace_attach, SPDK_RPC_RUNTIME)

struct nvme_spdk_rpc_namespace_detach_attr {
    uint32_t nsid;
    char *emu_name;
    char *subnqn;
    int cntlid;
};

static const struct spdk_json_object_decoder
nvme_spdk_rpc_namespace_detach_decoder[] = {
    {
        "nsid",
        offsetof(struct nvme_spdk_rpc_namespace_detach_attr, nsid),
        spdk_json_decode_uint32
    },
    {
        "ctrl",
        offsetof(struct nvme_spdk_rpc_namespace_detach_attr, emu_name),
        spdk_json_decode_string,
        true
    },
    {
        "subnqn",
        offsetof(struct nvme_spdk_rpc_namespace_detach_attr, subnqn),
        spdk_json_decode_string,
        true
    },
    {
        "cntlid",
        offsetof(struct nvme_spdk_rpc_namespace_detach_attr, cntlid),
        spdk_json_decode_int32,
        true
    },
};

static void nvme_spdk_rpc_namespace_detach_end(void *arg)
{
    struct spdk_jsonrpc_request *request = arg;
    struct spdk_json_write_ctx *w;

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_bool(w, true);
    spdk_jsonrpc_end_result(request, w);
}

static void nvme_spdk_rpc_namespace_detach(struct spdk_jsonrpc_request *request,
                                           const struct spdk_json_val *params)
{
    struct nvme_spdk_rpc_namespace_detach_attr *attr = NULL;
    nvme_spdk_ctrl_t *ctrl;
    nvme_spdk_namespace_t *ns;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    // Set invalid index, to identify when value was not decoded
    attr->cntlid = -1;

    if (spdk_json_decode_object(params,
                             nvme_spdk_rpc_namespace_detach_decoder,
                             SPDK_COUNTOF(nvme_spdk_rpc_namespace_detach_decoder),
                             attr)) {
        nvmx_error("Failed to decode nvme_spdk_rpc_namespace_detach parameters");
        goto free_attr;
    }

    pthread_mutex_lock(&nvme_spdk_ctrl_list_lock);
    ctrl = nvme_spdk_find_ctrl(attr->subnqn, attr->cntlid, attr->emu_name);
    if (!ctrl) {
        nvmx_error("No controller instance detected");
        pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);
        goto free_attr_fields;
    }

    pthread_spin_lock(&ctrl->namespaces_lock);
    ns = nvme_spdk_namespace_find(ctrl, attr->nsid);
    if (!ns) {
        nvmx_error("Namespace %d cannot be found", attr->nsid);
        pthread_spin_unlock(&ctrl->namespaces_lock);
        pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);
        goto free_attr_fields;
    }

    if (ctrl->expose_namespaces)
        nvme_spdk_namespace_detach(ns, ctrl);
    nvme_spdk_namespace_delete(ns, nvme_spdk_rpc_namespace_detach_end,
                               request);
    pthread_spin_unlock(&ctrl->namespaces_lock);
    pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);

    free(attr->emu_name);
    free(attr->subnqn);
    free(attr);
    return;

free_attr_fields:
    free(attr->emu_name);
    free(attr->subnqn);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}
SPDK_RPC_REGISTER("controller_nvme_namespace_detach",
                  nvme_spdk_rpc_namespace_detach, SPDK_RPC_RUNTIME)

struct nvme_spdk_rpc_namespace_list_attr {
    char *emu_name;
    char *subnqn;
    int cntlid;
};

static const struct spdk_json_object_decoder
nvme_spdk_rpc_namespace_list_decoder[] = {
    {
        "ctrl",
        offsetof(struct nvme_spdk_rpc_namespace_list_attr, emu_name),
        spdk_json_decode_string,
        true
    },
    {
        "subnqn",
        offsetof(struct nvme_spdk_rpc_namespace_list_attr, subnqn),
        spdk_json_decode_string,
        true
    },
    {
        "cntlid",
        offsetof(struct nvme_spdk_rpc_namespace_list_attr, cntlid),
        spdk_json_decode_int32,
        true
    },
};

static void nvme_spdk_rpc_namespace_list(struct spdk_jsonrpc_request *request,
                                         const struct spdk_json_val *params)
{
    struct nvme_spdk_rpc_namespace_list_attr *attr = NULL;
    struct spdk_json_write_ctx *w;
    nvme_spdk_ctrl_t *ctrl;
    nvme_spdk_namespace_t *ns;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    // Set invalid index, to identify when value was not decoded
    attr->cntlid = -1;

    if (spdk_json_decode_object(params,
                             nvme_spdk_rpc_namespace_list_decoder,
                             SPDK_COUNTOF(nvme_spdk_rpc_namespace_list_decoder),
                             attr)) {
        nvmx_error("Failed to decode nvme_spdk_rpc_namespace list parameters");
        goto free_attr;
    }

    pthread_mutex_lock(&nvme_spdk_ctrl_list_lock);
    ctrl = nvme_spdk_find_ctrl(attr->subnqn, attr->cntlid, attr->emu_name);
    if (!ctrl) {
        nvmx_error("No controller instance detected");
        pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);
        goto free_attr_fields;
    }

    w = spdk_jsonrpc_begin_result(request);

    spdk_json_write_object_begin(w);
    spdk_json_write_named_string(w, "name", ctrl->name);
    spdk_json_write_named_int32(w, "cntlid", ctrl->cntlid);
    spdk_json_write_named_array_begin(w, "Namespaces");
    pthread_spin_lock(&ctrl->namespaces_lock);
    TAILQ_FOREACH(ns, &ctrl->namespaces, entry) {
        spdk_json_write_object_begin(w);
        spdk_json_write_named_uint32(w, "nsid", ns->nsid);
        spdk_json_write_named_string(w, "bdev", ns->bdev_name);
        spdk_json_write_named_string(w, "bdev_type", ns->bdev_type);
        spdk_json_write_named_string(w, "qn", ns->qn);
        spdk_json_write_named_string(w, "protocol", ns->protocol);
        spdk_json_write_named_bool(w, "snap-direct", ns->zcopy);
        spdk_json_write_named_bool(w, "efficient-dma", ns->effdma);
        spdk_json_write_object_end(w);
    }
    pthread_spin_unlock(&ctrl->namespaces_lock);
    spdk_json_write_array_end(w);
    spdk_json_write_object_end(w);

    spdk_jsonrpc_end_result(request, w);

    pthread_mutex_unlock(&nvme_spdk_ctrl_list_lock);

    free(attr->emu_name);
    free(attr->subnqn);
    free(attr);
    return;

free_attr_fields:
    free(attr->emu_name);
    free(attr->subnqn);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}
SPDK_RPC_REGISTER("controller_nvme_namespace_list",
                  nvme_spdk_rpc_namespace_list, SPDK_RPC_RUNTIME)
#else /* HAVE_SPDK */
int nvme_spdk_create_ctrl(nvme_emu_ep_t *ep)
{
    return NVMX_ERR_NOTSUPPORTED;
}

void nvme_spdk_delete_ctrl(nvme_spdk_ctrl_t *ctrl)
{
}

int nvme_spdk_start_ctrl(nvme_emu_ep_t *ep)
{
    return NVMX_ERR_NOTSUPPORTED;
}

void nvme_spdk_stop_ctrl(nvme_emu_ep_t *ep)
{
}
#endif
