#include "config.h"
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>

#include <spdk/rpc.h>
#include <spdk/env.h>
#include <spdk/event.h>
#include <spdk/json.h>
#include <spdk/jsonrpc.h>
#include <spdk/bdev_module.h>
#ifdef HAVE_SPDK_UTIL_H
#include <spdk/util.h>
#endif

#include <snap.h>
#include "snap_nvme_emulator.h"
#include "nvme_spdk_subsystem.h"
#include "nvme_controller.h"
#include "mlnx_snap_pci_manager.h"
#include "spdk_emu_mgr.h"
#include "utils.h"

#define NVME_CTRL_SUBNQN_PREFIX "nqn.2021-06.mlnx.snap:"
#define SNAP_NVME_INVALID_NSID 0xFFFFFFFF

struct nvme_subsystem_attr {
    char *nqn;
    char *mn;
    char *sn;
    uint32_t nn;
    uint32_t mnan;
};

LIST_HEAD(, nvme_subsystem) nvme_subsystem_list = LIST_HEAD_INITIALIZER(nvme_subsystem_list);
pthread_mutex_t nvme_subsystem_lock = PTHREAD_MUTEX_INITIALIZER;
int nvme_spdk_subsystem;

static bool taken_subsysids[NVME_MAX_SUBSYSTEMS];

static const struct spdk_json_object_decoder
nvme_subsystem_decoder[] = {
    {
        "nqn",
        offsetof(struct nvme_subsystem_attr, nqn),
        spdk_json_decode_string,
        true
    },
    {
        "serial_number",
        offsetof(struct nvme_subsystem_attr, sn),
        spdk_json_decode_string
    },
    {
        "model_number",
        offsetof(struct nvme_subsystem_attr, mn),
        spdk_json_decode_string
    },
    {
        "number_namespaces",
        offsetof(struct nvme_subsystem_attr, nn),
        spdk_json_decode_uint32,
        true
    },
    {
        "maximum_namespaces",
        offsetof(struct nvme_subsystem_attr, mnan),
        spdk_json_decode_uint32,
        true
    },
};

static struct nvme_subsystem *subsystem_nvme_find_internal(const char *nqn)
{
    struct nvme_subsystem *subsys;

    LIST_FOREACH(subsys, &nvme_subsystem_list, entry) {
        if (strncmp(subsys->subnqn, nqn, SPDK_NVME_NQN_FIELD_SIZE) == 0)
            return subsys;
    }
    return NULL;
}

static struct nvme_subsystem *subsystem_nvme_find_internal_by_id(const int subsys_id)
{
    struct nvme_subsystem *subsys;

    LIST_FOREACH(subsys, &nvme_subsystem_list, entry) {
        if (subsys->subsysid == subsys_id)
            return subsys;
    }
    return NULL;
}

struct nvme_subsystem *subsystem_nvme_find(const char *nqn)
{
    struct nvme_subsystem *subsys;

    pthread_mutex_lock(&nvme_subsystem_lock);
    subsys = subsystem_nvme_find_internal(nqn);
    pthread_mutex_unlock(&nvme_subsystem_lock);

    return subsys;
}

#ifdef HAVE_RTE_HASH_H
uint32_t ns_hash_func(const void *key, uint32_t key_len, uint32_t init_val)
{
    //SPDK_ERRLOG("checking key %p %d\n", key, *(uint32_t *)key);
    return *(uint32_t *)key % NVME_SUBSYS_NS_HASH_SIZE;
}
#endif

static int nvme_spdk_subsystem_acquire_subsysid()
{
    int i;

    for (i = 0; i < NVME_MAX_SUBSYSTEMS; i++) {
        if (!taken_subsysids[i]) {
            taken_subsysids[i] = true;
            nvmx_debug("subsysid %d was acquired", i);
            return i;
        }
    }

    /* Not found */
    nvmx_error("Failed to find unused subsysid value");
    return -EINVAL;
}

static void nvme_spdk_subsystem_release_subsysid(int subsysid)
{
    taken_subsysids[subsysid] = false;
    nvmx_debug("subsysid %d was released", subsysid);
}

static bool subsystem_nvme_get_default_nqn(char *nqn, int subsysid)
{
    FILE *file;
    char bf_id[NVME_EMU_MAX_NQN_LEN - strlen(NVME_CTRL_SUBNQN_PREFIX) - 3];
    char *res;

    file = fopen("/etc/machine-id", "r");
    if (file) {
        res = fgets(bf_id, sizeof(bf_id), file);
        fclose(file);
        if (res) {
            bf_id[strcspn(bf_id, "\n")] = '\0';
            snprintf(nqn, NVME_EMU_MAX_NQN_LEN, "%s%s:%d",
                     NVME_CTRL_SUBNQN_PREFIX, bf_id, subsysid);
            nqn[NVME_EMU_MAX_NQN_LEN - 1] = '\0';
            return true;
        }
    }
    SPDK_ERRLOG("Failed to read from /etc/machine-id\n");
    return false;
}

static void subsystem_nvme_create(struct spdk_jsonrpc_request *request,
                                  const struct spdk_json_val *params)
{
    struct nvme_subsystem_attr *attr;
    struct nvme_subsystem *subsys;
    struct spdk_json_write_ctx *w;
    bool result;
#ifdef HAVE_RTE_HASH_H
    static int subsys_count;
    char hash_name[32];
    struct rte_hash_parameters hash_params = {0};
#endif
    int subsysid;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        return;

    // Set invalid nqn, to identify when value was not decoded
    attr->nqn = NULL;
    attr->nn = SNAP_NVME_INVALID_NSID;
    attr->mnan = SNAP_NVME_INVALID_NSID;

    if (spdk_json_decode_object(params, nvme_subsystem_decoder,
                                SPDK_COUNTOF(nvme_subsystem_decoder),
                                attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        result = false;
        goto write_result;
    }

    subsys = calloc(1, sizeof(*subsys));
    if (!subsys) {
        result = false;
        goto write_result;
    }

    subsysid = nvme_spdk_subsystem_acquire_subsysid();
    if (subsysid < 0) {
        result = false;
        free(subsys);
        goto write_result;
    }
    subsys->subsysid = subsysid;

    if (!attr->nqn) {
        attr->nqn = calloc(NVME_EMU_MAX_NQN_LEN, sizeof(char));
        if (!attr->nqn) {
            result = false;
            nvme_spdk_subsystem_release_subsysid(subsys->subsysid);
            free(subsys);
            goto write_result;
        }

        if (!subsystem_nvme_get_default_nqn(attr->nqn, subsys->subsysid)) {
            result = false;
            nvme_spdk_subsystem_release_subsysid(subsys->subsysid);
            free(subsys);
            goto write_result;
        }
    }

    if (subsystem_nvme_find_internal(attr->nqn)) {
        SPDK_ERRLOG("NVMe subsystem %s already exists\n", attr->nqn);
        result = false;
        goto write_result;
    }

    if (attr->nn == SNAP_NVME_INVALID_NSID)
        attr->nn = SNAP_CONFIG_DEFAULT_NN;

    if (attr->mnan == SNAP_NVME_INVALID_NSID)
        attr->mnan = nvmx_min(attr->nn, SNAP_CONFIG_DEFAULT_MNAN);

    if (attr->nn < attr->mnan) {
        SPDK_ERRLOG("Num. namespaces must be GTE max of allowed namespaces\n");
        result = false;
        goto write_result;
    }

    strncpy(subsys->subnqn, attr->nqn, SPDK_NVME_NQN_FIELD_SIZE);
    strncpy(subsys->sn, attr->sn, SPDK_NVME_CTRLR_SN_LEN);
    strncpy(subsys->mn, attr->mn, SPDK_NVME_CTRLR_MN_LEN);
    subsys->nn = attr->nn;
    subsys->mnan = attr->mnan;
    subsys->num_ns = 0;

    pthread_mutex_lock(&nvme_subsystem_lock);
#ifdef HAVE_RTE_HASH_H
    /* hash name is about 32 bytes long and must be unique,
     * subnqn is too long to be used here
     */
    snprintf(hash_name, sizeof(hash_name), "snap_ns_hash:%d", subsys_count++);
    hash_params.name = hash_name;
    hash_params.entries = NVME_SUBSYS_NS_HASH_SIZE;
    hash_params.key_len = sizeof(uint32_t);
    hash_params.hash_func = ns_hash_func;

    subsys->ns_hash = rte_hash_create(&hash_params);
    if (!subsys->ns_hash) {
        result = false;
        nvme_spdk_subsystem_release_subsysid(subsys->subsysid);
        free(subsys);
        pthread_mutex_unlock(&nvme_subsystem_lock);
        goto write_result;
    }
#else
    if (pthread_mutex_init(&subsys->ns_list_lock, NULL)) {
        result = false;
        nvme_spdk_subsystem_release_subsysid(subsys->subsysid);
        free(subsys);
        pthread_mutex_unlock(&nvme_subsystem_lock);
        goto write_result;
    }
    LIST_INIT(&subsys->ns_list);
#endif

    LIST_INSERT_HEAD(&nvme_subsystem_list, subsys, entry);
    pthread_mutex_unlock(&nvme_subsystem_lock);
    result = true;

write_result:
    nvmx_info("%s NVMe subsystem \"%s\" model \"%s\" serial \"%s\"",
              result ? "Created" : "Failed to create",
              attr->nqn, attr->mn, attr->sn);
    if (result) {
        w = spdk_jsonrpc_begin_result(request);
        spdk_json_write_object_begin(w);
        spdk_json_write_named_string(w, "nqn", attr->nqn);
        spdk_json_write_named_int32(w, "subsys_id", subsys->subsysid);
        spdk_json_write_object_end(w);
        spdk_jsonrpc_end_result(request, w);
    } else
        spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");

    free(attr->nqn);
    free(attr->mn);
    free(attr->sn);
    free(attr);
    return;
}

SPDK_RPC_REGISTER("subsystem_nvme_create", subsystem_nvme_create, SPDK_RPC_RUNTIME)

static void subsystem_nvme_delete(struct spdk_jsonrpc_request *request,
                                  const struct spdk_json_val *params)
{
    struct nvme_subsystem_attr *attr;
    struct nvme_subsystem *subsys;
    bool result;
    struct spdk_json_write_ctx *w;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        return;

    if (spdk_json_decode_object(params, nvme_subsystem_decoder, 1,
                                attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    pthread_mutex_lock(&nvme_subsystem_lock);

    subsys = subsystem_nvme_find_internal(attr->nqn);
    if (!subsys) {
        result = false;
        SPDK_ERRLOG("NVMe subsystem %s not found\n", attr->nqn);
        goto write_result;
    }

#ifdef HAVE_RTE_HASH_H
    if (rte_hash_count(subsys->ns_hash)) {
#else
    if (!LIST_EMPTY(&subsys->ns_list)) {
#endif
        result = false;
        SPDK_ERRLOG("NVMe subsystem %s has attached namespaces\n", attr->nqn);
        goto write_result;
    }

    if (!LIST_EMPTY(&subsys->ctrl_list)) {
        result = false;
        SPDK_ERRLOG("NVMe subsystem %s has attached controllers\n", attr->nqn);
        goto write_result;
    }

    LIST_REMOVE(subsys, entry);
#ifdef HAVE_RTE_HASH_H
    rte_hash_free(subsys->ns_hash);
#else
    pthread_mutex_destroy(&subsys->ns_list_lock);
#endif
    nvme_spdk_subsystem_release_subsysid(subsys->subsysid);
    free(subsys);
    result = true;

    SPDK_NOTICELOG("destroying NVMe subsystem %s\n", attr->nqn);

write_result:
    pthread_mutex_unlock(&nvme_subsystem_lock);
    if (result) {
        w = spdk_jsonrpc_begin_result(request);
        spdk_json_write_bool(w, result);
        spdk_jsonrpc_end_result(request, w);
    } else
        spdk_jsonrpc_send_error_response(request,
                    SPDK_JSONRPC_ERROR_INTERNAL_ERROR,
                    "Invalid parameters");
    free(attr->nqn);
free_attr:
    free(attr);
    return;
}

SPDK_RPC_REGISTER("subsystem_nvme_delete", subsystem_nvme_delete, SPDK_RPC_RUNTIME)

static void subsystem_nvme_list(struct spdk_jsonrpc_request *request,
                                const struct spdk_json_val *params)
{
    struct spdk_json_write_ctx *w;
    struct nvme_subsystem *subsys;
    struct spdk_emu_ctx *ctrl;

    if (params) {
        spdk_jsonrpc_send_error_response(request, SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                                         "No parameters required");
        return;
    }

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_array_begin(w);

    pthread_mutex_lock(&nvme_subsystem_lock);
    LIST_FOREACH(subsys, &nvme_subsystem_list, entry) {
        spdk_json_write_object_begin(w);
        spdk_json_write_named_string(w, "nqn", subsys->subnqn);
        spdk_json_write_named_string(w, "serial_number", subsys->sn);
        spdk_json_write_named_string(w, "model_number", subsys->mn);
        spdk_json_write_named_int32(w, "subsys_id", subsys->subsysid);
        spdk_json_write_named_array_begin(w, "controllers");

        /* write all controllers and namespaces */
        LIST_FOREACH(ctrl, &subsys->ctrl_list, nvme_subsys_entry) {
            spdk_json_write_object_begin(w);
            spdk_json_write_named_string(w, "name", ctrl->emu_name);
            spdk_json_write_named_uint32(w, "max_nsid", subsys->nn);
            spdk_json_write_named_uint32(w, "max_namespaces", subsys->mnan);
            spdk_json_write_named_uint32(w, "num_namespaces", subsys->num_ns);
            spdk_json_write_named_int32(w, "cntlid",
                                   snap_config_get_cntlid(ctrl->sconfig));
            spdk_json_write_named_string(w, "pci_bdf", ctrl->spci->pci_number);
            spdk_json_write_named_uint32(w, "pci_index", ctrl->spci->id);
            spdk_json_write_object_end(w);
        }

        spdk_json_write_array_end(w);
        spdk_json_write_object_end(w);
    }
    pthread_mutex_unlock(&nvme_subsystem_lock);

    spdk_json_write_array_end(w);

    spdk_jsonrpc_end_result(request, w);
}

SPDK_RPC_REGISTER("subsystem_nvme_list", subsystem_nvme_list, SPDK_RPC_RUNTIME)

static int nvme_spdk_subsystem_acquire_cntlid(struct nvme_subsystem *subsys, int cntlid)
{
    int i;

    if (cntlid > -1) {
        if (!subsys->taken_cntlids[cntlid]) {
            subsys->taken_cntlids[cntlid] = true;
            nvmx_debug("cntlid %d was acquired", cntlid);
            return cntlid;
        }
        nvmx_error("Controller id %d already in use in subsystem.", cntlid);
        return -EINVAL;
    }

    for (i = 0; i < NVME_SUBSYS_MAX_CONTROLLERS; i++) {
        if (!subsys->taken_cntlids[i]) {
            subsys->taken_cntlids[i] = true;
            nvmx_debug("cntlid %d was acquired", i);
            return i;
        }
    }

    /* Not found */
    nvmx_error("Failed to find unused cntlid value");
    return -EINVAL;
}

static void nvme_spdk_subsystem_release_cntlid(struct nvme_subsystem *subsys,
                                               int cntlid)
{
    subsys->taken_cntlids[cntlid] = false;
    nvmx_debug("cntlid %d was released", cntlid);
}

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_nvme_create_decoder[] = {
    {
        "nqn",
        offsetof(struct spdk_emu_controller_nvme_create_attr, nqn),
        spdk_json_decode_string,
        true
    },
    {
        "subsys_id",
        offsetof(struct spdk_emu_controller_nvme_create_attr, subsys_id),
        spdk_json_decode_int32,
        true
    },
    {
        "emulation_manager",
        offsetof(struct spdk_emu_controller_nvme_create_attr, emu_manager),
        spdk_json_decode_string
    },
    {
        "pci_bdf",
        offsetof(struct spdk_emu_controller_nvme_create_attr, pci_bdf),
        spdk_json_decode_string,
        true
    },
    {
        "pf_id",
        offsetof(struct spdk_emu_controller_nvme_create_attr, pf_id),
        spdk_json_decode_int32,
        true
    },
    {
        "vf_id",
        offsetof(struct spdk_emu_controller_nvme_create_attr, vf_id),
        spdk_json_decode_int32,
        true
    },
    {
        "conf_file",
        offsetof(struct spdk_emu_controller_nvme_create_attr, conf_file),
        spdk_json_decode_string,
        true
    },
    {
        "nr_io_queues",
        offsetof(struct spdk_emu_controller_nvme_create_attr, nr_io_queues),
        spdk_json_decode_int32,
        true
    },
    {
        "mdts",
        offsetof(struct spdk_emu_controller_nvme_create_attr, mdts),
        spdk_json_decode_int32,
        true
    },
    {
        "max_namespaces",
        offsetof(struct spdk_emu_controller_nvme_create_attr, max_namespaces),
        spdk_json_decode_int32,
        true
    },
    {
        "quirks",
        offsetof(struct spdk_emu_controller_nvme_create_attr, quirks),
        spdk_json_decode_int32,
        true
    },
    {
        "rdma_device",
        offsetof(struct spdk_emu_controller_nvme_create_attr, rdma_device),
        spdk_json_decode_string,
        true
    },
    {
        "mem",
        offsetof(struct spdk_emu_controller_nvme_create_attr, mem),
        spdk_json_decode_string,
        true
    },
    {
        "cntlid",
        offsetof(struct spdk_emu_controller_nvme_create_attr, cntlid),
        spdk_json_decode_int32,
        true
    },
    {
        "iova_mgmt",
        offsetof(struct spdk_emu_controller_nvme_create_attr, iova_mgmt),
        spdk_json_decode_bool,
        true
    },
    {
        "vuid",
        offsetof(struct spdk_emu_controller_nvme_create_attr, vuid),
        spdk_json_decode_string,
        true
    }
};

static snap_conf_t *
nvme_ctrl_config_create(struct nvme_subsystem *subsys,
                        struct spdk_emu_controller_nvme_create_attr *attr,
                        struct snap_pci *spci)
{
    snap_conf_t *sconfig;
    int cntlid;

    /* create controller specific config. Some params can be set
     * in the subsystem. And some params can be later overridden
     * in the controller specific configuration
     *
     * NOTE: spdk_emu_ctx_create takes config owenership and
     * should free it.
     */
    sconfig = snap_config_create(attr->conf_file);
    if (!sconfig)
        goto out;

    /* take mn & sn values from the subsystem */
    snap_config_set_sn(sconfig, subsys->sn);
    snap_config_set_mn(sconfig, subsys->mn);
    snap_config_set_nn(sconfig, subsys->nn);
    snap_config_set_mnan(sconfig, subsys->mnan);
    sconfig->spdk.subsys = subsys;
    sconfig->spdk.subnqn = subsys->subnqn;

    cntlid = nvme_spdk_subsystem_acquire_cntlid(subsys, attr->cntlid);
    if (cntlid < 0)
        goto sconfig_destroy;
    snap_config_set_cntlid(sconfig, cntlid);

    /* Last, update with values given by RPC command */
    if (spci->parent) {
        snap_config_set_pci_vfunc(sconfig, spci->id);
        snap_config_set_pci_func(sconfig, spci->parent->id);
    } else
        snap_config_set_pci_func(sconfig, spci->id);

    snap_config_set_pci_bdf(sconfig, spci->pci_number);

    if (attr->emu_manager) {
        if (snap_config_set_emu_manager(sconfig, attr->emu_manager))
            goto release_cntlid;
    } else {
        nvmx_error("emu_manager must be defined");
        goto release_cntlid;
    };

    if (attr->rdma_device) {
        if (snap_config_set_rdma_dev(sconfig, attr->rdma_device))
            goto release_cntlid;
    }

    if (strlen(snap_config_get_rdma_dev(sconfig)) == 0) {
        if (snap_config_set_rdma_dev(sconfig, attr->emu_manager))
            goto release_cntlid;
    }

    if (attr->nr_io_queues != -1) {
        if (snap_config_set_num_queues(sconfig, attr->nr_io_queues))
            goto sconfig_destroy;
    }

    if (attr->mdts != -1)
        snap_config_set_mdts(sconfig, attr->mdts);

    if (attr->max_namespaces != -1) {
        if (snap_config_set_max_namespaces(sconfig, attr->max_namespaces))
            goto sconfig_destroy;
    }

    if (attr->quirks != -1) {
        if (snap_config_set_quirks(sconfig, attr->quirks))
            goto sconfig_destroy;
    }

    if (attr->mem)
        if (snap_config_set_mem(sconfig, attr->mem))
            goto sconfig_destroy;

    if (attr->iova_mgmt != -1)
        snap_config_set_iova_mgmt(sconfig, !!attr->iova_mgmt);

    return sconfig;

release_cntlid:
    nvme_spdk_subsystem_release_cntlid(subsys, cntlid);
sconfig_destroy:
    snap_config_destroy(sconfig);
out:
    return NULL;
}

void nvme_ctrl_config_destroy(snap_conf_t *sconfig)
{
    struct nvme_subsystem *subsys;

    subsys = subsystem_nvme_find_internal(snap_config_get_subnqn(sconfig));
    if (!subsys) {
        nvmx_warn("Failed to find subnqn %s", snap_config_get_subnqn(sconfig));
        return;
    }

    nvme_spdk_subsystem_release_cntlid(subsys, snap_config_get_cntlid(sconfig));
    snap_config_destroy(sconfig);
}

static void
spdk_emu_rpc_controller_nvme_create(struct spdk_jsonrpc_request *request,
                                    const struct spdk_json_val *params)
{
    struct spdk_emu_controller_nvme_create_attr *attr = NULL;
    struct spdk_emu_ctx_create_attr emu_attr = {};
    struct nvme_ctrl_caps nvme_caps = {};
    struct spdk_emu_ctx *ctx;
    struct spdk_json_write_ctx *w;
    struct nvme_subsystem *subsys;
    struct snap_pci *spci;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    // Set invalid index, to identify when value was not decoded
    attr->pf_id = -1;
    attr->vf_id = -1;
    attr->nr_io_queues = -1;
    attr->mdts = -1;
    attr->max_namespaces = -1;
    attr->quirks = -1;
    attr->cntlid = -1;
    attr->iova_mgmt = -1;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_nvme_create_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_nvme_create_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    spci = mlnx_snap_get_pf_snap_pci_by_vuid(SNAP_NVME, attr->emu_manager, attr->vuid);
    if (!spci && !attr->vuid)
        spci = mlnx_snap_get_snap_pci(SNAP_NVME, attr->emu_manager,
                                      attr->pci_bdf, attr->pf_id, attr->vf_id);
    if (!spci)
        goto free_attr_strings;

    pthread_mutex_lock(&spdk_emu_list_lock);

    if (attr->pci_bdf) {
        ctx = spdk_emu_ctx_find_by_pci_bdf(attr->emu_manager,
                           spci->pci_number);
    } else {
        ctx = spdk_emu_ctx_find_by_pci_id(attr->emu_manager,
                           attr->pf_id, attr->vf_id,
                           SPDK_EMU_PROTOCOL_NVME);
    }

    if (ctx) {
        pthread_mutex_unlock(&spdk_emu_list_lock);
        SPDK_ERRLOG("PCI function is already in use by %s\n", ctx->emu_name);
        goto free_attr_strings;
    }

    pthread_mutex_lock(&nvme_subsystem_lock);

    if (attr->nqn)
        subsys = subsystem_nvme_find_internal(attr->nqn);
    else
        subsys = subsystem_nvme_find_internal_by_id(attr->subsys_id);

    if (!subsys) {
        pthread_mutex_unlock(&nvme_subsystem_lock);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        SPDK_ERRLOG("NVMe subsystem %s does not exist\n", attr->nqn);
        goto free_attr_strings;
    }
    attr->sconfig = nvme_ctrl_config_create(subsys, attr, spci);
    if (!attr->sconfig) {
        pthread_mutex_unlock(&nvme_subsystem_lock);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    emu_attr.priv = attr;
    emu_attr.protocol = SPDK_EMU_PROTOCOL_NVME;
    emu_attr.emu_manager = attr->emu_manager;
    emu_attr.spci = spci;
    ctx = spdk_emu_ctx_create(&emu_attr);
    if (!ctx) {
        pthread_mutex_unlock(&nvme_subsystem_lock);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_sconfig;
    }

    LIST_INSERT_HEAD(&spdk_emu_list, ctx, entry);
    LIST_INSERT_HEAD(&subsys->ctrl_list, ctx, nvme_subsys_entry);
    ctx->nvme_subsys = subsys;

    pthread_mutex_unlock(&nvme_subsystem_lock);
    pthread_mutex_unlock(&spdk_emu_list_lock);

    nvme_ctrl_query_caps((nvme_ctrl_t *)ctx->ctrl, &nvme_caps);
    if (spdk_emu_rpc_finished) {
        nvmx_warn("spdk rpc is closed. Not returning response for rpc request");
    } else {
        w = spdk_jsonrpc_begin_result(request);
        spdk_json_write_object_begin(w);
        spdk_json_write_named_string(w, "name", ctx->emu_name);
        spdk_json_write_named_int32(w, "cntlid",
                          snap_config_get_cntlid(ctx->sconfig));
        spdk_json_write_named_string(w, "version", nvme_caps.version);
        spdk_json_write_named_bool(w, "offload", nvme_caps.offload);
        spdk_json_write_named_bool(w, "mempool", nvme_caps.mempool);
        spdk_json_write_named_uint32(w, "max_nsid", nvme_caps.max_nsid);
        spdk_json_write_named_uint32(w, "max_namespaces",
                                 nvme_caps.max_namespaces);
        spdk_json_write_object_end(w);
        spdk_jsonrpc_end_result(request, w);
    }

    free(attr->nqn);
    free(attr->emu_manager);
    free(attr->pci_bdf);
    free(attr->conf_file);
    free(attr->rdma_device);
    free(attr->mem);
    free(attr->vuid);
    free(attr);
    return;

free_sconfig:
    snap_config_destroy(attr->sconfig);
free_attr_strings:
    free(attr->nqn);
    free(attr->emu_manager);
    free(attr->pci_bdf);
    free(attr->conf_file);
    free(attr->rdma_device);
    free(attr->mem);
    free(attr->vuid);
free_attr:
    free(attr);
invalid:
    if (spdk_emu_rpc_finished) {
        nvmx_warn("spdk rpc is closed. Not returning error response for rpc request");
    } else {
        spdk_jsonrpc_send_error_response(request,
                         SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                         "Invalid parameters");
    }
}

SPDK_RPC_REGISTER("controller_nvme_create",
                  spdk_emu_rpc_controller_nvme_create, SPDK_RPC_RUNTIME)


static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_nvme_delete_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_nvme_delete_attr, name),
        spdk_json_decode_string,
        true
    },
    {
        "subnqn",
        offsetof(struct spdk_emu_controller_nvme_delete_attr, subnqn),
        spdk_json_decode_string,
        true
    },
    {
        "cntlid",
        offsetof(struct spdk_emu_controller_nvme_delete_attr, cntlid),
        spdk_json_decode_int32,
        true
    },
    {
        "vuid",
        offsetof(struct spdk_emu_controller_nvme_delete_attr, vuid),
        spdk_json_decode_string,
        true
    },
};

void nvme_subsys_ctrl_del(struct spdk_emu_ctx *ctx)
{
    pthread_mutex_lock(&nvme_subsystem_lock);
    LIST_REMOVE(ctx, nvme_subsys_entry);
    pthread_mutex_unlock(&nvme_subsystem_lock);
}

static void
spdk_emu_rpc_controller_nvme_delete(struct spdk_jsonrpc_request *request,
                                    const struct spdk_json_val *params)
{
    struct spdk_emu_controller_nvme_delete_attr *attr = NULL;
    struct spdk_emu_ctx *ctx;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    // Set invalid index, to identify when value was not decoded
    attr->cntlid = -1;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_nvme_delete_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_nvme_delete_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    /* Find SPDK context matching the requested PF id */
    pthread_mutex_lock(&spdk_emu_list_lock);
    if (attr->vuid)
        ctx = spdk_emu_ctx_find_by_vuid(attr->vuid);
    else
        ctx = spdk_emu_ctx_find(attr->subnqn, attr->cntlid, attr->name);
    if (!ctx) {
        if (attr->cntlid != -1) {
            SPDK_ERRLOG("emulation ctrl(id:%d) cannot be found in"
                         "NVMe subsystem %s\n", attr->cntlid, attr->subnqn);
        } else if (attr->name) {
            SPDK_ERRLOG("emulation ctrl %s cannot be found\n", attr->name);
        }
        else
            SPDK_ERRLOG("emulation ctrl with vuid %s cannot be found\n", attr->vuid);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    if (ctx->protocol != SPDK_EMU_PROTOCOL_NVME) {
        SPDK_ERRLOG("Given controller is of wrong type (%s)\n",
                    spdk_emu_protocol_to_string(ctx->protocol));
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    LIST_REMOVE(ctx, entry);
    nvme_subsys_ctrl_del(ctx);
    pthread_mutex_unlock(&spdk_emu_list_lock);

    free(attr->name);
    free(attr->subnqn);
    free(attr);

    spdk_emu_ctx_destroy(ctx, spdk_emu_ctx_fini_cb_rpc, request);
    return;

free_attr_strings:
    free(attr->name);
    free(attr->subnqn);
    free(attr->vuid);
free_attr:
    free(attr);
invalid:
    if (spdk_emu_rpc_finished) {
        nvmx_warn("spdk rpc is closed. Not returning error response for rpc request");
    } else {
        spdk_jsonrpc_send_error_response(request,
                         SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                         "Invalid parameters");
    }
}

SPDK_RPC_REGISTER("controller_nvme_delete",
                  spdk_emu_rpc_controller_nvme_delete, SPDK_RPC_RUNTIME)

static const struct spdk_json_object_decoder

spdk_emu_rpc_controller_nvme_get_iostat_decoder[] = {
        {
            "name",
            offsetof(struct spdk_emu_controller_nvme_get_iostat_attr, name),
            spdk_json_decode_string,
            true
        }
};

static void
spdk_emu_rpc_controller_nvme_get_iostat(struct spdk_jsonrpc_request *request,
                                    const struct spdk_json_val *params)
{
    struct spdk_emu_controller_nvme_get_iostat_attr *attr;
    struct spdk_emu_ctx *ctx;
    struct spdk_json_write_ctx *w;
    nvme_ctrl_t *ctrl;
    nvme_namespace_t *ns;
    nvme_spdk_namespace_t *spdk_ns;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (params && spdk_json_decode_object(params,
            spdk_emu_rpc_controller_nvme_get_iostat_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_nvme_get_iostat_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_object_begin(w);
    spdk_json_write_named_array_begin(w, "controllers");

    pthread_mutex_lock(&spdk_emu_list_lock);
    LIST_FOREACH(ctx, &spdk_emu_list, entry) {

        if (ctx->protocol != SPDK_EMU_PROTOCOL_NVME)
            continue;

        if (attr->name && *attr->name && strncmp(ctx->emu_name, attr->name, SPDK_EMU_NAME_MAXLEN))
            continue;

        ctrl = ctx->ctrl;

        nvme_ctrl_pgs_suspend(ctrl);

        spdk_json_write_object_begin(w);
        spdk_json_write_named_string(w, "name", ctx->emu_name);

        spdk_json_write_named_array_begin(w, "bdevs");

        if (snap_config_get_emu_mode(ctrl->sconfig) != SNAP_EMU_CC_REMOTE) {

            pthread_spin_lock(&ctrl->namespaces_lock);
            TAILQ_FOREACH(ns, &ctrl->namespaces, entry) {

                nvme_stat_t stat = {{0}};
                int sqid;

                spdk_ns = ns->backend_ns;
                spdk_json_write_object_begin(w);

                for (sqid = 1; sqid < ctrl->num_queues; sqid++) {
                    const nvme_stat_t *s = &ns->sq_stat[sqid];
                    stat.read.total += s->read.total;
                    stat.read.completed += s->read.completed;
                    stat.read.err += s->read.err;
                    stat.write.total += s->write.total;
                    stat.write.completed += s->write.completed;
                    stat.write.err += s->write.err;
                    stat.flush.total += s->flush.total;
                    stat.flush.completed += s->flush.completed;
                    stat.flush.err += s->flush.err;
                }

                spdk_json_write_named_string(w, "bdev_name", spdk_ns->bdev_name);
                spdk_json_write_named_uint64(w, "read_ios", stat.read.total);
                spdk_json_write_named_uint64(w, "completed_read_ios", stat.read.completed);
                spdk_json_write_named_uint64(w, "write_ios", stat.write.total);
                spdk_json_write_named_uint64(w, "completed_write_ios", stat.write.completed);
                spdk_json_write_named_uint64(w, "flush_ios", stat.flush.total);
                spdk_json_write_named_uint64(w, "completed_flush_ios", stat.flush.completed);
                spdk_json_write_named_uint64(w, "err_read_ios", stat.read.err);
                spdk_json_write_named_uint64(w, "err_write_ios", stat.write.err);
                spdk_json_write_named_uint64(w, "err_flush_ios", stat.flush.err);
                spdk_json_write_object_end(w);
            }

            pthread_spin_unlock(&ctrl->namespaces_lock);
        }

        spdk_json_write_array_end(w);
        spdk_json_write_object_end(w);
        nvme_ctrl_pgs_resume(ctrl);
    }

    pthread_mutex_unlock(&spdk_emu_list_lock);

    spdk_json_write_array_end(w);
    spdk_json_write_object_end(w);
    spdk_jsonrpc_end_result(request, w);

    free(attr->name);
    free(attr);

    return;

free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
            SPDK_JSONRPC_ERROR_INVALID_PARAMS,
            "Invalid parameters");
}

SPDK_RPC_REGISTER("controller_nvme_get_iostat",
        spdk_emu_rpc_controller_nvme_get_iostat, SPDK_RPC_RUNTIME)

static const struct spdk_json_object_decoder

spdk_emu_rpc_controller_nvme_debugstat_decoder[] = {
        {
            "name",
            offsetof(struct spdk_emu_controller_nvme_get_debugstat_attr, name),
            spdk_json_decode_string,
            true
        },
        {
            "fw_counters",
            offsetof(struct spdk_emu_controller_nvme_get_debugstat_attr, fw_counters),
            spdk_json_decode_bool,
            false
        }
};

static void
spdk_emu_rpc_controller_nvme_get_debugstat(struct spdk_jsonrpc_request *request,
                                    const struct spdk_json_val *params)
{
    struct spdk_emu_controller_nvme_get_debugstat_attr *attr;
    struct spdk_emu_ctx *ctx;
    struct spdk_json_write_ctx *w;
    nvme_emu_counters_t emu_counters;
    nvme_ctrl_t *ctrl;
    nvme_stat_t stat;
    int sqid;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_nvme_debugstat_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_nvme_debugstat_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_object_begin(w);
    spdk_json_write_named_array_begin(w, "controllers");

    pthread_mutex_lock(&spdk_emu_list_lock);
    LIST_FOREACH(ctx, &spdk_emu_list, entry) {

        if (ctx->protocol != SPDK_EMU_PROTOCOL_NVME)
            continue;

        if (attr->name && *attr->name && strncmp(ctx->emu_name, attr->name, SPDK_EMU_NAME_MAXLEN))
            continue;

        ctrl = ctx->ctrl;

        nvme_ctrl_pgs_suspend(ctrl);

        spdk_json_write_object_begin(w);
        spdk_json_write_named_string(w, "name", ctx->emu_name);

        if (snap_config_get_emu_mode(ctrl->sconfig) != SNAP_EMU_CC_REMOTE && attr->fw_counters == false) {

            spdk_json_write_named_object_begin(w, "global");

            memset(&stat, 0, sizeof(stat));

            for (sqid = 1; sqid < ctrl->num_queues; sqid++) {
                stat.read.total += ctrl->sq_stat[sqid].read.total;
                stat.read.completed_bytes_count += ctrl->sq_stat[sqid].read.completed_bytes_count;
                stat.write.total += ctrl->sq_stat[sqid].write.total;
                stat.write.completed_bytes_count += ctrl->sq_stat[sqid].write.completed_bytes_count;
            }

            spdk_json_write_named_uint64(w, "hw_read_bytes", stat.read.completed_bytes_count);
            spdk_json_write_named_uint64(w, "hw_write_bytes", stat.write.completed_bytes_count);
            spdk_json_write_named_uint64(w, "hw_read_cmds", stat.read.total);
            spdk_json_write_named_uint64(w, "hw_write_cmds", stat.write.total);

            spdk_json_write_object_end(w);

            spdk_json_write_named_array_begin(w, "queues");

            for (sqid = 1; sqid < ctrl->num_queues; sqid++) {

                nvme_stat_t *stat = &ctrl->sq_stat[sqid];

                if (stat->valid) {
                    spdk_json_write_object_begin(w);
                    spdk_json_write_named_uint32(w, "qid", sqid);
                    spdk_json_write_named_uint64(w, "hw_read_bytes", stat->read.completed_bytes_count);
                    spdk_json_write_named_uint64(w, "hw_write_bytes", stat->write.completed_bytes_count);
                    spdk_json_write_named_uint64(w, "hw_read_cmds", stat->read.total);
                    spdk_json_write_named_uint64(w, "hw_write_cmds", stat->write.total);
                    spdk_json_write_object_end(w);
                }
            }

            spdk_json_write_array_end(w);
        }
        else {

            spdk_json_write_named_object_begin(w, "global");

            if (!nvme_emu_query_ctrl_counters(ctrl->dev, &emu_counters))
                SPDK_ERRLOG("Failed to query SQ counters\n");

            spdk_json_write_named_uint64(w, "hw_read_bytes", emu_counters.data_read);
            spdk_json_write_named_uint64(w, "hw_write_bytes", emu_counters.data_write);
            spdk_json_write_named_uint64(w, "hw_read_cmds", emu_counters.cmd_read);
            spdk_json_write_named_uint64(w, "hw_write_cmds",emu_counters.cmd_write);

            spdk_json_write_object_end(w);

            spdk_json_write_named_array_begin(w, "queues");

            for (sqid = 1; sqid < ctrl->num_queues; sqid++) {
                if (nvme_emu_query_sq_counters(ctrl->dev, sqid, &emu_counters)) {
                    spdk_json_write_object_begin(w);
                    spdk_json_write_named_uint32(w, "qid", sqid);
                    spdk_json_write_named_uint64(w, "hw_read_bytes", emu_counters.data_read);
                    spdk_json_write_named_uint64(w, "hw_write_bytes", emu_counters.data_write);
                    spdk_json_write_named_uint64(w, "hw_read_cmds", emu_counters.cmd_read);
                    spdk_json_write_named_uint64(w, "hw_write_cmds", emu_counters.cmd_write);
                    spdk_json_write_object_end(w);
                }
            }

            spdk_json_write_array_end(w);
        }

        spdk_json_write_object_end(w);

        nvme_ctrl_pgs_resume(ctrl);
    }

    pthread_mutex_unlock(&spdk_emu_list_lock);

    spdk_json_write_array_end(w);
    spdk_json_write_object_end(w);
    spdk_jsonrpc_end_result(request, w);

    free(attr->name);
    free(attr);

    return;

free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
            SPDK_JSONRPC_ERROR_INVALID_PARAMS,
            "Invalid parameters");
}

SPDK_RPC_REGISTER("controller_nvme_get_debugstat",
        spdk_emu_rpc_controller_nvme_get_debugstat, SPDK_RPC_RUNTIME)

int nvme_subsys_ns_list_allocated_nsids(struct nvme_subsystem *subsys,
                                        uint32_t *ns_list, size_t ns_list_sz,
                                        uint32_t min_nsid)
{
    nvme_spdk_namespace_t *ns;
    int j = 0;

    if (!subsys)
        return -1;

#ifdef HAVE_RTE_HASH_H
    uint32_t hash_iter;
    uint32_t *nsid_key;
    while (rte_hash_iterate(subsys->ns_hash, (const void **)&nsid_key,
                            (void **)&ns, &hash_iter) >= 0) {
        if (!ns)
            continue;

        if (ns->nsid != *nsid_key) {
            nvmx_error("Key %lu does not match saved value!", *nsid_key);
            continue;
        }
#else
    pthread_mutex_lock(&subsys->ns_list_lock);
    LIST_FOREACH(ns, &subsys->ns_list, nvme_subsys_entry) {
#endif
        /*
         * By NVMe spec., we should return only nsids that are greater than
         * the given min_nsid
         */
        if (ns->nsid <= min_nsid)
            continue;

        ns_list[j++] = cpu_to_le32(ns->nsid);
        if (j == ns_list_sz)
            break;
    }
#ifndef HAVE_RTE_HASH_H
    pthread_mutex_unlock(&subsys->ns_list_lock);
#endif

    /* According to NVMe spec., we have to return list in increasing order */
    qsort(ns_list, j, sizeof(uint32_t), compare_uint32_t);
    return 0;
}

bool nvme_subsys_ns_exists(struct nvme_subsystem *subsys,
                           uint32_t nsid)
{
#ifdef HAVE_RTE_HASH_H
    return rte_hash_lookup(subsys->ns_hash, &nsid) >= 0 ? true : false;
#else
    nvme_spdk_namespace_t *ns;
    bool found = false;

    pthread_mutex_lock(&subsys->ns_list_lock);
    LIST_FOREACH(ns, &subsys->ns_list, nvme_subsys_entry) {
        if (nsid == ns->nsid) {
            found = true;
            break;
        }
    }
    pthread_mutex_unlock(&subsys->ns_list_lock);
    return found;
#endif
}

int nvme_subsys_ns_add(struct nvme_subsystem *subsys,
                       nvme_spdk_namespace_t *ns)
{
    if ((ns->nsid > 0 && subsys->mnan > 0) &&
        (ns->nsid > subsys->nn || subsys->num_ns >= subsys->mnan)) {
        nvmx_info("Namespace nsid %d exceeded number of allowed namespaces", ns->nsid);
        return -1;
    }
#ifdef HAVE_RTE_HASH_H
    subsys->num_ns++;
    return rte_hash_add_key_data(subsys->ns_hash, &ns->nsid, ns);
#else
    pthread_mutex_lock(&subsys->ns_list_lock);
    LIST_INSERT_HEAD(&subsys->ns_list, ns, nvme_subsys_entry);
    subsys->num_ns++;
    pthread_mutex_unlock(&subsys->ns_list_lock);
    return 0;
#endif
}

void nvme_subsys_ns_del(struct nvme_subsystem *subsys,
                        nvme_spdk_namespace_t *ns)
{
#ifdef HAVE_RTE_HASH_H
    subsys->num_ns--;
    rte_hash_del_key(subsys->ns_hash, &ns->nsid);
#else
    pthread_mutex_lock(&subsys->ns_list_lock);
    LIST_REMOVE(ns, nvme_subsys_entry);
    subsys->num_ns--;
    pthread_mutex_unlock(&subsys->ns_list_lock);
#endif
}

struct nvme_spdk_rpc_async_event_attr {
    char *emu_name;
    char *subnqn;
    int cntlid;
    int event_bit;
};

static const struct spdk_json_object_decoder
nvme_spdk_rpc_async_event_decoder[] = {
    {
        "ctrl",
        offsetof(struct nvme_spdk_rpc_async_event_attr, emu_name),
        spdk_json_decode_string,
        true
    },
    {
        "subnqn",
        offsetof(struct nvme_spdk_rpc_async_event_attr, subnqn),
        spdk_json_decode_string,
        true
    },
    {
        "cntlid",
        offsetof(struct nvme_spdk_rpc_async_event_attr, cntlid),
        spdk_json_decode_int32,
        true
    },
    {
        "event_bit",
        offsetof(struct nvme_spdk_rpc_async_event_attr, event_bit),
        spdk_json_decode_int32,
        true
    },
};

static void nvme_spdk_rpc_async_event_enable(struct spdk_jsonrpc_request *request,
                                         const struct spdk_json_val *params)
{
    struct nvme_spdk_rpc_async_event_attr *attr = NULL;
    struct spdk_json_write_ctx *w;
    enum NvmeAsyncEventConfig event;
    struct spdk_emu_ctx *ctx;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    // Set invalid index, to identify when value was not decoded
    attr->cntlid = -1;
    attr->event_bit = -1;

    if (spdk_json_decode_object(params,
                             nvme_spdk_rpc_async_event_decoder,
                             SPDK_COUNTOF(nvme_spdk_rpc_async_event_decoder),
                             attr)) {
        nvmx_error("Failed to decode nvme_spdk_rpc_async_event_enable parameters");
        goto free_attr;
    }

    /* try to enable all events if 'event_bit' not provided */
    if (attr->event_bit != -1)
        event = 1 << attr->event_bit;
    else
        event = 0xffffffff;

    pthread_mutex_lock(&spdk_emu_list_lock);
    ctx = spdk_emu_ctx_find(attr->subnqn, attr->cntlid, attr->emu_name);
    if (!ctx) {
        if (attr->cntlid != -1) {
            SPDK_ERRLOG("emulation ctrl(id:%d) cannot be found in"
                         "NVMe subsystem %s\n", attr->cntlid, attr->subnqn);
        } else {
            SPDK_ERRLOG("emulation ctrl %s cannot be found\n", attr->emu_name);
        }
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_fields;
    }
    pthread_mutex_unlock(&spdk_emu_list_lock);

    nvme_ctrl_mask_unmask_async_event((nvme_ctrl_t *)ctx->ctrl, event, false);

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_object_begin(w);
    spdk_json_write_named_string(w, "Enable async event ", "Succeed");
    spdk_json_write_object_end(w);
    spdk_jsonrpc_end_result(request, w);

    free(attr->emu_name);
    free(attr->subnqn);
    free(attr);
    return;

free_attr_fields:
    free(attr->emu_name);
    free(attr->subnqn);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}
SPDK_RPC_REGISTER("controller_nvme_async_event_enable",
                  nvme_spdk_rpc_async_event_enable, SPDK_RPC_RUNTIME)

static void nvme_spdk_rpc_async_event_disable(struct spdk_jsonrpc_request *request,
                                         const struct spdk_json_val *params)
{
    struct nvme_spdk_rpc_async_event_attr *attr = NULL;
    struct spdk_json_write_ctx *w;
    enum NvmeAsyncEventConfig event;
    struct spdk_emu_ctx *ctx;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    // Set invalid index, to identify when value was not decoded
    attr->cntlid = -1;
    attr->event_bit = -1;

    if (spdk_json_decode_object(params,
                             nvme_spdk_rpc_async_event_decoder,
                             SPDK_COUNTOF(nvme_spdk_rpc_async_event_decoder),
                             attr)) {
        nvmx_error("Failed to decode nvme_spdk_rpc_async_event_enable parameters");
        goto free_attr;
    }

    /* try to enable all events if 'event_bit' not provided */
    if (attr->event_bit != -1)
        event = 1 << attr->event_bit;
    else
        event = 0xffffffff;

    pthread_mutex_lock(&spdk_emu_list_lock);
    ctx = spdk_emu_ctx_find(attr->subnqn, attr->cntlid, attr->emu_name);
    if (!ctx) {
        if (attr->cntlid != -1) {
            SPDK_ERRLOG("emulation ctrl(id:%d) cannot be found in"
                         "NVMe subsystem %s\n", attr->cntlid, attr->subnqn);
        } else {
            SPDK_ERRLOG("emulation ctrl %s cannot be found\n", attr->emu_name);
        }
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_fields;
    }
    pthread_mutex_unlock(&spdk_emu_list_lock);

    nvme_ctrl_mask_unmask_async_event((nvme_ctrl_t *)ctx->ctrl, event, true);

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_object_begin(w);
    spdk_json_write_named_string(w, "Disable async event ", "Succeed");
    spdk_json_write_object_end(w);
    spdk_jsonrpc_end_result(request, w);

    free(attr->emu_name);
    free(attr->subnqn);
    free(attr);
    return;

free_attr_fields:
    free(attr->emu_name);
    free(attr->subnqn);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}
SPDK_RPC_REGISTER("controller_nvme_async_event_disable",
                  nvme_spdk_rpc_async_event_disable, SPDK_RPC_RUNTIME)
