#include <stdint.h>

#include "config.h"
#include "nvme_emu.h"
#include "nvme_prp.h"
#include "memp.h"

/* read/write up to two pages of data
 */
static inline void nvme_emu_dma_rw(nvme_queue_ops_t *ops, void *buf, uint64_t addr,
                                   size_t len, int dir)
{
    switch (dir) {
        case PRP_DIR_WRITE:
            nvme_emu_dma_write(ops, buf, addr, len, 0);
            return;
        case PRP_DIR_READ:
            nvme_emu_dma_read(ops, buf, addr, len, 0);
            return;
        default:
            nvmx_fatal("invalid direction %d\n", dir);
    }
}

/* write data to prp */
enum nvme_status_code
nvme_prp_rw_short(nvme_queue_ops_t *ops, uint64_t prp1, uint64_t prp2,
                  void *buf, size_t len, int dir)
{
    size_t to_read, offset;
    /* validate length */
    nvmx_debug("total length %d[B]", (int)len);
    if (unlikely(len > 64 * PRP_PAGE_SIZE(ops->page_bits))) {
        nvmx_error("len %lu is bigger than 256K", len);
        return NVME_SC_DATA_TRAS_ERROR;
    }

    //1st PRP
    to_read = PRP_PAGE_SIZE(ops->page_bits) - PRP_PAGE_OFFSET(prp1, ops->page_bits);
    nvmx_debug("prp1 read size %d[B]", (int)to_read);

    if (to_read >= len) {
        nvme_emu_dma_rw(ops, buf, prp1, len, dir);
        return NVME_SC_SUCCESS;
    }

    nvme_emu_dma_rw(ops, buf, prp1, to_read, dir);
    len   -= to_read;
    offset = to_read;

    //2nd PRP
    /* According to section 4.3 in NVMe specification:
     * If PRP2 is a list entry then it shall be QWord aligned and may have non zero offset.
     * If PRP2 is not a list entry then it shall have zero offset.
     */

    if (len <= PRP_PAGE_SIZE(ops->page_bits)) {
        //PRP2 isn't a list entry - offset should be zero
        if (unlikely(PRP_PAGE_OFFSET(prp2, ops->page_bits) != 0)) {
            nvmx_error("prp2 entry with offset 0x%lx not zero",
                       PRP_PAGE_OFFSET(prp2, ops->page_bits));
            return NVME_SC_PRP_OFFSET_INVALID;
        }
        nvmx_debug("prp2 isn't a list entry. remaining length %d[B]", (int)len);
        nvme_emu_dma_rw(ops, (char *)buf + offset, prp2, len, dir);
        return NVME_SC_SUCCESS;
    }

    uint64_t prp_base[64];
    unsigned num_of_list_entries;
    int i;

    //PRP2 is a list entry - Qword aligned
    if (unlikely((PRP_PAGE_BASE(prp2, ops->page_bits) & 0x7) != 0)) {
        nvmx_error("prp2 list entry with address is not Qword aligned");
        return NVME_SC_DATA_TRAS_ERROR;
    }

    num_of_list_entries = (len + PRP_PAGE_SIZE(ops->page_bits) - 1)/PRP_PAGE_SIZE(ops->page_bits);
    if (unlikely(num_of_list_entries > 64)) {
        nvmx_error("more than 64 entries in PRP list");
        return NVME_SC_DATA_TRAS_ERROR;
    }
    nvmx_debug("prp2 is a list entry. remaining length %lu[B] (equals %d pages)",
               len, num_of_list_entries);

    //PRP list
    nvme_emu_dma_rw(ops, prp_base, prp2, sizeof(uint64_t) * num_of_list_entries , PRP_DIR_READ);
    for(i = 0; i < num_of_list_entries; i++) {
        if (unlikely(PRP_PAGE_OFFSET(prp_base[i], ops->page_bits) != 0)) {
            nvmx_error("prp list entry %d has non-zero offset", i);
            return NVME_SC_PRP_OFFSET_INVALID;
        }
        to_read = len <= PRP_PAGE_SIZE(ops->page_bits) ? len : PRP_PAGE_SIZE(ops->page_bits);
        nvme_emu_dma_rw(ops, (char *)buf + offset, prp_base[i], to_read, dir);
        len    -= to_read;
        offset += to_read;
    }

    nvmx_assertv_always(len == 0, "not all length was read");
    return NVME_SC_SUCCESS;
}

static inline void prp_to_iovec(nvme_async_req_t *req, uint64_t *list_base)
{
    int i;
    int li;
    size_t total_len, to_read;

    to_read = PRP_PAGE_SIZE(req->ops->page_bits);
    /* iov[0] already points to prp1, iov_cnt = 1 */
    total_len = req->iov[0].iov_len;
    i = li = 0;

    while (li < req->prp.list_size) {
        if (li + 1 < req->prp.list_size)
            to_read = PRP_PAGE_SIZE(req->ops->page_bits);
        else
            to_read = req->len - total_len;

        nvmx_debug("prp_list[%d] = 0x%llx read %d", li, list_base[li], to_read);

        if (req->iov[i].iov_base + req->iov[i].iov_len == (void *)list_base[li]) {
            req->iov[i].iov_len += to_read;
            nvmx_debug("     added to iov[%d]", i);
        } else {
            i++;
            req->iov[i].iov_base = (void *)list_base[li];
            req->iov[i].iov_len = to_read;
            req->iov_cnt++;
            nvmx_debug("     new iov[%d]", i);
        }
        li++;
        total_len += to_read;
    }
    nvmx_debug("iov list: %d total_len %d", req->iov_cnt, total_len);
}

static void prp_bcopy_operation_done(void *user, enum nvme_emu_status rc)
{
    nvme_async_req_t *req = user;

    if (req->dma_cmd.op == NVME_REQ_DMA_FROM_HOST)
        memcpy(req->bdata, req->io_buf, req->len);

    if (req->memp) {
        memp_free(req->memp, req->io_buf);
        req->io_buf = NULL;
        req->aux_buf = NULL;
    }
    req->bdata = NULL;

    req->comp_cb(req->comp_cb_user, rc);
}

static int prp_submit_iov(nvme_async_req_t *req)
{
    int rc;
    int i;

    req->dma_cmd.laddr = req->io_buf;
    req->dma_cmd.comp_cb = NULL;

    for (i = 0; i < req->iov_cnt; i++) {
        req->dma_cmd.raddr = (uintptr_t)req->iov[i].iov_base;
        req->dma_cmd.len = req->iov[i].iov_len;
        if (i == req->iov_cnt - 1) {
            if (unlikely(req->bdata)) {
                req->dma_cmd.comp_cb = prp_bcopy_operation_done;
                req->dma_cmd.comp_cb_user = req;
            }
            else {
                req->dma_cmd.comp_cb = req->comp_cb;
                req->dma_cmd.comp_cb_user = req->comp_cb_user;
            }
        }

        nvmx_debug("iov[%d] = 0x%llx len %d", i, req->dma_cmd.raddr, req->dma_cmd.len);
        /* TODO: optimize submit request to work with iovs */
        rc = nvme_emu_request_submit(req);
        if (rc != NVMX_INPROGRESS) {
            req->comp_cb(req->comp_cb_user, rc);
            return rc;
        }

        req->dma_cmd.laddr += req->iov[i].iov_len;
    }

    return NVMX_INPROGRESS;
}

static void read_prp_list_completion(void *user, int status)
{
    nvme_async_req_t *req = user;

    if (unlikely(status != NVMX_OK)) {
        nvmx_error("error reading prp list in prp2");
        req->comp_cb(req->comp_cb_user, status);
        return;
    }

    prp_to_iovec(req, (uint64_t *)req->dma_cmd.laddr);
    req->dma_cmd.op = req->prp.saved_op;

    nvmx_debug("PRP Check: PRP 2 List was used");

    if(req->is_tcp_rx_zcopy) {
        req->dma_cmd.comp_cb_user = req;
        req->dma_cmd.comp_cb = req->comp_cb;
        req->dma_cmd.writev2v = true;
        req->dma_cmd.op = NVME_REQ_DMA_TO_HOST;
        nvme_emu_request_submit(req);
    }
    else {
        if (req->dma_cmd.op == NVME_REQ_IOV)
            req->comp_cb(req->comp_cb_user, NVMX_OK);
        else
            prp_submit_iov(req);
    }
}

static inline int prp_submit_single(nvme_async_req_t *req)
{
    int rc;

    /* prp1 points to the phys contig memory block,
     * all can be done in one dma op */
    if (req->dma_cmd.op == NVME_REQ_IOV) {
        req->iov_cnt = 1;
        req->iov[0].iov_base = (void *)req->prp.prp1;
        req->iov[0].iov_len  = req->len;
        req->comp_cb(req->comp_cb_user, NVMX_OK);
        rc = NVMX_INPROGRESS;
    } else {
        req->dma_cmd.raddr   = req->prp.prp1;
        req->dma_cmd.len     = req->len;
        if (unlikely(req->bdata)) {
            req->dma_cmd.comp_cb = prp_bcopy_operation_done;
            req->dma_cmd.comp_cb_user = req;
        }
        else {
            req->dma_cmd.comp_cb = req->comp_cb;
            req->dma_cmd.comp_cb_user = req->comp_cb_user;
        }

        rc = nvme_emu_request_submit(req);
    }

    return rc;
}

int nvme_prp_rw(nvme_async_req_t *req)
{
    size_t to_read;

    /*
     * TODO: validation checks on prp offsets as required by spec
     *
     * The function reads/writes from/to prp list. It tries to combine
     * contig address and minimize amount of DMA transfers.
     * The algorithm is:
     * - check if data fit in prp1. If so read prp1. all done.
     * - check if prp2 has a prp list or data.
     * - if data
     *       check if prp1 + prp2 describe contig memory. If so do one DMA
     *       else do two DMAs
     *       all done
     * - if list
     *       read all list entries.
     *       use prp1 and list entries to build iovec. if prp[N] + 1 == prp[N+1]
     *       prp[N+1] is folded into the same iovec. See prp_to_iovec().
     * - for each iov read data
     */

    /* validate length. At the moment we and the FW support only prp_page_size/64
     * list entries.
     */

    nvmx_debug("total length %d[B]", (int)req->len);
    if (unlikely(req->len > 64*PRP_PAGE_SIZE(req->ops->page_bits))) {
        nvmx_error("len %d is bigger than 256K", (int)req->len);
        return NVMX_ERR;
    }

    req->err_code = NVME_SC_INTERNAL_DEV_ERROR;
    to_read = PRP_PAGE_SIZE(req->ops->page_bits) -
              PRP_PAGE_OFFSET(req->prp.prp1, req->ops->page_bits);
    nvmx_debug("prp1 read size %d[B]", (int)to_read);

    req->dma_cmd.rkey  = 0;
    req->dma_cmd.laddr = req->io_buf;

    req->iov_cnt = 1;
    req->iov[0].iov_base = (void *)req->prp.prp1;
    req->iov[0].iov_len  = to_read;

    if (to_read >= req->len) {
        /* data are in the single prp. Good for us */
        if(req->is_tcp_rx_zcopy) {
            req->iov[0].iov_len = req->len;
            return nvme_emu_request_submit(req);
        } else
            return prp_submit_single(req);
    }

    if (unlikely(!req->prp.prp2)) {
        nvmx_error("prp2 must not be NULL when reach to here.");
        req->err_code = NVME_SC_INVALID_FIELD;
        return NVMX_ERR;
    }

    /* try to compact prp1 + prp2 */
    if (req->prp.prp1 + to_read == req->prp.prp2 && req->len - to_read <= 
            PRP_PAGE_SIZE(req->ops->page_bits)) {
        nvmx_debug("compacted 0x%lx and 0x%lx", req->prp.prp1, req->prp.prp2);
        if(req->is_tcp_rx_zcopy) {
            req->iov[0].iov_len  = req->len;
            return nvme_emu_request_submit(req);
        }
        else {
            return prp_submit_single(req);
        }
    }

    if (req->len - to_read <= PRP_PAGE_SIZE(req->ops->page_bits)) {
        nvmx_debug("data are fit into prp1 and prp2 but are not contig");
        /* data are fit into prp1 and prp2 but are not contig. In this case
         * we have to do two operations */
        req->iov_cnt = 2;
        req->iov[1].iov_base = (void *)req->prp.prp2;
        req->iov[1].iov_len  = req->len - to_read;
        if(req->is_tcp_rx_zcopy) {
            return nvme_emu_request_submit(req);
        }
        else {
            if (req->dma_cmd.op == NVME_REQ_IOV) {
                req->comp_cb(req->comp_cb_user, NVMX_OK);
                return NVMX_INPROGRESS;
            } else {
                return prp_submit_iov(req);
            }
         }
    }

    /* note: the check implies that list_size > 1 */
    /* prp list, read list entries first then try to read data in contig chunks */
    req->prp.list_size = (req->len - to_read + PRP_PAGE_SIZE(req->ops->page_bits) - 1)/PRP_PAGE_SIZE(req->ops->page_bits);
    
    if(req->is_tcp_rx_zcopy) { 
        req->dma_cmd.laddr = req->io_buf;
    }
    else {
        /* If we read from host, we can use data buffer to receive prp list first, when convert it to IOVs and read. So no need for aux buffer */
        req->dma_cmd.laddr = req->dma_cmd.op == NVME_REQ_DMA_FROM_HOST ? req->io_buf : req->aux_buf;
    }

    /* According to the NVMe spec prp list addresses are 8 bytes */
    req->dma_cmd.raddr = req->prp.prp2;
    req->dma_cmd.len = sizeof(uint64_t) * req->prp.list_size;
    req->dma_cmd.comp_cb = read_prp_list_completion;
    req->dma_cmd.comp_cb_user = req;
    req->dma_cmd.writev2v = false;
    req->prp.saved_op = req->dma_cmd.op;
    req->dma_cmd.op = NVME_REQ_DMA_FROM_HOST;
    nvmx_debug("prp2 is a list entry. remaining length %d[B] (equals %d pages)",
            req->len - to_read, req->prp.list_size);

    return nvme_emu_request_submit(req);
}

static void prp_mem_ready(void *data, void *user)
{
    int rc;
    nvme_async_req_t *req = user;
    const size_t aux_size = nvme_prp_needs_aux(req) ? 1024ULL : 0;

    req->io_buf = data;
    req->aux_buf = aux_size ? req->io_buf + aux_size : NULL;

    if (req->dma_cmd.op == NVME_REQ_DMA_TO_HOST) {
        memcpy(req->io_buf, req->bdata, req->len);
        free(req->bdata);
    }

    rc = nvme_prp_rw(req);

    if (rc != NVMX_INPROGRESS) {
        req->bdata = NULL;
        req->comp_cb(req->comp_cb_user, rc);
    }
}

/**
 * nvme_prp_write_bcopy() - Start 'bcopy' prp write
 * @req:  async request
 * @prp1: prp1 physical memory address as specified by the NVMe spec
 * @prp2: prp2 physical memory address ad specified by the NVMe spec
 * @buf:  data buffer
 * @len:  data length
 *
 * The function copies @buf into the supplied @req->data_buf and starts
 * prp write to the host memory. The @req->comp_cb() will be called once the
 * request is completed.
 *
 * If request is to be done asynchronously, we do a shadow copy of @buf
 * so that client can release it's buffer right when this function returns
 *
 * Return: NVMX_INPROGRESS or error code
 */
int nvme_prp_write_bcopy(nvme_async_req_t *req, uint64_t prp1, uint64_t prp2,
                                       void *buf, size_t len)
{
    void *mem;
    int status;
    size_t aux_size;
    req->len = len;
    req->bdata = buf;
    req->is_prp = true;
    req->prp.prp1 = prp1;
    req->prp.prp2 = prp2;
    req->dma_cmd.op = NVME_REQ_DMA_TO_HOST;
    req->err_code = NVME_SC_INTERNAL_DEV_ERROR;

    if (req->memp) {
        aux_size = nvme_prp_needs_aux(req) ? 1024ULL : 0;
        mem = memp_alloc(req->memp, req->len + aux_size, req, prp_mem_ready,
                                                NULL, req->thread_id);
        if (mem) {
            req->io_buf = mem;
            req->aux_buf = aux_size ? req->io_buf + req->len : NULL;
        }
        else
        if (errno) {
            req->bdata = NULL;
            return NVMX_ERR;
        }
        else {
            req->bdata = malloc(len);
            if (!req->bdata)
                return NVMX_ERR;

            memcpy(req->bdata, buf, len);
            return NVMX_INPROGRESS;
        }
    }

    memcpy(req->io_buf, buf, len);
    status = nvme_prp_rw(req);
    if (status != NVMX_INPROGRESS)
        req->bdata = NULL;

    return status;
}

/**
 * nvme_prp_read_bcopy() - Start 'bcopy' prp read
 * @req:  async request
 * @prp1: prp1 physical memory address as specified by the NVMe spec
 * @prp2: prp2 physical memory address ad specified by the NVMe spec
 * @buf:  data buffer
 * @len:  data length
 *
 * The function starts prp read from the host memory.
 * When finish, the read data is copied to @buf and the @req->comp_cb()
 * is called.
 *
 * Return: NVMX_INPROGRESS or error code
 */
int nvme_prp_read_bcopy(nvme_async_req_t *req, uint64_t prp1, uint64_t prp2,
                                        void *buf, size_t len)
{
    void *mem;
    int status;
    size_t aux_size;
    req->len = len;
    req->bdata = buf;
    req->is_prp = true;
    req->prp.prp1 = prp1;
    req->prp.prp2 = prp2;
    req->dma_cmd.op = NVME_REQ_DMA_FROM_HOST;
    req->err_code = NVME_SC_INTERNAL_DEV_ERROR;

    if (req->memp) {
        aux_size = nvme_prp_needs_aux(req) ? 1024ULL : 0;
        mem = memp_alloc(req->memp, req->len + aux_size, req, prp_mem_ready,
                                                NULL, req->thread_id);
        if (mem) {
            req->io_buf = mem;
            req->aux_buf = aux_size ? req->io_buf + req->len : NULL;
        }
        else
        if (errno) {
            req->bdata = NULL;
            return NVMX_ERR;
        }
        else
            return NVMX_INPROGRESS;
    }

    status = nvme_prp_rw(req);
    if (status != NVMX_INPROGRESS)
        req->bdata = NULL;

    return status;
}
