#include <stdio.h>
#include <signal.h>

#include "compiler.h"
#include "config.h"
#include "nvme_emu.h"
#include "snap_conf.h"
#include "json/json.h"
#include "json/nvme_json.h"

#define MAX_NS_TYPE_LEN 16
#define MAX_EMU_TYPE_LEN 16
#define MAX_PARAM_LEN 64
#define MAX_HOSTID_UUID_LEN 37

#define DEFAULT_BE_NAME "snap_backend"
#define DEFAULT_BE_BLOCK_ORDER 9
#define OFFLOAD_EMU_MODE SNAP_EMU_CC_REMOTE
#define NON_OFFLOAD_EMU_MODE SNAP_EMU_SQE_ONLY

/* Configuration parameters names */
#define MN_PARAM "ctrl.mn"
#define SN_PARAM "ctrl.sn"
#define NN_PARAM "ctrl.nn"
#define MNAN_PARAM "ctrl.mnan"
#define PCI_FUNC_PARAM "ctrl.func_num"
#define SQES_PARAM "ctrl.sqes"
#define CQES_PARAM "ctrl.cqes"
#define CQ_PERIOD_PARAM "ctrl.cq_period"
#define CQ_MAX_COUNT_PARAM "ctrl.cq_max_count"
#define IO_QUEUES_PARAM "ctrl.nr_io_queues"
#define MDTS_PARAM "ctrl.mdts"
#define ONCS_PARAM "ctrl.oncs"
#define MAX_NS_PARAM "ctrl.max_namespaces"
#define EMU_MODE_PARAM "ctrl.emu_mode"
#define RDMA_DEV_PARAM "ctrl.rdma_device"
#define RPC_SERVER_PARAM "ctrl.rpc_server"
#define OFFLOAD_PARAM "ctrl.offload"
#define ZCOPY_ALIGN_PARAM "ctrl.zcopy_align"
#define MEM_PARAM "ctrl.mem"
#define IOVA_MGMT_PARAM "ctrl.iova_mgmt"
#define BACKENDS_PARAM "backends"
#define BACKEND_ID_PARAM "backends[%d].id"
#define BACKEND_HOSTID_PARAM "backends[%d].hostid"
#define BACKEND_NAME_PARAM "backends[%d].name"
#define BACKEND_TYPE_PARAM "backends[%d].type"
#define BACKEND_SIZE_PARAM "backends[%d].size_mb"
#define BACKEND_BLOCK_ORDER_PARAM "backends[%d].block_order"
#define BACKEND_PATHS_PARAM "backends[%d].paths"
#define PATH_ADDR_PARAM "backends[%d].paths[%d].addr"
#define PATH_PORT_PARAM "backends[%d].paths[%d].port"
#define PATH_NQN_PARAM "backends[%d].paths[%d].nqn"
#define PATH_KA_TO_PARAM "backends[%d].paths[%d].ka_timeout_ms"

/* Returns true if is a valid ASCII string as defined by the NVMe spec */
static bool snap_config_valid_ascii_string(const char *str, size_t size)
{
    size_t i;

    for (i = 0; i < size; i++) {
        if (str[i] < 0x20 || str[i] > 0x7E) {
            return false;
        }
    }

    return true;
}

static int snap_config_set_ascii_string(const char *val, char *dst,
                                        size_t max_len, char *param)
{
    size_t val_len;

    if (!val) {
        nvmx_error("No value to set");
        return -1;
    }
    val_len = strnlen(val, max_len);

    if (!snap_config_valid_ascii_string(val, val_len)) {
        nvmx_error("Value (%s) is not a valid ASCII string", val);
        return -1;
    }

    snprintf(dst, max_len, "%s", val);
    nvmx_debug("Setting %s to %s", param, val);

    return 0;
}

bool snap_config_get_offload(snap_conf_t *config)
{
    return config->emu_mode == OFFLOAD_EMU_MODE;
}

int snap_config_set_sn(snap_conf_t *config, const char *val)
{
    return snap_config_set_ascii_string(val, config->sn, MAX_SN_LEN, "sn");
}

int snap_config_set_mn(snap_conf_t *config, const char *val)
{
    return snap_config_set_ascii_string(val, config->mn, MAX_MN_LEN, "mn");
}

int snap_config_set_mem(snap_conf_t *config, const char *mem)
{
    if (strncmp(mem, "pool", 4) && strncmp(mem, "static", 6)) {
        nvmx_error("Invalid memory mode (%s)", mem);
        return -1;
    }
    return snap_config_set_ascii_string(mem, config->mem, MAX_MEM_LEN, "mem");
}

void snap_config_set_nn(snap_conf_t *config, uint32_t nn)
{
    config->nn = nn;
    nvmx_debug("Setting Number Of Namespaces to %u", config->nn);
}

void snap_config_set_mnan(snap_conf_t *config, uint32_t mnan)
{
    config->mnan = mnan;
    nvmx_debug("Setting Maximum Number of Allowed Namespaces to %u", config->mnan);
}

int snap_config_set_emu_mode(snap_conf_t *config, const char *val)
{
    enum snap_config_emu_mode mode;

    if (!val) {
        nvmx_error("Invalid emulation mode (%s)", val);
        return -1;
    } else {
        config->emu_mode = SNAP_EMU_SQE_ONLY;
        for (mode = 0; mode < SNAP_EMU_MAX; mode++) {
            if (!strncmp(val, emu_modes_str[mode],
                         strlen(emu_modes_str[mode]))) {
                config->emu_mode = mode;
                goto found;
            }
        }
    }

    nvmx_error("Invalid emulation mode (%s)", val);
    return -1;
found:
    nvmx_debug("Setting emulation mode to %s", emu_modes_str[config->emu_mode]);
    return 0;
}

void snap_config_set_mdts(snap_conf_t *config, uint32_t mdts)
{
    /*
     * MDTS = 0 means no limit should apply. So instead, set
     * the maximal limit possible
     */
    if (!mdts || mdts > SNAP_CONFIG_MAX_MDTS) {
        config->mdts = SNAP_CONFIG_MAX_MDTS;
        nvmx_info("Replacing MDTS (%u) to Max value (%u)", mdts,
                  SNAP_CONFIG_MAX_MDTS);
    } else {
        config->mdts = mdts;
        nvmx_debug("Setting MDTS to %u", config->mdts);
    }
}

void set_mdts_using_defaults(snap_conf_t *sconfig) {
    if (sconfig->emu_mode == SNAP_EMU_CC_REMOTE)
        snap_config_set_mdts(sconfig, SNAP_CONFIG_DEFAULT_MDTS_CC_REMOTE);
    else
        snap_config_set_mdts(sconfig, SNAP_CONFIG_DEFAULT_MDTS_OTHER);
}

int snap_config_set_num_queues(snap_conf_t *config, uint32_t val)
{
    if (val) {
        config->num_queues = val;
        nvmx_debug("Setting num_queues to %u", config->num_queues);
        return 0;
    } else {
        nvmx_error("Invalid num_queues \'%u\'", val);
        return -1;
    }
}

void snap_config_set_pci_func(snap_conf_t *config,  int index)
{
    nvmx_debug("Setting pci_index to %d", index);
    config->pci_index = index;
}

void snap_config_set_pci_vfunc(snap_conf_t *config,  int index)
{
    nvmx_debug("Setting VF pci_index to %d", index);
    config->vf_pci_index = index;
}

void snap_config_set_pci_bdf(snap_conf_t *config, const char *bdf)
{
    nvmx_debug("Setting pci_bdf to %s", bdf);
    strncpy(config->pci_bdf, bdf, MAX_PCI_FUNC_LEN);
}

void snap_config_set_cq_period(snap_conf_t *config, uint32_t cq_period)
{
    nvmx_debug("Setting cq_period to %u", cq_period);
    config->cq_period = cq_period;
}

void
snap_config_set_cq_max_count(snap_conf_t *config, uint32_t cq_max_count)
{
    nvmx_debug("Setting cq_max_count to %u", cq_max_count);
    config->cq_max_count = cq_max_count;
}

int snap_config_set_oncs(snap_conf_t *config, uint32_t oncs)
{
    if (oncs && oncs != 0x4) {
        nvmx_error("Invalid oncs (%u)", oncs);
        return -1;
    }

    nvmx_debug("Setting oncs to %u", oncs);
    config->oncs = oncs;
    return 0;
}

int snap_config_set_max_namespaces(snap_conf_t *config, uint32_t max_ns)
{
    config->max_namespaces = max_ns;
    nvmx_debug("Setting max_namespaces to %u", config->max_namespaces);
    return 0;
}

int snap_config_set_quirks(snap_conf_t *config,
                           enum nvme_quirks quirk_mask)
{
    /* Given quirk mask must be a subset of all available quirks,
     * so if it contains bits that are not supported as valid quirks,
     * fail on bad configuration
     */
    if (quirk_mask & ~SNAP_CONFIG_ALLOWED_QUIRKS) {
        nvmx_error("Invalid quirks (%d)", quirk_mask);
        return -1;
    }

    nvmx_debug("Setting quirks to %d", quirk_mask);
    config->quirks = quirk_mask;
    return 0;
}

void snap_config_set_sqes(snap_conf_t *config, uint32_t sqes)
{
    nvmx_debug("Setting sqes to %u", sqes);
    config->sqes = sqes ? : SNAP_CONFIG_DEFAULT_SQES;
}

void snap_config_set_cqes(snap_conf_t *config, uint32_t cqes)
{
    nvmx_debug("Setting cqes to %u", cqes);
    config->cqes = cqes ? : SNAP_CONFIG_DEFAULT_CQES;
}

void snap_config_set_cntlid(snap_conf_t *config, uint16_t cntlid)
{
    config->cntlid = cntlid;
}

int snap_config_set_emu_manager(snap_conf_t *config, const char *emu_manager)
{
    if (strnlen(emu_manager, MAX_RDMA_DEV_LEN + 1) > MAX_RDMA_DEV_LEN) {
        nvmx_error("emu_manager (%s) too long!", emu_manager);
        return -1;
    } else {
        snprintf(config->emu_manager, MAX_RDMA_DEV_LEN, "%s", emu_manager);
        nvmx_debug("Setting emu_manager to %s", emu_manager);
        return 0;
    }
}

void snap_config_set_zcopy_align(snap_conf_t *config, uint32_t zcopy_align)
{
    nvmx_debug("Setting zcopy_align to %u", zcopy_align);
    config->zcopy_align = zcopy_align;
}

void snap_config_set_iova_mgmt(snap_conf_t *config, bool iova_mgmt)
{
    nvmx_debug("Setting iova_mgmt to %u", iova_mgmt);
    config->iova_mgmt = iova_mgmt;
}

int snap_config_set_rdma_dev(snap_conf_t *config, const char *rdma_dev)
{
    if (strnlen(rdma_dev, MAX_RDMA_DEV_LEN + 1) > MAX_RDMA_DEV_LEN) {
        nvmx_error("rdma_dev (%s) too long!", rdma_dev);
        return -1;
    } else {
        snprintf(config->rdma_dev, MAX_RDMA_DEV_LEN, "%s", rdma_dev);
        nvmx_debug("Setting rdma_dev to %s", rdma_dev);
        return 0;
    }
}

int snap_config_set_rpc_server(snap_conf_t *config, const char *rpc_server)
{
    if (strnlen(rpc_server, MAX_RPC_SERVER_LEN + 1) > MAX_RPC_SERVER_LEN) {
        nvmx_error("rpc_server (%s) too long!", rpc_server);
        return -1;
    } else {
        snprintf(config->rpc_server, MAX_RPC_SERVER_LEN, "%s", rpc_server);
        nvmx_debug("Setting rpc_server to %s", rpc_server);
        return 0;
    }
}

int snap_config_set_backends(snap_conf_t *config)
{
    snap_config_be_path_t *path;
    snap_config_backend_t *backend;
    char be_hostid_str[MAX_HOSTID_UUID_LEN] = {0};

    nvmx_debug("Setting backend count to 1");
    config->backends_cnt = 1;

    /* set default value to backend */
    config->backends = calloc(1, sizeof(snap_config_backend_t));
    if (!config->backends) {
        nvmx_error("Allocate memory for backends failed");
        return -1;
    }

    backend = config->backends;
    nvmx_debug("Setting backend type to `spdk_bdev`");
    memcpy(backend->be_type, "spdk_bdev", sizeof("spdk_bdev"));
    uuid_generate(backend->hostid);
    uuid_unparse(backend->hostid, be_hostid_str);
    nvmx_debug("Using ctrl Host Identifier %s", be_hostid_str);
    nvmx_debug("Setting backend path count to 1");
    backend->paths_cnt = 1;

    /* set default value to backend path */
    backend->paths = calloc(1, sizeof(snap_config_be_path_t));
    if (!backend->paths) {
        free(backend);
        nvmx_error("Allocate memory for backends path failed");
        return -1;
    }

    path = backend->paths;
    path->ka_timeout_ms = SNAP_CONFIG_DEFAULT_PATH_KA_TO_MS;
    nvmx_debug("Setting backend path timeout to %d ms", path->ka_timeout_ms);

    return 0;
}

snap_config_backend_t *get_backend(snap_conf_t *config, int id)
{
    if ((id > -1) && (id < config->backends_cnt))
        return &config->backends[id];
    else
        return NULL;
}

const snap_config_backend_t *snap_config_get_backend(snap_conf_t *config, int id)
{
    return get_backend(config, id);
}

const snap_config_be_path_t *snap_config_get_path(snap_conf_t *config, int be_id,
                                                  int path_id)
{
    const snap_config_backend_t *backend = get_backend(config, be_id);

    if (backend)
        return snap_config_get_be_path(backend, path_id);
    else
        return NULL;
}

static int load_path(snap_config_backend_t *backend, int be_id, snap_conf_t *sconfig,
                     nvme_config_t *config, int path_id)
{
    snap_config_be_path_t *path;
    char addr[MAX_PARAM_LEN] = {0};
    char port[MAX_PARAM_LEN] = {0};
    char nqn[MAX_PARAM_LEN] = {0};
    char ka_to[MAX_PARAM_LEN] = {0};

    path = &(backend)->paths[path_id];

    snprintf(ka_to, sizeof(ka_to), PATH_KA_TO_PARAM, be_id, path_id);
    if (nvme_json_get_value(config, ka_to, &path->ka_timeout_ms,
                            sizeof(path->ka_timeout_ms))) {
        nvmx_debug("Failed loading %s from conf file, Using default (%u)",
                  ka_to, SNAP_CONFIG_DEFAULT_PATH_KA_TO_MS);
        path->ka_timeout_ms = SNAP_CONFIG_DEFAULT_PATH_KA_TO_MS;
    }

    snprintf(addr, sizeof(addr), PATH_ADDR_PARAM, be_id, path_id);
    if (nvme_json_get_value(config, addr, path->addr,
                            sizeof(path->addr))) {
        if (sconfig->emu_mode == OFFLOAD_EMU_MODE) {
            nvmx_error("For offload mode, valid backend addr must be provided");
            goto fail;
        } else {
            nvmx_debug("Failed loading %s from conf file, Ignoring.", addr);
            memset(path->addr, 0, sizeof(path->addr));
        }
    }

    snprintf(nqn, sizeof(nqn), PATH_NQN_PARAM, be_id, path_id);
    if (nvme_json_get_value(config, nqn, path->nqn,
                            sizeof(path->nqn))) {
        nvmx_debug("Failed loading %s from conf file, Ignoring.", nqn);
        memset(path->nqn, 0, sizeof(path->nqn));
    }

    snprintf(port, sizeof(port), PATH_PORT_PARAM, be_id, path_id);
    if (nvme_json_get_value(config, port, &path->port,
                            sizeof(path->port))) {
        if (sconfig->emu_mode == OFFLOAD_EMU_MODE) {
            nvmx_error("For offload mode, valid backend port must be provided");
            goto fail;
        } else {
            nvmx_debug("Failed loading %s from conf file, Ignoring.", port);
            path->port = 0;
        }
    }

    return 0;

fail:
    return -1;
}

static int load_backend(snap_conf_t *sconfig, nvme_config_t *config, int id)
{
    snap_config_backend_t *backend;
    char be_id[MAX_PARAM_LEN] = {0};
    char be_hostid[MAX_PARAM_LEN] = {0};
    char be_hostid_str[MAX_HOSTID_UUID_LEN] = {0};
    char be_type[MAX_PARAM_LEN] = {0};
    char be_name[MAX_PARAM_LEN] = {0};
    char be_size_mb[MAX_PARAM_LEN] = {0};
    char be_block_order[MAX_PARAM_LEN] = {0};
    char paths[MAX_PARAM_LEN] = {0};
    int i;

    backend = get_backend(sconfig, id);
    if (!backend) {
        nvmx_error("Failed to get backend %d", id);
        return -1;
    }

    snprintf(be_type, sizeof(be_type), BACKEND_TYPE_PARAM, id);
    if (nvme_json_get_value(config, be_type, backend->be_type,
                            sizeof(backend->be_type))) {
        nvmx_error("Failed loading %s from conf file!", be_type);
        goto fail;
    } else if (sconfig->emu_mode == OFFLOAD_EMU_MODE) {
        if (strncmp(backend->be_type, "nvmf_rdma", 9)) {
            nvmx_error("backend type '%s' does not support offload mode",
                       backend->be_type);
            goto fail;
        }
    }

    snprintf(be_name, sizeof(be_name), BACKEND_NAME_PARAM, id);
    if (nvme_json_get_value(config, be_name, backend->name,
                            sizeof(backend->name))) {
        if (sconfig->emu_mode == OFFLOAD_EMU_MODE) {
            nvmx_error("For offload mode, valid backend name must be provided");
            goto fail;
        } else {
            snprintf(backend->name, sizeof(backend->name), DEFAULT_BE_NAME);
            nvmx_debug("Failed loading %s from conf file, Using default name (%s)",
                       be_name, backend->name);
        }
    }

    snprintf(be_id, sizeof(be_id), BACKEND_ID_PARAM, id);
    if (nvme_json_get_value(config, be_id, backend->id, sizeof(backend->id))) {
        memset(backend->id, 0, sizeof(backend->id));
        nvmx_debug("Failed loading %s from conf file, Ignoring.", be_id);
    }

    snprintf(be_hostid, sizeof(be_hostid), BACKEND_HOSTID_PARAM, id);
    if (nvme_json_get_value(config, be_hostid, be_hostid_str,
                            sizeof(be_hostid_str))) {
        uuid_generate(backend->hostid);
        uuid_unparse(backend->hostid, be_hostid_str);
    } else {
        if (uuid_parse(be_hostid_str, backend->hostid)) {
            nvmx_error("Invalid Host Identifier %s. "
                       "Identifier UUID is a hex string of the form "
                       "1b4e28ba-2fa1-11d2-883f-b9a761bde3fb",
                       be_hostid_str);
            goto fail;
        }
    }
    nvmx_debug("Using ctrl Host Identifier %s", be_hostid_str);

    /*
     * In memdisk/posix_io backends, if user wishes for namespace to exist,
     * he must explicitly set its size and block size. In case not specified,
     * no namespaces will be configured.
     */
    if (!strncmp(backend->be_type, "memdisk", 7) ||
        !strncmp(backend->be_type, "posix_io", 8)) {
        snprintf(be_size_mb, sizeof(be_size_mb), BACKEND_SIZE_PARAM, id);
        if (nvme_json_get_value(config, be_size_mb, &backend->size_b,
                                sizeof(backend->size_b))) {
            nvmx_warn("Using %s backend with no namespaces", backend->be_type);
            backend->size_b = 0;
        } else {
            nvmx_debug("Setting backend size to %lu MB", backend->size_b);
            backend->size_b <<= MB_TO_BYTES_SHIFT;

            snprintf(be_block_order, sizeof(be_block_order),
                     BACKEND_BLOCK_ORDER_PARAM, id);
            if (nvme_json_get_value(config, be_block_order, &backend->block_order,
                                sizeof(backend->block_order)))
                backend->block_order = DEFAULT_BE_BLOCK_ORDER;
            nvmx_debug("Setting backend block order to %u", backend->block_order);
        }
    }

    snprintf(paths, sizeof(paths), BACKEND_PATHS_PARAM, id);
    backend->paths_cnt = nvme_json_get_array_size(config, paths);
    if (backend->paths_cnt > 0) {
        backend->paths = calloc(backend->paths_cnt,
                                sizeof(snap_config_be_path_t));
        for (i = 0 ; i < backend->paths_cnt ; i++) {
            if (load_path(backend, id, sconfig, config, i))
                goto fail;
	    }
    } else {
        nvmx_error("Invalid number of paths (%d) for %s!", backend->paths_cnt,
                   paths);
        goto fail;
    }
    return 0;

fail:
    return -1;
}

static void free_backends(snap_conf_t *sconfig)
{
    int i;
    snap_config_backend_t *backend;

    for (i = 0; i < sconfig->backends_cnt; i++) {
        backend = &(sconfig)->backends[i];
        if (backend->paths)
            free(backend->paths);
    }

    free(sconfig->backends);
}

static int load_backends(snap_conf_t *sconfig, nvme_config_t *config)
{
    int backends_cnt, i;

    if (sconfig->backends_cnt)
        free_backends(sconfig);

    backends_cnt = nvme_json_get_array_size(config, BACKENDS_PARAM);
    if (backends_cnt > 0) {
        sconfig->backends_cnt = backends_cnt;
        sconfig->backends = calloc(backends_cnt, sizeof(snap_config_backend_t));
        for (i = 0 ; i < backends_cnt ; i++)
            if (load_backend(sconfig, config, i))
                goto fail;
    } else {
        nvmx_error("Invalid number of backends (%d)", backends_cnt);
        goto fail;
    }
    return 0;
fail:
    return -1;
}

static int snap_config_parse_json_conf(snap_conf_t *sconfig,
                                       const char *json_file)
{
    nvme_config_t config;
    char sn[MAX_SN_LEN];
    char mn[MAX_MN_LEN];
    uint32_t nn;
    uint32_t mnan;
    char rdma_dev[MAX_RDMA_DEV_LEN];
    char rpc_server[MAX_RPC_SERVER_LEN];
    char emu_mode[MAX_EMU_TYPE_LEN];
    bool offload;
    char mem[MAX_MEM_LEN];
    const char *conf_emu_mode;
    uint32_t func_num;
    uint32_t sqes,cqes;
    uint32_t cq_period, cq_max_cnt;
    uint32_t nr_io_queues;
    uint32_t mdts, oncs;
    uint32_t max_ns;
    int ret;
    uint32_t zcopy_align;
    bool iova_mgmt;

    /* No config was provided, assume we will be running with defaults only */
    if (!json_file)
        return 0;

    memset(&config, 0, sizeof(config));

    if (nvme_json_parse_config(json_file, &config) < 0) {
        nvmx_error("Failed loading json file");
        ret = -1;
        goto out;
    }

    nvme_json_get_quirks(&config, &sconfig->quirks);

    if (nvme_json_get_value(&config, MN_PARAM, mn, MAX_MN_LEN)) {
        nvmx_debug("Failed loading mn from conf file, using default (%s)",
                  snap_config_get_mn(sconfig));
    } else {
        if (snap_config_set_mn(sconfig, mn))
            nvmx_debug("Failed setting mn from conf file (%s),"
                      " using default (%s)", mn, snap_config_get_mn(sconfig));
    }

    if (nvme_json_get_value(&config, SN_PARAM, sn, MAX_SN_LEN)) {
        nvmx_debug("Failed loading sn from conf file, using default (%s)",
                  snap_config_get_sn(sconfig));
    } else {
        if (snap_config_set_sn(sconfig, sn))
            nvmx_debug("Failed setting sn from conf file (%s),"
                       " using default (%s)", sn, snap_config_get_sn(sconfig));
    }


    if (nvme_json_get_value(&config, NN_PARAM, &nn, sizeof(nn))) {
        nvmx_debug("Failed setting number of namespaces from conf file,"
                  " using default (%u)", snap_config_get_nn(sconfig));
    } else {
        snap_config_set_nn(sconfig, nn);
    }

    if (nvme_json_get_value(&config, MNAN_PARAM, &mnan, sizeof(mnan))) {
        nvmx_debug("Failed setting maximum number of namespaces from conf file,"
                  " using default (%u)", snap_config_get_mnan(sconfig));
    } else {
        snap_config_set_mnan(sconfig, mnan);
    }

    if (!nvme_json_get_value(&config, PCI_FUNC_PARAM, &func_num,
                             sizeof(func_num)))
        snap_config_set_pci_func(sconfig, func_num);

    if (nvme_json_get_value(&config, SQES_PARAM, &sqes, sizeof(sqes))) {
        nvmx_debug("Failed loading sqes from conf file, using default (%u)",
                  snap_config_get_sqes(sconfig));
    } else {
        snap_config_set_sqes(sconfig, sqes);
    }

    if (nvme_json_get_value(&config, CQES_PARAM, &cqes, sizeof(cqes))) {
        nvmx_debug("Failed loading cqes from conf file, using default (%u)",
                  snap_config_get_cqes(sconfig));
    } else {
        snap_config_set_cqes(sconfig, cqes);
    }

    if (nvme_json_get_value(&config, CQ_PERIOD_PARAM, &cq_period,
                            sizeof(cq_period))) {
        nvmx_debug("Failed loading cq_period from conf file, using default (%u)",
                  snap_config_get_sqes(sconfig));
    } else {
        snap_config_set_cq_period(sconfig, cq_period);
    }

    if (nvme_json_get_value(&config, CQ_MAX_COUNT_PARAM, &cq_max_cnt,
                            sizeof(cq_max_cnt))) {
        nvmx_debug("Failed loading cq_max_count from conf file, using default "
                  "(%u)", snap_config_get_cq_max_count(sconfig));
    } else {
        snap_config_set_cq_max_count(sconfig, cq_max_cnt);
    }

    if (nvme_json_get_value(&config, IO_QUEUES_PARAM, &nr_io_queues,
                            sizeof(nr_io_queues))) {
        nvmx_debug("Failed loading nr_io_queues from conf file, using default "
                  "(%u)", snap_config_get_num_queues(sconfig));
    } else {
        snap_config_set_num_queues(sconfig, nr_io_queues);
    }

    if (nvme_json_get_value(&config, ZCOPY_ALIGN_PARAM, &zcopy_align,
                            sizeof(zcopy_align))) {
        nvmx_debug("Failed loading zcopy_align from conf file, using default (%u)",
                  snap_config_get_zcopy_align(sconfig));
    } else {
        if (zcopy_align == 0) {
            /* probably the user wants to disable alignment check */
            nvmx_warn("zcopy alignment must be greater than zero. Forcing it to 1");
            zcopy_align = 1;
        }
        snap_config_set_zcopy_align(sconfig, zcopy_align);
    }

    if (nvme_json_get_value(&config, MEM_PARAM, mem, MAX_MEM_LEN)) {
        nvmx_debug("Failed loading mem from conf file, using default (%s)",
                  snap_config_get_mem(sconfig));
    } else {
        if (snap_config_set_mem(sconfig, mem))
            nvmx_debug("Failed setting mem from conf file (%s),"
                      " using default (%s)", mem, snap_config_get_mem(sconfig));
    }

    if (nvme_json_get_value(&config, EMU_MODE_PARAM, emu_mode,
                            sizeof(emu_mode))) {
       /*
        * Instead of using emu_mode directly, users can set boolean
        * "offload" JSON parameter, which can be translated to the
        * matching emu_mode.
        */
        if (nvme_json_get_value(&config, OFFLOAD_PARAM, &offload,
                                sizeof(offload))) {
            nvmx_error("Failed load offload param from conf file");
            ret = -1;
            goto free_config;
	} else {
            conf_emu_mode = offload ? emu_modes_str[OFFLOAD_EMU_MODE] :
                                      emu_modes_str[NON_OFFLOAD_EMU_MODE];
            if (snap_config_set_emu_mode(sconfig, conf_emu_mode))
                nvmx_debug("Failed setting offload as (%s),"
                          " using default emu_mode (%s)", conf_emu_mode,
                          snap_config_get_emu_mode_str(sconfig));
        }
    } else {
        if (snap_config_set_emu_mode(sconfig, emu_mode))
            nvmx_debug("Failed setting emulation mode from conf file (%s),"
                      " using default (%s)", emu_mode,
                      emu_modes_str[sconfig->emu_mode]);
    }

    if (nvme_json_get_value(&config, MDTS_PARAM, &mdts, sizeof(mdts))) {
        set_mdts_using_defaults(sconfig);
        nvmx_debug("Failed loading mdts from conf file, using default (%u)",
                  snap_config_get_mdts(sconfig));
    } else {
        snap_config_set_mdts(sconfig, mdts);
    }

    if (nvme_json_get_value(&config, ONCS_PARAM, &oncs, sizeof(oncs))) {
        nvmx_debug("Failed loading oncs from conf file, using default (%u)",
                  snap_config_get_oncs(sconfig));
    } else {
        if (snap_config_set_oncs(sconfig, oncs))
            nvmx_debug("Failed setting oncs from conf file (%d),"
                      " using default (%d)", oncs,
                      snap_config_get_oncs(sconfig));
    }

    if (nvme_json_get_value(&config, MAX_NS_PARAM, &max_ns, sizeof(max_ns))) {
        snap_config_set_max_namespaces(sconfig, SNAP_CONFIG_DEFAULT_MAX_NAMESPACES);
        nvmx_debug("Failed setting max namespaces from conf file (%d),"
                  " using default (%d)", max_ns,
                  snap_config_get_max_namespaces(sconfig));
    } else {
        snap_config_set_max_namespaces(sconfig, max_ns);
    }

    if (!nvme_json_get_value(&config, RDMA_DEV_PARAM, rdma_dev,
                             MAX_RDMA_DEV_LEN)) {
        if (snap_config_set_rdma_dev(sconfig, rdma_dev)) {
            nvmx_error("Failed setting rdma_dev from conf file (%s), Snap conf"
                       " failed!", rdma_dev);
            ret = -1;
            goto free_config;
        }
    }

    if (nvme_json_get_value(&config, RPC_SERVER_PARAM, rpc_server,
                            MAX_RPC_SERVER_LEN)) {
        nvmx_debug("Failed loading rpc_server from conf file, Using default "
                  "(%s)", snap_config_get_rpc_server(sconfig));
    } else {
        if (snap_config_set_rpc_server(sconfig, rpc_server)) {
            nvmx_debug("Failed setting rpc_server from conf file (%s), Using "
                      " default (%s)", rpc_server,
                      snap_config_get_rpc_server(sconfig));
        }
    }

    if (nvme_json_get_value(&config, IOVA_MGMT_PARAM, &iova_mgmt, sizeof(iova_mgmt))) {
        nvmx_info("Failed loading iova_mgmt from conf file, Using default "
                  "(%s)", snap_config_get_iova_mgmt(sconfig));
    }
    else
        snap_config_set_iova_mgmt(sconfig, iova_mgmt);

    if (load_backends(sconfig, &config)) {
        ret = -1;
        goto free_config;
    }

    ret = 0;

free_config:
    nvme_json_config_close(&config);
out:
    return ret;
}

void snap_config_destroy(snap_conf_t *config)
{
    if (config) {
        if (config->backends_cnt)
            free_backends(config);

        free(config);
    }
}

snap_conf_t *snap_config_create(const char *json_file)
{
    snap_conf_t *config;

    config = calloc(1, sizeof(snap_conf_t));
    if (!config) {
        nvmx_error("Failed to allocate snap config");
        goto fail;
    }

    if (snap_config_set_sn(config, SNAP_CONFIG_DEFAULT_SN))
        goto fail;

    if (snap_config_set_mn(config, SNAP_CONFIG_DEFAULT_MN))
        goto fail;

    if (snap_config_set_num_queues(config, SNAP_CONFIG_DEFAULT_NUM_QUEUES))
        goto fail;

    if (snap_config_set_oncs(config, SNAP_CONFIG_DEFAULT_ONCS))
        goto fail;

    if (snap_config_set_quirks(config, SNAP_CONFIG_DEFAULT_QUIRKS))
        goto fail;

    if (snap_config_set_emu_mode(config, SNAP_CONFIG_DEFAULT_TYPE))
        goto fail;

    if (snap_config_set_rpc_server(config, SNAP_CONFIG_DEFAULT_RPC_SERVER))
        goto fail;

    if (snap_config_set_backends(config))
        goto fail;

    if (snap_config_set_mem(config, SNAP_CONFIG_DEFAULT_MEM))
        goto fail;

    set_mdts_using_defaults(config);
    snap_config_set_cq_period(config, SNAP_CONFIG_DEFAULT_CQ_PERIOD);
    snap_config_set_cq_max_count(config, SNAP_CONFIG_DEFAULT_CQ_MAX_COUNT);
    snap_config_set_sqes(config, SNAP_CONFIG_DEFAULT_SQES);
    snap_config_set_cqes(config, SNAP_CONFIG_DEFAULT_CQES);
    snap_config_set_pci_func(config, 0);
    snap_config_set_pci_vfunc(config, -1);
    snap_config_set_zcopy_align(config, SNAP_CONFIG_DEFAULT_ZCOPY_ALIGN);
    snap_config_set_nn(config, SNAP_CONFIG_DEFAULT_NN);
    snap_config_set_mnan(config, SNAP_CONFIG_DEFAULT_MNAN);
    snap_config_set_max_namespaces(config, SNAP_CONFIG_DEFAULT_MAX_NAMESPACES);

    if (snap_config_parse_json_conf(config, json_file))
        goto fail;

    nvmx_debug("Configuration loaded successfully");
    return config;

fail:
    nvmx_error("Configuration loading failed (%s)!!!", json_file);
    snap_config_destroy(config);
    return NULL;
}
