#include "config.h"
#include "snap_nvme_emulator.h"
#include "nvme_regs.h"
#include "nvme_prp.h"
#include "nvme_sgl.h"
#include "snap_conf.h"
#include "snap_env.h"
#include "zcopy_snap.h"
#include "memp.h"

#if HAVE_SNAP
#include "mlnx_snap_pci_manager.h"

#define MINIMAL_SNAP_Q_SIZE 4

/* read from host bar space */
int nvme_emu_mmio_read(nvme_emulator_t *dev, void *buf,
                              uint32_t bar_addr, unsigned len)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    struct snap_nvme_device_attr attr = {};
    nvme_bar_t *bar;
    union nvme_cc_register *cc;
    int ret;

    dev->pci_hotplug_state_prev = dev->pci_hotplug_state_cur;
    ret = snap_nvme_query_device(snap_emu->sdev, &attr);
    if (ret)
        return NVMX_ERR;

    snap_emu->curr_enabled = attr.enabled;
    dev->pci_hotplug_state_cur = attr.pci_hotplug_state;
    if (snap_emu->curr_enabled != snap_emu->prev_enabled) {
        nvmx_debug("enable change: %d -> %d", snap_emu->prev_enabled, snap_emu->curr_enabled);
        if (!snap_emu->curr_enabled && snap_emu->prev_enabled) {
            bar = (nvme_bar_t *)snap_emu->sdev->pci->bar.data;
            cc = (union nvme_cc_register *)&bar->cc;
            nvmx_debug("FLR detected!!! cc = 0x%0x", cc->raw);
            /*
             * We must clear the cc.en bit in case the HW device state changed
             * to disabled. In this case, probably caused by FLR, the host
             * will eventually set the cc.en to 1 and SW controller might miss
             * this reset/FLR case. This is not good since all the HW resources
             * that where created on the NVMe function where destroyed but the
             * parallel SW resources are still alive. SW must destroy (and
             * later re-create) those resources to avoid having stale resources
             * and to configure the HW correctly. For that we clear the cc.en
             * bit so the NVMe SW controller will be notified and perform the
             * needed destruction.
             */
            cc->bits.en = 0;
            snap_emu->flr_active = true;
        }
        snap_emu->prev_enabled = snap_emu->curr_enabled;
    }

    if (!snap_emu->flr_active && snap_emu->attr.type == SNAP_NVME_PF)
        dev->dev_caps.max_emulated_vfs = attr.num_of_vfs;

    memcpy(buf, snap_emu->sdev->pci->bar.data + bar_addr, len);
    return NVMX_OK;
}

static int regidx2mask[SNAP_NVME_REG_CMBLOC_IDX] = {
    SNAP_NVME_DEV_MOD_BAR_CAP_VS_CSTS,
    SNAP_NVME_DEV_MOD_BAR_CAP_VS_CSTS,
    0,
    0,
    SNAP_NVME_DEV_MOD_BAR_CC,
    SNAP_NVME_DEV_MOD_BAR_CAP_VS_CSTS,
    0,
    SNAP_NVME_DEV_MOD_BAR_AQA_ASQ_ACQ,
    SNAP_NVME_DEV_MOD_BAR_AQA_ASQ_ACQ,
    SNAP_NVME_DEV_MOD_BAR_AQA_ASQ_ACQ
};

static uint64_t
snap_emu_mmio_modify_mask_get(uint32_t bar_addr, unsigned len)
{
    int i;

    for (i = SNAP_NVME_REG_CAP_IDX; i <= SNAP_NVME_REG_ACQ_IDX; i++) {
        if (bar_addr == nvme_regs[i].reg_base)
            break;
    }

    if (i > SNAP_NVME_REG_ACQ_IDX)
        return 0;

    if (len == nvme_regs[i].reg_size) {
        return regidx2mask[i];
    } else {
        uint64_t mask = 0;
        while (len && i <= SNAP_NVME_REG_ACQ_IDX) {
            mask |= regidx2mask[i];
            len -= nvme_regs[i].reg_size;
            i++;
        }
        return mask;
    }
}

/* write to host bar space */
int nvme_emu_mmio_write(nvme_emulator_t *dev, void *buf,
                               uint32_t bar_addr, unsigned len)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    struct snap_device *sdev = snap_emu->sdev;
    struct snap_nvme_device_attr attr = {};
    int ret;
    uint64_t modify_mask;

    /* copy old value to attributes */
    memcpy((uint8_t *)&attr.bar, sdev->pci->bar.data, sdev->pci->bar.size);
    /* modify needed register values */
    memcpy((uint8_t *)&attr.bar + bar_addr, buf, len);

    modify_mask = snap_emu_mmio_modify_mask_get(bar_addr, len);

    ret = snap_nvme_modify_device(sdev, modify_mask, &attr);
    if (ret) {
        nvmx_error("dev 0x%p modify SNAP BAR. ret=%d", dev, ret);
        return NVMX_ERR;
    }

    return NVMX_OK;
}

static void snap_ctx_close(struct snap_context *sctx)
{
    /* Empty function, for symmetry */
}

nvme_emulator_t *nvme_emu_dev_open(const char *hca_manager,
                                   const char *hca_transport,
                                   struct snap_conf *config,
                                   memp_t *memp,
                                   nvme_emu_mmio_write_cb_t bar_write_ev_handler,
                                   void *dev_ctx,
                                   uint32_t *num_queues)
{
    snap_nvme_emulator_t *snap_emu;
    nvme_emulator_t *dev;
    struct ibv_context *ibctx;
    int pf_id, vf_id;
    int ret;

    snap_emu = calloc(1, sizeof(*snap_emu));
    if (!snap_emu)
        return NULL;

    dev = &snap_emu->emulator;

    snap_emu->sctx = mlnx_snap_get_snap_context(hca_manager);
    if (!snap_emu->sctx) {
        nvmx_error("failed to open snap ctx");
        goto out_free;
    }

    pf_id = snap_config_get_pci_func(config);
    if (pf_id >= snap_emu->sctx->nvme_pfs.max_pfs) {
        nvmx_error("dev 0x%p supports only %d emulated physical functions. \
                   Function id %d is needed",
                   dev, snap_emu->sctx->nvme_pfs.max_pfs, pf_id);
        goto out_dev_close;
    }

    dev->dev_caps.max_reg_size = snap_emu->sctx->nvme_caps.reg_size;
    dev->dev_caps.max_emulated_cq_num = snap_emu->sctx->nvme_caps.max_emulated_nvme_cqs;
    dev->dev_caps.max_emulated_sq_num = snap_emu->sctx->nvme_caps.max_emulated_nvme_sqs;
    dev->dev_caps.max_namespaces = snap_emu->sctx->nvme_caps.max_nvme_namespaces;
    dev->dev_caps.max_nsid = snap_emu->sctx->nvme_caps.max_nvme_nsid;
    dev->dev_caps.max_emulated_pfs = snap_emu->sctx->nvme_pfs.max_pfs;

    vf_id = snap_config_get_pci_vfunc(config);
    if (vf_id >= 0) {
        if (vf_id >= snap_emu->sctx->nvme_pfs.pfs[pf_id].num_vfs) {
            nvmx_error("dev 0x%p supports only %d emulated virtual functions. "
                       "Function id %d is needed",
                       dev, dev->dev_caps.max_emulated_vfs, vf_id);
            goto out_dev_close;
        }
        snap_emu->attr.type = SNAP_NVME_VF;
    } else
        snap_emu->attr.type = SNAP_NVME_PF;
    snap_emu->attr.pf_id = pf_id;
    snap_emu->attr.vf_id = vf_id;

    snap_emu->ctrlc = snap_nvme_create_ctrl_counters(snap_emu->sctx);
    if (!snap_emu->ctrlc) {
        nvmx_error("dev 0x%p can't create ctrl counters with id 0x%x", dev,
                    pf_id);
        goto out_dev_close;
    }
    snap_emu->attr.counter_set_id = snap_emu->ctrlc->id;

    snap_emu->sdev = snap_open_device(snap_emu->sctx, &snap_emu->attr);
    if (!snap_emu->sdev) {
        nvmx_error("dev 0x%p can't create snap device with id 0x%x", dev,
                    pf_id);
        goto out_destroy_ctrlc;
    }

    dev->bar_write_host_cb = bar_write_ev_handler;
    dev->ctrl = dev_ctx;
    dev->config = config;
    dev->vid = snap_emu->sdev->pci->pci_attr.vendor_id;
    dev->ssvid = snap_emu->sdev->pci->pci_attr.subsystem_vendor_id;

    nvme_emu_set_num_queues(&dev->dev_caps, config, num_queues);
    snap_emu->num_queues = *num_queues;
    snap_emu->sqs = calloc(snap_emu->num_queues, sizeof(*snap_emu->sqs));
    if (!snap_emu->sqs) {
        nvmx_error("dev 0x%p failed to allocated SQs array", dev);
        goto out_snap_close;
    }
    snap_emu->sqcs = calloc(snap_emu->num_queues, sizeof(*snap_emu->sqcs));
    if (!snap_emu->sqcs) {
        nvmx_error("dev 0x%p failed to allocated SQs array", dev);
        goto out_free_sqs;
    }

    ret = nvme_bar_init((nvme_bar_read_func_t)nvme_emu_mmio_read, &dev->bar, dev);
    if (ret != NVMX_OK) {
        nvmx_error("dev 0x%p failed to init BAR", dev);
        goto out_free_sqcs;
    }

    ret = nvme_bar_init_modify((nvme_bar_write_func_t)nvme_emu_mmio_write,
                               &dev->bar, dev);
    if (ret != NVMX_OK) {
        nvmx_error("dev 0x%p failed to init modify BAR", dev);
        goto out_free_sqcs;
    }

    ibctx = mlnx_snap_get_ibv_context(hca_transport);
    if (!ibctx) {
        nvmx_error("failed to find ibv context for %s", hca_transport);
        goto out_free_sqcs;
    }

    snap_emu->pd = ibv_alloc_pd(ibctx);
    if (!snap_emu->pd) {
        nvmx_error("failed to alloc pd for %s", hca_transport);
        goto out_free_sqcs;
    }

    // When memory pool is provided - we register a global mr for
    // the whole back buffer
    if (memp) {
        snap_emu->mr = snap_reg_mr(snap_emu->pd, memp_get_buffer(memp),
                memp_get_size(memp));

        if (!snap_emu->mr) {
            nvmx_error("failed to alloc mr for %s", hca_transport);
            goto out_free_pd;
        }
    }

    snap_emu->last_bar_cb = clock();
    snap_emu->is_started = false;
    strncpy(dev->name, SNAP_EMU_NAME, sizeof(dev->name) - 1);
    /* it will be updated during the next bar read */
    dev->dev_caps.max_emulated_vfs = 0;

    if (vf_id < 0) {
        nvmx_info("snap dev 0x%p on PF%d is opened successfully", dev, pf_id);
    } else
        nvmx_info("snap dev 0x%p on PF%d VF%d is opened successfully",
                  dev, pf_id, vf_id);

    return dev;

out_free_pd:
    ibv_dealloc_pd(snap_emu->pd);
out_free_sqcs:
    free(snap_emu->sqcs);
out_free_sqs:
    free(snap_emu->sqs);
out_snap_close:
    snap_close_device(snap_emu->sdev);
out_destroy_ctrlc:
    snap_nvme_destroy_ctrl_counters(snap_emu->ctrlc);
out_dev_close:
    snap_ctx_close(snap_emu->sctx);
out_free:
    free(snap_emu);
    return NULL;
}

void nvme_emu_dev_close(nvme_emulator_t *dev)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    int i;

    if (snap_emu->mr)
        ibv_dereg_mr(snap_emu->mr);

    ibv_dealloc_pd(snap_emu->pd);

    for (i = 1; i < snap_emu->num_queues; i++)
        if (snap_emu->sqcs[i])
            snap_nvme_destroy_sq_counters(snap_emu->sqcs[i]);

    free(snap_emu->sqcs);
    free(snap_emu->sqs);
    snap_close_device(snap_emu->sdev);
    snap_nvme_destroy_ctrl_counters(snap_emu->ctrlc);
    snap_ctx_close(snap_emu->sctx);
    free(snap_emu);
}

bool nvme_emu_dev_is_started(nvme_emulator_t *dev)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);

    return snap_emu->is_started;
}

int nvme_emu_dev_stop(nvme_emulator_t *dev)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    int ret;

    if (!nvme_emu_dev_is_started(dev))
        return NVMX_OK;

    /* If teardown fails (and it can because of FLR) the only way to
     * recover is to init device.
     */
    snap_emu->is_started = false;
    ret = snap_nvme_teardown_device(snap_emu->sdev);
    if (ret) {
        nvmx_error("dev %p failed to teardown nvme device ret=%d", dev, ret);
        return NVMX_ERR;
    }

    return NVMX_OK;
}

int nvme_emu_dev_start(nvme_emulator_t *dev)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    int ret, i;

    if (nvme_emu_dev_is_started(dev))
        return NVMX_OK;

    /*
     * On BF2 we have to close and open emulation manager object
     * in order to clear FLR status. Fortunately this approach
     * is backward compatible with the BF1.
     */
    if (snap_emu->flr_active) {
        snap_emu->prev_enabled = snap_emu->curr_enabled = false;
        nvmx_debug("reopening emulation object");

        // During FLR, on BlueField-1 the SQ counters object is being destroyed
        // so we cannot reuse it
        // nvme_need_tunnel is active for BF-1 only
        if (snap_emu->sctx->mctx.nvme_need_tunnel) {
            nvmx_debug("destroying SQ counters");
            for (i = 1; i < snap_emu->num_queues; i++) {
                if (snap_emu->sqcs[i]) {
                    snap_nvme_destroy_sq_counters(snap_emu->sqcs[i]);
                    snap_emu->sqcs[i] = NULL;
                }
            }
        }
        snap_close_device(snap_emu->sdev);

        snap_emu->sdev = snap_open_device(snap_emu->sctx, &snap_emu->attr);
        if (!snap_emu->sdev) {
            nvmx_error("dev 0x%p can't create snap device with id 0x%x", dev,
                       snap_emu->attr.pf_id);
            return NVMX_ERR;
        }
        snap_emu->flr_active = false;
    }

    ret = snap_nvme_init_device(snap_emu->sdev);
    if (ret) {
        nvmx_error("dev %p failed INIT snap nvme device ret=%d", dev, ret);
        return NVMX_ERR;
    }

    snap_emu->prev_enabled = snap_emu->curr_enabled = true;
    snap_emu->is_started = true;
    return NVMX_OK;
}

void nvme_emu_dev_remove_namespace(nvme_emu_dev_ns_t *ns)
{
    snap_nvme_destroy_namespace(ns->ns);
    free(ns);
}

nvme_emu_dev_ns_t *nvme_emu_dev_add_namespace(nvme_emulator_t *dev,
                                                 uint32_t src_nsid,
                                                 uint32_t dst_nsid,
                                                 uint16_t md_size,
                                                 uint8_t lba_size)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    struct snap_nvme_namespace_attr attr = {};
    nvme_emu_dev_ns_t *ns;

    ns = calloc(1, sizeof(*ns));
    if (!ns)
        return NULL;

    attr.src_nsid = src_nsid;
    attr.dst_nsid = dst_nsid;
    attr.lba_size = lba_size;
    attr.md_size = md_size;
    ns->ns = snap_nvme_create_namespace(snap_emu->sdev, &attr);
    if (!ns->ns) {
        nvmx_error("failed to add namespace (%d -> %d) ms_size %d lba_size %d",
                   src_nsid, dst_nsid, md_size, lba_size);
        goto free_ns;
    }

    ns->dev = dev;

    return ns;

free_ns:
    free(ns);
    return NULL;
}

static void snap_nvme_rx(struct snap_dma_q *dma_q, const void *cmd, uint32_t len,
                         uint32_t imm_data)
{
    snap_nvme_emu_sq_t *emu_sq = (snap_nvme_emu_sq_t *)snap_dma_q_ctx(dma_q);

    nvme_emu_process_sq(emu_sq->ops, emu_sq->db_addr, cmd, len);
}

static int to_snap_q_type(int nvmx_q_type, int *snap_q_type)
{
    switch (nvmx_q_type) {
        case NVME_EMU_QUEUE_SQE:
            *snap_q_type = SNAP_NVME_RAW_MODE;
            return 0;
        case NVME_EMU_QUEUE_NVME2NVMF_LOCAL:
        case NVME_EMU_QUEUE_NVME2NVMF_REMOTE:
            *snap_q_type = SNAP_NVME_TO_NVMF_MODE;
            return 0;
    }

    return -1;
}

static int nvme_emu_modify_sq(snap_nvme_emu_sq_t *emu_sq, struct ibv_qp *qp,
                              enum snap_nvme_sq_state network_state)
{
    struct snap_nvme_sq_attr sq_attr = {};
    int ret;

    nvmx_debug("dev %p attempt to modify SQ %p", emu_sq->emulator, emu_sq);

    sq_attr.qp = qp;
    sq_attr.state = network_state;
    ret = snap_nvme_modify_sq(emu_sq->sq,
                              SNAP_NVME_SQ_MOD_QPN | SNAP_NVME_SQ_MOD_STATE,
                              &sq_attr);
    if (ret) {
        nvmx_error("dev %p failed to modify SQ %p ret=%d",
                   emu_sq->emulator, emu_sq, ret);
        return NVMX_ERR;
    }

    nvmx_info("NVMe SQ 0x%x is linked with QPN 0x%x",
              emu_sq->ops->qid, !qp ? 0 : qp->qp_num);
    return NVMX_OK;
}

static int nvme_emu_create_sq_be(snap_nvme_emu_sq_t *emu_sq, struct ibv_qp *qp)
{
    struct snap_nvme_sq_be_attr attr = {};
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(emu_sq->emulator);

    attr.sq = emu_sq->sq;
    attr.qp = qp;
    emu_sq->sq_be = snap_nvme_create_sq_be(snap_emu->sdev, &attr);
    if (!emu_sq->sq_be)
        return NVMX_ERR;

    nvmx_info("NVMe SQ backend 0x%x is created with QPN 0x%x",
              emu_sq->ops->qid, qp->qp_num);
    return NVMX_OK;
}

static int nvme_emu_destroy_sq_be(snap_nvme_emu_sq_t *emu_sq)
{
    if (!emu_sq->sq_be)
        return NVMX_OK;

    snap_nvme_destroy_sq_be(emu_sq->sq_be);
    emu_sq->sq_be = NULL;
    nvmx_info("NVMe SQ backend 0x%x is destroyed", emu_sq->ops->qid);
    return NVMX_OK;
}

int nvme_emu_attach_sq_qp(nvme_emulator_t *dev, int qid, struct ibv_qp *qp)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    snap_nvme_emu_sq_t *emu_sq = snap_emu->sqs[qid];

    if (snap_nvme_sq_is_fe_only(emu_sq->sq))
        return nvme_emu_create_sq_be(emu_sq, qp);
    else
        return nvme_emu_modify_sq(emu_sq, qp, SNAP_NVME_SQ_STATE_RDY);
}

int nvme_emu_detach_sq_qp(nvme_emulator_t *dev, int qid)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    snap_nvme_emu_sq_t *emu_sq = snap_emu->sqs[qid];

    if (snap_nvme_sq_is_fe_only(emu_sq->sq))
        return nvme_emu_destroy_sq_be(emu_sq);
    else
        return nvme_emu_modify_sq(emu_sq, NULL, SNAP_NVME_SQ_STATE_ERR);
}

static inline int nvme_emu_get_qsize_multiplier(nvme_emulator_t *dev)
{
    int mdts, mpsmin;
    int iov_cnt, multiplier;

    if (snap_env_getenv("NVME_SNAP_TCP_RX_ZCOPY") == 0)
        return 1;

    mdts = snap_config_get_mdts(dev->config);
    mpsmin = nvme_reg_cap_mpsmin(dev->bar.curr.cap);

    /* assume each iov elements is 4k size */
    iov_cnt = ((1 << mdts) * (1 << (mpsmin + 12))) / 4096;
    /* without this multiplier, the size of created QP is 4 times of SQ size */
    multiplier = (iov_cnt + 3) / 4;

    return multiplier;
}

/* ops is an output param */
int nvme_emu_create_sq(nvme_emulator_t *dev,
                       const struct nvme_emu_create_sq_attr *attr,
                       nvme_queue_ops_t *ops)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    struct snap_nvme_device *ndev = (struct snap_nvme_device *)snap_emu->sdev->dd_data;
    struct snap_nvme_sq_attr sq_attr = {};
    struct snap_dma_q_create_attr dma_attr = {};
    struct snap_cross_mkey_attr cm_attr = {};
    snap_nvme_emu_sq_t *emu_sq;
    struct ibv_qp *fw_qp;
    int ret, mode, qid = attr->id;
    struct snap_cross_mkey *snap_cross_mkey;

    ret = to_snap_q_type(ops->q_type, &mode);
    if (ret) {
        nvmx_error("dev 0x%p failed to create SQ %d. type %d not supported",
                   dev, qid, ops->q_type);
        return NVMX_ERR;
    }

    emu_sq = calloc(1, sizeof(*emu_sq));
    if (!emu_sq)
        return NVMX_ERR;

    emu_sq->emulator = dev;
    emu_sq->ops = ops;
    ops->q_impl = emu_sq;
    snap_emu->sqs[qid] = emu_sq;

    if (qid) {
        if (!snap_emu->sqcs[qid])
            snap_emu->sqcs[qid] = snap_nvme_create_sq_counters(snap_emu->sdev);

        if (snap_emu->sqcs[qid])
            sq_attr.counter_set_id = snap_emu->sqcs[qid]->obj->obj_id;
    }

    sq_attr.type = mode;
    sq_attr.id = attr->id;
    sq_attr.queue_depth = attr->size;
    sq_attr.base_addr = attr->base_addr;
    sq_attr.cq = (struct snap_nvme_cq *)attr->cq_ops->q_impl;
    sq_attr.fe_only = attr->fe_only;
    emu_sq->sq = snap_nvme_create_sq(snap_emu->sdev, &sq_attr);
    if (!emu_sq->sq) {
        nvmx_error("dev 0x%p ops 0x%p failed to create SQ %d", dev, ops, qid);
        goto free_sq;
    }

    if (ops->q_type == NVME_EMU_QUEUE_NVME2NVMF_REMOTE) {
        nvmx_info("dev 0x%p ops 0x%p created offloaded SQ %d", dev, ops, qid);
        return NVMX_OK;
    }

    ret = posix_memalign((void **)&emu_sq->bbuf, 4096, SNAP_EMU_MAX_BCOPY);
    if (ret) {
        nvmx_error("failed allocating bbuf for sq %d", qid);
        goto out_destroy_sq;
    }

    emu_sq->mr = ibv_reg_mr(snap_emu->pd, emu_sq->bbuf, SNAP_EMU_MAX_BCOPY,
                            IBV_ACCESS_LOCAL_WRITE |
                            IBV_ACCESS_REMOTE_READ |
                            IBV_ACCESS_REMOTE_WRITE);
    if (!emu_sq->mr) {
        nvmx_error("failed reg mr for bbuf for sq %d", qid);
        goto free_bbuf;
    }

    dma_attr.mode = snap_env_getenv(SNAP_DMA_Q_OPMODE);
    dma_attr.tx_qsize = nvmx_max(attr->size, MINIMAL_SNAP_Q_SIZE * 2) *
                                             nvme_emu_get_qsize_multiplier(dev);
    dma_attr.rx_qsize = attr->size;
    dma_attr.tx_elem_size = 16;
    dma_attr.rx_elem_size = 64;
    dma_attr.uctx = emu_sq;
    dma_attr.rx_cb = snap_nvme_rx;
    dma_attr.iov_enable = snap_env_getenv("NVME_SNAP_TCP_RX_ZCOPY") > 0;
    emu_sq->dma_q = snap_dma_q_create(snap_emu->pd, &dma_attr);
    if (!emu_sq->dma_q)
        goto dereg_mr;

    memset(&sq_attr, 0, sizeof(sq_attr));

    /* Link the SQ with its loopback counterpart by default */
    fw_qp = snap_dma_q_get_fw_qp(emu_sq->dma_q);
    if (!fw_qp) {
        nvmx_error("dev 0x%p ops 0x%p failed to get fw QP %d",
                   dev, ops, qid);
        goto free_dma_q;
    }
    ret = nvme_emu_attach_sq_qp(dev, qid, fw_qp);
    if (ret) {
        nvmx_error("dev 0x%p ops 0x%p failed to attach QP to SQ %d ret=%d",
                   dev, ops, qid, ret);
        goto free_dma_q;
    }

    cm_attr.vtunnel = snap_emu->sdev->mdev.vtunnel;
    cm_attr.dma_rkey = snap_emu->sdev->dma_rkey;
    cm_attr.vhca_id = snap_get_vhca_id(snap_emu->sdev);
    cm_attr.crossed_vhca_mkey = snap_emu->sdev->crossed_vhca_mkey;

    snap_cross_mkey = snap_create_cross_mkey_by_attr(snap_emu->pd, &cm_attr);
    if (!snap_cross_mkey) {
        nvmx_error("dev 0x%p ops 0x%p failed to create cross mkey for SQ %d",
                   dev, ops, qid);
        goto free_dma_q;
    }

    emu_sq->null_mr = ibv_alloc_null_mr(snap_emu->pd);
    if (!emu_sq->null_mr) {
        nvmx_error("dev 0x%p ops 0x%p failed to create null_mr for SQ %d",
                   dev, ops, qid);
        goto out_destroy_cross_mkey;
    }

    emu_sq->mkey = snap_cross_mkey;
    emu_sq->db_addr = ndev->db_base + qid * 8;
    emu_sq->dma_rkey = snap_cross_mkey->mkey;

    nvmx_debug("dev 0x%p ops 0x%p created SQ %d qpn 0x%x dma_rkey 0x%x", dev, ops, qid,
               snap_dma_q_get_fw_qp(emu_sq->dma_q)->qp_num, emu_sq->dma_rkey);
    return NVMX_OK;

out_destroy_cross_mkey:
    snap_destroy_cross_mkey(snap_cross_mkey);
free_dma_q:
    snap_dma_q_destroy(emu_sq->dma_q);
dereg_mr:
    ibv_dereg_mr(emu_sq->mr);
free_bbuf:
    free(emu_sq->bbuf);
out_destroy_sq:
    snap_nvme_destroy_sq(emu_sq->sq);
free_sq:
    free(emu_sq);
    ops->q_impl = NULL;
    return NVMX_ERR;
}

int nvme_emu_delete_sq(nvme_queue_ops_t *ops, int sqid)
{
    snap_nvme_emulator_t *snap_emu;
    snap_nvme_emu_sq_t *emu_sq;
    int ret;

    if (!ops->q_impl)
        goto out;

    emu_sq = ops->q_impl;
    if (emu_sq->dma_q)
        snap_dma_q_flush(emu_sq->dma_q);

    ret = nvme_emu_detach_sq_qp(emu_sq->emulator, sqid);
    if (ret)
        nvmx_debug("ops 0x%p SQ_TYPE %d qid %d - modify failed (%d)",
                   ops, ops->q_type, sqid, ret);

    snap_emu = to_snap_nvme_emulator(emu_sq->emulator);
    snap_emu->sqs[sqid] = NULL;
    ops->q_impl = NULL;

    ret = snap_nvme_destroy_sq(emu_sq->sq);
    if (ret)
        nvmx_debug("ops 0x%p SQ_TYPE %d qid %d - destroy SQ failed (%d)",
                   ops, ops->q_type, sqid, ret);

    if (ops->q_type != NVME_EMU_QUEUE_NVME2NVMF_REMOTE) {
        ret = snap_destroy_cross_mkey(emu_sq->mkey);
        if (ret)
            nvmx_warn("ops 0x%p SQ_TYPE %d qid %d - destroy mkey failed (%d)",
                      ops, ops->q_type, sqid, ret);

        snap_dma_q_destroy(emu_sq->dma_q);
        ibv_dereg_mr(emu_sq->mr);
        free(emu_sq->bbuf);
    }

    if (emu_sq->null_mr)
        ibv_dereg_mr(emu_sq->null_mr);

    nvmx_info("ops 0x%p SQ_TYPE %d qid %d - delete", ops, ops->q_type, sqid);
    free(emu_sq);
out:
    return NVMX_OK;
}

/* ops is an output param */
int nvme_emu_create_cq(nvme_emulator_t *dev, nvme_queue_ops_t *ops,
                       const struct nvme_emu_create_cq_attr *qattr)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    struct snap_nvme_cq_attr cq_attr = {};
    struct snap_nvme_cq *cq;
    const struct snap_nvme_caps *hw_caps = &snap_emu->sctx->nvme_caps;
    int ret, mode;

    ret = to_snap_q_type(ops->q_type, &mode);
    if (ret) {
        nvmx_error("dev 0x%p failed to create CQ %d. type %d not supported",
                   dev, qattr->qid, ops->q_type);
        return NVMX_ERR;
    }

    if (qattr->irq_disable && !hw_caps->cq_interrupt_disabled) {
        nvmx_error("HW does not support CQs without interrupts");
        return NVMX_ERR;
    }

    /* Support RAW mode only */
    cq_attr.type = mode;
    cq_attr.id = qattr->qid;
    cq_attr.msix = qattr->vector;
    cq_attr.queue_depth = qattr->size;
    cq_attr.base_addr = qattr->base_addr;
    cq_attr.cq_period = qattr->cq_period;
    cq_attr.cq_max_count = qattr->cq_max_count;
    cq_attr.interrupt_disable = qattr->irq_disable;
    cq = snap_nvme_create_cq(snap_emu->sdev, &cq_attr);
    if (!cq) {
        nvmx_error("dev 0x%p ops 0x%p failed to create CQ %d", dev, ops,
                   qattr->qid);
        return NVMX_ERR;
    }

    ops->q_impl = cq;
    nvmx_debug("dev 0x%p ops 0x%p created CQ %d", dev, ops, qattr->qid);

    return NVMX_OK;
}

int nvme_emu_delete_cq(nvme_queue_ops_t *ops, int cqid)
{
    struct snap_nvme_cq *cq;
    int ret;

    if (!ops->q_impl)
        goto out;

    cq = ops->q_impl;
    ops->q_impl = NULL;

    ret = snap_nvme_destroy_cq(cq);
    if (ret)
        return NVMX_ERR;

    nvmx_info("ops 0x%p CQ_TYPE %d qid %d - delete", ops, ops->q_type, cqid);
out:
    return NVMX_OK;
}

int nvme_emu_progress_sq(nvme_queue_ops_t *ops)
{
    snap_nvme_emu_sq_t *emu_sq = ops->q_impl;

    if (ops->q_type == NVME_EMU_QUEUE_NVME2NVMF_REMOTE)
        return 0;

    if (emu_sq)
        return snap_dma_q_progress(emu_sq->dma_q);

    return 0;
}

void nvme_emu_progress_mmio(nvme_emulator_t *dev)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    clock_t diff;

    diff = clock() - snap_emu->last_bar_cb;
    if (diff >= NVME_EMU_BAR_CB_TRIGGER) {
        dev->bar_write_host_cb(dev->ctrl);
        snap_emu->last_bar_cb = clock();
    }
}

struct ibv_pd *nvme_emu_get_pd(nvme_emulator_t *emulator)
{
    snap_nvme_emulator_t *snap_emu = container_of(emulator, snap_nvme_emulator_t, emulator);

    return snap_emu->pd;
}

void nvme_emu_interrupt(nvme_queue_ops_t *ops)
{
}

/* read from the host memory */
void nvme_emu_dma_read(nvme_queue_ops_t *ops, void *dst_buf, uint64_t srcaddr,
                              size_t len, uint32_t rmkey)
{
    snap_nvme_emu_sq_t *emu_sq = ops->q_impl;
    struct snap_dma_q *dma_q = emu_sq->dma_q;
    int ret;

    nvmx_assertv_always(len <= SNAP_EMU_MAX_BCOPY, "can not fit %lu into bounce buffer", len);
    ret = snap_dma_q_flush(dma_q);
    nvmx_assertv_always(ret >= 0, "dma read: flush failed");
    ret = snap_dma_q_read(dma_q, emu_sq->bbuf, len, emu_sq->mr->lkey, srcaddr,
                          rmkey ? rmkey : emu_sq->dma_rkey, NULL);
    nvmx_assertv_always(ret >= 0, "dma read failed");
    ret = snap_dma_q_flush(dma_q);
    nvmx_assertv_always(ret >= 0, "dma read: wait failed");
    memcpy(dst_buf, emu_sq->bbuf, len);
}

/* write to the host memory */
void nvme_emu_dma_write(nvme_queue_ops_t *ops, void *src_buf, uint64_t dstaddr,
                               size_t len, uint32_t rmkey)
{
    snap_nvme_emu_sq_t *emu_sq = ops->q_impl;
    struct snap_dma_q *dma_q = emu_sq->dma_q;
    int ret;

    nvmx_assertv_always(len <= SNAP_EMU_MAX_BCOPY, "can not fit %lu into bounce buffer", len);
    ret = snap_dma_q_flush(dma_q);
    memcpy(emu_sq->bbuf, src_buf, len);
    nvmx_assertv_always(ret >= 0, "dma write: flush failed");
    ret = snap_dma_q_write(dma_q, emu_sq->bbuf, len, emu_sq->mr->lkey, dstaddr,
                           rmkey ? rmkey : emu_sq->dma_rkey, NULL);
    nvmx_assertv_always(ret >= 0, "dma write failed");
    ret = snap_dma_q_flush(dma_q);
    nvmx_assertv_always(ret >= 0, "dma write: wait failed");
}

/*
 * Make sure that there is at least @bb_cnt send wqe BBs in the QP.
 * This is a temporary solution until 'pending queue' mechanism
 * is implemented.
 * There are several problems here:
 * - blocking calling thread
 * - additional progress calls from inside progress. With a possibility
 *   of a deep recursion
 * - fairness both with respect to other ops on the same queue and with
 *   respect to other queues on the same thread context
 */
static inline void snap_emu_wait4tx(struct snap_dma_q *dma_q, int bb_cnt)
{
    while (dma_q->tx_available <= bb_cnt)
        snap_dma_q_progress(dma_q);
}

/* send to host memory */
int nvme_emu_send(nvme_queue_ops_t *ops, void *src_buf, size_t len)
{
    snap_nvme_emu_sq_t *emu_sq = ops->q_impl;
    struct snap_dma_q *dma_q = emu_sq->dma_q;
    int ret;

    ret = snap_dma_q_send_completion(dma_q, src_buf, len);
    if (unlikely(ret))
        return ret == -EAGAIN ? NVMX_ERR_NOTREADY : NVMX_ERR;

    return NVMX_OK;
}

void snap_emu_dma_cb(struct snap_dma_completion *comp, int status)
{
    snap_async_req_t *snap_req = container_of(comp, snap_async_req_t, comp);
    enum nvme_emu_status req_status;

    if (unlikely(status != IBV_WC_SUCCESS))
        req_status = NVMX_ERR;
    else
        req_status = NVMX_OK;

    snap_req->base_req.dma_cmd.comp_cb(snap_req->base_req.dma_cmd.comp_cb_user, req_status);
}

int nvme_emu_request_size(nvme_emulator_t *dev)
{
    return sizeof(snap_async_req_t);
}

struct ibv_mr *nvme_emu_sq_init(nvme_queue_ops_t *ops, uint8_t *io_buf, size_t len)
{
    snap_nvme_emu_sq_t *emu_sq = ops->q_impl;
    snap_nvme_emulator_t *snap_emu = (snap_nvme_emulator_t *)emu_sq->emulator;
    struct ibv_mr *mr;

    mr = snap_reg_mr(snap_emu->pd, io_buf ,len);
    if (!mr) {
            nvmx_error("failed to register sq memory");
    }
    return mr;
}

int nvme_emu_request_init(nvme_queue_ops_t *ops, nvme_async_req_t *req, struct ibv_mr *mr)
{
    snap_nvme_emu_sq_t *emu_sq = ops->q_impl;
    snap_nvme_emulator_t *snap_emu = (snap_nvme_emulator_t *)emu_sq->emulator;
    snap_async_req_t *snap_req = (snap_async_req_t *)req;

    /* prp1 also can take iovec, hence +1 */
    req->iov_max = nvmx_max(1 + req->data_len / PRP_PAGE_SIZE(req->ops->page_bits),
            SGL_MAX_DESCRIPTORS);

    req->iov = calloc(req->iov_max, sizeof(struct iovec));
    if (!req->iov)
        return NVMX_ERR;

    // If we have global mr from memory pool - use it
    if (snap_emu->mr)
        snap_req->mr = snap_emu->mr;
    else
        snap_req->mr = mr;

    snap_req->comp.func = snap_emu_dma_cb;
    return NVMX_OK;
}

void nvme_emu_sq_reset(nvme_async_req_t *req)
{
    snap_async_req_t *snap_req = (snap_async_req_t *)req;
    snap_nvme_emu_sq_t *emu_sq = req->ops->q_impl;
    snap_nvme_emulator_t *snap_emu = (snap_nvme_emulator_t *)emu_sq->emulator;

    // Don't release global mr from memory pool
    if (snap_req->mr && snap_req->mr != snap_emu->mr)
        ibv_dereg_mr(snap_req->mr);
}

void nvme_emu_request_reset(nvme_async_req_t *req)
{
    free(req->iov);
    req->iov = NULL;
}

int nvme_emu_request_submit(nvme_async_req_t *req)
{
    snap_nvme_emu_sq_t *emu_sq = req->ops->q_impl;
    struct snap_dma_q *dma_q = emu_sq->dma_q;
    snap_async_req_t *snap_req = (snap_async_req_t *)req;
    struct snap_dma_completion  *comp;
    int ret, i;
    uint32_t rkey;

    snap_emu_wait4tx(dma_q, MINIMAL_SNAP_Q_SIZE);

    if (req->dma_cmd.comp_cb) {
        snap_req->comp.count++;
        comp = &snap_req->comp;
    } else
        comp = NULL;

    rkey = req->dma_cmd.rkey ? req->dma_cmd.rkey : emu_sq->dma_rkey;
    switch (req->dma_cmd.op) {
        case NVME_REQ_DMA_TO_HOST:
            if(!req->dma_cmd.writev2v) {
                ret = snap_dma_q_write(dma_q, req->dma_cmd.laddr,
                                       req->dma_cmd.len, snap_req->mr->lkey,
                                       req->dma_cmd.raddr, rkey, comp);
            } else {
                /* RX ZCOPY:
                 * This flow handles the RX ZCOPY data placement only.
                 * The read of the SGL/PRP list should not be done by this path.
                 */

                for(i = 0; i < req->tcp_rx_zcopy.iov_cnt; i++) {
                    nvmx_debug("prp_submit_writev2v: req->tcp_rx_zcopy.iov[%d].iov_len = %d, req->tcp_rx_zcopy.iov[%d].iov_base = %p", 
                               i, req->tcp_rx_zcopy.iov[i].iov_len, i, (void*) req->tcp_rx_zcopy.iov[i].iov_base);
                    req->tcp_rx_zcopy.lkey[i] = req->tcp_rx_zcopy.sgl[i].lkey;
                }
                snap_req->comp.count = 1;
                comp = &snap_req->comp;

                req->dma_cmd.comp_cb = req->comp_cb;
                req->dma_cmd.comp_cb_user = req->comp_cb_user;

                /* snap_dma_q_writev2v() below need number of `req->iov_cnt` WQE BB cnt */
                snap_emu_wait4tx(dma_q, req->iov_cnt);
                ret = snap_dma_q_writev2v(dma_q,
                                         (uint32_t *) &req->tcp_rx_zcopy.lkey, req->tcp_rx_zcopy.iov, req->tcp_rx_zcopy.iov_cnt,
                                         &rkey, req->iov, req->iov_cnt,
                                         false, true,
                                         comp);
            }
            break;
        case NVME_REQ_DMA_FROM_HOST:
            ret = snap_dma_q_read(dma_q, req->dma_cmd.laddr,
                                  req->dma_cmd.len, snap_req->mr->lkey,
                                  req->dma_cmd.raddr, rkey, comp);
            break;
        default:
            nvmx_error("unsupported dma operation %d", req->dma_cmd.op);
            ret = -1;
    }

    return ret == 0 ? NVMX_INPROGRESS : NVMX_ERR;
}

bool nvme_emu_query_ctrl_counters(nvme_emulator_t *dev, nvme_emu_counters_t *counters)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    struct snap_nvme_ctrl_counters* ctrlc = snap_emu->ctrlc;

    if (ctrlc && snap_nvme_query_ctrl_counters(ctrlc) == 0) {
        counters->data_read = ctrlc->data_read * 512;
        counters->data_write = ctrlc->data_write * 512;
        counters->cmd_read = ctrlc->cmd_read;
        counters->cmd_write = ctrlc->cmd_write;
        return true;
    }

    return false;
}

bool nvme_emu_query_sq_counters(nvme_emulator_t *dev, int qid, nvme_emu_counters_t *counters)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    struct snap_nvme_sq_counters* sqcs = snap_emu->sqcs[qid];

    if (sqcs && snap_nvme_query_sq_counters(sqcs) == 0) {
        counters->data_read = sqcs->data_read * 512;
        counters->data_write = sqcs->data_write * 512;
        counters->cmd_read = sqcs->cmd_read;
        counters->cmd_write = sqcs->cmd_write;
        return true;
    }

    return false;
}

bool nvme_emu_is_pf(nvme_emulator_t *dev)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    return snap_emu->attr.type == SNAP_NVME_PF;
}

int snap_nvme_ctrl_hotunplug(nvme_emulator_t *dev)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    struct snap_device *sdev = snap_emu->sdev;
    struct snap_nvme_device_attr attr = {};
    uint64_t mask = SNAP_NVME_DEV_MOD_HOTPLUG_STATE;

    attr.pci_hotplug_state = MLX5_EMULATION_HOTPLUG_STATE_HOTUNPLUG_PREPARE;
    return snap_nvme_modify_device(sdev, mask, &attr);
}

void nvme_emu_destroy_zcopy_key_table(void *key_table)
{
    zcopy_snap_destroy_key_table(key_table);
}

void *nvme_emu_create_zcopy_key_table(char *name)
{
    return zcopy_snap_create_key_table(name);
}

struct snap_nvme_emu_memzero_comp {
    struct snap_dma_completion comp;
    void *cpl_cb;
    void *cpl_cb_arg;
    memzero_write_comp_cb_t write_comp;
    bool in_use;
};

#define SNAP_NVME_EMU_NUM_COMPS 1000

struct snap_nvme_emu_memzero_ctx {
    memzero_ops_t ops;
    snap_nvme_emu_sq_t *emu_sq;
    struct snap_nvme_emu_memzero_comp comps[SNAP_NVME_EMU_NUM_COMPS];
};

static void snap_emu_dma_memzero_cb(struct snap_dma_completion *comp, int status)
{
    struct snap_nvme_emu_memzero_comp *memzero_comp =
            container_of(comp, struct snap_nvme_emu_memzero_comp, comp);

    memzero_comp->write_comp(memzero_comp->cpl_cb, memzero_comp->cpl_cb_arg,
                             status);
    if (memzero_comp->in_use)
        memzero_comp->in_use = false;
    else
        free(memzero_comp);
}

static int memzero_write(memzero_ops_t *ops, struct iovec *iov, uint32_t iovcnt,
                         void *cpl_cb, void *cpl_cb_arg,
                         memzero_write_comp_cb_t comp)
{
    struct snap_nvme_emu_memzero_ctx *memzero_ctx =
            container_of(ops, struct snap_nvme_emu_memzero_ctx, ops);
    snap_nvme_emu_sq_t *emu_sq = memzero_ctx->emu_sq;
    struct snap_nvme_emu_memzero_comp *memzero_comp = NULL;
    int ret, i;

    // TODO: alloc io_cxt per request and not per queue.
    for (i = 0; i < SNAP_NVME_EMU_NUM_COMPS; i++) {
        if (!memzero_ctx->comps[i].in_use) {
                memzero_comp = &memzero_ctx->comps[i];
                memzero_comp->in_use = true;
                break;
        }
    }

    if (!memzero_comp) {
        memzero_comp = calloc(1, sizeof(*memzero_comp));
        if (!memzero_comp)
            return -ENOMEM;
    }

    memzero_comp->comp.func = snap_emu_dma_memzero_cb;
    memzero_comp->comp.count = 1;
    memzero_comp->cpl_cb = cpl_cb;
    memzero_comp->cpl_cb_arg = cpl_cb_arg;
    memzero_comp->write_comp = comp;

    /* The source iov can be anything when using a null mkey */
    ret = snap_dma_q_writev2v(emu_sq->dma_q, (uint32_t *)&emu_sq->null_mr->lkey,
                              iov, iovcnt, &emu_sq->dma_rkey, iov, iovcnt, true,
                              true, &memzero_comp->comp);

    return ret;
}

static void destroy_ops(memzero_ops_t *ops)
{
    struct snap_nvme_emu_memzero_ctx *memzero_ctx =
            container_of(ops, struct snap_nvme_emu_memzero_ctx, ops);

    free(memzero_ctx);
}

memzero_ops_t *nvme_emu_dma_create_memzero(nvme_emulator_t *dev, int qid)
{
    snap_nvme_emulator_t *snap_emu = to_snap_nvme_emulator(dev);
    snap_nvme_emu_sq_t *emu_sq = snap_emu->sqs[qid];
    struct snap_nvme_emu_memzero_ctx *memzero_ctx;

    memzero_ctx = calloc(1, sizeof(*memzero_ctx));
    if (!memzero_ctx)
        return NULL;

    memzero_ctx->ops.write = memzero_write;
    memzero_ctx->ops.destroy = destroy_ops;
    memzero_ctx->emu_sq = emu_sq;

    return &memzero_ctx->ops;
}

#else
nvme_emulator_t *nvme_emu_open(const char *iface_name,
                                      struct snap_conf *sconfig,
                                      nvme_emu_mmio_write_cb_t bar_write_ev_handler,
                                      void *dev_ctx,
                                      uint32_t *num_queues)
{
    return NULL;
}

void nvme_emu_close(nvme_emulator_t *dev)
{
}

/* read from host bar space */
int nvme_emu_mmio_read(nvme_emulator_t *dev, void *buf,
                              uint32_t bar_addr, unsigned len)
{
    return NVMX_OK;
}

/* write to host bar space */
int nvme_emu_mmio_write(nvme_emulator_t *dev, void *buf,
                               uint32_t bar_addr, unsigned len)
{
    return NVMX_OK;
}

int nvme_emu_start(nvme_emulator_t *dev)
{
    return NVMX_OK;
}

int nvme_emu_stop(nvme_emulator_t *dev)
{
    return NVMX_OK;
}

void nvme_emu_remove_namespace(nvme_emu_dev_ns_t *ns)
{
}

nvme_emu_dev_ns_t *nvme_emu_add_namespace(nvme_emulator_t *dev,
                                                 uint32_t src_nsid,
                                                 uint32_t dst_nsid,
                                                 uint16_t md_size,
                                                 uint8_t lba_size)
{
    return NULL;
}

bool nvme_emu_is_started(nvme_emulator_t *dev)
{
    return false;
}

/* ops is an output param */
int nvme_emu_create_sq(nvme_emulator_t *dev,
                       const struct nvme_emu_create_sq_attr *attr,
                       nvme_queue_ops_t *ops)
{
    return NVMX_ERR;
}

int nvme_emu_delete_sq(nvme_queue_ops_t *ops, int sqid)
{
    return NVMX_ERR;
}

/* ops is an output param */
int nvme_emu_create_cq(nvme_emulator_t *dev, nvme_queue_ops_t *ops,
                       const struct nvme_emu_create_cq_attr *qattr)
{
    return NVMX_ERR;
}

int nvme_emu_delete_cq(nvme_queue_ops_t *ops, int cqid)
{
    return NVMX_ERR;
}

int nvme_emu_attach_sq_qp(nvme_emulator_t *dev, int qid, struct ibv_qp *qp,
                          enum nvme_emu_sq_state network_state)
{
    return NVMX_ERR;
}

int nvme_emu_detach_sq_qp(nvme_emulator_t *dev, int qid)
{
    return NVMX_ERR;
}

int nvme_emu_progress_sq(nvme_queue_ops_t *ops)
{
    return 0;
}

void nvme_emu_progress_mmio(nvme_emulator_t *dev)
{
}

void nvme_emu_interrupt(nvme_queue_ops_t *ops)
{
}

/* read from the host memory */
void nvme_emu_dma_read(nvme_queue_ops_t *ops, void *dst_buf, uint64_t srcaddr,
                              size_t len, uint32_t rmkey)
{
}

/* write to the host memory */
void nvme_emu_dma_write(nvme_queue_ops_t *ops, void *src_buf, uint64_t dstaddr,
                               size_t len, uint32_t rmkey)
{
}

/* send to host memory */
int nvme_emu_send(nvme_queue_ops_t *ops, void *src_buf, size_t len)
{
    return -1;
}

int nvme_emu_request_size(nvme_emulator_t *dev)
{
    return sizeof(nvme_async_req_t);
}

int nvme_emu_request_init(nvme_queue_ops_t *ops, nvme_async_req_t *req, struct ibv_mr *mr)
{
    return NVMX_ERR;
}

struct ibv_mr *nvme_emu_sq_init(nvme_queue_ops_t *ops, uint8_t *io_buf, size_t len)
{
    return NULL;
}

void nvme_emu_sq_reset(nvme_async_req_t *req)
{
}

void nvme_emu_request_reset(nvme_async_req_t *req)
{
}

int nvme_emu_request_submit(nvme_async_req_t *req)
{
    return NVMX_ERR;
}

bool nvme_emu_query_ctrl_counters(nvme_emulator_t *dev, nvme_emu_counters_t *counters)
{
    return false;
}

bool nvme_emu_query_sq_counters(nvme_emulator_t *dev, int qid, nvme_emu_counters_t *counters)
{
    return false;
}

bool nvme_emu_is_pf(nvme_emulator_t *dev)
{
    return false;
}

int snap_nvme_ctrl_hotunplug(nvme_emulator_t *dev)
{
    return 0;
}

void nvme_emu_destroy_zcopy_key_table(void *key_table)
{
    return;
}

void *nvme_emu_create_zcopy_key_table(char *name)
{
    return NULL;
}

memzero_ops_t *nvme_emu_dma_create_memzero(nvme_emulator_t *dev, int qid)
{
    return NULL;
}
#endif
