#include <stdlib.h>
#include <unistd.h>
#include <infiniband/verbs.h>
#include <spdk/bdev_module.h>
#include <spdk/env.h>
#include <snap_virtio_blk_ctrl.h>
#include "snap_env.h"
#include <linux/virtio_blk.h>
#include <snap_blk_ops.h>

#include "config.h"
#include "compiler.h"
#include "virtio_blk_controller.h"
#include "nvme_emu.h"
#include "nvme_emu_log.h"
#include "virtio_blk_regs.h"
#include "mlnx_snap_pci_manager.h"
#include "memp.h"
#include "zcopy_snap.h"
#include "spdk_ext_io/spdk_ext_io.h"

#define VIRTIO_SPDK_BUF_ALIGN_MIN 4096
#define VIRTIO_NONE_BDEV_NAME "None"
#define VIRTIO_BLK_CTRL_ID_DEFAULT "00000000000000000000"
#define VIRTIO_BLK_CTRL_BDEV_NAME_MAXLEN 16

#define VIRTIO_BLK_SNAP_ZCOPY "VIRTIO_BLK_SNAP_ZCOPY"
SNAP_ENV_REG_ENV_VARIABLE(VIRTIO_BLK_SNAP_ZCOPY, 1);

static void virtio_blk_ctrl_bdev_close(void *arg);

static struct snap_context *
virtio_blk_ctrl_find_snap_context(const char *emu_manager, int pf_id)
{
    struct snap_context *ctx, *found = NULL;
    struct ibv_device **ibv_list;
    struct snap_pci **pf_list;
    int ibv_list_sz, pf_list_sz;
    int i, j;

    ibv_list = ibv_get_device_list(&ibv_list_sz);
    if (!ibv_list)
        return NULL;

    for (i = 0; i < ibv_list_sz; i++) {
        if (strncmp(ibv_get_device_name(ibv_list[i]), emu_manager, 16))
            continue;

        ctx = mlnx_snap_get_snap_context(ibv_get_device_name(ibv_list[i]));
        if (!ctx)
            continue;

        if (!(ctx->emulation_caps & SNAP_VIRTIO_BLK))
            continue;

        pf_list = calloc(ctx->virtio_blk_pfs.max_pfs, sizeof(*pf_list));
        if (!pf_list)
            continue;

        pf_list_sz = snap_get_pf_list(ctx, SNAP_VIRTIO_BLK, pf_list);
        for (j = 0; j < pf_list_sz; j++) {
            if (pf_list[j]->plugged && pf_list[j]->id == pf_id) {
                found = ctx;
                break;
            }
        }
        free(pf_list);

        if (found)
            break;
    }

    ibv_free_device_list(ibv_list);
    return found;
}

static inline struct virtio_spdk_bdev_io_ctx*
virtio_blk_ctrl_get_io_context(struct virtio_blk_ctrl *ctrl, uint32_t thread_id)
{
    struct virtio_spdk_bdev_ctx *bctx = &ctrl->bctx;
    struct virtio_spdk_bdev_io_ctx *io_ctx;

    if (unlikely(!bctx->io_channels))
        return NULL;

    io_ctx = &bctx->io_channels[thread_id];
    if (unlikely(!io_ctx->spdk_channel)) {
        void *key_table = NULL;

        io_ctx->spdk_channel = spdk_bdev_get_io_channel(bctx->desc);
        io_ctx->spdk_thread = spdk_get_thread();

        if (virtio_blk_ctrl_is_zcopy(ctrl)) {
            char name[32];

            if (ctrl->vf_id < 0) {
                snprintf(name, sizeof(name), "pd-mkey-pf%d-tid%d",
                        ctrl->pf_id, thread_id);
            } else {
                snprintf(name, sizeof(name), "pd-mkey-pf%dvf%d-tid%d",
                        ctrl->pf_id, ctrl->vf_id, thread_id);
            }

            key_table = zcopy_snap_create_key_table(name);
        }

        spdk_ext_io_context_init(&io_ctx->ext_io_ctx, ctrl->sctrl->common.sdev,
                                 key_table, NULL, NULL);
    }
    return io_ctx;
}

static void virtio_blk_ctrl_spdk_op_done(struct spdk_bdev_io *bdev_io,
                                         bool success, void *cb_arg)
{
    struct snap_bdev_io_done_ctx *done_ctx = cb_arg;

    done_ctx->cb(success ? SNAP_BDEV_OP_SUCCESS : SNAP_BDEV_OP_IO_ERROR,
                 done_ctx->user_arg);
    spdk_bdev_free_io(bdev_io);
}

static int virtio_blk_ctrl_backend_read(void *ctx,
                                        void *buf,
                                        uint64_t offset,
                                        uint64_t len,
                                        struct snap_bdev_io_done_ctx *done_ctx,
                                        int thread_id)
{
    struct virtio_blk_ctrl *ctrl = ctx;
    struct virtio_spdk_bdev_io_ctx *io_ctx;

    if (unlikely(ctrl->bdev_type == VIRTIO_BLK_CTRL_BDEV_NONE))
        return -ENODEV;

    io_ctx = virtio_blk_ctrl_get_io_context(ctrl, thread_id);
    if (!io_ctx->spdk_channel)
        return -EINVAL;

    return spdk_bdev_read(ctrl->bctx.desc,
                           io_ctx->spdk_channel,
                           buf,
                           offset,
                           len,
                           virtio_blk_ctrl_spdk_op_done,
                           done_ctx);
}

static int virtio_blk_ctrl_backend_write(void *ctx,
                                         void *buf,
                                         uint64_t offset,
                                         uint64_t len,
                                         struct snap_bdev_io_done_ctx *done_ctx,
                                         int thread_id)
{
    struct virtio_blk_ctrl *ctrl = ctx;
    struct virtio_spdk_bdev_io_ctx *io_ctx;

    if (unlikely(ctrl->bdev_type == VIRTIO_BLK_CTRL_BDEV_NONE))
        return -ENODEV;

    io_ctx = virtio_blk_ctrl_get_io_context(ctrl, thread_id);
    if (!io_ctx->spdk_channel)
        return -EINVAL;

    return spdk_bdev_write(ctrl->bctx.desc,
                            io_ctx->spdk_channel,
                            buf,
                            offset,
                            len,
                            virtio_blk_ctrl_spdk_op_done,
                            done_ctx);
}

#ifdef HAVE_SPDK_EXT_IO
static int virtio_blk_ctrl_backend_readv_blocks(void *ctx,
                                        struct iovec *iov, int iovcnt,
                                        uint64_t offset_blocks,
                                        uint64_t num_blocks,
                                        struct snap_bdev_io_done_ctx *done_ctx,
                                        int thread_id)
{
    struct virtio_blk_ctrl *ctrl = ctx;
    struct virtio_spdk_bdev_io_ctx *io_ctx;

    if (unlikely(ctrl->bdev_type == VIRTIO_BLK_CTRL_BDEV_NONE))
        return -ENODEV;

    io_ctx = virtio_blk_ctrl_get_io_context(ctrl, thread_id);
    if (!io_ctx->spdk_channel)
        return -EINVAL;

    return spdk_bdev_readv_blocks_ext(ctrl->bctx.desc,
                           io_ctx->spdk_channel,
                           iov,
                           iovcnt,
                           offset_blocks,
                           num_blocks,
                           virtio_blk_ctrl_spdk_op_done,
                           done_ctx,
                           &io_ctx->ext_io_ctx.ext_io_opts);
}

static int virtio_blk_ctrl_backend_writev_blocks(void *ctx,
                                         struct iovec *iov, int iovcnt,
                                         uint64_t offset_blocks,
                                         uint64_t num_blocks,
                                         struct snap_bdev_io_done_ctx *done_ctx,
                                         int thread_id)
{
    struct virtio_blk_ctrl *ctrl = ctx;
    struct virtio_spdk_bdev_io_ctx *io_ctx;

    if (unlikely(ctrl->bdev_type == VIRTIO_BLK_CTRL_BDEV_NONE))
        return -ENODEV;

    io_ctx = virtio_blk_ctrl_get_io_context(ctrl, thread_id);
    if (!io_ctx->spdk_channel)
        return -EINVAL;

    return spdk_bdev_writev_blocks_ext(ctrl->bctx.desc,
                            io_ctx->spdk_channel,
                            iov,
                            iovcnt,
                            offset_blocks,
                            num_blocks,
                            virtio_blk_ctrl_spdk_op_done,
                            done_ctx,
                            &io_ctx->ext_io_ctx.ext_io_opts);
}

static bool virtio_blk_ctrl_backend_zcopy_validate_params(void *ctx,
        struct iovec *iov, size_t iov_cnt, uint64_t offset, uint64_t len)
{
    struct virtio_blk_ctrl *ctrl = ctx;
    struct spdk_bdev *bdev = ctrl->bctx.bdev;
    uint32_t block_size = spdk_bdev_get_block_size(bdev);

    if (!spdk_ext_io_validate_iov(bdev, iov, iov_cnt, VIRTIO_SPDK_BUF_ALIGN_MIN))
        return false;

    if (unlikely(offset & (block_size - 1))) {
        nvmx_debug("ZCOPY_FAIL: offset %llu is not aligned to bdev block size %u\n",
                   offset, block_size);
        return false;
    }

    if (unlikely(len & (block_size - 1))) {
        nvmx_debug("ZCOPY_FAIL: len %llu is not aligned to bdev block size %u\n",
                   len, block_size);
        return false;
    }

    return true;
}
#else
static int virtio_blk_ctrl_backend_readv_blocks(void *ctx,
                                        struct iovec *iov, int iovcnt,
                                        uint64_t offset_blocks,
                                        uint64_t num_blocks,
                                        struct snap_bdev_io_done_ctx *done_ctx,
                                        int thread_id)
{
    return -1;
}

static int virtio_blk_ctrl_backend_writev_blocks(void *ctx,
                                         struct iovec *iov, int iovcnt,
                                         uint64_t offset_blocks,
                                         uint64_t num_blocks,
                                         struct snap_bdev_io_done_ctx *done_ctx,
                                         int thread_id)
{
    return -1;
}

static bool virtio_blk_ctrl_backend_zcopy_validate_params(void *ctx,
        struct iovec *iov, size_t iov_cnt, uint64_t offset, uint64_t len)
{
    return false;
}
#endif

static int virtio_blk_ctrl_backend_flush(void *ctx,
                                         uint64_t offset,
                                         uint64_t len,
                                         struct snap_bdev_io_done_ctx *done_ctx,
                                         int thread_id)
{
    struct virtio_blk_ctrl *ctrl = ctx;
    struct virtio_spdk_bdev_io_ctx *io_ctx;

    if (unlikely(ctrl->bdev_type == VIRTIO_BLK_CTRL_BDEV_NONE))
        return -ENODEV;

    io_ctx = virtio_blk_ctrl_get_io_context(ctrl, thread_id);
    if (!io_ctx->spdk_channel)
        return -EINVAL;

    return spdk_bdev_flush(ctrl->bctx.desc,
                           io_ctx->spdk_channel,
                           offset,
                           len,
                           virtio_blk_ctrl_spdk_op_done,
                           done_ctx);
}

static int virtio_blk_ctrl_backend_write_zeroes(void *ctx,
                                                uint64_t offset,
                                                uint64_t len,
                                                struct snap_bdev_io_done_ctx *done_ctx,
                                                int thread_id)
{
    done_ctx->cb(SNAP_BDEV_OP_IO_ERROR, done_ctx->user_arg);
    return 0;
}

static int virtio_blk_ctrl_backend_discard(void *ctx,
                                           uint64_t offset,
                                           uint64_t len,
                                           struct snap_bdev_io_done_ctx *done_ctx,
                                           int thread_id)
{
    done_ctx->cb(SNAP_BDEV_OP_IO_ERROR, done_ctx->user_arg);
    return 0;
}

static void *virtio_blk_ctrl_backend_dma_malloc(size_t size)
{
    return spdk_dma_malloc(size, VIRTIO_SPDK_BUF_ALIGN_MIN, NULL);
}

static void virtio_blk_ctrl_backend_dma_free(void *buf)
{
    spdk_dma_free(buf);
}

static uint64_t virtio_blk_ctrl_backend_get_num_blocks(void *ctx)
{
    struct virtio_blk_ctrl *ctrl = ctx;

    if (unlikely(ctrl->bdev_type == VIRTIO_BLK_CTRL_BDEV_NONE))
        return 0;

    return spdk_bdev_get_num_blocks(ctrl->bctx.bdev);
}

static uint32_t virtio_blk_ctrl_backend_get_block_size(void *ctx)
{
    struct virtio_blk_ctrl *ctrl = ctx;

    /*
     * On NONE bdev there is no real logic in block size, so
     * just return a reasonable value (better not to return 0
     * to avoid crashes elsewhere)
     */
    if (unlikely(ctrl->bdev_type == VIRTIO_BLK_CTRL_BDEV_NONE))
        return 4096;

    return spdk_bdev_get_block_size(ctrl->bctx.bdev);
}

static const char *virtio_blk_ctrl_backend_get_bdev_name(void *ctx)
{
    struct virtio_blk_ctrl *ctrl = ctx;

    return ctrl->serial;
}

static bool virtio_blk_ctrl_backend_is_zcopy(void *ctx)
{
    struct virtio_blk_ctrl *ctrl = ctx;

    return ctrl->zcopy && ctrl->bdev_type != VIRTIO_BLK_CTRL_BDEV_NONE &&
            ctrl->bctx.zcopy;
}

bool virtio_blk_ctrl_is_zcopy(struct virtio_blk_ctrl *ctrl)
{
    return virtio_blk_ctrl_backend_is_zcopy(ctrl);
}

static void virtio_blk_ctrl_backend_mem_ready(void *data, void *user)
{
    struct snap_blk_mempool_ctx *dma_pool_ctx = user;
    struct virtio_blk_ctrl *ctrl = dma_pool_ctx->ctx;
    dma_pool_ctx->callback(data, ctrl->mr, dma_pool_ctx->user);
}

static int virtio_blk_ctrl_backend_dma_pool_malloc(size_t size,
                            struct snap_blk_mempool_ctx *dma_pool_ctx)
{
    void *p;
    struct virtio_blk_ctrl *ctrl = dma_pool_ctx->ctx;

    if (ctrl->memp) {
        p = memp_alloc(ctrl->memp, size, dma_pool_ctx,
                                    virtio_blk_ctrl_backend_mem_ready,
                                    &dma_pool_ctx->tag, dma_pool_ctx->thread_id);
        if (!p && errno)
            return -1;

        if (p)
            virtio_blk_ctrl_backend_mem_ready(p, dma_pool_ctx);

        return 0;
    }
    else
        return -1;
}

static void virtio_blk_ctrl_backend_dma_pool_cancel(
                            struct snap_blk_mempool_ctx *dma_pool_ctx)
{
    struct virtio_blk_ctrl *ctrl = dma_pool_ctx->ctx;
    if (ctrl->memp)
        memp_cancel(ctrl->memp, dma_pool_ctx->tag, 0);
}

static void virtio_blk_ctrl_backend_dma_pool_free(
                    struct snap_blk_mempool_ctx *dma_pool_ctx, void *p)
{
    struct virtio_blk_ctrl *ctrl = dma_pool_ctx->ctx;
    if (ctrl->memp)
        memp_free(ctrl->memp, p);
}

static bool virtio_blk_ctrl_backend_dma_pool_enabled(void *ctx)
{
    struct virtio_blk_ctrl *ctrl = ctx;
    return !!ctrl->memp;
}

bool virtio_blk_ctrl_dma_pool_enabled(struct virtio_blk_ctrl *ctrl)
{
    return virtio_blk_ctrl_backend_dma_pool_enabled(ctrl);
}

static struct snap_bdev_ops virtio_blk_ctrl_backend_ops = {
    .read = virtio_blk_ctrl_backend_read,
    .write = virtio_blk_ctrl_backend_write,
    .readv_blocks = virtio_blk_ctrl_backend_readv_blocks,
    .writev_blocks = virtio_blk_ctrl_backend_writev_blocks,
    .flush = virtio_blk_ctrl_backend_flush,
    .write_zeroes = virtio_blk_ctrl_backend_write_zeroes,
    .discard = virtio_blk_ctrl_backend_discard,
    .dma_malloc = virtio_blk_ctrl_backend_dma_malloc,
    .dma_free = virtio_blk_ctrl_backend_dma_free,
    .get_num_blocks = virtio_blk_ctrl_backend_get_num_blocks,
    .get_block_size = virtio_blk_ctrl_backend_get_block_size,
    .get_bdev_name = virtio_blk_ctrl_backend_get_bdev_name,
    .is_zcopy = virtio_blk_ctrl_backend_is_zcopy,
    .zcopy_validate_params = virtio_blk_ctrl_backend_zcopy_validate_params,
    .dma_pool_malloc = virtio_blk_ctrl_backend_dma_pool_malloc,
    .dma_pool_cancel = virtio_blk_ctrl_backend_dma_pool_cancel,
    .dma_pool_free = virtio_blk_ctrl_backend_dma_pool_free,
    .dma_pool_enabled = virtio_blk_ctrl_backend_dma_pool_enabled,
};

void virtio_blk_ctrl_progress(void *arg)
{
    struct virtio_blk_ctrl *ctrl = arg;
    struct snap_virtio_ctrl sctrl = ctrl->sctrl->common;
    time_t cur_time;

    snap_virtio_blk_ctrl_progress(ctrl->sctrl);

    if (sctrl.bar_curr->pci_hotplug_state == MLX5_EMULATION_HOTPLUG_STATE_POWER_OFF &&
        sctrl.bar_prev->pci_hotplug_state != MLX5_EMULATION_HOTPLUG_STATE_POWER_OFF) {
        ctrl->hotunplug_device_cb(ctrl->hotunplug_device_ctx);
    } else if (sctrl.bar_curr->pci_hotplug_state == MLX5_EMULATION_HOTPLUG_STATE_HOTUNPLUG_PREPARE) {
        if (time(&cur_time) == (time_t)(-1)) {
            nvmx_error("Failed to get system time");
            return;
        }

        if ((ctrl->hotunplug_timer_expire > 0 && ctrl->hotunplug_timer_expire <= cur_time) ||
             ctrl->hotunplug_timer_expire == -1) {
            ctrl->hotunplug_timer_expire = 0;
            if (ctrl->force_hotunplug_device)
                ctrl->hotunplug_device_cb(ctrl->hotunplug_device_ctx);
            else if (sctrl.bar_curr->status & VIRTIO_CONFIG_S_DRIVER_OK)
                ctrl->hotunplug_timeout_cb(ctrl->hotunplug_device_ctx);
            else
                ctrl->hotunplug_device_cb(ctrl->hotunplug_device_ctx);
        }
    }
}

#ifndef HAVE_SPDK_POLLER_BUSY
enum spdk_thread_poller_rc { SPDK_POLLER_IDLE , SPDK_POLLER_BUSY };
#endif

int virtio_blk_ctrl_progress_all_io(void *arg)
{
    struct virtio_blk_ctrl *ctrl = arg;

    return snap_virtio_blk_ctrl_io_progress(ctrl->sctrl) ? SPDK_POLLER_BUSY : SPDK_POLLER_IDLE;
}

int virtio_blk_ctrl_progress_io(void *arg, int thread_id)
{
    struct virtio_blk_ctrl *ctrl = arg;

    return snap_virtio_blk_ctrl_io_progress_thread(ctrl->sctrl, thread_id) ? SPDK_POLLER_BUSY : SPDK_POLLER_IDLE;
}

void virtio_blk_ctrl_suspend(void *arg)
{
    struct virtio_blk_ctrl *ctrl = arg;

    snap_virtio_ctrl_suspend(&ctrl->sctrl->common);
}

bool virtio_blk_ctrl_is_suspended(void *arg)
{
    struct virtio_blk_ctrl *ctrl = arg;

    if (!ctrl->sctrl)
        return true;

    return snap_virtio_ctrl_is_suspended(&ctrl->sctrl->common) ||
           snap_virtio_ctrl_is_stopped(&ctrl->sctrl->common);
}

int virtio_blk_ctrl_resume(void *arg)
{
    struct virtio_blk_ctrl *ctrl = arg;

    return snap_virtio_ctrl_resume(&ctrl->sctrl->common);
}

#ifdef HAVE_SPDK_BDEV_OPEN_EXT
static void virtio_blk_ctrl_bdev_resize(struct virtio_blk_ctrl *ctrl,
                                        struct spdk_bdev *bdev)
{
    struct snap_virtio_blk_registers regs = {};
    uint32_t blk_size;
    uint64_t num_blocks;

    num_blocks = spdk_bdev_get_num_blocks(bdev);
    blk_size = spdk_bdev_get_block_size(bdev);
    regs.capacity = (num_blocks * blk_size) / 512;

    nvmx_info("New size detected for bdev %s: %lu bytes",
              spdk_bdev_get_name(bdev), num_blocks * blk_size);
    if (snap_virtio_blk_ctrl_bar_setup(ctrl->sctrl, &regs,
                                       SNAP_VIRTIO_MOD_DEV_CFG))
        nvmx_warn("Failed to resize bdev %s", spdk_bdev_get_name(bdev));
}
#endif

static void virtio_blk_ctrl_bdev_close_end(void *arg)
{
    struct virtio_blk_ctrl *ctrl = arg;
    bool free_ctrl = false;

    pthread_mutex_lock(&ctrl->bdev_lock);
    ctrl->bctx.num_io_channels--;

    /* For the last IO channel, close bdev */
    if (!ctrl->bctx.num_io_channels) {
        virtio_blk_ctrl_bdev_close(ctrl);
        free_ctrl = true;
    }
    pthread_mutex_unlock(&ctrl->bdev_lock);

    if (free_ctrl && ctrl->bdev_close_cb)
        ctrl->bdev_close_cb(ctrl);
}

static void virtio_blk_ctrl_bdev_put_channel(void *arg)
{
    struct virtio_spdk_bdev_io_ctx *io_ctx = arg;

    if (io_ctx->spdk_channel) {
        spdk_put_io_channel(io_ctx->spdk_channel);
        io_ctx->spdk_channel = NULL;
        io_ctx->spdk_thread = NULL;
    }

    spdk_ext_io_context_clear(&io_ctx->ext_io_ctx);

#ifdef HAVE_SPDK_EXT_IO
    if (virtio_blk_ctrl_is_zcopy(io_ctx->ctrl)) {
        zcopy_snap_destroy_key_table(io_ctx->ext_io_ctx.zcopy_ctx.key_table);
        io_ctx->ext_io_ctx.zcopy_ctx.key_table = NULL;
    }
#endif

    if (io_ctx->caller_thread && io_ctx->caller_cb_fn)
        spdk_thread_send_msg(io_ctx->caller_thread,
                             io_ctx->caller_cb_fn, io_ctx->ctrl);
}

static void virtio_blk_ctrl_flr_progress(struct virtio_blk_ctrl *ctrl);
static void virtio_blk_ctrl_flr_state_change(struct virtio_blk_ctrl *ctrl, enum virtio_blk_ctrl_flr_state state);


static void virtio_blk_ctrl_io_channels_cleared(void *arg)
{
    struct virtio_blk_ctrl *ctrl = arg;

    ctrl->bctx.num_io_channels--;

    /* For the last IO channel, close bdev */
    if (!ctrl->bctx.num_io_channels) {
        nvmx_debug("IO channels are all cleared");
        virtio_blk_ctrl_flr_state_change(ctrl, VIRTIO_BLK_CTRL_FLR_END);
        virtio_blk_ctrl_flr_progress(ctrl);
    }
}

static void virtio_blk_ctrl_io_channels_clear(struct virtio_blk_ctrl *ctrl,
                                              spdk_msg_fn fn,
                                              struct spdk_thread *thread)
{
    int i;
    struct virtio_spdk_bdev_io_ctx *io_ctx;
    size_t num_io_channels = ctrl->bctx.num_io_channels;

    nvmx_debug("Ctrl %p: Clearing IO channels (and closing cross-mkeys)", ctrl);

    for (i = 0; i < num_io_channels; i++) {
        io_ctx = &ctrl->bctx.io_channels[i];

        if (io_ctx->spdk_thread) {
            io_ctx->caller_thread = thread;
            io_ctx->caller_cb_fn = fn;
            spdk_thread_send_msg(io_ctx->spdk_thread,
                                 virtio_blk_ctrl_bdev_put_channel, io_ctx);
        } else {
           /*
            * TODO: we only need to decrease num_io_channels here in case
            * of full controller closure. Currently there are only 2 flows
            * which reach here - and only controller closure provides
            * thread & fn. Need to rearrange code better.
            */
            // Channel wasn't opened, no need to close it
            if (thread && fn)
                ctrl->bctx.num_io_channels--;
        }
    }
}

static void virtio_blk_ctrl_bdev_close_begin(void *arg)
{
    struct virtio_blk_ctrl *ctrl = arg;
    bool run_close_cb = false;

    pthread_mutex_lock(&ctrl->bdev_lock);
    if (ctrl->bdev_type == VIRTIO_BLK_CTRL_BDEV_NONE) {
        run_close_cb = true;
        goto out;
    }

    virtio_blk_ctrl_io_channels_clear(ctrl, virtio_blk_ctrl_bdev_close_end,
                                      spdk_get_thread());

    // If all IO channels wasn't opened, close bdev immediately
    if (ctrl->bctx.num_io_channels == 0) {
        run_close_cb = true;
        virtio_blk_ctrl_bdev_close(ctrl);
    }
out:
    pthread_mutex_unlock(&ctrl->bdev_lock);

    if (ctrl->bdev_close_cb && run_close_cb)
        ctrl->bdev_close_cb(ctrl);
}

static void virtio_blk_ctrl_bdev_detach_begin(struct virtio_blk_ctrl *ctrl);

#ifdef HAVE_SPDK_BDEV_OPEN_EXT
static void virtio_spdk_bdev_handle_event(enum spdk_bdev_event_type type,
                                          struct spdk_bdev *bdev,
                                          void *event_ctx)
{
    struct virtio_blk_ctrl *ctrl = event_ctx;

    switch (type) {
    case SPDK_BDEV_EVENT_REMOVE:
        nvmx_info("closing SPDK bdev %s\n", spdk_bdev_get_name(bdev));
        virtio_blk_ctrl_bdev_detach_begin(ctrl);
        break;
    case SPDK_BDEV_EVENT_RESIZE:
        nvmx_info("resizing SPDK bdev %s\n", spdk_bdev_get_name(bdev));
        virtio_blk_ctrl_bdev_resize(ctrl, bdev);
        break;
    default:
        nvmx_error("Unsupported bdev event: type %d\n", type);
        return;
    }
}

#else
static void virtio_spdk_bdev_close_cb(void *remove_ctx)
{
    struct virtio_blk_ctrl *ctrl = remove_ctx;

    virtio_blk_ctrl_bdev_close_begin(ctrl);
}

#endif

static void virtio_blk_ctrl_bdev_close_spdk(struct virtio_blk_ctrl *ctrl)
{
    struct virtio_spdk_bdev_ctx *bctx = &ctrl->bctx;

    spdk_bdev_close(bctx->desc);
    bctx->desc = NULL;
    free(bctx->io_channels);
    bctx->io_channels = NULL;

    if (!virtio_blk_ctrl_is_suspended(ctrl))
        nvmx_warn("BDEV was removed while host driver is still up");
}

static int virtio_blk_ctrl_bdev_open_spdk(struct virtio_blk_ctrl *ctrl,
                                          const char *bdev_name,
                                          size_t num_io_channels)
{
    struct virtio_spdk_bdev_ctx *bctx = &ctrl->bctx;
    int ret;
    int i;

    bctx->io_channels = calloc(snap_max(1, num_io_channels),
                               sizeof(*bctx->io_channels));
    if (!bctx->io_channels) {
        nvmx_error("failed to allocate io channels");
        ret = -ENOMEM;
        goto err;
    }
    bctx->num_io_channels = num_io_channels;
    for (i = 0; i < num_io_channels; i++)
        bctx->io_channels[i].ctrl = ctrl;

#ifdef HAVE_SPDK_BDEV_OPEN_EXT
    ret = spdk_bdev_open_ext(bdev_name,
                             true,
                             virtio_spdk_bdev_handle_event,
                             ctrl,
                             &bctx->desc);
#else
    ret = spdk_bdev_open(spdk_bdev_get_by_name(bdev_name),
                         true,
                         virtio_spdk_bdev_close_cb,
                         ctrl,
                         &bctx->desc);
#endif
    if (ret) {
        nvmx_error("failed to open bdev %s err %d", bdev_name, ret);
        goto free_io_channels;
    }

    bctx->bdev = spdk_bdev_desc_get_bdev(bctx->desc);
    if (!bctx->bdev) {
        nvmx_error("failed to get bdev from bdev_desc");
        goto close_bdev;
    }

    bctx->zcopy = spdk_ext_io_zcopy_status(bctx->bdev) == 0;

    return 0;

close_bdev:
    spdk_bdev_close(bctx->desc);
free_io_channels:
    free(bctx->io_channels);
    bctx->io_channels = NULL;
    bctx->num_io_channels = 0;
err:
    return ret;
}

static int virtio_blk_ctrl_num_vfs_changed(void *arg, uint16_t new_numvfs)
{
    struct virtio_blk_ctrl *ctrl = arg;
    struct snap_virtio_blk_ctrl *sctrl = ctrl->sctrl;

    nvmx_debug("num_vfs_changed event was detected");

    if (ctrl->vf_change_cb)
        ctrl->vf_change_cb(ctrl->vf_change_cb_arg, ctrl->pf_id, new_numvfs);

    /* Create or destroy array for VFs controllers on this PF */
    if (new_numvfs) {
           sctrl->vfs_ctrl = calloc(new_numvfs, sizeof(struct snap_virtio_blk_ctrl*));
        if (!sctrl->vfs_ctrl)
                return -ENOMEM;
    } else if (sctrl->vfs_ctrl) {
            free(sctrl->vfs_ctrl);
            sctrl->vfs_ctrl = NULL;
    }

    return 0;
}

static void virtio_blk_ctrl_bdev_reset_done(struct spdk_bdev_io *bdev_io, bool success, void *cb_arg)
{
    struct virtio_blk_ctrl *ctrl = cb_arg;

    if (!success) {
        nvmx_error("Failed to execute reset for bdev %s", spdk_bdev_get_name(bdev_io->bdev));
        virtio_blk_ctrl_flr_state_change(ctrl, VIRTIO_BLK_CTRL_FLR_END);
        return;
    } else {
        nvmx_info("Finished resetting bdev %s successfully", spdk_bdev_get_name(bdev_io->bdev));
        virtio_blk_ctrl_flr_state_change(ctrl, VIRTIO_BLK_CTRL_FLR_IO_CHANNEL_CLEAR);
    }

    virtio_blk_ctrl_flr_progress(ctrl);
}

struct virtio_blk_ctrl_flr_context {
    struct virtio_blk_ctrl *ctrl;
    struct spdk_io_channel *io_channel;
};

static struct virtio_blk_ctrl_flr_context *virtio_blk_ctrl_flr_start(struct virtio_blk_ctrl *ctrl)
{
    struct virtio_blk_ctrl_flr_context *context;
    struct spdk_io_channel *io_channel;

    context = calloc(1, sizeof(*context));
    if (!context) {
        nvmx_error("Failed to allocate FLR context");
        return NULL;
    }

    io_channel = spdk_bdev_get_io_channel(ctrl->bctx.desc);
    if (!io_channel) {
        nvmx_error("Failed to get IO channel");
        free(context);
        return NULL;
    }

    context->ctrl = ctrl;
    context->io_channel = io_channel;
    return context;
}

static void virtio_blk_ctrl_flr_end(struct virtio_blk_ctrl_flr_context *flr_context)
{
    spdk_put_io_channel(flr_context->io_channel);
    free(flr_context);
}

#ifdef DEBUG_ENABLED
static const char *virtio_blk_ctrl_flr_state_to_string(enum virtio_blk_ctrl_flr_state state)
{
    switch (state) {
    case VIRTIO_BLK_CTRL_FLR_IDLE:
        return "VIRTIO_BLK_CTRL_FLR_IDLE";
    case VIRTIO_BLK_CTRL_FLR_START:
        return "VIRTIO_BLK_CTRL_FLR_START";
    case VIRTIO_BLK_CTRL_FLR_BDEV_RESET:
        return "VIRTIO_BLK_CTRL_FLR_BDEV_RESET";
    case VIRTIO_BLK_CTRL_FLR_BDEV_RESET_WAIT_DONE:
        return "VIRTIO_BLK_CTRL_FLR_BDEV_RESET_WAIT_DONE";
    case VIRTIO_BLK_CTRL_FLR_IO_CHANNEL_CLEAR:
        return "VIRTIO_BLK_CTRL_FLR_IO_CHANNEL_CLEAR";
    case VIRTIO_BLK_CTRL_FLR_IO_CHANNEL_WAIT_CLEARED:
        return "VIRTIO_BLK_CTRL_FLR_IO_CHANNEL_WAIT_CLEARED";
    case VIRTIO_BLK_CTRL_FLR_END:
        return "VIRTIO_BLK_CTRL_FLR_END";
    default:
        return "UNKNNOWN";
    }
}
#endif

static void virtio_blk_ctrl_flr_state_change(struct virtio_blk_ctrl *ctrl, enum virtio_blk_ctrl_flr_state state)
{
    nvmx_debug("%s: ctrl %p moved from state %s (%d) to state %s (%d)", __func__, ctrl,
            virtio_blk_ctrl_flr_state_to_string(ctrl->flr_state), ctrl->flr_state,
            virtio_blk_ctrl_flr_state_to_string(state), state);
    ctrl->flr_state = state;
}

static void virtio_blk_ctrl_flr_progress(struct virtio_blk_ctrl *ctrl)
{
    int rc;
    bool repeat;

    do {
        repeat = false;

        switch (ctrl->flr_state) {
        case VIRTIO_BLK_CTRL_FLR_IDLE:
            break;

        case VIRTIO_BLK_CTRL_FLR_START:
            ctrl->flr_context = virtio_blk_ctrl_flr_start(ctrl);
            if (!ctrl->flr_context) {
                nvmx_error("Failed to create FLR context for ctrl %p", ctrl);
                virtio_blk_ctrl_flr_state_change(ctrl, VIRTIO_BLK_CTRL_FLR_IDLE);
                break;
            }

            virtio_blk_ctrl_flr_state_change(ctrl, VIRTIO_BLK_CTRL_FLR_BDEV_RESET);
            /* Fallthrough */
        case VIRTIO_BLK_CTRL_FLR_BDEV_RESET:
            rc = spdk_bdev_reset(ctrl->bctx.desc, ctrl->flr_context->io_channel, virtio_blk_ctrl_bdev_reset_done, ctrl);
            if (rc != 0) {
                nvmx_error("Failed to initiate reset for bdev %s", spdk_bdev_get_name(ctrl->bctx.bdev));
                virtio_blk_ctrl_flr_state_change(ctrl, VIRTIO_BLK_CTRL_FLR_END);
                repeat = true;
                break;
            }

            virtio_blk_ctrl_flr_state_change(ctrl, VIRTIO_BLK_CTRL_FLR_BDEV_RESET_WAIT_DONE);
            break;

        case VIRTIO_BLK_CTRL_FLR_BDEV_RESET_WAIT_DONE:
            break;

        case VIRTIO_BLK_CTRL_FLR_IO_CHANNEL_CLEAR:
            /*
            * Upon FLR, we must cleanup all created mkeys, which happens
            * during spdk_ext_io_context_clear() call. As there might still
            * be IOs in-flight, we should do it asynchronously from the
            * IO threads context for an orderly cleanup.
            */
            virtio_blk_ctrl_io_channels_clear(ctrl, virtio_blk_ctrl_io_channels_cleared, spdk_get_thread());
            virtio_blk_ctrl_flr_state_change(ctrl, (ctrl->bctx.num_io_channels > 0) ?
                              VIRTIO_BLK_CTRL_FLR_IO_CHANNEL_WAIT_CLEARED : VIRTIO_BLK_CTRL_FLR_END);
            repeat = true;
            break;

        case VIRTIO_BLK_CTRL_FLR_IO_CHANNEL_WAIT_CLEARED:
            break;

        case VIRTIO_BLK_CTRL_FLR_END:
            virtio_blk_ctrl_flr_end(ctrl->flr_context);
            ctrl->flr_context = NULL;
            virtio_blk_ctrl_flr_state_change(ctrl, VIRTIO_BLK_CTRL_FLR_IDLE);
            break;

        default:
            break;
        }
    } while (repeat == true);
}

static int virtio_blk_ctrl_notify_flr(void *arg)
{
    struct virtio_blk_ctrl *ctrl = arg;

    nvmx_debug("ctrl %p serial '%s' pf_id %d : PCI FLR detected (before)",
            ctrl, ctrl->serial, ctrl->pf_id);

    if (ctrl->bctx.num_io_channels == 0)
        return 0;

    if (ctrl->flr_state != VIRTIO_BLK_CTRL_FLR_IDLE) {
        nvmx_debug("Already under FLR process, ignoring...");
        return 0;
    }

    virtio_blk_ctrl_flr_state_change(ctrl, VIRTIO_BLK_CTRL_FLR_START);
    virtio_blk_ctrl_flr_progress(ctrl);
    return 0;
}

static int virtio_blk_ctrl_pre_flr(void *arg)
{
    struct virtio_blk_ctrl *ctrl = arg;

    virtio_blk_ctrl_flr_progress(ctrl);
    if (ctrl->flr_state != VIRTIO_BLK_CTRL_FLR_IDLE) {
        nvmx_debug("Backend FLR is still in progress");
        return EAGAIN;
    }

    nvmx_debug("Backend FLR is finished successfully");
    return 0;
}

static int virtio_blk_ctrl_post_flr(void *arg)
{
    #ifdef DEBUG_ENABLED
    struct virtio_blk_ctrl *ctrl = arg;
    #endif

    nvmx_debug("ctrl %p serial '%s' pf_id %d : PCI FLR detected (after)",
            ctrl, ctrl->serial, ctrl->pf_id);
    return 0;
}

static void virtio_blk_ctrl_bdev_close(void *arg)
{
    struct virtio_blk_ctrl *ctrl = arg;

    switch (ctrl->bdev_type) {
    case VIRTIO_BLK_CTRL_BDEV_SPDK:
        virtio_blk_ctrl_bdev_close_spdk(ctrl);
        break;
    case VIRTIO_BLK_CTRL_BDEV_NONE:
        break;
    default:
        break;
    }
    ctrl->bdev_type = VIRTIO_BLK_CTRL_BDEV_NONE;
}

static int virtio_blk_ctrl_set_bdev_type(const char *type_str,
                                         enum virtio_blk_ctrl_bdev_type *type)
{
    if (!strncmp(type_str, "spdk", 4)) {
        *type = VIRTIO_BLK_CTRL_BDEV_SPDK;
        return 0;
    } else if (!strncmp(type_str, "none", 4)) {
        *type = VIRTIO_BLK_CTRL_BDEV_NONE;
        return 0;
    }

    return -EINVAL;
}

static int virtio_blk_ctrl_bdev_open(struct virtio_blk_ctrl *ctrl,
                                     const char *bdev_type,
                                     const char *bdev_name,
                                     struct virtio_blk_regs_set_attr *regs)
{
    pthread_mutex_lock(&ctrl->bdev_lock);
    if (virtio_blk_ctrl_set_bdev_type(bdev_type, &ctrl->bdev_type))
        goto err;

    switch (ctrl->bdev_type) {
    case VIRTIO_BLK_CTRL_BDEV_SPDK:
        if (!bdev_name) {
            nvmx_error("No bdev name provided for bdev_type %d", bdev_type);
            goto err;
        } else if (virtio_blk_ctrl_bdev_open_spdk(ctrl, bdev_name,
                                                  ctrl->nthreads)) {
            goto err;
        }
        regs->num_blocks = spdk_bdev_get_num_blocks(ctrl->bctx.bdev);
        regs->blk_size = spdk_bdev_get_block_size(ctrl->bctx.bdev);
        break;
    case VIRTIO_BLK_CTRL_BDEV_NONE:
        regs->num_blocks = 0;
        regs->blk_size = 0;
        break;
    default:
        nvmx_error("Unknown bdev_type %d", bdev_type);
        goto err;
    }
    pthread_mutex_unlock(&ctrl->bdev_lock);
    return 0;

err:
    ctrl->bdev_type = VIRTIO_BLK_CTRL_BDEV_NONE;
    pthread_mutex_unlock(&ctrl->bdev_lock);
    return -EINVAL;
}

static void virtio_blk_ctrl_free(struct virtio_blk_ctrl *ctrl)
{
    pthread_mutex_destroy(&ctrl->bdev_lock);
    if (ctrl->mr)
        ibv_dereg_mr(ctrl->mr);
    ibv_dealloc_pd(ctrl->pd);

    if (ctrl->destroy_done_cb)
        ctrl->destroy_done_cb(ctrl->destroy_done_cb_arg);

    free(ctrl);
}

static void
virtio_blk_ctrl_bdev_detach_end(struct virtio_blk_ctrl *ctrl)
{
    struct virtio_blk_regs_set_attr regs_attr = {};
    struct snap_virtio_blk_registers regs = {};
    bool success = true;

    /* Clear callback pointer after it is used */
    ctrl->bdev_close_cb = NULL;

    /*
     * We must always present some bdev to host, so after detaching
     * a "real" bdev, we open a new dummy one to replace it.
     */
    if (virtio_blk_ctrl_bdev_open(ctrl, "none", 0, &regs_attr)) {
        nvmx_error("Failed to open `none` bdev");
        success = false;
        goto end;
    }

    virtio_blk_regs_set(&regs_attr, &regs);
    if (snap_virtio_blk_ctrl_bar_setup(ctrl->sctrl, &regs,
                                       SNAP_VIRTIO_MOD_DEV_CFG)) {
        nvmx_error("Failed to update BAR with new bdev params");
        success = false;
        goto bdev_close;
    }
    goto end;

bdev_close:
    virtio_blk_ctrl_bdev_close_begin(ctrl);
end:
    ctrl->sctrl->pending_bdev_detach = false;

    /* return rpc response if needed */
    if (ctrl->bdev_detach_done) {
            ctrl->bdev_detach_done(ctrl->bdev_detach_done_arg, success);
            ctrl->bdev_detach_done = NULL;
    }
}

static void
virtio_blk_ctrl_bdev_detach_begin(struct virtio_blk_ctrl *ctrl)
{
    ctrl->bdev_close_cb = virtio_blk_ctrl_bdev_detach_end;

    if (virtio_blk_ctrl_is_suspended(ctrl))
        virtio_blk_ctrl_bdev_close_begin(ctrl);
    else {
        ctrl->sctrl->bdev_detach_cb_arg = ctrl;
        ctrl->sctrl->bdev_detach_cb = virtio_blk_ctrl_bdev_close_begin;
        ctrl->sctrl->pending_bdev_detach = true;
    }
}

void virtio_blk_ctrl_bdev_detach(void *ctrl_arg, void *done_cb_arg,
                                 bdev_detach_cb_t done_cb)
{
    struct virtio_blk_ctrl *ctrl = ctrl_arg;

    ctrl->bdev_detach_done_arg =  done_cb_arg;
    ctrl->bdev_detach_done = done_cb;
    virtio_blk_ctrl_bdev_detach_begin(ctrl);
}

int virtio_blk_ctrl_bdev_attach(void *ctrl_arg, const char *bdev_type,
                                const char *bdev_name, uint32_t size_max,
                                uint32_t seg_max, const char *serial)
{
    struct virtio_blk_ctrl *ctrl = ctrl_arg;
    struct virtio_blk_regs_set_attr regs_attr = {};
    struct snap_virtio_blk_registers regs = {};

    /*
     * No need to suspend the controller before attaching new bdev.
     * The old bdev has to be none, io cannot run until new bdev is attached
     */
    if (ctrl->bdev_type != VIRTIO_BLK_CTRL_BDEV_NONE) {
        nvmx_error("There is already a bdev attached to the controller");
        goto err;
    }

    if (serial && strnlen(serial, VIRTIO_BLK_ID_BYTES) > 0) {
        size_t serial_len = strnlen(serial, VIRTIO_BLK_ID_BYTES);
        if (serial_len == VIRTIO_BLK_ID_BYTES &&
            serial[VIRTIO_BLK_ID_BYTES] != '\0') {
            nvmx_error("Serial value too long (max %d chars)",
                       VIRTIO_BLK_ID_BYTES);
            goto err;
        }
        memcpy(ctrl->serial, serial, VIRTIO_BLK_ID_BYTES + 1);
    } else {
        if (strnlen(ctrl->serial, VIRTIO_BLK_ID_BYTES) == 0) {
            memcpy(ctrl->serial, VIRTIO_BLK_CTRL_ID_DEFAULT,
                   VIRTIO_BLK_ID_BYTES + 1);
        }
    }

    if (virtio_blk_ctrl_bdev_open(ctrl, bdev_type, bdev_name, &regs_attr)) {
        nvmx_error("Failed to open bdev of type %d", bdev_type);
        goto err;
    }

    regs_attr.size_max = size_max;
    regs_attr.seg_max = seg_max;
    virtio_blk_regs_set(&regs_attr, &regs);
    if (snap_virtio_blk_ctrl_bar_setup(ctrl->sctrl, &regs,
                                       SNAP_VIRTIO_MOD_DEV_CFG)) {
        nvmx_error("Failed to update BAR with new bdev params");
        goto bdev_close;
    }

    return 0;

bdev_close:
    virtio_blk_ctrl_bdev_close_begin(ctrl);
err:
    return -EINVAL;
}

int virtio_blk_ctrl_get_debugstat(void *ctrl_arg,
                        struct snap_virtio_ctrl_debugstat *ctrl_debugstat)
{
    struct virtio_blk_ctrl *ctrl = ctrl_arg;

    return snap_virtio_blk_ctrl_get_debugstat(ctrl->sctrl, ctrl_debugstat);
}

struct virtio_blk_ctrl *
virtio_blk_ctrl_init(const struct virtio_blk_ctrl_init_attr *attr)
{
    struct virtio_blk_ctrl *ctrl;
    struct virtio_blk_regs_set_attr regs_attr = {};
    struct snap_virtio_blk_ctrl_attr sctrl_attr = {};
    struct snap_virtio_ctrl_bar_cbs bar_cbs = {
        .num_vfs_changed = virtio_blk_ctrl_num_vfs_changed,
        .notify_flr = virtio_blk_ctrl_notify_flr,
        .pre_flr = virtio_blk_ctrl_pre_flr,
        .post_flr = virtio_blk_ctrl_post_flr,
    };
    int ret, i;
    size_t serial_len;

    ctrl = calloc(1, sizeof(*ctrl));
    if (!ctrl)
        goto err;
    ctrl->nthreads = attr->nthreads;

    if (attr->serial && strnlen(attr->serial, VIRTIO_BLK_ID_BYTES) > 0) {
        serial_len = strnlen(attr->serial, VIRTIO_BLK_ID_BYTES);
        if (serial_len == VIRTIO_BLK_ID_BYTES &&
            attr->serial[VIRTIO_BLK_ID_BYTES] != '\0') {
            nvmx_error("Serial value too long (max %d chars)",
                       VIRTIO_BLK_ID_BYTES);
            goto free_ctrl;
        }
        memcpy(ctrl->serial, attr->serial, VIRTIO_BLK_ID_BYTES + 1);
    } else {
        memcpy(ctrl->serial, VIRTIO_BLK_CTRL_ID_DEFAULT,
               VIRTIO_BLK_ID_BYTES + 1);
    }

    ctrl->sctx = virtio_blk_ctrl_find_snap_context(attr->emu_manager_name,
                                                   attr->pf_id);
    if (!ctrl->sctx)
        goto free_ctrl;

    ctrl->pd = ibv_alloc_pd(ctrl->sctx->context);
    if (!ctrl->pd)
        goto free_ctrl;

    if (attr->use_mem_pool) {
        if (attr->memp) {
            ctrl->memp = attr->memp;

            ctrl->mr = snap_reg_mr(ctrl->pd, memp_get_buffer(attr->memp), memp_get_size(attr->memp));
            if (!ctrl->mr) {
                nvmx_error("failed to alloc");
                goto dealloc_pd;
            }
        }
        else
            nvmx_warn("Mempool mode requested but memory pool is not provided");
    }

    ret = pthread_mutex_init(&ctrl->bdev_lock, NULL);
    if (ret)
        goto dereg_mr;

    ret = virtio_blk_ctrl_bdev_open(ctrl, attr->bdev_type, attr->bdev_name,
                                    &regs_attr);
    if (ret) {
        nvmx_error("Failed to attach bdev of type %d", attr->bdev_type);
        goto destroy_mutex;
    }

    regs_attr.max_queues = attr->num_queues;
    regs_attr.queue_depth = attr->queue_depth;
    regs_attr.size_max = attr->size_max;
    regs_attr.seg_max = attr->seg_max;
    regs_attr.legacy_mode = false;
    if (attr->admin_q && attr->vf_id < 0)
       regs_attr.admin_queue = true;
    virtio_blk_regs_set(&regs_attr, &sctrl_attr.regs);

    ctrl->vf_change_cb = attr->vf_change_cb;
    ctrl->vf_change_cb_arg = attr->vf_change_cb_arg;
    ctrl->hotunplug_device_cb = attr->hotunplug_device_cb;
    ctrl->hotunplug_device_ctx = attr->hotunplug_device_ctx;
    ctrl->hotunplug_timeout_cb = attr->hotunplug_timeout_cb;

    if (snap_env_getenv(VIRTIO_BLK_SNAP_ZCOPY) != 0 &&
                        ctrl->sctx->virtio_blk_caps.crossing_vhca_mkey)
        ctrl->zcopy = true;

    sctrl_attr.common.bar_cbs = &bar_cbs;
    sctrl_attr.common.cb_ctx = ctrl;
    sctrl_attr.common.pf_id = attr->pf_id;
    sctrl_attr.common.vf_id = attr->vf_id;
    sctrl_attr.common.pci_type = (attr->vf_id >= 0) ? SNAP_VIRTIO_BLK_VF :
                                                      SNAP_VIRTIO_BLK_PF;
    sctrl_attr.common.pd = ctrl->pd;
    sctrl_attr.common.npgs = attr->nthreads;
    sctrl_attr.common.force_in_order = attr->force_in_order;
    sctrl_attr.common.suspended = attr->suspended;
    sctrl_attr.common.recover = attr->recover;
    sctrl_attr.common.force_recover = attr->force_recover;

    /* [00988428]BF2: controller_virtio_blk_create failed
     * The root cause: BusMaster is disabled and FLR is present on VF.
     * In this case creation of controller failed.
     *
     * We did 3 tests:
     * 1. FLR on VF X in parallel create controller on VF Y - creation succeeded
     * 2. FLR on VF X in parallel create controller on VF X - creation failed
     * 3. Set BusMaster disable on VF Y,
     *    FLR on VF X in parallel create controller on VF X - creation failed
     *
     * As the FW doesn't set status & syndrome (verified this), we cannot base
     * our solution on status/syndrome.
     *
     * Work-around: retry creation of controller on failure.
     */

    for (i = 0; i < 5; i++) {
        ctrl->sctrl = snap_virtio_blk_ctrl_open(ctrl->sctx, &sctrl_attr,
                                                &virtio_blk_ctrl_backend_ops,
                                                ctrl);
        if (ctrl->sctrl)
            break;

        sleep(1);
    }

    if (!ctrl->sctrl) {
        if (attr->vf_id >= 0) {
            nvmx_error("Failed to open VirtIO BLK controller %d [in order %d recover %d]"
                " over RDMA device %s, PF %d ZCOPY %d, MEMPOOL %d",
                attr->vf_id, attr->force_in_order, attr->recover,
                attr->emu_manager_name, attr->pf_id, ctrl->zcopy, !!ctrl->memp);
        } else {
            nvmx_error("Failed to open VirtIO BLK controller %d [in order %d recover %d]"
                " over RDMA device %s, PF %d ZCOPY %d, MEMPOOL %d",
                attr->pf_id, attr->force_in_order, attr->recover,
                attr->emu_manager_name, attr->pf_id, ctrl->zcopy, !!ctrl->memp);
        }

        goto close_bdev;
    }

    ctrl->pf_id = attr->pf_id;
    ctrl->vf_id = -1;
    if (attr->vf_id >= 0) {
        ctrl->vf_id = attr->vf_id;
        nvmx_info("new VirtIO BLK controller %d [in order %d recover %d]"
                  " was opened successfully over RDMA device %s, PF %d ZCOPY %d,"
                  " MEMPOOL %d",
                  attr->vf_id, attr->force_in_order, attr->recover,
                  attr->emu_manager_name, attr->pf_id, ctrl->zcopy, !!ctrl->memp);
        snprintf(ctrl->name, VBLK_EMU_NAME_MAXLEN,
                "%s%dpf%dvf%d", VBLK_EMU_NAME_PREFIX,
                rdma_dev_name_to_id(attr->emu_manager_name), attr->pf_id, attr->vf_id);
    } else {
        nvmx_info("new VirtIO BLK controller %d [in order %d recover %d]"
                  " was opened successfully over RDMA device %s ZCOPY %d,"
                  " MEMPOOL %d",
                  attr->pf_id, attr->force_in_order, attr->recover,
                  attr->emu_manager_name, ctrl->zcopy, !!ctrl->memp);
        snprintf(ctrl->name, VBLK_EMU_NAME_MAXLEN,
                "%s%dpf%d", VBLK_EMU_NAME_PREFIX,
                rdma_dev_name_to_id(attr->emu_manager_name), attr->vf_id);
    }
    return ctrl;

close_bdev:
    ctrl->bdev_close_cb = virtio_blk_ctrl_free;
    virtio_blk_ctrl_bdev_close_begin(ctrl);
    /* The rest of resources will be cleaned in the callback */
    return NULL;
destroy_mutex:
    pthread_mutex_destroy(&ctrl->bdev_lock);
dereg_mr:
    if (ctrl->mr)
        ibv_dereg_mr(ctrl->mr);
dealloc_pd:
    ibv_dealloc_pd(ctrl->pd);
free_ctrl:
    free(ctrl);
err:
    return NULL;
}

void virtio_blk_ctrl_destroy(void *arg, void (*done_cb)(void *arg),
                             void *done_cb_arg)
{
    struct virtio_blk_ctrl *ctrl = arg;

    /* bdev_close is async, so free ctrl once done */
    snap_virtio_blk_ctrl_close(ctrl->sctrl);
    ctrl->sctrl = NULL;

    /*
     * bdev_close_cb() marks the end of the process started
     * with calling virtio_blk_ctrl_bdev_close_begin().
     * destroy_done_cb() marks the end of the process started
     * with calling virtio_blk_ctrl_destroy().
     */
    ctrl->bdev_close_cb = virtio_blk_ctrl_free;
    ctrl->destroy_done_cb = done_cb;
    ctrl->destroy_done_cb_arg = done_cb_arg;
    virtio_blk_ctrl_bdev_close_begin(ctrl);
}

void virtio_blk_ctrl_get_stat(struct virtio_blk_ctrl *ctrl, struct virtio_blk_ctrl_io_stat *io_stat)
{
    int i;
    struct snap_virtio_blk_ctrl *sctrl = ctrl->sctrl;
    const struct snap_virtio_ctrl_queue_stats *q_stat;

    memset(io_stat, 0, sizeof(*io_stat));

    for (i = 0; i < sctrl->common.max_queues; i++) {
        q_stat = snap_virtio_ctrl_q_io_stats(&sctrl->common, i);
        if (!q_stat)
            continue;
        io_stat->read_ios += q_stat->read.total;
        io_stat->completed_read_ios += q_stat->read.success;
        io_stat->completed_unord_r_ios += q_stat->read.unordered;
        io_stat->write_ios += q_stat->write.total;
        io_stat->completed_write_ios += q_stat->write.success;
        io_stat->completed_unord_w_ios += q_stat->write.unordered;
        io_stat->flush_ios += q_stat->flush.total;
        io_stat->completed_flush_ios += q_stat->flush.success;
        io_stat->completed_unord_f_ios += q_stat->flush.unordered;
        io_stat->err_read_os += q_stat->read.fail;
        io_stat->err_write_os += q_stat->write.fail;
        io_stat->err_flush_os += q_stat->flush.fail;
        io_stat->outstand_in_ios += q_stat->outstanding.outstanding_total;
        io_stat->outstand_in_bdev_ios += q_stat->outstanding.outstanding_in_bdev;
        io_stat->outstand_to_host_ios += q_stat->outstanding.outstanding_to_host;
        io_stat->fatal_ios += q_stat->outstanding.fatal;
    }
}

static void *virtio_blk_ctrl_state_alloc(struct virtio_blk_ctrl *ctrl, int *len)
{
    size_t common_cfg, queue_cfg, dev_cfg;
    void *buf;

    *len = snap_virtio_ctrl_state_size(&ctrl->sctrl->common, &common_cfg,
                                       &queue_cfg, &dev_cfg);
    if (*len <= 0) {
        nvmx_error("Failed to get controller state size");
        return NULL;
    }

    buf = calloc(1, *len);
    if (!buf) {
        nvmx_error("Not enough memory to save state");
        return NULL;
    }
    return buf;
}

int virtio_blk_ctrl_state_save(void *arg, char *file)
{
    struct virtio_blk_ctrl *ctrl = arg;
    FILE *fp;
    char *buf;
    int len;
    int ret;

    buf = virtio_blk_ctrl_state_alloc(ctrl, &len);
    if (!buf)
        return -ENOMEM;

    ret = snap_virtio_ctrl_state_save(&ctrl->sctrl->common, buf, len);
    if (ret < 0) {
        nvmx_error("Failed to get controller state");
        goto free_state;
    }

    fp = fopen(file, "w");
    if (!fp) {
        nvmx_error("Failed to open %s : %m", file);
        goto free_state;
    }

    ret = fwrite(buf, 1, len, fp);
    if (ret != len) {
        nvmx_error("Failed to save state");
        goto close_file;
    }

    fclose(fp);
    return 0;

close_file:
    fclose(fp);
free_state:
    free(buf);
    return -EINVAL;
}

int virtio_blk_ctrl_state_restore(void *arg, char *file)
{
    struct virtio_blk_ctrl *ctrl = arg;
    FILE *fp;
    char *buf;
    int len;
    int ret;

    buf = virtio_blk_ctrl_state_alloc(ctrl, &len);
    if (!buf)
        return -ENOMEM;

    fp = fopen(file, "r");
    if (!fp) {
        nvmx_error("Failed to open %s : %m", file);
        goto free_state;
    }

    ret = fread(buf, 1, len, fp);
    if (ret != len) {
        nvmx_error("Failed to save state");
        goto close_file;
    }

    fclose(fp);

    ret = snap_virtio_ctrl_state_restore(&ctrl->sctrl->common, buf, len);
    if (ret < 0) {
        nvmx_error("Failed to get controller state");
        goto free_state;
    }

    return 0;

close_file:
    fclose(fp);
free_state:
    free(buf);
    return -EINVAL;
}

int virtio_blk_ctrl_lm_enable(void *arg, char *lm_channel_name)
{
    struct virtio_blk_ctrl *ctrl = arg;

    return snap_virtio_ctrl_lm_enable(&ctrl->sctrl->common, lm_channel_name);
}

void virtio_blk_ctrl_lm_disable(void *arg)
{
    struct virtio_blk_ctrl *ctrl = arg;

    snap_virtio_ctrl_lm_disable(&ctrl->sctrl->common);
}
