#include <stdlib.h>
#include <unistd.h>
#include <infiniband/verbs.h>
#include <snap_virtio_fs_ctrl.h>
#include "compiler.h"
#include "virtio_fs_controller.h"
#include "nvme_emu_log.h"
#include "virtio_fs_regs.h"
#include "mlnx_snap_pci_manager.h"

#define VIRTIO_FS_BUF_ALIGNMENT 4096

static struct snap_context *
virtio_fs_ctrl_find_snap_context(const char *emu_manager, int pf_id)
{
    struct snap_context *ctx, *found = NULL;
    struct ibv_device **ibv_list;
    struct snap_pci **pf_list;
    int ibv_list_sz, pf_list_sz;
    int i, j;

    ibv_list = ibv_get_device_list(&ibv_list_sz);
    if (!ibv_list)
        return NULL;

    for (i = 0; i < ibv_list_sz; i++) {
        if (strncmp(ibv_get_device_name(ibv_list[i]), emu_manager, 16))
            continue;

        ctx = mlnx_snap_get_snap_context(ibv_get_device_name(ibv_list[i]));
        if (!ctx)
            continue;

        if (!(ctx->emulation_caps & SNAP_VIRTIO_FS))
            continue;

        pf_list = calloc(ctx->virtio_fs_pfs.max_pfs, sizeof(*pf_list));
        if (!pf_list)
            continue;

        pf_list_sz = snap_get_pf_list(ctx, SNAP_VIRTIO_FS, pf_list);
        for (j = 0; j < pf_list_sz; j++) {
            if (pf_list[j]->plugged && pf_list[j]->id == pf_id) {
                found = ctx;
                break;
            }
        }
        free(pf_list);

        if (found)
            break;
    }

    ibv_free_device_list(ibv_list);
    return found;
}

static int virtio_fs_ctrl_fs_dev_handle_req(void *ctx,
                                            struct iovec *fuse_in_iov, int in_iovcnt,
                                            struct iovec *fuse_out_iov, int out_iovcnt,
                                            struct snap_fs_dev_io_done_ctx *done_ctx)
{
    struct virtio_fs_ctrl *ctrl = ctx;

    if (unlikely(ctrl->dev_type == VIRTIO_FS_CTRL_FS_DEV_NONE))
        return -ENODEV;

    return -ENOSYS;
}

static void *virtio_fs_ctrl_fs_dev_dma_malloc(size_t size)
{
    void *buf;
    if (!posix_memalign(&buf, VIRTIO_FS_BUF_ALIGNMENT, size))
        return buf;

    return NULL;
}

static void virtio_fs_ctrl_fs_dev_dma_free(void *buf)
{
    free(buf);
}

static struct snap_fs_dev_ops fs_ctrl_dev_ops = {
    .handle_req = virtio_fs_ctrl_fs_dev_handle_req,
    .dma_malloc = virtio_fs_ctrl_fs_dev_dma_malloc,
    .dma_free = virtio_fs_ctrl_fs_dev_dma_free
};

void virtio_fs_ctrl_progress(void *arg)
{
    struct virtio_fs_ctrl *ctrl = arg;

    snap_virtio_fs_ctrl_progress(ctrl->sctrl);
}

int virtio_fs_ctrl_progress_all_io(void *arg)
{
    struct virtio_fs_ctrl *ctrl = arg;

    snap_virtio_fs_ctrl_io_progress(ctrl->sctrl);
    return -1;
}

int virtio_fs_ctrl_progress_io(void *arg, int thread_id)
{
    struct virtio_fs_ctrl *ctrl = arg;

    snap_virtio_fs_ctrl_io_progress_thread(ctrl->sctrl, thread_id);
    return -1;
}

void virtio_fs_ctrl_suspend(void *arg)
{
    struct virtio_fs_ctrl *ctrl = arg;

    snap_virtio_ctrl_suspend(&ctrl->sctrl->common);
}

bool virtio_fs_ctrl_is_suspended(void *arg)
{
    struct virtio_fs_ctrl *ctrl = arg;

    if (!ctrl->sctrl)
        return true;

    return snap_virtio_ctrl_is_suspended(&ctrl->sctrl->common) ||
           snap_virtio_ctrl_is_stopped(&ctrl->sctrl->common);
}

int virtio_fs_ctrl_resume(void *arg)
{
    struct virtio_fs_ctrl *ctrl = arg;

    return snap_virtio_ctrl_resume(&ctrl->sctrl->common);
}

static void virtio_fs_ctrl_fs_dev_close_begin(struct virtio_fs_ctrl *ctrl)
{
    bool run_close_cb = false;

    pthread_mutex_lock(&ctrl->dev_lock);
    if (ctrl->dev_type == VIRTIO_FS_CTRL_FS_DEV_NONE) {
        run_close_cb = true;
        goto out;
    }

out:
    pthread_mutex_unlock(&ctrl->dev_lock);

    if (ctrl->fsdev_close_cb && run_close_cb)
        ctrl->fsdev_close_cb(ctrl);
}

static int virtio_fs_ctrl_num_vfs_changed(void *arg, uint16_t new_numvfs)
{
    struct virtio_fs_ctrl *ctrl = arg;

    nvmx_debug("num_vfs_changed event was detected");

    if (ctrl->vf_change_cb)
        ctrl->vf_change_cb(ctrl->vf_change_cb_arg, ctrl->pf_id, new_numvfs);

    return 0;
}

static int virtio_fs_ctrl_set_fs_dev_type(const char *type_str,
                                          enum virtio_fs_ctrl_dev_type *type)
{
    if (!strncmp(type_str, "none", 4)) {
        *type = VIRTIO_FS_CTRL_FS_DEV_NONE;
        return 0;
    }

    return -EINVAL;
}

static int virtio_fs_ctrl_fs_dev_open(struct virtio_fs_ctrl *ctrl,
                                      const char *fs_dev_type,
                                      struct virtio_fs_regs_set_attr *regs)
{
    pthread_mutex_lock(&ctrl->dev_lock);

    if (virtio_fs_ctrl_set_fs_dev_type(fs_dev_type, &ctrl->dev_type))
        goto err;

    switch (ctrl->dev_type) {
        case VIRTIO_FS_CTRL_FS_DEV_NONE:
            break;
        default:
            nvmx_error("Unknown fs_dev_type %d", fs_dev_type);
            goto err;
    }

    pthread_mutex_unlock(&ctrl->dev_lock);
    return 0;

err:
    ctrl->dev_type = VIRTIO_FS_CTRL_FS_DEV_NONE;
    pthread_mutex_unlock(&ctrl->dev_lock);
    return -EINVAL;
}

static void virtio_fs_ctrl_free(struct virtio_fs_ctrl *ctrl)
{
    pthread_mutex_destroy(&ctrl->dev_lock);
    ibv_dealloc_pd(ctrl->pd);
    free(ctrl);
}

struct virtio_fs_ctrl *
virtio_fs_ctrl_init(const struct virtio_fs_ctrl_init_attr *attr)
{
    struct virtio_fs_ctrl *ctrl;
    struct virtio_fs_regs_set_attr regs_attr = {};
    struct snap_virtio_fs_ctrl_attr sctrl_attr = {};
    struct snap_virtio_ctrl_bar_cbs bar_cbs = {
        .num_vfs_changed = virtio_fs_ctrl_num_vfs_changed,
    };
    int ret;
    size_t tag_len;

    ctrl = calloc(1, sizeof(*ctrl));
    if (!ctrl)
        goto err;
    ctrl->nthreads = attr->nthreads;

    /* 5.11.4 Device configuration layout
     * The tag is encoded in UTF-8 and padded with NUL bytes if shorter
     * than the available space. This field is not NUL-terminated if
     * the encoded bytes take up the entire field.
     */
    if (attr->tag) {
        tag_len = strnlen(attr->tag, SNAP_VIRTIO_FS_DEV_CFG_TAG_LEN + 1);
        if (tag_len > SNAP_VIRTIO_FS_DEV_CFG_TAG_LEN) {
            nvmx_error("Tag value too long (max %d chars)",
                       SNAP_VIRTIO_FS_DEV_CFG_TAG_LEN);
            goto free_ctrl;
        }
    } else {
        nvmx_error("Please specify the 'tag' value.");
        goto free_ctrl;
    }

    ctrl->sctx = virtio_fs_ctrl_find_snap_context(attr->emu_manager_name,
                                                  attr->pf_id);
    if (!ctrl->sctx)
        goto free_ctrl;

    ctrl->pd = ibv_alloc_pd(ctrl->sctx->context);
    if (!ctrl->pd)
        goto free_ctrl;

    ret = pthread_mutex_init(&ctrl->dev_lock, NULL);
    if (ret)
        goto dealloc_pd;

    ret = virtio_fs_ctrl_fs_dev_open(ctrl, attr->dev_type, &regs_attr);
    if (ret) {
        nvmx_error("Failed to attach fsdev of type %s", attr->dev_type);
        goto destroy_mutex;
    }

    regs_attr.max_queues = attr->num_queues;
    regs_attr.queue_depth = attr->queue_depth;
    regs_attr.tag = attr->tag;
    regs_attr.legacy_mode = false;
    virtio_fs_regs_set(&regs_attr, &sctrl_attr.regs);

    ctrl->vf_change_cb = attr->vf_change_cb;
    ctrl->vf_change_cb_arg = attr->vf_change_cb_arg;

    sctrl_attr.common.pd = ctrl->pd;
    sctrl_attr.common.npgs = attr->nthreads;

    sctrl_attr.common.bar_cbs = &bar_cbs;
    sctrl_attr.common.cb_ctx = ctrl;
    sctrl_attr.common.pf_id = attr->pf_id;
    sctrl_attr.common.vf_id = attr->vf_id;
    sctrl_attr.common.pci_type = (attr->vf_id >= 0) ? SNAP_VIRTIO_FS_VF :
                                                      SNAP_VIRTIO_FS_PF;
    sctrl_attr.common.pd = ctrl->pd;
    sctrl_attr.common.npgs = attr->nthreads;
    sctrl_attr.common.force_in_order = attr->force_in_order;
    sctrl_attr.common.suspended = attr->suspended;
    sctrl_attr.common.recover = attr->recover;

    ctrl->sctrl = snap_virtio_fs_ctrl_open(ctrl->sctx, &sctrl_attr,
                                           &fs_ctrl_dev_ops,
                                           ctrl);
    if (!ctrl->sctrl)
        goto close_bdev;

    ctrl->pf_id = attr->pf_id;
    if (attr->vf_id >= 0) {
        nvmx_info("new virtual VirtIO FS controller %d [in order %d recover %d]"
                  " was opened successfully over RDMA device %s, PF %d",
                  attr->vf_id, attr->force_in_order, attr->recover,
                  attr->emu_manager_name, attr->pf_id);

    } else {
        nvmx_info("new VirtIO FS controller %d [in order %d recover %d]"
                  " was opened successfully over RDMA device %s",
                  attr->pf_id, attr->force_in_order, attr->recover,
                  attr->emu_manager_name);
    }
    return ctrl;

close_bdev:
    ctrl->fsdev_close_cb = virtio_fs_ctrl_free;
    virtio_fs_ctrl_fs_dev_close_begin(ctrl);
    /* The rest of resources will be cleaned in the callback */
    return NULL;
destroy_mutex:
    pthread_mutex_destroy(&ctrl->dev_lock);
dealloc_pd:
    ibv_dealloc_pd(ctrl->pd);
free_ctrl:
    free(ctrl);
err:
    return NULL;
}

void virtio_fs_ctrl_destroy(void *arg)
{
    struct virtio_fs_ctrl *ctrl = arg;

    /* fsdev_close is async, so free ctrl once done */
    snap_virtio_fs_ctrl_close(ctrl->sctrl);
    ctrl->sctrl = NULL;
    ctrl->fsdev_close_cb = virtio_fs_ctrl_free;
    virtio_fs_ctrl_fs_dev_close_begin(ctrl);
}

static void *virtio_fs_ctrl_state_alloc(struct virtio_fs_ctrl *ctrl, int *len)
{
    size_t common_cfg, queue_cfg, dev_cfg;
    void *buf;

    *len = snap_virtio_ctrl_state_size(&ctrl->sctrl->common, &common_cfg,
                                       &queue_cfg, &dev_cfg);
    if (*len <= 0) {
        nvmx_error("Failed to get controller state size");
        return NULL;
    }

    buf = calloc(1, *len);
    if (!buf) {
        nvmx_error("Not enough memory to save state");
        return NULL;
    }
    return buf;
}

int virtio_fs_ctrl_state_save(void *arg, char *file)
{
    struct virtio_fs_ctrl *ctrl = arg;
    FILE *fp;
    char *buf;
    int len;
    int ret;

    buf = virtio_fs_ctrl_state_alloc(ctrl, &len);
    if (!buf)
        return -ENOMEM;

    ret = snap_virtio_ctrl_state_save(&ctrl->sctrl->common, buf, len);
    if (ret < 0) {
        nvmx_error("Failed to get controller state");
        goto free_state;
    }

    fp = fopen(file, "w");
    if (!fp) {
        nvmx_error("Failed to open %s : %m", file);
        goto free_state;
    }

    ret = fwrite(buf, 1, len, fp);
    if (ret != len) {
        nvmx_error("Failed to save state");
        goto close_file;
    }

    fclose(fp);
    return 0;

close_file:
    fclose(fp);
free_state:
    free(buf);
    return -EINVAL;
}

int virtio_fs_ctrl_state_restore(void *arg, char *file)
{
    struct virtio_fs_ctrl *ctrl = arg;
    FILE *fp;
    char *buf;
    int len;
    int ret;

    buf = virtio_fs_ctrl_state_alloc(ctrl, &len);
    if (!buf)
        return -ENOMEM;

    fp = fopen(file, "r");
    if (!fp) {
        nvmx_error("Failed to open %s : %m", file);
        goto free_state;
    }

    ret = fread(buf, 1, len, fp);
    if (ret != len) {
        nvmx_error("Failed to save state");
        goto close_file;
    }

    fclose(fp);

    ret = snap_virtio_ctrl_state_restore(&ctrl->sctrl->common, buf, len);
    if (ret < 0) {
        nvmx_error("Failed to get controller state");
        goto free_state;
    }

    return 0;

close_file:
    fclose(fp);
free_state:
    free(buf);
    return -EINVAL;
}

int virtio_fs_ctrl_lm_enable(void *arg, char *lm_channel_name)
{
    struct virtio_fs_ctrl *ctrl = arg;

    return snap_virtio_ctrl_lm_enable(&ctrl->sctrl->common, lm_channel_name);
}

void virtio_fs_ctrl_lm_disable(void *arg)
{
    struct virtio_fs_ctrl *ctrl = arg;

    snap_virtio_ctrl_lm_disable(&ctrl->sctrl->common);
}
